package se.su.dsv.scipro;

import com.google.inject.persist.Transactional;
import org.joda.time.DateTime;
import se.su.dsv.scipro.checklist.ChecklistCategory;
import se.su.dsv.scipro.match.ApplicationPeriod;
import se.su.dsv.scipro.match.Keyword;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.milestones.dataobjects.MilestonePhaseTemplate;
import se.su.dsv.scipro.milestones.service.MilestoneActivityTemplateService;
import se.su.dsv.scipro.profiles.CurrentProfile;
import se.su.dsv.scipro.profiles.Profiles;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.report.GradingCriterionPointTemplate;
import se.su.dsv.scipro.report.GradingReportTemplate;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.system.*;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.*;

public class DataInitializer implements Lifecycle {
    public static final int APPLICATION_PERIOD_START_MINUS_DAYS = 1;
    public static final int APPLICATION_PERIOD_END_PLUS_DAYS = 3;
    public static final int APPLICATION_PERIOD_COURSE_START_PLUS_DAYS = 5;
    public static final long RESEARCH_AREA_ID = 12L;

    @Inject
    private UserService userService;
    @Inject
    private PasswordService passwordService;
    @Inject
    private MilestoneActivityTemplateService milestoneActivityTemplateService;
    @Inject
    private CurrentProfile profile;
    @Inject
    private Provider<EntityManager> em;

    private static final String MAIL = "@example.com";

    private static final String ADMIN = "admin";
    private static final String EMPLOYEE_1 = "Eric";
    private static final String EMPLOYEE_2 = "Eve";
    private static final String EMPLOYEE_LAST = "Employee";
    private static final String STUDENT_1 = "Sture";
    private static final String STUDENT_2 = "Stina";
    private static final String STUDENT_3 = "Sid";
    private static final String STUDENT_4 = "Simon";
    private static final String STUDENT_LAST = "Student";
    private static final String PROJECT_1 = "Putting the it in supervising";
    private static final String PROJECT_2 = "Preserving quality in thesis work";
    private static final String RESEARCH_AREA_1 = "Digital Flames";

    private static final String RESEARCH_AREA_2 = "Intellectual Morphism";
    private User eric_employee;
    private User eve_employee;
    private User sture_student;
    private User stina_student;
    private User sid_student;
    private User simon_student;
    private ProjectType bachelorClass;
    private Set<ResearchArea> researchAreas;
    private Long researchAreaId = RESEARCH_AREA_ID;
    private Set<Language> languages;
    private ResearchArea researchArea1;
    private ResearchArea researchArea2;
    private ProjectType masterClass;

    @Transactional
    @Override
    public void start() {
        if (profile.getCurrentProfile().equals(Profiles.DEV) && noUsers()) {
            createDefaultProjectTypesIfNotDone();
            createDefaultChecklistCategoriesIfNotDone();
            createDefaultLanguagesIfNotDone();
            createApplicationPeriodIfNotDone();
            createGradingCriterionTemplateIfNotDone();
            createResearchAreasForDemo();
            createKeywordsIfNotDone();
            createMilestonesIfNotDone();
            createUsers();
            createProjects();
        }
        if (profile.getCurrentProfile().equals(Profiles.DEV) && noAdminUser()) {
            createAdmin();
        }
    }

    @Override
    public void stop() {
    }

    private boolean noAdminUser() {
        return userService.findByUsername(ADMIN + MAIL) == null;
    }


    private void createApplicationPeriodIfNotDone() {
        ApplicationPeriod applicationPeriod = new ApplicationPeriod("HT 2014");
        applicationPeriod.setStartDate(new DateTime().minusDays(APPLICATION_PERIOD_START_MINUS_DAYS).toDate());
        applicationPeriod.setEndDate(new DateTime().plusDays(APPLICATION_PERIOD_END_PLUS_DAYS).toDate());
        applicationPeriod.setCourseStartDate(new DateTime().plusDays(APPLICATION_PERIOD_COURSE_START_PLUS_DAYS).toDate());
        applicationPeriod = save(applicationPeriod);
        applicationPeriod.setProjectTypes(new HashSet<>(Collections.singletonList(bachelorClass)));
        save(applicationPeriod);
    }

    private void createKeywordsIfNotDone() {
        Keyword keyword1 = new Keyword("IT");
        keyword1.addResearchArea(researchArea1);
        keyword1.addResearchArea(researchArea2);
        save(keyword1);

        Keyword keyword2 = new Keyword("Computers");
        keyword2.addResearchArea(researchArea1);
        keyword2.addResearchArea(researchArea2);
        save(keyword2);
    }

    private void createResearchAreasForDemo() {
        researchAreas = new HashSet<>();
        researchArea1 = createResearchArea(RESEARCH_AREA_1);
        researchArea2 = createResearchArea(RESEARCH_AREA_2);
    }

    private ResearchArea createResearchArea(String title) {
        ResearchArea researchArea = new ResearchArea();
        researchArea.setTitle(title);
        researchArea.setIdentifier(researchAreaId++);
        researchArea = save(researchArea);
        researchAreas.add(researchArea);
        return researchArea;
    }

    private boolean noUsers() {
        return userService.count() == 0;
    }

    private void createProjects() {
        createProject(PROJECT_1, eric_employee, sture_student, stina_student, eve_employee);
        createProject(PROJECT_2, eve_employee, sid_student, simon_student, eric_employee);
    }

    private void createProject(String title, User headSupervisor, User student1, User student2, User reviewer) {
        Project project = Project.builder().title(title).projectType(bachelorClass).build();
        project.setHeadSupervisor(headSupervisor);
        project.addProjectParticipant(student2);
        project.addProjectParticipant(student1);
        project.addReviewer(reviewer);
        save(project);
    }

    private void createUsers() {
        createAdmin();

        eric_employee = createEmployee(EMPLOYEE_1);
        eve_employee = createEmployee(EMPLOYEE_2);

        sture_student = createStudent(STUDENT_1);
        stina_student = createStudent(STUDENT_2);
        sid_student = createStudent(STUDENT_3);
        simon_student = createStudent(STUDENT_4);
    }

    private User createStudent(String firstName) {
        User user = createUser(firstName, STUDENT_LAST);
        createBeta(user);
        return user;
    }

    private User createEmployee(String firstName) {
        User user = createUser(firstName, EMPLOYEE_LAST);
        Unit u = createUnit();

        user.addRole(Roles.SUPERVISOR);
        user.addRole(Roles.REVIEWER);
        user.addRole(Roles.EXAMINER);
        user.setUnit(u);
        user.setResearchAreas(researchAreas);
        user.setLanguages(languages);
        createBeta(user);
        return user;
    }

    private User createUser(String firstName, String lastName) {
        User user = save(User.builder().firstName(firstName).lastName(lastName).emailAddress(firstName + MAIL).build());
        addUserName(user);
        user = addPassword(user);
        return user;
    }

    private void createAdmin() {
        User admin = createUser(ADMIN, ADMIN);
        admin.addRole(Roles.SYSADMIN);
        createBeta(admin);
        passwordService.updatePassword(admin, "aey7ru8aefei0jaW2wo9eX8EiShi0aan");
    }

    private void createBeta(User user) {
        user.addRole(Roles.BETA);
        save(user);
    }

    private void addUserName(User user) {
        Username username = new Username();
        username.setUser(user);
        username.setUsername(user.getEmailAddress());
        save(username);
    }

    private User addPassword(User user) {
        char[] passwordInput = user.getFirstName().toLowerCase().toCharArray();
        Password password = new Password();
        password.setSalt(PasswordHandler.generateSalt());
        password.setHash(PasswordHandler.generateHash(passwordInput, password.getSalt()));
        password.setUser(user);
        password.setDeleted(false);
        password = save(password);
        user.setPassword(password);
        return save(user);
    }

    private Unit createUnit() {
        Unit u = new Unit();
        u.setTitle("unit");
        u = save(u);
        return u;
    }

    private void createGradingCriterionTemplateIfNotDone() {
        save(getBachelorTemplate(0));
        save(getBachelorTemplate(ProjectType.HP_15));
        save(getBachelorTemplate(ProjectType.HP_30));
        save(getMaster15Template());
        save(getMaster30Template(0));
        save(getMaster30Template(ProjectType.HP_30));
    }

    private GradingReportTemplate getMaster30Template(int credits) {
        GradingReportTemplate gradingReportTemplateMaster30 = new GradingReportTemplate(masterClass, credits);

        List<GradingCriterionPointTemplate> gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the abstract of the thesis adequately reflects its contents by describing the problem, the research question, the choice and application of the research methods, the result, and conclusions and that it can be read and understood separately from the thesis.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U1 Abstract", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis introduces the subject and problem of the thesis.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U2 Introduction", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that a problem of general interest exists, that entirely or partially can be solved by answering the research question.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U3 Problem", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: a clearly formulated and delimited research question that is of general interest. The student shall independently identify and formulate the research question. The research question should derive from the presented problem so that it is clear how the answer to the research question solves a portion or the entire problem.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: an innovative research question that provides necessary conditions so that the thesis could provide a significant scientific contribution.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U4 Research question", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis provides a base for the topic of the thesis based on previous scientific research. The area within computer and system sciences to which the thesis contributes should also be named by presenting the scientific research to which the thesis refers.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that a deep and critical discussion is made about how the thesis builds upon previous scientific research.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that a comprehensive literature study is presented that is the basis for placing and evaluating the research contribution of the thesis in a scientific context.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U5 Scientific base", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the choice of a research strategy and research methods is clearly described and justified based on the research question, that at least one alternative research strategy and method that could be used to solve the research question are discussed, as well as that relevant ethical considerations are discussed.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that alternative, applicable research strategies and methods are comprehensively discussed and that a profound reasoning about the chosen strategies and methods is made, where the motives for choices made are clearly evident.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that the choice of method is discussed in relation to the research strategies and methods that are used in current, related research studies that can be regarded as state-of-the-art.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U6 Choice of research method", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the application of the chosen scientific strategies and methods are clearly described, that the use of software tools is described, and that relevant ethical aspects are discussed.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the application of research strategies and methods are done in accordance to the demands of said methods and strategies and that a clear argumentation exists for this.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that there is a real depth to the data analysis.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U7 Application of research method", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the results are of such a magnitude and quality and are presented in such a way that the research question can to some extent be answered.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the results are of sufficient magnitude and high quality so that the research question can to a great extent be answered.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that the results are well written and of considerable magnitude and high quality so that well grounded conclusions relevant for the research question can be made.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U8 Result", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the research question has a clear answer, that the limitations in the composition of the thesis and their impact on the conclusions are discussed as well as how the results relate to previous research, that possible future research based on the study in the thesis is discussed, and that ethical and societal consequences of the conclusions in the thesis are discussed.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the limitations of the study are thoroughly discussed and that a profound reasoning about possible and relevant future studies is made.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U9 Conclusions and discussion", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis is divided into distinct and logical, coherent sections and fulfills the fundamental layout requirements and that the text is written with an adequate and professional language").build());
        gradingReportTemplateMaster30.addProjectCriterion("U10 Form, structure and language", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the argumentation is well grounded, logically coherent, concise, clear and easily understood.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U11 Argumentation", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that references in the thesis to previous research are made in a correct way according to a recognized reference system, that a clear listing of used references is made in the same system, that all quotes from previous work are clearly specified, and that relevant supplemental attachments are included.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U12 References and documentation", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis contributes with significant or original research contributions, in the form of new ideas, artifacts, products, or services.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the research contributions are of such quality that the thesis could be presented in a scientific workshop of good quality.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that the research contributions are of such quality that the thesis could be presented at a scientific conference of good quality or that they could be a basis for useful solutions, for example commercializable products.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(4).description("is also required: that the research contributions are of such quality that the thesis could be presented in a scientific journal of good quality.").build());
        gradingReportTemplateMaster30.addProjectCriterion("U13 Originality and significance", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the opposition report provides a short summary of the evaluated thesis, that it deliberates about the scientific basis, originality, significance, and formulation of the problem and research question, as well as that it contains clear suggestions for improvements.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the opposition report thoroughly and in a well-balanced way describes from numerous aspects the strengths and weaknesses of the evaluated thesis and that it offers clear and well-motivated suggestions for improvements.").build());
        gradingReportTemplateMaster30.addIndividualCriterion("Ö1 Opposition report", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that oral presentations are of sufficient quality, that they have been given on the assigned dates, and that the ability to orally defend one’s own thesis has been shown.").build());
        gradingReportTemplateMaster30.addIndividualCriterion("Ö2 Presentations", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the ability to orally discuss and provide constructive criticism regarding others’ work has been shown in seminars and meetings.").build());
        gradingReportTemplateMaster30.addIndividualCriterion("Ö3 Participation in seminars and meetings", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the ability to prepare and deliver material and presentations on time has been demonstrated at all the necessary occasions.").build());
        gradingReportTemplateMaster30.addIndividualCriterion("Ö4 Deadlines", 0, gradingCriterionPointTemplates).setFx(false);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that only one smaller revision of the thesis is needed after the final seminar.").build());
        gradingReportTemplateMaster30.addIndividualCriterion("Ö5 Revisions after the final seminar", 0, gradingCriterionPointTemplates).setFx(false);

        return gradingReportTemplateMaster30;
    }

    private List<GradingCriterionPointTemplate> initPointTemplates() {
        List<GradingCriterionPointTemplate> gradingCriterionPointTemplates = new ArrayList<>();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(0).build());
        return gradingCriterionPointTemplates;
    }

    //Magister = Master15hp
    private GradingReportTemplate getMaster15Template() {
        GradingReportTemplate gradingReportTemplateMaster15 = new GradingReportTemplate(masterClass, 15);

        List<GradingCriterionPointTemplate> gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the abstract of the thesis adequately reflects its contents by describing the problem, the research question, the choice and application of the research methods, the result, and conclusions and that it can be read and understood separately from the thesis.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U1 Abstract", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis introduces the subject and problem of the thesis.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U2 Introduction", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that a problem of general interest exists, that entirely or partially can be solved by answering the research question.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U3 Problem", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: a clearly formulated and delimited research question that is of general interest. The student shall independently identify and formulate the research question. The research question should derive from the presented problem so that it is clear how the answer to the research question solves a portion or the entire problem.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: an innovative research question that provides necessary conditions so that the thesis could provide a significant scientific contribution.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U4 Research question", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis provides a base for the topic of the thesis based on previous scientific research. The area within computer and system sciences to which the thesis contributes should also be named by presenting the scientific research to which the thesis refers.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that a deep and critical discussion is made about how the thesis builds upon previous scientific research.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U5 Scientific base", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the choice of a research strategy and research methods is clearly described and justified based on the research question, that at least one alternative research strategy and method that could be used to solve the research question are discussed, as well as that relevant ethical considerations are discussed.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that alternative, applicable research strategies and methods are comprehensively discussed and that a profound reasoning about the chosen strategies and methods is made, where the motives for choices made are clearly evident.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U6 Choice of research method", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the application of the chosen scientific strategies and methods are clearly described, that the use of software tools is described, and that relevant ethical aspects are discussed.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the application of research strategies and methods are done in accordance to the demands of said methods and strategies and that a clear argumentation exists for this.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U7 Application of research method", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the results are of such a magnitude and quality and are presented in such a way that the research question can to some extent be answered.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the results are of sufficient magnitude and high quality so that the research question can to a great extent be answered.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that the results are well written and of considerable magnitude and high quality so that well grounded conclusions relevant for the research question can be made.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U8 Result", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the research question has a clear answer, that the limitations in the composition of the thesis and their impact on the conclusions are discussed as well as how the results relate to previous research, that possible future research based on the study in the thesis is discussed, and that ethical and societal consequences of the conclusions in the thesis are discussed.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the limitations of the study are thoroughly discussed and that a profound reasoning about possible and relevant future studies is made.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U9 Conclusions and discussion", 2, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis is divided into distinct and logical, coherent sections and fulfills the fundamental layout requirements and that the text is written with an adequate and professional language.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U10 Form, structure and language", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the argumentation is well grounded, logically coherent, concise, clear and easily understood.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U11 Argumentation", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that references in the thesis to previous research are made in a correct way according to a recognized reference system, that a clear listing of used references is made in the same system, that all quotes from previous work are clearly specified, and that relevant supplemental attachments are included.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U12 References and documentation", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the thesis contributes with significant or original research contributions, in the form of new ideas, artifacts, products, or services.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the research contributions are of such quality that the thesis could be presented in a scientific workshop of good quality.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("is also required: that the research contributions are of such quality that the thesis could be presented at a scientific conference of good quality or that they could be a basis for useful solutions, for example commercializable products.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(4).description("is also required: that the research contributions are of such quality that the thesis could be presented in a scientific journal of good quality.").build());
        gradingReportTemplateMaster15.addProjectCriterion("U13 Originality and significance", 1, gradingCriterionPointTemplates);


        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the opposition report provides a short summary of the evaluated thesis, that it deliberates about the scientific basis, originality, significance, and formulation of the problem and research question, as well as that it contains clear suggestions for improvements.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("is also required: that the opposition report thoroughly and in a well-balanced way describes from numerous aspects the strengths and weaknesses of the evaluated thesis and that it offers clear and well-motivated suggestions for improvements.").build());
        gradingReportTemplateMaster15.addIndividualCriterion("Ö1 Opposition report", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that oral presentations are of sufficient quality, that they have been given on the assigned dates, and that the ability to orally defend one’s own thesis has been shown.").build());
        gradingReportTemplateMaster15.addIndividualCriterion("Ö2 Presentations", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the ability to orally discuss and provide constructive criticism regarding others’ work has been shown in seminars and meetings.").build());
        gradingReportTemplateMaster15.addIndividualCriterion("Ö3 Participation in seminars and meetings", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that the ability to prepare and deliver material and presentations on time has been demonstrated at all the necessary occasions.").build());
        gradingReportTemplateMaster15.addIndividualCriterion("Ö4 Deadlines", 0, gradingCriterionPointTemplates).setFx(false);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("is required: that only one smaller revision of the thesis is needed after the final seminar.").build());
        gradingReportTemplateMaster15.addIndividualCriterion("Ö5 Revisions after the final seminar", 0, gradingCriterionPointTemplates).setFx(false);
        return gradingReportTemplateMaster15;
    }

    private GradingReportTemplate getBachelorTemplate(int credits) {
        GradingReportTemplate gradingReportTemplate = new GradingReportTemplate(bachelorClass, credits);

        List<GradingCriterionPointTemplate> gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att uppsatsens sammanfattning korrekt återspeglar uppsatsens innehåll genom att beskriva problem, frågeställning, metodval, metodtillämpning, resultat och slutsatser samt att den kan läsas och förstås fristående från uppsatsen.").build());
        gradingReportTemplate.addProjectCriterion("U1 Sammanfattning", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att uppsatsen ger en introduktion till uppsatsens ämne och problem.").build());
        gradingReportTemplate.addProjectCriterion("U2 Introduktion", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att det finns ett problem av generellt intresse som helt eller delvis kan lösas genom att frågeställningen besvaras.").build());
        gradingReportTemplate.addProjectCriterion("U3 Problem", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: en tydligt formulerad och väl avgränsad frågeställning som är av generellt intresse. Studenten skall självständigt identifiera och formulera frågeställningen. Frågeställningen skall motiveras utifrån problemet så att det är tydligt hur svaret på frågan löser en del av eller hela problemet.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: en innovativ frågeställning som ger förutsättningar för att arbetet skall kunna ge ett signifikant bidrag.").build());
        gradingReportTemplate.addProjectCriterion("U4 Frågeställning", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att uppsatsen ger en ämnesmässig förankring utifrån tidigare vetenskapliga arbeten. Det skall också framgå till vilket område inom data- och systemvetenskap som arbetet bidrar genom en redovisning av de vetenskapliga arbeten som uppsatsen relaterar till.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att en djupgående och kritisk diskussion förs om hur arbetet bygger vidare på tidigare vetenskapliga arbeten.").build());
        gradingReportTemplate.addProjectCriterion("U5 Vetenskaplig förankring", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att valet av forskningsstrategier och forskningsmetoder tydligt beskrivs och motiveras utifrån forskningsfrågan, att minst en alternativ forskningsstrategi och forskningsmetod som kan användas för att angripa frågeställningen diskuteras, samt att relevanta etiska överväganden diskuteras.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att alternativa tillämpliga forskningsstrategier och forskningsmetoder diskuteras utförligt och att ett djupgående resonemang kring strategi- och metodval förs, där motiven för gjorda val tydligt framgår.").build());
        gradingReportTemplate.addProjectCriterion("U6 Metodval", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att tillämpningen av valda forskningsstrategier och forskningsmetoder är tydligt beskriven, att användning av programvaruverktyg beskrivs, samt att relevanta etiska aspekter diskuteras.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att tillämpningen av forskningsstrategier och forskningsmetoder är genomförd i enlighet med de krav som dessa ställer och att det finns en tydlig argumentation för detta.").build());
        gradingReportTemplate.addProjectCriterion("U7 Metodtillämpning", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att resultaten är av sådan omfattning och kvalitet och presenteras på ett sådant sätt att frågeställningen till viss del kan besvaras.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att resultaten är av tillräcklig omfattning och av hög kvalitet, så att frågeställningen till stor del kan besvaras.").build());
        gradingReportTemplate.addProjectCriterion("U8 Resultat", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att frågeställningen ges ett tydligt svar, att begränsningar i studiens upplägg och deras påverkan på slutsatserna diskuteras liksom hur resultaten relaterar till tidigare studier, att möjliga framtida studier med utgångspunkt från den aktuella studien diskuteras, samt att etiska och samhälleliga konsekvenser av studiens slutsatser diskuteras.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att studiens begränsningar diskuteras utförligt och att ett djupgående resonemang om möjliga och relevanta framtida studier förs.").build());
        gradingReportTemplate.addProjectCriterion("U9 Slutsatser och diskussion", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att uppsatsen är indelad i tydliga och väl sammanhållna avsnitt och uppfyller grundläggande krav på layout samt att texten är skriven med ett adekvat och professionellt språkbruk.").build());
        gradingReportTemplate.addProjectCriterion("U10 Struktur och språk", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att argumentationen är välgrundad, logiskt sammanhållen, koncis, tydlig och lättbegriplig.").build());
        gradingReportTemplate.addProjectCriterion("U11 Argumentation", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att tidigare arbeten refereras till på ett korrekt sätt enligt ett vedertaget referenssystem, att en tydlig förteckning över använda källor anges enligt samma system, att samtliga citat från tidigare arbeten anges tydligt, samt att relevanta bilagor inkluderas.").build());
        gradingReportTemplate.addProjectCriterion("U12 Källhänvisningar och dokumentation", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att arbetet ger originella och signifikanta kunskapsbidrag, till exempel i form av idéer, artefakter, produkter eller tjänster.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att kunskapsbidragen är av sådan kvalitet att arbetet skulle kunna presenteras på en vetenskaplig workshop av god kvalitet.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(3).description("krävs dessutom: att kunskapsbidragen är av sådan kvalitet att arbetet skulle kunna presenteras på en vetenskaplig konferens av god kvalitet eller kunna ligga till grund för användbara lösningar, till exempel i form av kommersialiserbara produkter.").build());
        gradingReportTemplate.addProjectCriterion("U13 Originalitet och signifikans", 0, gradingCriterionPointTemplates).setFx(false);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att oppositionsrapporten ger en kort sammanfattning av det utvärderade arbetet, resonerar kring uppsatsens vetenskapliga förankring, originalitet, signifikans, formulering av problem och frågeställning, metodval och metodtillämpning, samt innehåller tydliga förslag till förbättringar.").build());
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(2).description("krävs dessutom: att oppositionsrapporten ingående och välbalanserat beskriver styrkor och svagheter hos det utvärderade arbetet ur flera aspekter samt att den innehåller tydliga och välmotiverade förslag till förbättringar.").build());
        gradingReportTemplate.addIndividualCriterion("Ö1 Oppositionsrapport", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att muntliga presentationer av tillräcklig kvalitet har hållits vid anvisade tillfällen samt att förmåga att muntligt försvara det egna arbetet har uppvisats.").build());
        gradingReportTemplate.addIndividualCriterion("Ö2 Presentationer", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att förmåga att muntligt diskutera och ge konstruktiv kritik när det gäller andras arbete har uppvisats vid seminarier och möten.").build());
        gradingReportTemplate.addIndividualCriterion("Ö3 Aktivitet vid seminarer och möten", 1, gradingCriterionPointTemplates);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att förmåga har uppvisats att i tid förbereda och leverera material och presentationer vid alla tillfällen som detta krävs.").build());
        gradingReportTemplate.addIndividualCriterion("Ö4 Deadlines", 0, gradingCriterionPointTemplates).setFx(false);

        gradingCriterionPointTemplates = initPointTemplates();
        gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(1).description("krävs: att endast en mindre revision av uppsatsen krävs efter slutseminariet.").build());
        gradingReportTemplate.addIndividualCriterion("Ö5 Revision efter slutseminarium", 0, gradingCriterionPointTemplates).setFx(false);
        return gradingReportTemplate;
    }

    private void createDefaultProjectTypesIfNotDone() {
        bachelorClass = new ProjectType(ProjectType.BACHELOR, ProjectType.BACHELOR_STRING, "Bachelor degree thesis project");
        bachelorClass = save(bachelorClass);

        masterClass = new ProjectType(ProjectType.MASTER, ProjectType.MASTER_STRING, "Master degree thesis project");
        save(masterClass);

        final ProjectType phdClass = new ProjectType(ProjectType.PHD, ProjectType.PHD_STRING, "PhD project");
        save(phdClass);
    }

    private void createDefaultChecklistCategoriesIfNotDone() {
        final ChecklistCategory peer = new ChecklistCategory("Peer");
        save(peer);
    }

    private void createDefaultLanguagesIfNotDone() {
        Language english = new Language();
        save(english);
        languages = Collections.singleton(english);
    }

    private void createMilestonesIfNotDone() {
        MilestonePhaseTemplate milestonePhaseTemplate1 = createMileStonePhase("Phase 1 - Preparations", "This is phase 1.");
        MilestonePhaseTemplate milestonePhaseTemplate2 = createMileStonePhase("Phase 2 - Research question and method", "This is phase 2.");
        MilestonePhaseTemplate milestonePhaseTemplate3 = createMileStonePhase("Phase 3 - Result and discussion", "This is phase 3.");
        MilestonePhaseTemplate milestonePhaseTemplate4 = createMileStonePhase("Phase 4 - Final seminar", "This is phase 4.");
        MilestonePhaseTemplate milestonePhaseTemplate5 = createMileStonePhase("Phase 5 - Examination", "This is phase 5.");

        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "First meeting held", "First meeting with supervisor.", milestonePhaseTemplate1, null);
        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Project plan approved", "Project plan approved by supervisor.", milestonePhaseTemplate1, null);

        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Rough draft sent to reviewer for approval", "Rough draft approved by reviewer.", milestonePhaseTemplate2, null);
        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Rough draft approved by reviewer", "Rough draft approved.", milestonePhaseTemplate2, null);
        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "Peer review 1", "This is a recommendation of when to perform peer review 1.", milestonePhaseTemplate2, MilestoneActivityTemplate.PEER_REVIEW_ONE);

        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Result and discussion completed and approved", "Result and discussion.", milestonePhaseTemplate3, null);
        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "Peer review 2", "This is a recommendation of when to perform peer review 2.", milestonePhaseTemplate3, MilestoneActivityTemplate.PEER_REVIEW_TWO);

        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Thesis approved for final seminar presentation", "Thesis approved for final seminar.", milestonePhaseTemplate4, null);
        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Final seminar created", "Creation of final seminar.", milestonePhaseTemplate4, MilestoneActivityTemplate.CREATE_SEMINAR);
        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Final seminar thesis uploaded", "Final seminar thesis uploaded.", milestonePhaseTemplate4, MilestoneActivityTemplate.THESIS_UPLOADED);
        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "Perform an oral and written opposition", "Opposition.", milestonePhaseTemplate4, null);
        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "Active participation in a final seminar", "Active participation.", milestonePhaseTemplate4, null);
        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "Defend the thesis in a final seminar", "Defence of final thesis.", milestonePhaseTemplate4, null);

        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Revised final thesis of the submitted thesis", "Revised final thesis.", milestonePhaseTemplate5, null);
        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Originality report approved", "Originality report.", milestonePhaseTemplate5, null);
        createMileStone(MilestoneActivityTemplate.Type.PROJECT, "Supervisor and reviewer final grading report submitted", "Final grading report.", milestonePhaseTemplate5, null);
        createMileStone(MilestoneActivityTemplate.Type.STUDENT, "Grading completed", "Grading completed by examiner.", milestonePhaseTemplate5, null);
    }

    private void createMileStone(MilestoneActivityTemplate.Type type, String title, String description, MilestonePhaseTemplate milestonePhaseTemplate, String code) {
        MilestoneActivityTemplate milestoneActivityTemplate = new MilestoneActivityTemplate(type, title, description);
        milestoneActivityTemplate.addProjectType(bachelorClass);
        milestoneActivityTemplate.addProjectType(masterClass);
        milestoneActivityTemplate.setMilestonePhaseTemplate(milestonePhaseTemplate);
        milestoneActivityTemplate.setCode(code);
        milestoneActivityTemplateService.save(milestoneActivityTemplate, milestonePhaseTemplate);
    }

    private MilestonePhaseTemplate createMileStonePhase(String title, String description) {
        MilestonePhaseTemplate milestonePhaseTemplate1 = new MilestonePhaseTemplate(title, description);
        return save(milestonePhaseTemplate1);
    }

    private <T> T save(T entity) {
        em.get().persist(entity);
        return entity;
    }
}
