package se.su.dsv.scipro.finalseminar;

import com.google.common.eventbus.EventBus;
import com.google.inject.persist.Transactional;
import com.mysema.query.BooleanBuilder;
import com.mysema.query.types.expr.BooleanExpression;
import org.joda.time.*;
import org.springframework.data.domain.Pageable;
import scala.runtime.AbstractFunction0;
import scala.runtime.AbstractFunction1;
import scala.util.Either;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.misc.DaysService;
import se.su.dsv.scipro.nonworkperiod.NonWorkDayPeriodService;
import se.su.dsv.scipro.plagiarism.TurnItInCheck;
import se.su.dsv.scipro.plagiarism.TurnItInCheckRepo;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.report.OppositionReportService;
import se.su.dsv.scipro.system.AbstractServiceImpl;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.Date;
import java.util.List;

import static com.mysema.query.support.Expressions.allOf;

public abstract class FinalSeminarServiceImpl extends AbstractServiceImpl<FinalSeminar, Long> implements FinalSeminarService {

    @Inject
    FinalSeminarSettingsService finalSeminarSettingsService;
    @Inject
    DaysService daysService;
    @Inject
    FileDescriptionService fileDescriptionService;
    @Inject
    TurnItInCheckRepo turnItInCheckRepo;
    @Inject
    FinalSeminarOppositionService finalSeminarOppositionService;
    @Inject
    OppositionReportService oppositionReportService;
    @Inject
    NonWorkDayPeriodService nonWorkDayPeriodService;

    private final EventBus eventBus;

    @Inject
    public FinalSeminarServiceImpl(Provider<EntityManager> em, EventBus eventBus) {
        super(em, FinalSeminar.class, QFinalSeminar.finalSeminar);
        this.eventBus = eventBus;
    }

    @Override
    @Transactional
    public <S extends FinalSeminar> S save(S finalSeminar) {
        if (finalSeminar.getId()==null){
            S saved = super.save(finalSeminar);
            eventBus.post(new FinalSeminarCreatedEvent(saved));
            return saved;
        } else {
            return super.save(finalSeminar);
        }
    }

    @Override
    @Transactional
    public Either<SchedulingError, FinalSeminar> move(final FinalSeminar finalSeminar, final Date to) {
        final Date earliestCreationDate = daysService.workDaysAfter(new Date(),
                finalSeminarSettingsService.getInstance().getDaysAheadToCreate());

        Either<SchedulingError, FinalSeminar> move = ExtendedFinalSeminarServiceImpl.tryMove(finalSeminar, to,
                new AbstractFunction0<Date>() {
                    @Override
                    public Date apply() {
                        return earliestCreationDate;
                    }
                },
                new AbstractFunction1<Date, Object>() {
                    @Override
                    public Boolean apply(final Date date) {
                        return nonWorkDayPeriodService.isNonWorkDay(date);
                    }
                });
        if (move.isRight()) {
            FinalSeminar moved = save(move.right().get());
            eventBus.post(new FinalSeminarMovedEvent(moved, to));
        }
        return move;
    }

    @Override
    @Transactional
    public void delete(FinalSeminar seminar) {

        if (seminar.getDocument() != null) {
            TurnItInCheck turnItInCheck = turnItInCheckRepo.findByFileDescription(fileDescriptionService.findOne(seminar.getDocument().getId()));
            if (turnItInCheck != null) {
                turnItInCheckRepo.delete(turnItInCheck.getId());
            }
        }

        for (FinalSeminarOpposition opposition : seminar.getOppositions()) {
            oppositionReportService.deleteOppositionReport(opposition);
            finalSeminarOppositionService.delete(opposition.getId());
        }

        eventBus.post(new FinalSeminarDeletedEvent(seminar));
        super.delete(seminar);
    }

    @Override
    public void participate(User student, FinalSeminar finalSeminar, Project project) {
        FinalSeminarActiveParticipation participation = new FinalSeminarActiveParticipation();
        participation.setUser(student);
        participation.setFinalSeminar(finalSeminar);
        participation.setProject(project);
        finalSeminar.addActiveParticipant(participation);
        save(finalSeminar);
    }

    private boolean alreadyOpponent(User user, FinalSeminar finalSeminar) {
        for (FinalSeminarOpposition fso : finalSeminar.getOppositions()) {
            if (fso.getUser().equals(user)) {
                return true;
            }
        }
        return false;
    }

    private boolean alreadyParticipant(User user, FinalSeminar finalSeminar) {
        for (FinalSeminarActiveParticipation fsap : finalSeminar.getActiveParticipations()) {
            if (fsap.getUser().equals(user)) {
                return true;
            }
        }
        return false;
    }

    private boolean isAuthor(User user, FinalSeminar finalSeminar) {
        for (User s : finalSeminar.getProject().getProjectParticipants()) {
            if (s.equals(user)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public OppositionRegistrationStatus canOppose(User user, FinalSeminar finalSeminar, Project project) {
        if (finalSeminar.isCancelled()) {
            return new FinalSeminarCancelled();
        }
        if (!notAlreadyInSeminar(user, finalSeminar)) {
            return new AlreadyParticipating();
        }
        if (finalSeminar.getStartDate().before(new Date())) {
            return new AlreadyHappened(finalSeminar.getStartDate());
        }
        if (finalSeminar.getOppositions().size() >= finalSeminar.getMaxOpponents()) {
            return new TooManyOpponents(finalSeminar.getOppositions().size(), finalSeminar.getMaxOpponents());
        }
        FinalSeminarOppositionService.FilterParams params = new FinalSeminarOppositionService.FilterParams();
        params.setStudent(user);
        params.setProject(project);

        for (FinalSeminarOpposition opposition : finalSeminarOppositionService.findAll(params)) {
            if (opposition.getGrade() == null) {
                return new UngradedOpposition(opposition);
            }
            else if (opposition.getGrade() == FinalSeminarGrade.APPROVED) {
                return new AlreadyOpposed(opposition);
            }
        }

        int oppositionPriorityDays = finalSeminarSettingsService.getInstance().getOppositionPriorityDays();
        Period available = new Period(new Instant(finalSeminar.getDateCreated()), Instant.now());
        if (oppositionPriorityDays > available.getDays() && findByProject(project) == null) {
            return new PriorityForSeminarAuthors(oppositionPriorityDays, new DateTime(finalSeminar.getDateCreated()).plus(Days.days(oppositionPriorityDays)));
        }
        return new CanRegister();
    }

    private boolean notAlreadyInSeminar(User user, FinalSeminar finalSeminar) {
        return !alreadyOpponent(user, finalSeminar) && !isAuthor(user, finalSeminar) && !alreadyParticipant(user, finalSeminar);
    }

    @Override
    public boolean canParticipate(User user, FinalSeminar finalSeminar) {
        return finalSeminar.getActiveParticipations().size() < finalSeminar.getMaxParticipants() &&
                finalSeminar.getStartDate().after(new Date()) &&
                notAlreadyInSeminar(user, finalSeminar) &&
                !finalSeminar.isDeleted();
    }

    @Override
    public Iterable<FinalSeminar> findAll(Filter params) {
        return findAll(fromParams(params));
    }

    @Override
    public Date thesisUploadDeadline(FinalSeminar finalSeminar) {
        return daysService.workDaysAhead(finalSeminar.getStartDate(), finalSeminarSettingsService.getInstance().getDaysAheadToUploadThesis());
    }

    @Override
    public boolean hasDeadlinePassed(FinalSeminar finalSeminar) {
        return thesisUploadDeadline(finalSeminar).before(new Date());
    }

    @Override
    public boolean hasThesis(FinalSeminar finalSeminar) {
        FileDescription fileDescription = finalSeminar.getDocument();
        return fileDescription != null && fileDescription.isDataAvailable();
    }

    @Override
    public FinalSeminar findByProject(Project project) {
        return findOne(QFinalSeminar.finalSeminar.project.eq(project));
    }

    @Override
    public Iterable<FinalSeminar> findByStartDateAfter(Date date) {
        return findAll(startDateIsAfter(date));
    }

    private BooleanBuilder fromParams(Filter params) {
        BooleanBuilder bb = new BooleanBuilder();
        if (params.getHasDocument() != null && params.getHasDocument()) {
            bb.and(QFinalSeminar.finalSeminar.document.isNotNull());
        }
        if (params.getFromDate() != null) {
            bb.and(startDateIsAfter(params.getFromDate()));
        }
        if (params.getToDate() != null) {
            bb.and(startDateIsBefore(params.getToDate()));
        }
        if (params.getProjectType() != null) {
            bb.and(isProjectType(params.getProjectType()));
        }
        if (params.getReviewer() != null) {
            bb.and(hasReviewer(params.getReviewer()));
        }
        if (params.getLazyDeleted() != null) {
            bb.and(lazyDeleted(params.getLazyDeleted()));
        }
        if (params.getExempted() != null) {
            bb.and(exempted(params.getExempted()));
        }
        if (params.getOnlyActiveProjects() != null) {
            bb.and(onlyActiveProjectsFilter());
        }
        if (params.getOnlyActiveOrCompletedProjects() != null) {
            bb.and(onlyActiveOrCompletedProjectsFilter());
        }

        if (params.getHeadSupervisor() != null) {
            bb.and(hasHeadSupervisor(params.getHeadSupervisor()));
        }
        return bb;
    }


    private BooleanExpression hasHeadSupervisor(User headSupervisor) {
        return QFinalSeminar.finalSeminar.project.headSupervisor.eq(headSupervisor);
    }

    private BooleanExpression onlyActiveProjectsFilter() {
        return QFinalSeminar.finalSeminar.project.projectStatus.eq(ProjectStatus.ACTIVE);
    }

    private BooleanExpression onlyActiveOrCompletedProjectsFilter() {
        return QFinalSeminar.finalSeminar.project.projectStatus.in(ProjectStatus.ACTIVE, ProjectStatus.COMPLETED);
    }

    @Override
    public Iterable<FinalSeminar> findAll(Filter params, Pageable pageable) {
        if (params != null) {
            return findAll(fromParams(params), pageable);
        } else {
            return findAll(pageable);
        }
    }

    @Override
    public long count(Filter params) {
        if (params != null) {
            return count(fromParams(params));
        } else {
            return count();
        }
    }

    @Override
    public FinalSeminar findByFileDescription(FileDescription fileDescription) {
        return findOne(QFinalSeminar.finalSeminar.document.eq(fileDescription));
    }

    @Override
    public Iterable<FinalSeminar> findProjectOpposing(Project project, Pageable pageable) {
        return findAll(projectOpposing(project), pageable);
    }

    @Override
    public Iterable<FinalSeminar> findProjectParticipating(Project project, Pageable pageable) {
        return findAll(projectParticipating(project), pageable);
    }

    @Override
    public boolean isOpponent(FinalSeminar seminar, User opponent) {
        return alreadyOpponent(opponent, seminar);
    }

    @Override
    public boolean isActiveParticipant(FinalSeminar seminar, User activeParticipant) {
        return alreadyParticipant(activeParticipant, seminar);
    }

    @Override
    public boolean hasHadFinalSeminar(Project project) {
        FinalSeminar byProject = findByProject(project);
        return byProject != null && byProject.getStartDate().before(new Date());
    }

    @Override
    public List<FinalSeminar> findUnfinishedSeminars(Date after, Date before, Pageable pageable) {
        return findAll(unfinishedSeminars(after, before), pageable);
    }

    @Override
    public long countUnfinishedSeminars(Date after, Date before) {
        return count(unfinishedSeminars(after, before));
    }

    @Override
    @Transactional
    public FinalSeminar cancel(FinalSeminar finalSeminar) {
        finalSeminar.setCancelled(true);
        FinalSeminar cancelled = save(finalSeminar);
        eventBus.post(new FinalSeminarCancelledEvent(cancelled));
        return cancelled;
    }

    @Override
    public List<FinalSeminar> findUserOpposing(Project project, User user) {
        return findAll(oppositions(project, user));
    }

    @Override
    public List<FinalSeminar> findUserParticipating(Project project, User user) {
        return findAll(participations(project, user));
    }

    private BooleanExpression participations(Project project, User user) {
        return allOf(QFinalSeminar.finalSeminar.activeParticipations.any().project.eq(project), QFinalSeminar.finalSeminar.activeParticipations.any().user.eq(user));
    }

    private BooleanExpression oppositions(Project project, User user) {
        return allOf(QFinalSeminar.finalSeminar.oppositions.any().project.eq(project), QFinalSeminar.finalSeminar.oppositions.any().user.eq(user));
    }

    private BooleanExpression unfinishedSeminars(Date after, Date before) {
        QFinalSeminar seminar = QFinalSeminar.finalSeminar;
        if (after == null && before == null) {
            return seminar.oppositions.any().grade.isNull().or(
                            seminar.activeParticipations.any().grade.isNull().or(
                            seminar.respondents.any().grade.isNull()));
        }
        else {
            return seminar.startDate.between(after, before)
                .andAnyOf(seminar.oppositions.any().grade.isNull(),
                        seminar.activeParticipations.any().grade.isNull(),
                        seminar.respondents.any().grade.isNull());
        }
    }

    public BooleanExpression exempted(Boolean exempted) {
        return QFinalSeminar.finalSeminar.project.finalSeminarRuleExempted.eq(exempted);
    }

    public BooleanExpression lazyDeleted(Boolean deleted) {
        return QFinalSeminar.finalSeminar.deleted.eq(deleted);
    }

    public BooleanExpression projectOpposing(Project project) {
        return QFinalSeminar.finalSeminar.oppositions.any().project.eq(project);
    }

    public BooleanExpression projectParticipating(Project project) {
        return QFinalSeminar.finalSeminar.activeParticipations.any().project.eq(project);
    }

    private BooleanExpression startDateIsAfter(Date date) {
        return QFinalSeminar.finalSeminar.startDate.after(date);
    }

    private BooleanExpression startDateIsBefore(Date date) {
        return QFinalSeminar.finalSeminar.startDate.before(date);
    }

    private BooleanExpression isProjectType(ProjectType projectType) {
        return QFinalSeminar.finalSeminar.project.projectType.eq(projectType);
    }

    private BooleanExpression hasReviewer(User employee) {
        return QFinalSeminar.finalSeminar.project.reviewers.any().eq(employee);
    }

    @Override
    @Transactional
    public FinalSeminar deleteThesis(FinalSeminar seminar) {
        if (seminar.getDocument() == null) {
            return seminar;
        }

        if (turnItInCheckRepo.findByFileDescription(seminar.getDocument()) != null) {
            turnItInCheckRepo.delete(turnItInCheckRepo.findByFileDescription(seminar.getDocument()).getId());
        }

        seminar.setDocument(null);
        seminar.setDocumentUploadDate(null);

        FinalSeminar reloadedSeminar = save(seminar);

        eventBus.post(new FinalSeminarThesisDeletedEvent(reloadedSeminar));

        return reloadedSeminar;
    }
}
