package se.su.dsv.scipro.finalthesis;

import com.google.common.eventbus.EventBus;
import com.mysema.query.BooleanBuilder;
import com.mysema.query.types.Predicate;
import org.springframework.data.domain.Pageable;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.notifications.dataobject.ProjectEvent;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.AbstractServiceImpl;

import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.Date;

@Named
public class FinalThesisServiceImpl extends AbstractServiceImpl<FinalThesis, Long> implements FinalThesisService {
    protected final NotificationController notificationController;
    protected final FileDescriptionService fileDescriptionService;
    private final EventBus eventBus;

    @Inject
    public FinalThesisServiceImpl(Provider<EntityManager> em, NotificationController notificationController, FileDescriptionService fileDescriptionService, final EventBus eventBus) {
        super(em, FinalThesis.class, QFinalThesis.finalThesis);
        this.notificationController = notificationController;
        this.fileDescriptionService = fileDescriptionService;
        this.eventBus = eventBus;
    }

    @Override
    public FinalThesis upload(FileUpload fileUpload, Project project) {
        FileDescription fileDescription = storeFinalThesisFile(fileUpload);
        FinalThesis finalThesis = new FinalThesis();
        finalThesis.setFileDescription(fileDescription);
        finalThesis.setProject(project);
        finalThesis = save(finalThesis);
        notificationController.notifyProject(project, ProjectEvent.Event.FINAL_THESIS_UPLOADED, new NotificationSource());
        return finalThesis;
    }

    private FileDescription storeFinalThesisFile(FileUpload fileUpload) {
        return fileDescriptionService.storeFile(fileUpload);
    }

    @Override
    public boolean hasFinalThesis(Project project) {
        return !findAll(QFinalThesis.finalThesis.project.eq(project)).isEmpty();
    }

    @Override
    public FileDescription getFinalThesisFileDescription(Project project) {
        return hasFinalThesis(project) ? findByProject(project).getFileDescription() : null;
    }

    @Override
    public boolean isUploadAllowed(Project project) {
        return !hasFinalThesis(project) ||
                findLatestFinalThesis(project).isRejected() &&
                        !findLatestFinalThesis(project).getStatus().equals(FinalThesis.Status.APPROVED);
    }

    @Override
    public FinalThesis findByProject(Project project) {
        return hasFinalThesis(project) ? findByProject0(project) : null;
    }

    private FinalThesis findByProject0(Project project) {
        return from(QFinalThesis.finalThesis)
                .where(QFinalThesis.finalThesis.project.eq(project))
                .orderBy(QFinalThesis.finalThesis.dateCreated.desc())
                .list(QFinalThesis.finalThesis).get(0);
    }

    @Override
    public Date getRejectedDate(Project project) {
        if (!hasFinalThesis(project)) {
            throw new IllegalStateException("No final thesis");
        }
        return findLatestFinalThesis(project).getDateRejected();
    }

    @Override
    public void approve(Project project) {
        setStatus(project, FinalThesis.Status.APPROVED);
        eventBus.post(new FinalThesisApprovedEvent(project));
        notificationController.notifyProject(project, ProjectEvent.Event.FINAL_THESIS_APPROVED, new NotificationSource());
    }

    @Override
    public void reject(Project project) {
        setStatus(project, FinalThesis.Status.REJECTED);
    }

    private void setStatus(Project project, FinalThesis.Status status) {
        FinalThesis latestFinalThesis = findLatestFinalThesis(project);
        latestFinalThesis.setStatus(status);
        if (status.equals(FinalThesis.Status.APPROVED)) {
            latestFinalThesis.setDateApproved(new Date());
        }
        if (status.equals(FinalThesis.Status.REJECTED)) {
            latestFinalThesis.setDateRejected(new Date());
        }
        save(latestFinalThesis);
    }

    private FinalThesis findLatestFinalThesis(Project project) {
        return findByProject(project);
    }

    @Override
    public Iterable<FinalThesis> findAll(Filter filter, Pageable pageable) {
        return findAll(toPredicate(filter), pageable);
    }

    @Override
    public long count(Filter filter) {
        return count(toPredicate(filter));
    }

    private Predicate toPredicate(Filter filter) {
        BooleanBuilder predicate = new BooleanBuilder();
        if (filter.getStatus() != null) {
            predicate.and(QFinalThesis.finalThesis.status.eq(filter.getStatus()));
        }
        return predicate;
    }
}
