package se.su.dsv.scipro.match;

import com.google.common.base.Joiner;
import com.mysema.query.annotations.QueryInit;
import lombok.*;
import se.su.dsv.scipro.data.dataobjects.Member;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.*;

import javax.persistence.*;
import java.util.*;

@Data
@EqualsAndHashCode(callSuper = true, of = "id")
@Entity
@Table(name = "idea")
@Cacheable(true)
@ToString(of = {"id", "title", "match", "cachedStatus", "export"})
@SuppressWarnings({"PMD.UnusedPrivateField", "PMD.SingularField"})
public class Idea extends DomainObject {

    public static final int DESCRIPTION_LENGTH = 4000;
    public static final int PREREQUISITES_LENGTH = 4000;
    public static final int TITLE_LENGTH = 1024;
    @Id
    @GeneratedValue
    private Long id;

    @ManyToOne(optional = false)
    private ProjectType projectType;

    @Enumerated(EnumType.STRING)
    private Type type;

    @ManyToOne(optional = false)
    private User creator;

    @Column(nullable = true, length = DESCRIPTION_LENGTH)
    private String description;

    @Column(nullable = true, length = PREREQUISITES_LENGTH)
    private String prerequisites;

    @OneToMany(fetch = FetchType.LAZY, mappedBy = "idea", cascade = CascadeType.ALL, orphanRemoval = true)
    @QueryInit({"user", "program"})
    private Set<IdeaParticipation> ideaParticipations = new HashSet<>();

    @ManyToOne(optional = true)
    private ApplicationPeriod applicationPeriod;

    @Column(nullable = false, length = TITLE_LENGTH)
    private String title;

    @ManyToOne(optional = true)
    private ResearchArea researchArea;

    @ManyToMany
    @JoinTable(name = "idea_language", inverseJoinColumns = @JoinColumn(name = "language_id"))
    private Set<Language> languages = new HashSet<>();

    @Embedded
    private Watson watson = new Watson();

    @ManyToMany(fetch = FetchType.EAGER)
    private Set<Keyword> keywords = new HashSet<>();

    @OneToOne
    private Project project;

    @OneToOne(mappedBy = "idea", orphanRemoval = true, optional = true)
    private FirstMeeting firstMeeting;

    @OneToMany(fetch = FetchType.LAZY, mappedBy = "idea", cascade = CascadeType.ALL)
    @OrderBy("dateCreated DESC")
    private List<Match> matchHistory = new ArrayList<>();

    @OneToOne(optional = true, cascade = CascadeType.ALL)
    @QueryInit({"supervisor.user", "supervisor.unit"})
    private Match match;

    @Column(name = "published")
    private boolean published = true;

    @OneToOne(mappedBy = "idea", cascade = CascadeType.ALL)
    private IdeaExport export;

    @Enumerated(EnumType.STRING)
    @Setter(AccessLevel.NONE)
    @Getter(AccessLevel.NONE)
    private Status cachedStatus;

    @Basic
    private boolean inactive = false;

    public Idea copy(User supervisor) {
        Idea copy = new Idea();
        copy.projectType = this.projectType;
        copy.type = Type.SUPERVISOR;
        copy.creator = this.creator;
        copy.description = this.description;
        copy.prerequisites = this.prerequisites;
        copy.title = this.title;
        copy.researchArea = this.researchArea;
        copy.languages = new HashSet<>(this.languages);
        copy.keywords = new HashSet<>(this.keywords);
        copy.published = false;
        copy.inactive = this.inactive;

        if (this.watson != null) {
            copy.watson = new Watson(this.watson);
        }

        Match copyMatch = new Match();
        copyMatch.setIdea(copy);
        copyMatch.setStatus(Status.UNMATCHED);
        copyMatch.setChangedBy(supervisor);
        copyMatch.setSupervisor(supervisor);

        copy.match = copyMatch;
        copy.matchHistory.add(copyMatch);

        return copy;
    }

    public Match getMatch() {
        return match;
    }

    public Project getProject() {
        return project;
    }

    public ApplicationPeriod getApplicationPeriod() {
        return applicationPeriod;
    }

    public String getTitle(final int characters) {
        String result = title;
        if (title.length() > characters) {
            result = title.substring(0, characters) + "..";
        }
        return result;
    }

    public String getTitle() {
        return title;
    }

    public void addIdeaParticipation(IdeaParticipation ip) {
        this.ideaParticipations.add(ip);
    }

    public boolean isExported() {
        return export != null;
    }

    public boolean wasExportSuccessful() {
        return isExported() && getExport().wasSuccessful();
    }

    public Collection<IdeaParticipation> getCoAuthors() {
        Collection<IdeaParticipation> coAuthors = new HashSet<>();
        for (IdeaParticipation participation : getIdeaParticipations()) {
            if (!participation.getUser().equals(getCreator())) {
                coAuthors.add(participation);
            }
        }
        return coAuthors;
    }

    public List<Member> getMembers() {
        List<Member> members = new ArrayList<>();

        if (type == Type.STUDENT) {
            members.add(new Member(getCreator(), Member.Type.CREATOR));
        }

        for (IdeaParticipation coAuthor : getCoAuthors()) {
            members.add(new Member(coAuthor.getUser(), Member.Type.CO_AUTHOR));
        }

        if (getMatch().getSupervisor() != null) {
            members.add(new Member(getMatch().getSupervisor(), Member.Type.SUPERVISOR));
        }

        return members;
    }

    @PrePersist
    @PreUpdate
    void cacheStatus() {
        this.cachedStatus = getMatchStatus();
    }

    @PostLoad
    @PostPersist
    @PostUpdate
    void fuckHibernate() {
        if (this.watson == null) {
            this.watson = new Watson();
        }
    }

    public Status getMatchStatus() {
        if (hasProject()) {
            return Status.COMPLETED;
        }
        if (isInactive()) {
            return Status.INACTIVE;
        }
        if (isMatched()) {
            return Status.MATCHED;
        }
        return Status.UNMATCHED;
    }

    private boolean isMatched() {
        return hasSupervisor() && hasAuthors() && hasApplicationPeriod();
    }

    private boolean hasApplicationPeriod() {
        return applicationPeriod != null;
    }

    private boolean hasProject() {
        return project != null;
    }

    public boolean hasSupervisor() {
        return match != null && match.getSupervisor() != null;
    }

    private boolean hasAuthors() {
        return !ideaParticipations.isEmpty();
    }

    public int getMinAuthors() {
        return getProjectType().getMinAuthors();
    }

    public int getMaxAuthors() {
        return getProjectType().getMaxAuthors();
    }

    public Date getCourseStartDate() {
        return getApplicationPeriod().getCourseStartDate();
    }

    // Workaround for https://hibernate.atlassian.net/browse/HHH-7610
    public Watson getWatson() {
        return this.watson == null ? new Watson() : this.watson;
    }

    public String getAuthorNames() {
        List<String> authorNames = new ArrayList<>();
        for (IdeaParticipation ideaParticipation : getIdeaParticipations()) {
            authorNames.add(ideaParticipation.getUser().getFullName());
        }
        return Joiner.on(", ").join(authorNames);
    }

    public String getLanguageNames() {
        List<String> languageNames = new ArrayList<>();
        for (Language language : getLanguages()) {
            languageNames.add(language.getName());
        }
        return Joiner.on(", ").join(languageNames);
    }

    public String getKeywordNames() {
        List<String> keywordNames = new ArrayList<>();
        for (Keyword keyword: getKeywords()) {
            keywordNames.add(keyword.getKeyword());
        }
        return Joiner.on(", ").join(keywordNames);
    }

    public enum Type {
        STUDENT {
            public String toString() {
                return "Student idea";
            }
        }, SUPERVISOR {
            public String toString() {
                return "Supervisor idea";
            }
        },
    }

    public enum Status {
        UNMATCHED {
            public String toString() {
                return "Unmatched";
            }
        },
        MATCHED {
            public String toString() {
                return "Matched, no project";
            }
        },
        COMPLETED {
            public String toString() {
                return "Matched, has project";
            }
        },
        INACTIVE {
            public String toString() {
                return "Inactive";
            }
        }
    }
}
