package se.su.dsv.scipro.milestones.service.impl;

import com.mysema.query.BooleanBuilder;
import com.mysema.query.jpa.JPASubQuery;
import com.mysema.query.jpa.impl.JPAQuery;
import com.mysema.query.types.OrderSpecifier;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.Projections;
import com.mysema.query.types.expr.BooleanExpression;
import com.mysema.query.types.path.StringPath;
import lombok.Data;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.milestones.dataobjects.QMilestone;
import se.su.dsv.scipro.milestones.service.MilestoneStatisticsService;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.project.QProject;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.*;

public class MilestoneStatisticsServiceImpl implements MilestoneStatisticsService {

    private final Provider<EntityManager> entityManagerProvider;

    private Predicate toPredicate(MilestoneStatisticsService.Filter filter) {
        BooleanBuilder predicate = new BooleanBuilder();
        if (filter.getCreatedAfter() != null) {
            predicate.and(createdAfter(filter.getCreatedAfter()));
        }
        if (filter.getCreatedBefore() != null) {
            predicate.and(createdBefore(filter.getCreatedBefore()));
        }
        if (filter.getProjectTypes() != null) {
            predicate.and(hasAnyType(filter.getProjectTypes()));
        }
        if (filter.getStatuses() != null) {
            predicate.and(hasAnyStatus(filter.getStatuses()));
        }
        if (filter.getHeadSupervisor() != null) {
            predicate.and(headSupervisor(filter.getHeadSupervisor()));
        }
        return predicate;
    }

    private BooleanExpression createdAfter(Date date) {
        return QProject.project.dateCreated.after(date);
    }

    private BooleanExpression createdBefore(Date date) {
        return QProject.project.dateCreated.before(date);
    }

    private BooleanBuilder hasAnyType(Collection<ProjectType> types) {
        BooleanBuilder bb = new BooleanBuilder();
        if (types != null) {
            for (ProjectType type : types) {
                bb.or(hasType(type));
            }
        }
        return bb;
    }

    private BooleanExpression hasType(ProjectType type) {
        return QProject.project.projectType.eq(type);
    }

    private BooleanBuilder hasAnyStatus(Collection<ProjectStatus> statuses) {
        BooleanBuilder bb = new BooleanBuilder();
        if (statuses != null) {
            for (ProjectStatus status : statuses) {
                bb.or(hasStatus(status));
            }
        }
        return bb;
    }

    private BooleanExpression hasStatus(ProjectStatus status) {
        return QProject.project.projectStatus.eq(status);
    }

    private BooleanExpression headSupervisor(User user) {
        return QProject.project.headSupervisor.eq(user);
    }

    @Inject
    public MilestoneStatisticsServiceImpl(final Provider<EntityManager> entityManagerProvider) {
        this.entityManagerProvider = entityManagerProvider;
    }

    @Override
    public long countProgressedProjects(final MilestoneActivityTemplate completed, final MilestoneActivityTemplate notCompleted) {
        return progressedProjects(completed, notCompleted).count();
    }

    @Override
    public Long countNotCompletedBoth(Filter filter) {
        if (filter.getFirstMilestone() == null || filter.getSecondMilestone() == null) {
            return null;
        }
        return notCompletedBothQuery(filter).count();
    }

    public Long median(Filter filter) {

        if (filter.getFirstMilestone() == null || filter.getSecondMilestone() == null) {
            return null;
        }

        QMilestone firstMilestone = new QMilestone("firstMilestone");
        QMilestone secondMilestone = new QMilestone("secondMilestone");

        List<MyClass> dates = completedBothQuery(filter, firstMilestone, secondMilestone)
                .list(Projections.constructor(MyClass.class,
                        firstMilestone.lastModified,
                        secondMilestone.lastModified));

        List<Long> differences = new ArrayList<>();

        for (MyClass mc : dates) {
            differences.add(mc.secondDate.getTime() - mc.firstDate.getTime());
        }

        Collections.sort(differences);

        if (differences.isEmpty()) {
            return 0L;
        }

        if (differences.size() % 2 == 0) {
            long f = differences.get(differences.size() / 2);
            long s = differences.get((differences.size() / 2) - 1);
            double r = ((double) f + (double) s) / 2f;
            return Math.round(r / 1000 / 60 / 60 / 24);
        } else {
            double d = differences.get(differences.size() / 2);
            double r = d / 1000 / 60 / 60 / 24;
            return Math.round(r);
        }
    }

    @Override
    public Long countCompletedBoth(Filter filter) {
        if (filter.getFirstMilestone() == null || filter.getSecondMilestone() == null) {
            return null;
        }
        return completedBothQuery(filter, new QMilestone("firstMilestone"), new QMilestone("secondMilestone")).count();
    }

    @Data
    public static class MyClass {
        private Date firstDate;
        private Date secondDate;

        public MyClass(final Date firstDate, final Date secondDate) {
            this.firstDate = firstDate;
            this.secondDate = secondDate;
        }
    }

    private JPAQuery completedBothQuery(Filter filter,
                                        QMilestone firstMilestone,
                                        QMilestone secondMilestone) {
        return new JPAQuery(entityManagerProvider.get())
                .from(QProject.project, firstMilestone, secondMilestone)
                .where(toPredicate(filter))
                .where(firstMilestone.project.eq(QProject.project),
                        firstMilestone.activity.eq(filter.getFirstMilestone()),
                        firstMilestone.confirmed.eq(true))
                .where(secondMilestone.project.eq(QProject.project),
                        secondMilestone.activity.eq(filter.getSecondMilestone()),
                        secondMilestone.confirmed.eq(true));
    }

    private JPAQuery notCompletedBothQuery(Filter filter) {
        return new JPAQuery(entityManagerProvider.get())
                .from(QProject.project)
                .where(toPredicate(filter))
                .where((milestoneIsComplete(filter.getFirstMilestone()).notExists()).or(milestoneIsComplete(filter.getSecondMilestone()).notExists()));
    }

    @Override
    public List<Project> getProgressedProjects(final MilestoneActivityTemplate completed, final MilestoneActivityTemplate notCompleted, final Pageable pageable) {
        return progressedProjects(completed, notCompleted)
                .offset(pageable.getOffset())
                .limit(pageable.getLimit())
                .orderBy(fromSort(pageable.getSort()))
                .list(QProject.project);
    }

    private OrderSpecifier<?> fromSort(final Sort sort) {
        if (sort == null) {
            return QProject.project.title.asc();
        }
        StringPath path = new StringPath(QProject.project, sort.getProperty());
        return sort.isAscending() ? path.asc() : path.desc();
    }

    private JPAQuery progressedProjects(final MilestoneActivityTemplate completed, final MilestoneActivityTemplate notCompleted) {
        return new JPAQuery(entityManagerProvider.get())
                .from(QProject.project)
                .where(milestoneIsComplete(completed).exists(),
                        milestoneIsComplete(notCompleted).notExists());
    }

    private JPASubQuery milestoneIsComplete(final MilestoneActivityTemplate completed) {
        final QMilestone milestone = QMilestone.milestone;
        return new JPASubQuery()
                .from(milestone)
                .where(
                        milestone.project.eq(QProject.project),
                        milestone.activity.eq(completed),
                        milestone.confirmed.isTrue()
                );
    }
}
