package se.su.dsv.scipro.notifications;

import com.google.inject.ProvisionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.data.dataobjects.Member;
import se.su.dsv.scipro.finalseminar.FinalSeminar;
import se.su.dsv.scipro.group.Group;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.match.Idea;
import se.su.dsv.scipro.milestones.dataobjects.Milestone;
import se.su.dsv.scipro.notifications.dataobject.*;
import se.su.dsv.scipro.notifications.interfaces.NotificationMailFormatter;
import se.su.dsv.scipro.notifications.settings.entities.DeliveryMethod;
import se.su.dsv.scipro.notifications.settings.service.DeliveryConfigurationService;
import se.su.dsv.scipro.notifications.settings.service.ReceiverConfigurationService;
import se.su.dsv.scipro.peer.PeerRequest;
import se.su.dsv.scipro.peer.PeerReview;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.CurrentUser;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import javax.inject.Provider;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class NotificationControllerImpl implements NotificationController {

    private static final Logger LOGGER = LoggerFactory.getLogger(NotificationControllerImpl.class);
    private final NotificationService notificationService;
    private final NotificationMailFormatter notificationMailFormatter;
    private final MailEventService mailEventService;
    private final ReceiverConfigurationService receiverConfiguration;
    private final DeliveryConfigurationService deliveryConfiguration;
    private final Provider<CurrentUser> currentUserProvider;

    @Inject
    public NotificationControllerImpl(
            NotificationService notificationService,
            NotificationMailFormatter notificationMailFormatter,
            MailEventService mailEventService,
            ReceiverConfigurationService receiverConfiguration,
            DeliveryConfigurationService deliveryConfiguration, final Provider<CurrentUser> currentUserProvider) {
        this.notificationService = notificationService;
        this.notificationMailFormatter = notificationMailFormatter;
        this.mailEventService = mailEventService;
        this.receiverConfiguration = receiverConfiguration;
        this.deliveryConfiguration = deliveryConfiguration;
        this.currentUserProvider = currentUserProvider;
    }

    private Set<Notification> notify(Set<Member> recipients, NotificationEvent data) {
        Set<Notification> notifications = new HashSet<>();
        if (data.getCausedBy() == null){
            data.setCausedBy(getUser());
        }
        for (Member member : recipients) {
            User user = member.getUser();

            if (user.equals(data.getCausedBy())) {
                continue;
            }

            Notification notification = new Notification();
            notification.setUser(user);
            notification.setNotificationEvent(data);

            boolean mailDelivery;
            boolean webDelivery;

            if (member.getType() == Member.Type.SUPERVISOR) {
                mailDelivery = deliveryConfiguration.isDelivery(data.getType(), data.getEvent(), DeliveryMethod.MAIL, user);
                webDelivery = deliveryConfiguration.isDelivery(data.getType(), data.getEvent(), DeliveryMethod.WEB, user);
            } else {
                mailDelivery = deliveryConfiguration.isDelivery(data.getType(), data.getEvent(), DeliveryMethod.MAIL);
                webDelivery = deliveryConfiguration.isDelivery(data.getType(), data.getEvent(), DeliveryMethod.WEB);
            }

            if (webDelivery) {
                notificationService.save(notification);
                notifications.add(notification);
            }

            if (mailDelivery && !deliveryConfiguration.isMailCompilation(user)) {
                MailEvent mail = new MailEvent();
                NotificationMail notificationMail = notificationMailFormatter.format(notification);

                mail.setMessageBody(notificationMail.getContent());
                mail.setSubject(notificationMail.getSubject());
                mail.getRecipients().add(user);

                mailEventService.save(mail);
            }
        }
        return notifications;
    }

    private User getUser() {
        try {
            return currentUserProvider.get().get();
        }
        catch (ProvisionException ignored) {
            LOGGER.info("We are outside a session and therefore no current user exist.", ignored);
            return null;
        }
    }

    @Override
    public void notifyIdea(Idea idea, IdeaEvent.Event event, NotificationSource source) {
        IdeaEvent data = new IdeaEvent();
        data.setIdea(idea);
        data.setEvent(event);
        data.setSource(source.getMessage());
        data.setAdditionalSource(source.getAdditionalMessage());

        Set<Member> recipients = getRecipients(Notification.Type.IDEA, event, idea.getMembers(), source);

        notify(recipients, data);
    }

    @Override
    public void notifyMileStone(Milestone milestone, MileStoneEvent.Event event, NotificationSource source) {
        MileStoneEvent mileStoneEvent = new MileStoneEvent();
        mileStoneEvent.setSource(source.getMessage());
        mileStoneEvent.setMilestone(milestone);
        mileStoneEvent.setEvent(event);

        Set<Member> recipients = getRecipients(Notification.Type.MILESTONE, event, milestone.getMembers(), source);

        notify(recipients, mileStoneEvent);
    }

    @Override
    public Set<Notification> notifyGroup(Group group, GroupEvent.Event event, NotificationSource source) {
        return notifyGroup(group, event, source, null);
    }

    @Override
    public Set<Notification> notifyGroup(Group group, GroupEvent.Event event, NotificationSource source, User user) {
        GroupEvent data = new GroupEvent();
        data.setGroup(group);
        data.setEvent(event);
        data.setSource(source.getMessage());
        data.setAdditionalSource(source.getAdditionalMessage());
        data.setCausedBy(user);
        Set<Member> recipients = getRecipients(Notification.Type.GROUP, event, group.getMembers(), source);
        return notify(recipients, data);
    }

    @Override
    public void notifyProject(Project project, ProjectEvent.Event event, NotificationSource source) {
        ProjectEvent data = new ProjectEvent();
        data.setProject(project);
        data.setEvent(event);
        data.setSource(source.getMessage());
        data.setAdditionalSource(source.getAdditionalMessage());

        Set<Member> recipients = getRecipients(Notification.Type.PROJECT, event, project.getMembers(), source);

        notify(recipients, data);
    }

    @Override
    public Set<Notification> notifyProjectForum(ProjectForumEvent.Event event, NotificationSource source, Project project) {
        return notifyProjectForum(event, source, project, null);
    }

    @Override
    public Set<Notification> notifyProjectForum(ProjectForumEvent.Event event, NotificationSource source, Project project, User user) {
        ProjectForumEvent data = new ProjectForumEvent();
        data.setProject(project);
        data.setEvent(event);
        data.setSource(source.getMessage());
        data.setAdditionalSource(source.getAdditionalMessage());
        data.setCausedBy(user);
        Set<Member> recipients = getRecipients(Notification.Type.FORUM, event, project.getMembers(), source);
        return notify(recipients, data);
    }

    @Override
    public void notifySeminar(FinalSeminar seminar, SeminarEvent.Event event, NotificationSource source) {
        SeminarEvent data = new SeminarEvent();
        data.setSeminar(seminar);
        data.setEvent(event);
        data.setSource(source.getMessage());

        Set<Member> recipients;

            recipients = getRecipients(Notification.Type.FINAL_SEMINAR, event, seminar.getMembers(), source);
            notify(recipients, data);
    }

    @Override
    public void notifyPeer(PeerReview review, PeerEvent.Event event, NotificationSource source) {
        PeerEvent data = new PeerEvent();
        data.setReview(review);
        data.setEvent(event);
        data.setSource(source.getMessage());

        Set<Member> recipients = getRecipients(Notification.Type.PEER, event, review.getMembers(), source);

        notify(recipients, data);
    }

    @Override
    public void notifyPeer(PeerRequest request, PeerEvent.Event event, NotificationSource source) {
        PeerRequestEvent data = new PeerRequestEvent();
        data.setRequest(request);
        data.setEvent(event);
        data.setSource(source.getMessage());

        Set<Member> recipients = getRecipients(Notification.Type.PEER, event, request.getMembers(), source);

        notify(recipients, data);
    }

    private Set<Member> getRecipients(Notification.Type group, Enum event, List<Member> members, NotificationSource source) {
        Set<Member> recipients = new HashSet<>();
        for (Member member : members) {
            if (receiverConfiguration.isReceiving(group, event, member.getType())) {
                recipients.add(member);
            }
        }

        for (Member member : source.getAdditionalRecipients()) {
            if (receiverConfiguration.isReceiving(group, event, member.getType())) {
                recipients.add(member);
            }
        }
        return recipients;
    }
}
