package se.su.dsv.scipro.notifications.interfaces.impl;

import com.google.common.base.Joiner;
import se.su.dsv.scipro.notifications.NotificationModule;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.date.DateService;
import se.su.dsv.scipro.finalseminar.FinalSeminar;
import se.su.dsv.scipro.finalseminar.FinalSeminarService;
import se.su.dsv.scipro.finalseminar.FinalSeminarSettingsService;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.notifications.dataobject.*;
import se.su.dsv.scipro.notifications.interfaces.NotificationMailFormatter;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.util.PropsUtils;

import javax.inject.Inject;
import javax.inject.Provider;
import java.text.MessageFormat;
import java.util.*;

public class NotificationMailFormatterImpl implements NotificationMailFormatter {

    private static final Logger LOGGER = LoggerFactory.getLogger(NotificationMailFormatterImpl.class);

    @Inject
    GeneralSystemSettingsService systemSettingsService;
    @Inject
    FinalSeminarSettingsService service;
    @Inject
    DateService dateService;

    /**
     * Using a provider is required because of circular Spring dependencies between NotificationMailFormatter >
     * FinalSeminarService > NotificationController > back to NotificationMailFormatter. Of course, refactoring out the
     * circular dependencies would be preferable but I frankly have no idea how to go about that. It seems to me that
     * avoiding circular references sometimes just isnt possible.
     * <p/>
     * Allegedly, Services, Controllers etc should not contain dependencies on any other Services, but only on their own
     * repo, and any methods that require more than one service should be moved out to Facades. I dont see how this
     * solves the problem though because that just moves it up one level and there will be circular references in that
     * layer as well.
     */
    @Inject
    Provider<FinalSeminarService> finalSeminarService;

    protected String getFinalSeminarDeadline(FinalSeminar finalSeminar) {
        return dateService.format(finalSeminarService.get().thesisUploadDeadline(finalSeminar), DateStyle.DATETIME);
    }

    protected String getFinalSeminarUploadedOnString(FinalSeminar finalSeminar) {
        Date date = finalSeminar.getDocumentUploadDate();
        if (date != null) {
            return dateService.format(date, DateStyle.DATETIME);
        } else {
            return "no final seminar thesis";
        }
    }

    private Map<Class<? extends NotificationEvent>, EventFormatter<? extends NotificationEvent>> formatters =
            new HashMap<>();

    private Properties properties;

    public NotificationMailFormatterImpl() {
        Properties webProperties = null;
        try {
            webProperties = PropsUtils.load("se/su/dsv/scipro/notifications/notifications.properties");
        } catch (Exception ex) {
            LOGGER.warn("Could not read web notification properties", ex);
        }

        properties = new Properties(webProperties);
        try {
            PropsUtils.load("se/su/dsv/scipro/notifications/MailFormatter.properties", properties);
        } catch (Exception ex) {
            LOGGER.error("Could not read mail formatting properties", ex);
        }

        mapFormatters();
    }

    private void mapFormatters() {
        formatters.put(ProjectEvent.class, new ProjectEventFormatter(this));
        formatters.put(ProjectForumEvent.class, new ProjectForumEventFormatter(this));
        formatters.put(SeminarEvent.class, new SeminarEventFormatter(this));
        formatters.put(PeerEvent.class, new PeerEventFormatter(this));
        formatters.put(PeerRequestEvent.class, new PeerRequestFormatter(this));
        formatters.put(IdeaEvent.class, new IdeaEventFormatter(this));
        formatters.put(MileStoneEvent.class, new MileStoneEventFormatter(this));
        formatters.put(GroupEvent.class, new GroupEventFormatter(this));
    }

    @Override
    public NotificationMail format(Notification notification) {
        String title = notification.getTitle();
        if (title.length() > TITLE_TRUNCATION_LENGTH) {
            title = title.substring(0, TITLE_TRUNCATION_LENGTH) + "...";
        }
        if (title.length() == 0) {
            Project project = notification.getNotificationEvent().getProject();
            title = (project != null) ? project.getTitle() : "";
        }

        String subjectProperty = makeProperty(notification.getType().name(), notification.getEvent().name(), "title");
        String subject = MessageFormat.format(getProperty(subjectProperty), title, notification.getSource(), notification.getAdditionalSource());

        String bodyProperty = makeProperty(notification.getType().name(), notification.getEvent().name(), "body");
        String body = MessageFormat.format(getProperty(bodyProperty), notification.getSource(), getAbsoluteURL(notification), notification.getAdditionalSource());

        String details = "";
        if (notification.getNotificationEvent().hasEntity()) {
            try {
                Class<? extends NotificationEvent> eventClass = notification.getNotificationEvent().getClass();
                @SuppressWarnings("unchecked")
                EventFormatter<NotificationEvent> formatter = (EventFormatter<NotificationEvent>) formatters.get(eventClass);

                details = formatter.format(notification.getNotificationEvent());
            } catch (Exception ex) {
                details = "";
                LOGGER.error("Could not format event of type: " + notification.getNotificationEvent().getClass(), ex);
            }
        }

        StringBuilder content = new StringBuilder();
        content.append(MessageFormat.format(getProperty("single.body"), body, details));

        addAffixes(content);
        return new NotificationMail(subject, content.toString());
    }

    @Override
    public NotificationMail compile(Collection<Notification> notifications) {
        String subject = MessageFormat.format(getProperty("compilation.subject"), notifications.size());

        StringBuilder contents = new StringBuilder();

        contents.append(MessageFormat.format(getProperty("compilation.body"), notifications.size()));

        contents.append(getProperty("compilation.list.begin"));
        MessageFormat format = new MessageFormat(getProperty("compilation.list.each"));
        for (Notification notification : notifications) {
            String titleProperty = makeProperty(notification.getType(), notification.getEvent(), "title");
            String suffixProperty = makeProperty(notification.getType(), "compilationSuffix");
            String causedByName = notification.getCausedBy() != null ? "by " + notification.getCausedBy().getFullName() : "";
            contents.append(format.format(new Object[]{
                    MessageFormat.format(getProperty(titleProperty), notification.getTitle(), notification.getSource(), notification.getAdditionalSource()),
                    causedByName, MessageFormat.format(getProperty(suffixProperty), notification.getTitle())
            }));
        }
        contents.append(getProperty("compilation.list.end"));

        addAffixes(contents);
        return new NotificationMail(subject, contents.toString());
    }

    private String getAbsoluteURL(final Notification notification) {
        final String baseUrl = systemSettingsService.getGeneralSystemSettingsInstance().getSciproURL();
        long id = notification.getNotificationEvent().getId();
        return baseUrl + "/" + NotificationModule.NOTIFICATION_RELATIVE_PAGE_URL + "?id=" + id;
    }

    String makeProperty(Object... parts) {
        return Joiner.on('.').join(parts);
    }

    String getProperty(String property) {
        return properties.getProperty(property, "No property with key: " + property);
    }

    String getDateString(Date date) {
        return getDateString(date, DateStyle.DATE);
    }

    String getDateString(Date date, DateStyle style) {
        return dateService.format(date, style);
    }

    private void addAffixes(StringBuilder builder) {
        GeneralSystemSettings systemSettings = systemSettingsService.getGeneralSystemSettingsInstance();

        String baseUrl = systemSettings.getSciproURL();
        baseUrl = baseUrl + (!baseUrl.endsWith("/") ? "/" : "") + "notifications";

        builder.append(MessageFormat.format(getProperty("prefix"), baseUrl));
        builder.append(MessageFormat.format(getProperty("suffix"), baseUrl));
    }
}
