package se.su.dsv.scipro.peer;

import com.google.common.eventbus.EventBus;
import com.google.inject.persist.Transactional;
import org.joda.time.DateTime;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Option;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.ProjectTypeSettings;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;

import static com.mysema.query.types.ExpressionUtils.allOf;

public class PeerPortalImpl implements PeerPortal, PerformReviewService {

    private static final Logger LOGGER = LoggerFactory.getLogger(PeerPortalImpl.class);

    private final FileDescriptionService fileDescriptionService;
    private final PeerReviewRepository peerReviewRepository;
    private final PeerRequestRepository peerRequestRepository;
    private final EventBus eventBus;

    @Inject
    public PeerPortalImpl(FileDescriptionService fileDescriptionService,
                          PeerReviewRepository peerReviewRepository,
                          PeerRequestRepository peerRequestRepository,
                          EventBus eventBus) {
        this.fileDescriptionService = fileDescriptionService;
        this.peerReviewRepository = peerReviewRepository;
        this.peerRequestRepository = peerRequestRepository;
        this.eventBus = eventBus;
    }

    @Override
    @Transactional
    public PeerReview acceptReview(PeerRequest peerRequest, User student, Project project)
            throws Exception {
        if(peerRequest.getProject().isParticipant(student)) {
            throw new PartOfProjectException();
        }
        if(recentlyReviewedProject(peerRequest.getProject(), student, project)) {
            throw new RecentlyReviewedException();
        }
        if (peerRequest.getStatus() != RequestStatus.WAITING) {
            throw new RequestTakenException();
        }

        PeerReview review = new PeerReview();

        try {
            if (peerRequest.getChecklistTemplate() != null) {
                for (String question : peerRequest.getChecklistTemplate().getQuestions()) {
                    review.addAnswer(question);
                }
            }
            review.setPeerRequest(peerRequest);
            review.setProject(project);
            review.setReviewer(student);
            review = peerReviewRepository.save(review);

            peerRequest.setStatus(RequestStatus.TAKEN);
            peerRequestRepository.save(peerRequest);

            return review;
        } catch (Exception e) {
            LOGGER.error("Error while accepting peer request: " + e.getMessage());
            if (review.getId() != null) {
                review = peerReviewRepository.findOne(review.getId());
                peerReviewRepository.delete(review.getId());
            }
            if (peerRequest.getStatus() != RequestStatus.WAITING) {
                peerRequest.setStatus(RequestStatus.WAITING);
                peerRequestRepository.save(peerRequest);
            }
            throw e;
        }
    }

    private boolean recentlyReviewedProject(Project requestingProject, User student, Project activeProject) {
        ProjectTypeSettings settings = requestingProject.getProjectType().getProjectTypeSettings();
        int daysBetweenReviewsOnSameProject = settings.getNumDaysBetweenPeerReviewsOnSameProject();

        DateTime earliestPermissibleDateSinceReview = new DateTime();
        earliestPermissibleDateSinceReview = earliestPermissibleDateSinceReview
                .minusDays(daysBetweenReviewsOnSameProject);

        QPeerReview qReview = QPeerReview.peerReview;
        Iterable<PeerReview> performedReviews = peerReviewRepository.findAll(
                qReview.reviewer.eq(student)
                        .and(qReview.project.eq(activeProject)
                                .and(qReview.status.eq(PeerReview.ReviewStatus.EXPIRED).not())));
        for (PeerReview pr : performedReviews) {
            boolean sameProject = pr.getPeerRequest().getProject().equals(requestingProject);
            if (sameProject) {
                boolean reviewPerformedBeforeLimit = pr.getLastModified().compareTo(
                        earliestPermissibleDateSinceReview.toDate()) > 0;
                if (reviewPerformedBeforeLimit) {
                    return true;
                }
            }
        }
        return false;
    }


    @Transactional
    public PeerRequest storePeerRequest(FileUpload fileUpload, PeerRequest request) {
        PeerRequest peerRequest = peerRequestRepository.save(request);

        FileDescription fileDescription = fileDescriptionService.storeFile(fileUpload);
        return storePeerRequest(fileDescription, peerRequest);
    }

    @Override
    public Option<PeerReview> getInProgressReview(User peerReviewer, Project project) {
        QPeerReview q = QPeerReview.peerReview;
        PeerReview peerReview = peerReviewRepository.findOne(allOf(
                q.reviewer.eq(peerReviewer),
                q.status.eq(PeerReview.ReviewStatus.IN_PROGRESS),
                q.project.eq(project)));
        return Option.apply(peerReview);
    }

    @Override
    @Transactional
    public PeerReview save(PeerReview peerReview) {
        return peerReviewRepository.save(peerReview);
    }

    @Override
    @Transactional(ignore = {TooShortCommentException.class, MissingAnswerException.class})
    public void submit(PeerReview review, Option<FileUpload> upload) {
        review.submit();
        storePeerReviewFileUpload(review, upload);
        PeerRequest request = review.getPeerRequest();
        request.setStatus(review.isLate() ? RequestStatus.WAITING : RequestStatus.FINISHED);
        peerRequestRepository.save(request);
        PeerReview peerReview = peerReviewRepository.save(review);

        if (peerReview.isCompleted()) {
            postCompletionEvent(peerReview);
        }
    }

    private void postCompletionEvent(PeerReview peerReview) {
        long completed = peerReviewRepository.count(allOf(
                QPeerReview.peerReview.reviewer.eq(peerReview.getReviewer()),
                QPeerReview.peerReview.status.eq(PeerReview.ReviewStatus.COMPLETED),
                QPeerReview.peerReview.project.eq(peerReview.getProject())));

        if (completed == 1) {
            eventBus.post(new FirstPeerReviewCompletedEvent(peerReview));
        }
        else if (completed == 2) {
            eventBus.post(new SecondPeerReviewCompletedEvent(peerReview));
        }
    }

    private PeerRequest storePeerRequest(FileDescription fileDescription, PeerRequest request) {
        request.setFile(fileDescription);
        return peerRequestRepository.save(request);
    }

    private void storePeerReviewFileUpload(PeerReview review, Option<FileUpload> fileUpload) {
        if (fileUpload.isDefined()) {
            FileDescription fileDescription = fileDescriptionService.storeFile(fileUpload.get());
            review.setFile(fileDescription);
        }
    }

}
