package se.su.dsv.scipro.reusable;

import org.joda.time.DateTime;

import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.*;

public final class SciProUtilities {

    public static final int LAST_HOUR = 23;
    public static final int LAST_MINUTE = 59;
    public static final int LAST_SECOND = 59;
    public static final int LAST_MILLISECOND = 999;

    /**
     * Suppress default constructor for noninstantiability. For details about this, see
     * http://stackoverflow.com/questions/1844355/java-static-class
     */
    private SciProUtilities() {
        throw new AssertionError();
    }

    /**
     * Some strings have control, non-printable or illegal characters in them which cause problems when displayed in
     * AJAX contexts. Project titles are a frequent culprit. This method can be used to clean such strings. E.g. It has
     * been hooked up to the getTitle() method in Project to clean the title string that we get from the db before it's
     * used.
     *
     * @param string .
     * @return .
     */
    public static String cleanString(String string) {
        if (string == null || string.isEmpty()) {
            return "";
        } else {
            String clean = string;
            clean = clean.replaceAll("\\p{C}|\\p{Cf}|\n|\t|\f|\n", "?");
            return clean;
        }
    }

    /**
     * "Safe" method for converting long to int.
     *
     * @param l .
     * @return .
     */
    public static int safeLongToInt(long l) {
        if (l < Integer.MIN_VALUE || l > Integer.MAX_VALUE) {
            throw new IllegalArgumentException
                    (l + " cannot be cast to int without changing its value.");
        }
        return (int) l;
    }

    public static Date setTimeOfDayTo0000(final Date date) {
        Calendar c = Calendar.getInstance();
        c.setTime(date);
        c.set(Calendar.HOUR_OF_DAY, 0);
        c.set(Calendar.MINUTE, 0);
        c.set(Calendar.SECOND, 0);
        c.set(Calendar.MILLISECOND, 0);
        return c.getTime();
    }

    public static Date setTimeOfDayTo2359(final Date date) {
        Calendar c = Calendar.getInstance();
        c.setTime(date);
        c.set(Calendar.HOUR_OF_DAY, LAST_HOUR);
        c.set(Calendar.MINUTE, LAST_MINUTE);
        c.set(Calendar.SECOND, LAST_SECOND);
        c.set(Calendar.MILLISECOND, LAST_MILLISECOND);
        return c.getTime();
    }

    public static <T> List<T> copyIterator(Iterator<T> iterator) {
        List<T> copy = new ArrayList<>();
        while (iterator.hasNext()) {
            copy.add(iterator.next());
        }
        return copy;
    }

    /**
     * Takes byte array of hash values and returns a string with the values displayed in hex. <p/> Hex: When e.g. a
     * string is hashed, the resultant hash is a series of numeral values. These values can be displayed using different
     * encodings. Encoding in this case refers not to character encoding such as ASCII or UTF-8 but rather positional
     * numeral systems with which to display the numeral values in the hash. One of the more common systems for
     * displaying hashes is hex, where symbols 0-9 are used for values zero to nine and symbols A-F are used for values
     * ten to fifteen. You could display the same hash in hex or some other system, and they would be displayed
     * differently but actually be the same hash values.
     */
    public static String convertToHex(byte[] data) {
        StringBuilder buf = new StringBuilder();
        for (byte aData : data) {
            int halfByte = (aData >>> 4) & 0x0F;
            int fullByte = aData & 0xFF;
            if (halfByte == 0) {
                buf.append('0');
            }
            buf.append(Integer.toHexString(fullByte));
        }
        return buf.toString();
    }

    /**
     * Takes a String and returns an md5 hash of it displayed in hex.
     */
    public static String md5(String text) {
        try {
            MessageDigest md = MessageDigest.getInstance("md5");
            md.update(text.getBytes(Charset.defaultCharset()));
            return convertToHex(md.digest());
        }
        catch (NoSuchAlgorithmException e) {
            throw new Error(e); // MD5 must exist according to the Java standard
        }
    }

    /**
     * Maybe use localdates instead? Maybe don't worry about chronology?
     */
    public static boolean isToday(Date input){
        DateTime compare = new DateTime(input);
        DateTime now = DateTime.now();
        if (!now.getChronology().equals(compare.getChronology())){
            throw new IllegalArgumentException("Input must use same chronology as whatever is running this code does");
        }
        return now.year().equals(compare.year()) &&
                now.monthOfYear().equals(compare.monthOfYear()) &&
                now.dayOfMonth().equals(compare.dayOfMonth());
    }
}
