package se.su.dsv.scipro.reviewing;

import com.google.common.eventbus.EventBus;
import com.mysema.query.BooleanBuilder;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.expr.BooleanExpression;
import org.springframework.data.domain.Pageable;
import scala.Option;
import scala.runtime.AbstractFunction1;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileDescriptionService;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.AbstractServiceImpl;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.util.Collection;

import static com.mysema.query.support.Expressions.allOf;

class ReviewingServiceImpl extends AbstractServiceImpl<ReviewerApproval, Long> implements MyReviewService, ReviewerDecisionService {

    private final FileDescriptionService fileDescriptionService;
    private final EventBus eventBus;

    @Inject
    ReviewingServiceImpl(Provider<EntityManager> em, FileDescriptionService fileDescriptionService, final EventBus eventBus) {
        super(em, ReviewerApproval.class, QReviewerApproval.reviewerApproval);
        this.fileDescriptionService = fileDescriptionService;
        this.eventBus = eventBus;
    }

    @Override
    public Iterable<ReviewerApproval> findAll(final Filter filter, final Pageable pageable) {
        return findAll(fromFilter(filter), pageable);
    }

    @Override
    public Iterable<ReviewerApproval> findAll(final Filter filter) {
        return findAll(fromFilter2(filter));
    }

    @Override
    public long count(final Filter filter) {
        return count(fromFilter(filter));
    }

    @Override
    public long countUndecidedRoughDraft(final User reviewer) {
        return count(forReviewer(reviewer).and(isUndecided()).and(forStep(ReviewerApproval.Step.ROUGH_DRAFT_APPROVAL)));
    }

    @Override
    public long countUndecidedFinalSeminar(final User reviewer) {
        return count(forReviewer(reviewer).and(isUndecided()).and(forStep(ReviewerApproval.Step.FINAL_SEMINAR_APPROVAL)));
    }

    @Override
    public ReviewerApproval findApproval(final Long id) {
        return findOne(id);
    }

    @Override
    public void reject(final ReviewerApproval reviewerApproval, final String reason, final Option<FileUpload> feedback, final Option<FileDescription> gradingReport) {
        Option<FileDescription> fileDescription = store(feedback);
        reviewerApproval.reject(reason, fileDescription, gradingReport);
        save(reviewerApproval);

        if (reviewerApproval instanceof FinalSeminarApproval) {
            eventBus.post(new FinalSeminarApprovalRejectedEvent(reviewerApproval.getProject()));
        }
        else if (reviewerApproval instanceof RoughDraftApproval) {
            eventBus.post(new RoughDraftApprovalRejectedEvent(reviewerApproval.getProject()));
        }
    }

    @Override
    public void approve(final ReviewerApproval process, final String reason, final Option<FileUpload> feedback, final Option<FileDescription> gradingReport) {
        Option<FileDescription> fileDescription = store(feedback);
        process.approve(reason, fileDescription, gradingReport);
        save(process);

        if (process instanceof FinalSeminarApproval) {
            eventBus.post(new FinalSeminarApprovalApprovedEvent(process));
        }
        else if (process instanceof RoughDraftApproval) {
            eventBus.post(new RoughDraftApprovalApprovedEvent(process));
        }
    }

    private static BooleanExpression forReviewer(final User reviewer) {
        return QReviewerApproval.reviewerApproval.project.reviewers.contains(reviewer);
    }

    private static BooleanExpression forStep(final ReviewerApproval.Step step) {
        if (step.equals(ReviewerApproval.Step.ROUGH_DRAFT_APPROVAL)) {
            return QReviewerApproval.reviewerApproval.instanceOf(RoughDraftApproval.class);
        }
        return QReviewerApproval.reviewerApproval.instanceOf(FinalSeminarApproval.class);
    }

    private static BooleanExpression forStatus(final Collection<Status> statuses) {
        return   QReviewerApproval.reviewerApproval.decisions.any().status.in(statuses);
    }

    private BooleanBuilder fromFilter(Filter filter) {
        BooleanBuilder bb = new BooleanBuilder();
        bb.and(forReviewer(filter.getUser()));
        bb.and(forStep(filter.getStep()));
        if (!filter.isShowAll()) {
            bb.and(isUndecided());
        }

        return bb;
    }

    private BooleanBuilder fromFilter2(Filter filter) {
        BooleanBuilder bb = new BooleanBuilder();
        bb.and(forStep(filter.getStep()));
        bb.and(forStatus(filter.getStatuses()));
        return bb;
    }


    private static Predicate isUndecided() {
        return allOf(
                QReviewerApproval.reviewerApproval.decisions.any().status.eq(Status.UNDECIDED),
                QReviewerApproval.reviewerApproval.project.projectStatus.eq(ProjectStatus.ACTIVE)
        );
    }

    private Option<FileDescription> store(final Option<FileUpload> feedback) {
        return feedback.map(new AbstractFunction1<FileUpload, FileDescription>() {
            @Override
            public FileDescription apply(final FileUpload input) {
                return fileDescriptionService.storeFile(input);
            }
        });
    }

}
