package se.su.dsv.scipro.system;

import com.google.inject.persist.Transactional;
import com.mysema.query.jpa.impl.JPAQuery;
import com.mysema.query.types.EntityPath;
import com.mysema.query.types.PathMetadataFactory;
import com.mysema.query.types.Predicate;
import com.mysema.query.types.path.StringPath;
import org.springframework.data.domain.Pageable;
import org.springframework.data.domain.Sort;

import javax.inject.Provider;
import javax.persistence.EntityManager;
import java.io.Serializable;
import java.util.List;

public class AbstractServiceImpl<T, ID extends Serializable> extends AbstractRepository implements GenericService<T, ID> {
    private final Class<T> clazz;
    private final EntityPath<T> path;

	protected AbstractServiceImpl(Provider<EntityManager> em, Class<T> clazz, EntityPath<T> path) {
        super(em);
        this.clazz = clazz;
        this.path = path;
    }

    protected JPAQuery createQuery() {
        return from(path);
    }

    private JPAQuery pagedQuery(Pageable pageable) {
        return applyPageable(createQuery(), pageable);
    }

    protected JPAQuery applyPageable(JPAQuery query, Pageable pageable) {
        JPAQuery pagedQuery = paged(query, pageable.getOffset(), pageable.getLimit());
        if (pageable.getSort() == null) {
            return pagedQuery;
        }
        return sorted(pagedQuery, pageable.getSort());
    }

    private JPAQuery paged(JPAQuery query, long offset, long limit) {
        return query.limit(limit).offset(offset);
    }

    protected JPAQuery sorted(JPAQuery query, Sort sort) {
        StringPath property = new StringPath(PathMetadataFactory.forProperty(path, sort.getProperty()));
        return query.orderBy(sort.isAscending() ? property.asc() : property.desc());
    }

    @Override
    public long count(Predicate predicate) {
        return createQuery().where(predicate).count();
    }

    @Override
    public T findOne(Predicate predicate) {
        return createQuery().where(predicate).uniqueResult(path);
    }

    @Override
    public List<T> findAll(Predicate predicate) {
        return createQuery().where(predicate).list(path);
    }

    @Override
    public List<T> findAll(Predicate predicate, Pageable pageable) {
        return pagedQuery(pageable).where(predicate).list(path);
    }

    @Override
    @Transactional
    public <S extends T> S save(S entity) {
        EntityManager entityManager = em();
        if (entityManager.contains(entity)) {
            return entityManager.merge(entity);
        }
        else {
            entityManager.persist(entity);
            return entity;
        }
    }

    @Override
    public T findOne(ID id) {
		return em().find(clazz, id);
	}

    @Override
    public List<T> findAll() {
        return from(path).list(path);
    }

    @Override
    public List<T> findAll(Pageable pageable) {
        return pagedQuery(pageable).list(path);
    }

    @Override
    public long count() {
		return createQuery().count();
	}

    @Override
    @Transactional
	public void delete(ID id) {
		delete(em().getReference(clazz, id));
	}

    @Override
    @Transactional
    public void delete(T entity) {
        em().remove(entity);
    }
}
