package se.su.dsv.scipro.system;

import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.ToString;
import se.su.dsv.scipro.security.auth.roles.Roles;

import javax.persistence.*;
import java.io.Serializable;
import java.util.*;

@Entity
@Table(name = "user")
@Cacheable(true)
@Data
@EqualsAndHashCode(of = {"id"}, callSuper = false)
@ToString(exclude = {"password", "usernames", "languages", "roles", "programs", "researchAreas", "unit"})
@SuppressWarnings("PMD.UnusedPrivateField")
public class User extends LazyDeletableDomainObject {

    @Id
    @GeneratedValue
    private Long id;

    @Column(unique = true)
    private Long identifier;

    @Basic(optional = false)
    private String firstName;
    @Basic(optional = false)
    private String lastName;
    @Basic(optional = false)
    private String emailAddress;

    @OneToMany(mappedBy = "user", cascade = CascadeType.REMOVE)
    private Set<Username> usernames = new HashSet<>();

    @ElementCollection
    @CollectionTable(name = "user_role")
    @Column(name = "role")
    @Enumerated(EnumType.STRING)
    private Set<Roles> roles = EnumSet.of(Roles.AUTHOR);

    @OneToOne(optional = true, cascade = CascadeType.REMOVE, mappedBy = "user")
    private Password password;

    @ManyToMany
    @JoinTable(name = "user_program", inverseJoinColumns = @JoinColumn(name = "program_id"))
    private Set<Program> programs = new HashSet<>();

    @ManyToMany
    @JoinTable(name = "user_languages", inverseJoinColumns = @JoinColumn(name = "language_id"))
    private Set<Language> languages = new HashSet<>();

    @ManyToMany
    @JoinTable(name = "user_research_area", inverseJoinColumns = @JoinColumn(name = "research_area_id"))
    private Set<ResearchArea> researchAreas = new HashSet<>();

    @OneToOne(optional = true)
    private Unit unit;

    public Unit getUnit() {
        return unit;
    }

    public Set<ResearchArea> getActiveResearchAreas() {
        Set<ResearchArea> activeAreas = new HashSet<>();
        for (ResearchArea area : researchAreas) {
            if (!area.isDeleted()) {
                activeAreas.add(area);
            }
        }
        return activeAreas;
    }

    @SuppressWarnings("UnusedDeclaration")
    public List<String> getResearchAreaNames() {
        List<String> result = new ArrayList<>();
        for (ResearchArea researchArea : getActiveResearchAreas()) {
            result.add(researchArea.getTitle());
        }
        return result;
    }

    public String getFullName() {
        return getFirstName() + " " + getLastName();
    }

    public void addRole(Roles role) {
        roles.add(role);
    }

    public void removeRole(Roles role) {
        roles.remove(role);
    }

    public boolean hasRole(Roles role) {
        return roles.contains(role);
    }

    public void addResearchArea(ResearchArea researchArea) {
        researchAreas.add(researchArea);
    }

    public void addProgram(Program program) {
        programs.add(program);
    }

    public boolean hasPrograms() {
        return !programs.isEmpty();
    }

    /**
     * Builder stuff. These top ones are mandatory.
     */

    public static IFirstName builder() {
        return new Builder();
    }

    public interface IFirstName {
        ILastName firstName(String firstName);
    }

    public interface ILastName {
        IEmailAddress lastName(String lastName);
    }

    public interface IEmailAddress {
        IBuild emailAddress(String emailAddress);
    }

    /**
     * These are the rest.
     */
    public interface IBuild {

        IBuild identifier(Long identifier);

        IBuild roles(Set<Roles> roles);

        IBuild password(Password password);

        IBuild deleted(boolean deleted);

        User build();
    }

    private static class Builder implements IFirstName, ILastName, IEmailAddress, IBuild {
        private final User instance = new User();

        @Override
        public ILastName firstName(String firstName) {
            instance.setFirstName(firstName);
            return this;
        }

        @Override
        public IEmailAddress lastName(String lastName) {
            instance.setLastName(lastName);
            return this;
        }

        @Override
        public IBuild emailAddress(String emailAddress) {
            instance.setEmailAddress(emailAddress.toLowerCase());
            return this;
        }

        @Override
        public IBuild identifier(Long identifier) {
            instance.setIdentifier(identifier);
            return this;
        }

        @Override
        public IBuild roles(Set<Roles> roles) {
            instance.setRoles(roles);
            return this;
        }

        @Override
        public IBuild password(Password password) {
            instance.setPassword(password);
            return this;
        }

        @Override
        public IBuild deleted(boolean deleted) {
            instance.setDeleted(deleted);
            return this;
        }

        @Override
        public User build() {
            return instance;
        }
    }

    public static class ByNameComparator implements Comparator<User>, Serializable {

        @Override
        public int compare(User o1, User o2) {
            int i = o1.getLastName().compareTo(o2.getLastName());
            if (i == 0)
                i = o1.getFirstName().compareTo(o2.getFirstName());
            if (i == 0)
                i = o1.getId().compareTo(o2.getId());
            return i;
        }
    }
}