package se.su.dsv.scipro.workerthreads;

import com.google.inject.persist.UnitOfWork;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import java.util.Date;

/**
 * Subclasses must be annotated with @Component or similar annotation in order for autowiring of dependencies to work
 *
 */
public abstract class AbstractWorker implements Worker {

    private static final Logger LOGGER = LoggerFactory.getLogger(AbstractWorker.class);

    private WorkerDataService workerDataService;

    private WorkerData wd = null;
    private Date lastSuccessfulRun = new Date(0);
    private boolean successfulWorker = true;
    private Provider<EntityManager> emProvider;
    private EntityTransaction transaction;
    private UnitOfWork unitOfWork;

    /**
     * Subclasses must be annotated with @Component or similar annotation in order for autowiring of dependencies to work
     */
    public AbstractWorker() {
    }

    @Inject
    public void setWorkerDataService(WorkerDataService workerDataService) {
        this.workerDataService = workerDataService;
    }

    @Inject
    public void setTxManager(Provider<EntityManager> em) {
        this.emProvider = em;
    }

    @Inject
    public void setUnitOfWork(UnitOfWork unitOfWork) {
        this.unitOfWork = unitOfWork;
    }

    @Override
    public void run() {
        unitOfWork.begin();
        EntityManager em = emProvider.get();
        try {
            transaction = em.getTransaction();

            wd = workerDataService.getWorkerDataByName(getName());
            if (wd != null) {
                lastSuccessfulRun = wd.getLastSuccessfulRun();
            } else { //No date found so assume we've never run before, set earliest possible Date
                lastSuccessfulRun = new Date(0);
                wd = new WorkerData();
                wd.setName(getName());
            }

            Date startedWorkingDate = new Date();
            wd.setLastRun(startedWorkingDate);
            saveWorkerData();

			/*
             * Do manually transaction-handled work
			 */
            try {
                doWork();
            }
            catch (RuntimeException ex) {
                LOGGER.info("Worker {} threw an exception", getClass().getSimpleName());
                LOGGER.error("Worker failed", ex);
            }
            finally {
                if (transaction.isActive() && em.isOpen()) {
                    transaction.rollback();
                }
            }

            if (successfulWorker) {
                wd.setLastSuccessfulRun(startedWorkingDate);
            }
            saveWorkerData();

        } finally {
            if (transaction.isActive() && em.isOpen()) {
                transaction.rollback();
            }
            unitOfWork.end();
        }
    }

    private void saveWorkerData() {
        this.beginTransaction();
        try {
            wd = workerDataService.save(wd);
            this.commitTransaction();
        } catch (Exception e) {
            this.rollbackTransaction();
            LOGGER.error(getName() + " failed to save it's workerdata!");
            LOGGER.error(e.getMessage(), e);
        }
    }

    /**
     * Set whether the worker was run successfully or if an error occurred. Default
     * is true.
     * This result will be stored in the database and the latest successful run will
     * be logged with a timestamp.
     */
    public void setSuccessfulWorker(boolean successfulWorker) {
        this.successfulWorker = successfulWorker;
    }

    /**
     * Method to get the name/key string to lookup when it was last run for a certain worker, based on it's class.
     *
     */
    @Override
    public String getName() {
        return getClass().getName();
    }

    /**
     * Returns a date when the work was last successfully run or the date January 01 1970 if it's never been run successfully before.
     *
     */
    public final Date getLastSuccessfulRun() {
        return lastSuccessfulRun;
    }

    /**
     * Use beginTransaction(), commitTransaction() and rollbackTransaction() for transaction work.
     * Use setSuccessfulRun() to specify whether a run is successful or not. It defaults to true.
     */
    protected abstract void doWork();

    protected void beginTransaction() {
        transaction.begin();
    }

    protected void commitTransaction() {
        transaction.commit();
    }

    protected void rollbackTransaction() {
        transaction.rollback();
    }

}
