package se.su.dsv.scipro.workerthreads;

import com.google.common.eventbus.EventBus;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.activityplan.Activity;
import se.su.dsv.scipro.activityplan.ActivityPlanFacade;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.match.*;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import java.text.MessageFormat;
import java.util.*;

import static se.su.dsv.scipro.match.IdeaExport.Result.FAIL;
import static se.su.dsv.scipro.match.IdeaExport.Result.SUCCESS;

public class IdeaExportWorker extends AbstractWorker {

    private static final Logger LOGGER = LoggerFactory.getLogger(IdeaExportWorker.class);

    private final IdeaService ideaService;
    private final MailEventService mailService;
    private final ProjectService projectService;
    private final Set<IdeaCreationJudge> ideaCreationJudges;
    private final EventBus eventBus;
    private final ActivityPlanFacade activityPlanFacade;

    @Inject
    public IdeaExportWorker(final IdeaService ideaService,
                            final MailEventService mailService,
                            final ProjectService projectService,
                            final Set<IdeaCreationJudge> ideaCreationJudges,
                            final EventBus eventBus,
                            final ActivityPlanFacade activityPlanFacade) {
        this.ideaService = ideaService;
        this.mailService = mailService;
        this.projectService = projectService;
        this.ideaCreationJudges = ideaCreationJudges;
        this.eventBus = eventBus;
        this.activityPlanFacade = activityPlanFacade;
    }

    @Override
    protected void doWork() {
        Collection<Idea> failed = new LinkedList<>();
        Iterable<Idea> ideas = getMatchIdeas();
        for (Idea idea : ideas) {
            Decision decision = isAllowedToStart(idea);
            if (decision.isAllowed()) {
                allow(idea);
                createProject(idea);
                eventBus.post(new ProjectStartedEvent(idea));
            } else {
                deny(idea, decision.getReason());
                failed.add(idea);
                eventBus.post(new ProjectStartDeniedEvent(idea));
            }
        }

        if (!failed.isEmpty()) {
            new SupportMail(failed).process();
        }
    }

    private Iterable<Idea> getMatchIdeas() {
        return ideaService.findReadyToStart();
    }

    private Decision isAllowedToStart(final Idea idea) {
        for (IdeaCreationJudge ideaCreationJudge : ideaCreationJudges) {
            Decision decision = ideaCreationJudge.ruling(idea);
            if (!decision.isAllowed()) {
                return decision;
            }
        }
        return Decision.allowed();
    }


    private void allow(final Idea idea) {
        store(idea, "", SUCCESS);
    }

    private void deny(final Idea idea, final String reason) {
        store(idea, reason, FAIL);
    }

    private void store(final Idea idea, final String reason, final IdeaExport.Result result) {
        IdeaExport export = new IdeaExport();
        export.setIdea(idea);
        export.setReason(reason);
        export.setResult(result);
        idea.setExport(export);
        ideaService.save(idea);
    }

    private void createProject(Idea idea) {
        beginTransaction();
        LOGGER.info("Exporting idea: {}", idea);
        Project project = Project.builder()
                .title(idea.getTitle())
                .projectType(idea.getProjectType())
                .headSupervisor(idea.getMatch().getSupervisor())
                .projectParticipants(getAuthors(idea))
                .dateCreated(idea.getCourseStartDate())
                .build();

        idea.setProject(project);
        projectService.save(project);
        ideaService.save(idea);
        createFirstMeetingEvent(idea, project);
        addCompletedMatch(idea);
        commitTransaction();
    }

    private void createFirstMeetingEvent(Idea idea, Project project) {
        if (!project.hasModule(ProjectModule.ACTIVITY_PLAN) || idea.getFirstMeeting() == null) {
            return;
        }
        Activity event = new Activity();
        event.setDate(idea.getFirstMeeting().getFirstMeetingDate());
        event.setTitle("First meeting");
        String description = idea.getFirstMeeting().getRoom();
        if (idea.getFirstMeeting().getDescription() != null) {
            description += " " + idea.getFirstMeeting().getDescription();
        }
        event.setDescription(description);
        event.setActivityPlan(activityPlanFacade.retrieveActivityPlan(project));
        activityPlanFacade.saveActivity(event);
    }

    private void addCompletedMatch(final Idea idea) {
        Match match = new Match();
        match.setIdea(idea);
        match.setSupervisor(idea.getMatch().getSupervisor());
        match.setStatus(idea.getMatchStatus());
        idea.setMatch(match);
    }

    private SortedSet<User> getAuthors(final Idea idea) {
        SortedSet<User> authors = new TreeSet<>(new User.ByNameComparator());
        for (IdeaParticipation ideaParticipation : idea.getIdeaParticipations()) {
            authors.add(ideaParticipation.getUser());
        }
        return authors;
    }

    private class SupportMail {

        String supportEmail = "thesissupport@dsv.su.se";
        String subjectFormat = "[SciPro] Automatic idea exporting {0, date}";
        String prefix = "The following {0, choice, 1#idea|1<{0, number} ideas} failed to export into Daisy.\n\n";
        String suffix = "***DO NOT REPLY TO THIS MESSAGE!*** This is an automated message from SciPro.";
        String ideaFormat =
                "{0}\n" +
                        "Error: {1}\n" +
                        " - Supervisor: {2}\n" +
                        " - Authors: {3}\n\n";


        Collection<Idea> failedIdeas;
        String subject;
        String body;
        MailEvent mail;

        SupportMail(Collection<Idea> failedIdeas) {
            this.failedIdeas = failedIdeas;
        }

        public void process() {
            generateSubject();
            generateBody();
            generateMail();
            send();
        }

        private void send() {
            mailService.save(mail);
        }

        private void generateMail() {
            mail = new MailEvent();

            mail.setSubject(subject);
            mail.setMessageBody(body);
            mail.getNonUserRecipients().add(supportEmail);
        }

        private void generateSubject() {
            subject = MessageFormat.format(subjectFormat, new Date());
        }

        private void generateBody() {
            StringBuilder sb = new StringBuilder();

            sb.append(MessageFormat.format(prefix, failedIdeas.size()));

            for (Idea idea : failedIdeas) {
                String title = idea.getTitle();
                String error = idea.getExport().getReason();
                String supervisor = idea.getMatch().getSupervisor().getFullName();
                String authors = getAuthors(idea);
                sb.append(MessageFormat.format(ideaFormat, title, error, supervisor, authors));
            }

            sb.append(suffix);

            body = sb.toString();
        }

        private String getAuthors(Idea idea) {
            StringBuilder authors = new StringBuilder();
            for (Iterator<IdeaParticipation> iterator = idea.getIdeaParticipations().iterator(); iterator.hasNext(); ) {
                authors.append(iterator.next().getUser().getFullName());
                if (iterator.hasNext()) {
                    authors.append(", ");
                }
            }
            return authors.toString();
        }
    }
}
