package se.su.dsv.scipro.workerthreads;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.notifications.NotificationService;
import se.su.dsv.scipro.notifications.dataobject.Notification;
import se.su.dsv.scipro.notifications.dataobject.NotificationMail;
import se.su.dsv.scipro.notifications.interfaces.NotificationMailFormatter;
import se.su.dsv.scipro.notifications.settings.entities.DeliveryMethod;
import se.su.dsv.scipro.notifications.settings.service.DeliveryConfigurationService;
import se.su.dsv.scipro.springdata.services.UserProfileService;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

public class NotificationCompilationWorker extends AbstractWorker {

    private static final Logger LOGGER = LoggerFactory.getLogger(NotificationCompilationWorker.class);

    @Inject
    NotificationService notificationService;
    @Inject
    NotificationMailFormatter notificationMailFormatter;
    @Inject
    MailEventService mailEventService;
    @Inject
    GeneralSystemSettingsService systemSettingsService;
    @Inject
    DeliveryConfigurationService deliveryConfiguration;
    @Inject
    UserProfileService userProfileService;

    @Override
    protected void doWork() {
        LOGGER.info("Starting NotificationCompilationWorker..");
        Map<User, List<Notification>> notifications = getUserListMap();

        for(Map.Entry<User, List<Notification>> entry : notifications.entrySet()) {
            sendNotifications(entry.getKey(), entry.getValue());
        }

        int notificationCount = 0;
        for (List<Notification> notificationList : notifications.values()) {
            notificationCount += notificationList.size();
        }
        LOGGER.info(String.format("Sent notification compilation to %d users covering %d notifications.%n", notifications.size(), notificationCount));
    }

    private void sendNotifications(User user, List<Notification> notifications) {
        beginTransaction();

        GeneralSystemSettings systemSettings = systemSettingsService.getGeneralSystemSettingsInstance();
        NotificationMail notificationMail = notificationMailFormatter.compile(notifications);

        String subject = notificationMail.getSubject();
        String mailBody = notificationMail.getContent();
        String fromName = systemSettings.getMailFromName();
        String fromMail = systemSettings.getSystemFromMail();

        MailEvent mailEvent = new MailEvent(subject, mailBody, user, fromName, fromMail);
        mailEventService.save(mailEvent);

        for (Notification notification : notifications) {
            notification.setMailed(true);
            notificationService.save(notification);
        }

        commitTransaction();
    }

    private Map<User, List<Notification>> getUserListMap() {
        Map<User, List<Notification>> notifications = new HashMap<>();

        beginTransaction();

        List<User> withMailCompilation = userProfileService.findUsersWithMailCompilation();
        Iterable<Notification> unreadNotifications = notificationService.findUnreadByUsers(withMailCompilation);

        for (Notification notification : unreadNotifications) {
            if (notification.isMailed()) {
                continue;
            }

            User user = notification.getUser();

            if (deliveryConfiguration.isDelivery(notification.getType(), notification.getEvent(), DeliveryMethod.MAIL, user)) {
                if (!notifications.containsKey(user)) {
                    notifications.put(user, new LinkedList<Notification>());
                }
                notifications.get(user).add(notification);
            }
        }
        commitTransaction();
        return notifications;
    }
}
