package se.su.dsv.scipro.workerthreads;

import se.su.dsv.scipro.system.Lifecycle;

import javax.inject.Inject;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;

@Singleton
class SchedulerImpl implements Lifecycle, Scheduler {

    private final ScheduledExecutorService scheduledExecutorService;

    private final Set<Task> tasks = new TreeSet<>(new Task.ByDescriptionComparator());
    private final Set<Task> runningWorkers = Collections.synchronizedSet(new HashSet<Task>());

    @Inject
    SchedulerImpl(ScheduledExecutorService scheduledExecutorService) {
        this.scheduledExecutorService = scheduledExecutorService;
    }

    @Override
    public void execute(Task task) {
        scheduledExecutorService.execute(tracked(task));
    }

    Job addTask(final String description, final Provider<? extends Worker> worker, final Schedule schedule) {
        Task task = new Task(description, worker, schedule);
        Runnable runnable = tracked(task);
        tasks.add(task);
        ScheduledFuture<?> scheduledFuture = scheduledExecutorService.scheduleAtFixedRate(runnable, schedule.getDelay(), schedule.getPeriod(), schedule.getTimeUnit());
        return new Job(this, task, scheduledFuture);
    }

    void removeTask(Task task) {
        tasks.remove(task);
    }

    private Runnable tracked(final Task task) {
        return new Runnable() {
            @Override
            public void run() {
                runningWorkers.add(task);
                task.execute();
                runningWorkers.remove(task);
            }
        };
    }

    @Override
    public WorkerBuilder schedule(String description) {
        return new TaskScheduling(this).withDescription(description);
    }

    @Override
    public void start() {
    }

    @Override
    public void stop() {
        scheduledExecutorService.shutdownNow();
    }

    @Override
    public Set<Task> getRunningWorkers() {
        return runningWorkers;
    }

    @Override
    public Set<Task> getTasks() {
        return Collections.unmodifiableSet(tasks);
    }
}
