package se.su.dsv.scipro.workerthreads;

import org.joda.time.DateTime;
import org.joda.time.Days;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.plagiarism.TurnitinSettings;
import se.su.dsv.scipro.plagiarism.TurnitinSettingsService;

import javax.inject.Inject;
import java.text.MessageFormat;
import java.util.Date;
import java.util.HashSet;

public class TurnItInExpirationWorker extends AbstractWorker {

    private static final Logger LOGGER = LoggerFactory.getLogger(TurnItInExpirationWorker.class);

    private static final int[] NOTIFY_DAYS_AHEAD = new int[]{0, 1, 7, 14, 21, 28};
    private static final String SUBJECT = "[SciPro] TurnItIn expiration reminder " +
            "({0, choice, -1# Expired {1, number} days ago|0#Expires today|1#Expires tomorrow|1<Expires in {0} days})";
    public static final int WEEK = 7;

    private final GeneralSystemSettingsService systemSettingsService;
    private final TurnitinSettingsService service;
    private final MailEventService mailEventService;

    @Inject
    public TurnItInExpirationWorker(GeneralSystemSettingsService systemSettingsService, TurnitinSettingsService service, MailEventService mailEventService) {
        this.systemSettingsService = systemSettingsService;
        this.service = service;
        this.mailEventService = mailEventService;
    }

    @Override
    protected void doWork() {
        try {
            beginTransaction();

            if (dataExists()&& shouldRemind()) {
                sendReminder();
            }

            commitTransaction();
        } catch (Exception ex) {
            setSuccessfulWorker(false);
            LOGGER.error("Could not send reminder about TurnItIn expiration", ex);
            rollbackTransaction();
        }
    }

    private void sendReminder() {
        MailEvent mail = buildMail();
        mailEventService.save(mail);
    }

    private MailEvent buildMail() {
        GeneralSystemSettings settings = systemSettingsService.getGeneralSystemSettingsInstance();
        TurnitinSettings turnitinSettings = service.getSettings();
        MailEvent mail = new MailEvent();

        mail.setFromEmail(settings.getSystemFromMail());
        mail.setFromName(settings.getMailFromName());
        mail.setMessageBody(turnitinSettings.getExpirationContent());
        mail.setSubject(getSubject());
        mail.setNonUserRecipients(new HashSet<>(turnitinSettings.getExpirationMails()));

        return mail;
    }

    private boolean dataExists(){
        GeneralSystemSettings settings = systemSettingsService.getGeneralSystemSettingsInstance();
        TurnitinSettings turnitinSettings = service.getSettings();
        return settings.getSystemFromMail() != null &&
                settings.getMailFromName() != null &&
                hasExpirationData(turnitinSettings);
    }

    private boolean hasExpirationData(TurnitinSettings turnitinSettings) {
        return turnitinSettings.getExpirationContent() != null &&
                turnitinSettings.getExpirationDate() != null &&
                turnitinSettings.getExpirationMails() != null &&
                !turnitinSettings.getExpirationMails().isEmpty();
    }

    private String getSubject() {
        return MessageFormat.format(SUBJECT, daysToExpiration(), -daysToExpiration());
    }

    private boolean shouldRemind() {
        for (int days : NOTIFY_DAYS_AHEAD) {
            if (days == daysToExpiration()) {
                return true;
            }
        }
        return expiredNotice();
    }

    private boolean expiredNotice() {
        return daysToExpiration() < 0 && daysToExpiration() % WEEK == 0;
    }

    private int daysToExpiration() {
        TurnitinSettings turnitinSettings = service.getSettings();
        Date expirationDate = turnitinSettings.getExpirationDate();

        DateTime now = new DateTime().withTimeAtStartOfDay();
        DateTime expiration = new DateTime(expirationDate).withTimeAtStartOfDay();

        return Days.daysBetween(now, expiration).getDays();
    }
}
