package se.su.dsv.scipro.milestones.service

import javax.inject.Inject

import com.google.common.eventbus.{EventBus, Subscribe}
import se.su.dsv.scipro.finalseminar._
import se.su.dsv.scipro.finalthesis.FinalThesisApprovedEvent
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate
import se.su.dsv.scipro.peer.{FirstPeerReviewCompletedEvent, SecondPeerReviewCompletedEvent}
import se.su.dsv.scipro.project.Project
import se.su.dsv.scipro.report.{ReviewerGradingReportSubmittedEvent, SupervisorGradingReportSubmittedEvent}
import se.su.dsv.scipro.reviewing.{RoughDraftApprovalApprovedEvent, RoughDraftApprovalRequestedEvent, FinalSeminarApprovalApprovedEvent}
import se.su.dsv.scipro.grading.GradingCompletedEvent
import se.su.dsv.scipro.system.User

import scala.collection.immutable.Set

trait ProjectAware[A] extends NamedEvent[A] {
  def project(a: A): Project
}

trait Authorship[A] extends ProjectAware[A] {
  def author(a: A): User
}

trait NamedEvent[A] {
  def name(a: A): Set[String]
}

class MilestoneActivator @Inject()(milestoneActivityTemplateService: MilestoneActivityTemplateService, mileStoneService: MileStoneService, eventBus: EventBus) {
  eventBus.register(this)

  @Subscribe def oppositionApproved(event: OppositionApprovedEvent): Unit = {
    activateIndividualMilestone(event)
  }

  @Subscribe def participationApproved(event: ParticipationApprovedEvent): Unit = {
    activateIndividualMilestone(event)
  }

  @Subscribe def respondentApproved(event: RespondentApprovedEvent): Unit = {
    activateIndividualMilestone(event)
  }

  @Subscribe def oppositionFailed(event: OppositionFailedEvent): Unit = {
    deactivateIndividualMilestone(event)
  }

  @Subscribe def respondentFailed(event: RespondentFailedEvent): Unit = {
    deactivateIndividualMilestone(event)
  }

  @Subscribe def participationFailed(event: ParticipationFailedEvent): Unit = {
    deactivateIndividualMilestone(event)
  }

  @Subscribe def peerReviewCompleted(event: FirstPeerReviewCompletedEvent): Unit = {
    activateIndividualMilestone(event)
  }

  @Subscribe def secondPeerReviewCompleted(event: SecondPeerReviewCompletedEvent): Unit = {
    activateIndividualMilestone(event)
  }

  @Subscribe def seminarThesisUploaded(event: FinalSeminarThesisUploadedEvent): Unit = {
    activateProjectMilestone(event)
  }

  @Subscribe def finalSeminarCreated(event: FinalSeminarCreatedEvent): Unit = {
    activateProjectMilestone(event)
  }

  @Subscribe def finalSeminarDeleted(event: FinalSeminarDeletedEvent): Unit = {
    deactivateProjectMilestone(event)
  }

  @Subscribe def reviewerGradingReportSubmitted(event: ReviewerGradingReportSubmittedEvent): Unit = {
    activateProjectMilestone(event)
  }

  @Subscribe def supervisorGradingReportSubmitted(event: SupervisorGradingReportSubmittedEvent): Unit = {
    activateIndividualMilestone(event)
  }

  @Subscribe def reviewerApprovalApproved(event: FinalSeminarApprovalApprovedEvent): Unit = {
    activateProjectMilestone(event)
  }

  @Subscribe def reviewerApprovalApproved(event: RoughDraftApprovalApprovedEvent): Unit = {
    activateProjectMilestone(event)
  }

  @Subscribe def finalSeminarThesisDeleted(event: FinalSeminarThesisDeletedEvent): Unit = {
    deactivateProjectMilestone(event)
  }

  @Subscribe def finalThesisApproved(event: FinalThesisApprovedEvent) = activateProjectMilestone(event)

  @Subscribe def roughDraftApprovalRequested(event: RoughDraftApprovalRequestedEvent) = activateProjectMilestone(event)

  @Subscribe def gradingCompleted(event: GradingCompletedEvent) : Unit = {
    activateIndividualMilestone(event)
  }

  private def activateProjectMilestone[A : ProjectAware](event: A): Unit = {
    changeProjectMilestoneActivation(event, activationState = true)
  }

  private def deactivateProjectMilestone[A : ProjectAware](event: A): Unit = {
    changeProjectMilestoneActivation(event, activationState = false)
  }

  private def changeProjectMilestoneActivation[A](event: A, activationState: Boolean)(implicit P: ProjectAware[A]): Unit = {
    getMilestoneActivity(event) foreach { activity =>
      mileStoneService.setConfirmed(P.project(event), activity, activationState)
    }
  }

  private def deactivateIndividualMilestone[A : Authorship](event: A): Unit = {
    confirmIndividualMilestone(event, confirmed = false)
  }

  private def activateIndividualMilestone[A : Authorship](event: A): Unit = {
    confirmIndividualMilestone(event, confirmed = true)
  }

  private def confirmIndividualMilestone[A](event: A, confirmed: Boolean)(implicit A: Authorship[A]): Unit = {
    getMilestoneActivity(event) foreach { activity =>
      mileStoneService.setConfirmed(A.project(event), A.author(event), activity, confirmed)
    }
  }

  private def getMilestoneActivity[A](event: A)(implicit N: NamedEvent[A]): Set[MilestoneActivityTemplate] = {
    N.name(event).flatMap(name => Option(milestoneActivityTemplateService.findByEvent(name)))
  }
}
