package se.su.dsv.scipro.finalseminar;

import com.google.common.collect.Lists;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.plagiarism.TurnItInCheck;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.report.GradingReportTemplate;
import se.su.dsv.scipro.report.OppositionReport;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.system.*;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;
import java.util.Date;
import java.util.List;

import static org.hamcrest.CoreMatchers.hasItem;
import static org.junit.Assert.*;

public class FinalSeminarServiceImplIntegrationTest extends IntegrationTest {

    @Inject
    private FinalSeminarService finalSeminarService;

    private ProjectType projectType;
    private FinalSeminar futureFinalSeminar;
    private User user;
    private Project otherProject;
    private FinalSeminar pastFinalSeminar;
    private Language english;

    @Before
    public void setUp() throws Exception {
        projectType = createProjectType();
        english = createLanguage();
        futureFinalSeminar = createFinalSeminar(createProject(), 1);
        user = createUser();
        otherProject = createProject();
        pastFinalSeminar = createFinalSeminar(createProject(), -1);
    }

    @Test
    public void has_had_final_seminar() throws Exception {
        assertTrue(finalSeminarService.hasHadFinalSeminar(pastFinalSeminar.getProject()));
    }

    @Test
    public void has_not_had_final_seminar() throws Exception {
        Project project = createProject();

        assertFalse(finalSeminarService.hasHadFinalSeminar(project));
    }

    @Test
    public void delete_seminar() throws Exception {
        finalSeminarService.delete(futureFinalSeminar);
        assertEquals(1, finalSeminarService.count());
    }

    @Test
    public void delete_seminar_with_document() throws Exception {
        futureFinalSeminar.setDocument(createFileDescription());
        futureFinalSeminar = save(futureFinalSeminar);
        finalSeminarService.delete(futureFinalSeminar);
        assertEquals(1, finalSeminarService.count());
    }

    @Test
    public void delete_seminar_with_document_and_turnitin_check() throws Exception {
        FileDescription fileDescription = createFileDescription();
        futureFinalSeminar.setDocument(fileDescription);
        futureFinalSeminar = save(futureFinalSeminar);
        createTurnitinCheck(fileDescription);
        finalSeminarService.delete(futureFinalSeminar);
        assertEquals(1, finalSeminarService.count());
    }

    @Test
    public void delete_seminar_with_opposition() throws Exception {
        FinalSeminarOpposition opposition = createOpposition(futureFinalSeminar);
        opposition.setOppositionReport(createOppositionReport(opposition));
        futureFinalSeminar.setOppositions(Lists.newArrayList(opposition));
        futureFinalSeminar = save(futureFinalSeminar);
        finalSeminarService.delete(futureFinalSeminar);
        assertEquals(1, finalSeminarService.count());
    }

    @Test
    public void delete_seminar_with_opposition_and_opposition_report() throws Exception {
        futureFinalSeminar.setOppositions(Lists.newArrayList(createOpposition(futureFinalSeminar)));
        futureFinalSeminar = save(futureFinalSeminar);
        finalSeminarService.delete(futureFinalSeminar);
        assertEquals(1, finalSeminarService.count());
    }

    @Test
    public void can_oppose_future_seminar() {
        assertTrue(finalSeminarService.canOppose(user, futureFinalSeminar, otherProject) instanceof CanRegister);
    }

    @Test
    public void can_not_oppose_past_seminar() {
        assertTrue(finalSeminarService.canOppose(user, pastFinalSeminar, otherProject) instanceof AlreadyHappened);
    }

    @Test
    public void oppose() {
        otherProject.addProjectParticipant(user);
        assertFalse(finalSeminarService.isOpponent(futureFinalSeminar, user));
        finalSeminarService.oppose(user, futureFinalSeminar, otherProject);
        assertTrue(finalSeminarService.isOpponent(futureFinalSeminar, user));
    }

    @Test
    public void can_not_oppose_if_already_opponent() {
        otherProject.addProjectParticipant(user);
        finalSeminarService.oppose(user, futureFinalSeminar, otherProject);
        assertTrue(finalSeminarService.canOppose(user, futureFinalSeminar, otherProject) instanceof AlreadyParticipating);
    }

    @Test
    public void can_not_oppose_if_participant() {
        finalSeminarService.participate(user, futureFinalSeminar, otherProject);
        assertTrue(finalSeminarService.canOppose(user, futureFinalSeminar, otherProject) instanceof AlreadyParticipating);
    }

    @Test
    public void can_participate_on_future_seminar() {
        assertTrue(finalSeminarService.canParticipate(user, futureFinalSeminar));
    }

    @Test
    public void can_not_participate_on_past_seminar() {
        assertFalse(finalSeminarService.canParticipate(user, pastFinalSeminar));
    }

    @Test
    public void can_not_participate_twice() {
        finalSeminarService.participate(user, futureFinalSeminar, otherProject);
        assertFalse(finalSeminarService.canParticipate(user, futureFinalSeminar));
    }

    @Test
    public void can_not_oppose_if_max_is_reached() {
        futureFinalSeminar.setMaxOpponents(0);
        assertTrue(finalSeminarService.canOppose(user, futureFinalSeminar, otherProject) instanceof TooManyOpponents);
    }

    @Test
    public void can_not_participate_if_max_is_reached() {
        futureFinalSeminar.setMaxParticipants(0);
        assertFalse(finalSeminarService.canParticipate(user, futureFinalSeminar));
    }

    @Test
    public void can_not_oppose_if_is_author() {
        futureFinalSeminar.getProject().addProjectParticipant(user);
        assertTrue(finalSeminarService.canOppose(user, futureFinalSeminar, otherProject) instanceof AlreadyParticipating);
    }

    @Test
    public void can_not_participate_if_is_author() {
        futureFinalSeminar.getProject().addProjectParticipant(user);
        assertFalse(finalSeminarService.canParticipate(user, futureFinalSeminar));
    }

    @Test
    public void can_not_oppose_on_deleted_seminar() {
        futureFinalSeminar.setDeleted(true);
        assertTrue(finalSeminarService.canOppose(user, futureFinalSeminar, otherProject) instanceof FinalSeminarCancelled);
    }

    @Test
    public void can_not_participate_on_deleted_seminar() {
        futureFinalSeminar.setDeleted(true);
        assertFalse(finalSeminarService.canParticipate(user, futureFinalSeminar));
    }

    @Test
    public void can_participate_if_previous_oppositions_are_failed() {
        FinalSeminar secondSeminar = createFinalSeminar(createProject(), 10);
        finalSeminarService.oppose(user, futureFinalSeminar, otherProject);
        for (FinalSeminarOpposition finalSeminarOpposition : futureFinalSeminar.getOppositions()) {
            finalSeminarOpposition.setGrade(FinalSeminarGrade.NOT_APPROVED);
        }
        assertTrue(finalSeminarService.canOppose(user, secondSeminar, otherProject) instanceof CanRegister);
    }

    @Test
    public void findByProject() throws Exception {
        assertEquals(futureFinalSeminar, finalSeminarService.findByProject(futureFinalSeminar.getProject()));
    }

    @Test
    public void filter_by_date() {
        FinalSeminarService.Filter filter = new FinalSeminarService.Filter();
        filter.setFromDate(new Date());
        assertThat(finalSeminarService.findAll(filter), hasItem(futureFinalSeminar));
    }

    @Test
    public void find_unfinished_seminars_handles_oppositions() throws Exception {
        FinalSeminar unfinished = createFinalSeminar(createProject(), -6);
        FinalSeminar finished = createFinalSeminar(createProject(), -6);
        FinalSeminar unfinishedTooEarly = createFinalSeminar(createProject(), -3);
        addOpposition(finished, FinalSeminarGrade.APPROVED);
        addOpposition(unfinished, null);
        addOpposition(unfinishedTooEarly, null);

        List<FinalSeminar> seminars = finalSeminarService.findUnfinishedSeminars(null,new DateTime().plusDays(-5).toDate() , new PageRequest(0, 5) );
        assertEquals(1, seminars.size());
        assertThat(seminars, hasItem(unfinished));
    }

    @Test
    public void find_unfinished_seminars_handles_active_participants() throws Exception {
        FinalSeminar unfinished = createFinalSeminar(createProject(), -6);
        FinalSeminar finished = createFinalSeminar(createProject(), -6);
        FinalSeminar unfinishedTooEarly = createFinalSeminar(createProject(), -3);
        addActiveParticipant(finished, FinalSeminarGrade.APPROVED);
        addActiveParticipant(unfinished, null);
        addActiveParticipant(unfinishedTooEarly, null);

        List<FinalSeminar> seminars = finalSeminarService.findUnfinishedSeminars(null, new DateTime().plusDays(-5).toDate(), new PageRequest(0, 5));
        assertEquals(1, seminars.size());
        assertThat(seminars, hasItem(unfinished));
    }

    @Test
    public void find_unfinished_seminars_handles_respondents() throws Exception {
        FinalSeminar unfinished = createFinalSeminar(createProject(), -6);
        FinalSeminar finished = createFinalSeminar(createProject(), -6);
        FinalSeminar unfinishedTooEarly = createFinalSeminar(createProject(), -3);
        addRespondent(finished, FinalSeminarGrade.APPROVED);
        addRespondent(unfinished, null);
        addRespondent(unfinishedTooEarly, null);

        List<FinalSeminar> seminars = finalSeminarService.findUnfinishedSeminars(null, new DateTime().plusDays(-5).toDate(), new PageRequest(0, 5));
        assertEquals(1, seminars.size());
        assertThat(seminars, hasItem(unfinished));
    }

    @Test
    public void count_unfinished_seminars() throws Exception {
        FinalSeminar unfinished = createFinalSeminar(createProject(), -6);
        FinalSeminar finished = createFinalSeminar(createProject(), -6);
        FinalSeminar unfinishedTooEarly = createFinalSeminar(createProject(), -3);
        addRespondent(finished, FinalSeminarGrade.APPROVED);
        addRespondent(unfinished, null);
        addRespondent(unfinishedTooEarly, null);

        assertEquals(1, finalSeminarService.countUnfinishedSeminars(null, new DateTime().plusDays(-5).toDate()));
    }

    @Test
    public void can_not_oppose_if_priority_set_and_no_own_seminar() throws Exception {
        setOppositionPriorityDaysTo(8);
        FinalSeminar finalSeminar = createFutureFinalSeminarSomeDaysAgo(5);

        assertTrue(finalSeminarService.canOppose(user, finalSeminar, otherProject) instanceof PriorityForSeminarAuthors);
    }

    @Test
    public void can_oppose_if_priority_has_lapsed() throws Exception {
        setOppositionPriorityDaysTo(3);
        FinalSeminar finalSeminar = createFutureFinalSeminarSomeDaysAgo(5);

        assertTrue(finalSeminarService.canOppose(user, finalSeminar, otherProject) instanceof CanRegister);
    }

    @Test
    public void can_oppose_if_priority_set_and_has_own_seminar() throws Exception {
        setOppositionPriorityDaysTo(8);
        FinalSeminar finalSeminar = createFutureFinalSeminarSomeDaysAgo(5);
        save(initFinalSeminar(otherProject, 30));

        assertTrue(finalSeminarService.canOppose(user, finalSeminar, otherProject) instanceof CanRegister);
    }

    private FinalSeminar createFutureFinalSeminarSomeDaysAgo(final int daysAgo) {
        FinalSeminar finalSeminar = initFinalSeminar(createProject(), 5);
        finalSeminar.setDateCreated(new DateTime().minusDays(daysAgo).toDate());
        return save(finalSeminar);
    }

    private void setOppositionPriorityDaysTo(final int oppositionPriorityDays) {
        FinalSeminarSettings finalSeminarSettings = new FinalSeminarSettings();
        finalSeminarSettings.setId(1L);
        finalSeminarSettings.setOppositionPriorityDays(oppositionPriorityDays);
        save(finalSeminarSettings);
    }

    private void addRespondent(FinalSeminar seminar, FinalSeminarGrade grade) {
        FinalSeminarRespondent respondent = new FinalSeminarRespondent(createUser(), seminar);
        respondent.setGrade(grade);
        save(respondent);
    }

    private void addActiveParticipant(FinalSeminar seminar, FinalSeminarGrade grade) {
        FinalSeminarActiveParticipation participation = new FinalSeminarActiveParticipation();
        participation.setProject(createProject());
        participation.setUser(createUser());
        participation.setFinalSeminar(seminar);
        participation.setGrade(grade);
        save(participation);
        seminar.addActiveParticipant(participation);
        save(seminar);
    }

    private void addOpposition(FinalSeminar seminar, FinalSeminarGrade grade) {
        FinalSeminarOpposition opposition = createOpposition(seminar);
        opposition.setGrade(grade);
        save(opposition);
    }

    private OppositionReport createOppositionReport(FinalSeminarOpposition opposition) {
        OppositionReport oppositionReport = new OppositionReport(createGradingReportTemplate(projectType), opposition);
        return save(oppositionReport);
    }

    private GradingReportTemplate createGradingReportTemplate(ProjectType projectType) {
        GradingReportTemplate template = new GradingReportTemplate(projectType, 30);
        return save(template);
    }

    private FinalSeminarOpposition createOpposition(FinalSeminar finalSeminar) {
        FinalSeminarOpposition opposition = new FinalSeminarOpposition();
        opposition.setProject(createProject());
        opposition.setUser(createUser());
        opposition.setFinalSeminar(finalSeminar);
        return save(opposition);
    }

    private TurnItInCheck createTurnitinCheck(FileDescription fileDescription) {
        TurnItInCheck turnitinCheck = new TurnItInCheck(fileDescription);
        return save(turnitinCheck);
    }

    private FinalSeminar initFinalSeminar(Project project, int relativeDays) {
        FinalSeminar finalSeminar = new FinalSeminar();
        finalSeminar.setProject(project);
        finalSeminar.setStartDate(new DateTime().plusDays(relativeDays).toDate());
        finalSeminar.setRoom("room");
        finalSeminar.setPresentationLanguage(english);
        finalSeminar.setReportLanguage(english);
        return finalSeminar;
    }

    private FileDescription createFileDescription() {
        FileDescription fileDescription = new FileDescription();
        return save(fileDescription);
    }

    private FinalSeminar createFinalSeminar(Project project, int relativeDays) {
        FinalSeminar finalSeminar = initFinalSeminar(project, relativeDays);
        return save(finalSeminar);
    }

    private Project createProject() {
        Unit unit = createUnit();
        User supervisor = createSupervisor(unit);
        User author = createUser();
        return createProject(supervisor, author, projectType);
    }

    private Project createProject(User supervisor, User author, ProjectType projectType) {
        Project project = Project.builder().title("Some title").projectType(projectType).headSupervisor(supervisor).build();
        return save(project);
    }

    private ProjectType createProjectType() {
        ProjectType projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");
        return save(projectType);
    }

    private User createUser() {
        User studentUser = User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build();
        studentUser = save(studentUser);
        return studentUser;
    }

    private User createSupervisor(Unit unit) {
        User supervisorUser = createUser();
        supervisorUser.addRole(Roles.SUPERVISOR);
        supervisorUser.setUnit(unit);
        return save(supervisorUser);
    }

    private Unit createUnit() {
        Unit unit = new Unit();
        unit.setTitle("title");
        return save(unit);
    }

    private Language createLanguage() {
        Language language = new Language();
        language.setName("English");
        return save(language);
    }
}
