package se.su.dsv.scipro.finalthesis;

import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.TypeSafeMatcher;
import org.junit.Before;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;

import static org.hamcrest.CoreMatchers.everyItem;
import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class FinalThesisServiceImplTest extends IntegrationTest {

    public static final String TEXT_PLAIN = "text/plain";

    @Inject
    private FinalThesisServiceImpl finalThesisService;

    private User uploader;
    private Project project;
    private NotificationSource source;
    private String name;
    private String contentType;

    @Before
    public void setUp() throws Exception {
        ProjectType bachelor = save(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"));
        User headSupervisor = save(User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build());
        uploader = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        project = save(Project.builder().title("Some title").projectType(bachelor).headSupervisor(headSupervisor).build());
        source = new NotificationSource();
        name = "name";
        contentType = TEXT_PLAIN;
    }

    @Test
    public void upload() throws Exception {
        assertNotNull(uploadThesis());
    }

    @Test
    public void upload_new_thesis_file_creates_final_thesis_entity() throws Exception {
        uploadThesis();
        assertTrue(finalThesisService.hasFinalThesis(project));
    }

    @Test
    public void testHasFinalThesisReturnsFalseWhenNoFinalThesis() throws Exception {
        assertFalse(finalThesisService.hasFinalThesis(project));
    }

    @Test
    public void when_no_final_thesis_getFinalThesisFileDescription_should_return_null() throws Exception {
        assertNull(finalThesisService.getFinalThesisFileDescription(project));
    }

    @Test
    public void when_project_has_only_one_final_thesis_getFinalThesisFileDescription_should_return_final_thesis() throws Exception {
        uploadThesis();
        assertNotNull(finalThesisService.getFinalThesisFileDescription(project));
    }

    @Test
    public void when_project_has_multiple_file_descriptions_should_return_latest() throws Exception {
        uploadThesis();
        FinalThesis finalThesis = uploadThesis();

        assertEquals(finalThesis.getFileDescription(), finalThesisService.getFinalThesisFileDescription(project));
    }

    @Test
    public void when_project_has_multiple_final_theses_should_return_latest() throws Exception {
        uploadThesis();
        FinalThesis secondThesis = uploadThesis();

        assertEquals(secondThesis, finalThesisService.findByProject(project));
    }

    @Test
    public void approve_final_thesis() throws Exception {
        uploadThesis();
        finalThesisService.approve(project);

        FinalThesis finalThesis = finalThesisService.findByProject(project);
        assertEquals(FinalThesis.Status.APPROVED, finalThesis.getStatus());
        assertNotNull(finalThesis.getDateApproved());
    }

    @Test
    public void reject_final_thesis() throws Exception {
        uploadThesis();

        finalThesisService.reject(project);

        FinalThesis finalThesis = finalThesisService.findByProject(project);
        assertEquals(finalThesis.getStatus(), FinalThesis.Status.REJECTED);
        assertNotNull(finalThesis.getDateRejected());
    }

    @Test
    public void uploading_is_allowed_if_no_final_thesis_exists() throws Exception {
        assertTrue(finalThesisService.isUploadAllowed(project));
    }

    @Test
    public void uploading_is_forbidden_if_latest_final_thesis_is_approved() throws Exception {
        uploadThesis();
        finalThesisService.approve(project);

        assertFalse(finalThesisService.isUploadAllowed(project));
    }

    @Test
    public void uploading_is_allowed_if_latest_final_thesis_is_rejected() throws Exception {
        uploadThesis();
        finalThesisService.reject(project);

        assertTrue(finalThesisService.isUploadAllowed(project));
    }

    @Test(expected = RuntimeException.class)
    public void get_reject_date_throws_exception() throws Exception {
        finalThesisService.getRejectedDate(project);
    }

    @Test
    public void find_all_with_filter() throws Exception {
        saveThesis(FinalThesis.Status.APPROVED);
        saveThesis(FinalThesis.Status.REJECTED);

        Iterable<FinalThesis> result = finalThesisService.findAll(approvedFilter(), new PageRequest(0, 5));
        assertThat(result, everyItem(hasStatus(FinalThesis.Status.APPROVED)));
    }

    @Test
    public void count_with_filter() throws Exception {
        saveThesis(FinalThesis.Status.APPROVED);
        saveThesis(FinalThesis.Status.REJECTED);

        assertEquals(finalThesisService.count(approvedFilter()), 1L);
    }

    private FinalThesisService.Filter approvedFilter() {
        FinalThesisService.Filter filter = new FinalThesisService.Filter();
        filter.setStatus(FinalThesis.Status.APPROVED);
        return filter;
    }

    private FinalThesis saveThesis(FinalThesis.Status status) {
        FinalThesis thesis = uploadThesis();
        thesis.setStatus(status);
        return save(thesis);
    }

    private FinalThesis uploadThesis() {
        FileUpload fileUpload = mock(FileUpload.class);
        when(fileUpload.getContentType()).thenReturn(contentType);
        when(fileUpload.getFileName()).thenReturn(name);
        when(fileUpload.getUploader()).thenReturn(uploader);
        return finalThesisService.upload(fileUpload, project);
    }

    private static Matcher<FinalThesis> hasStatus(final FinalThesis.Status status) {
        return new TypeSafeMatcher<FinalThesis>() {
            @Override
            protected boolean matchesSafely(FinalThesis item) {
                return item.getStatus().equals(status);
            }

            @Override
            public void describeTo(Description description) {
                description.appendText("has status").appendValue(status);
            }
        };
    }
}
