package se.su.dsv.scipro.forum;

import com.google.common.eventbus.EventBus;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.runners.MockitoJUnitRunner;
import org.mockito.stubbing.Answer;
import se.su.dsv.scipro.file.*;
import se.su.dsv.scipro.forum.dataobjects.ForumPost;
import se.su.dsv.scipro.forum.dataobjects.ForumPostReadState;
import se.su.dsv.scipro.forum.dataobjects.ForumThread;
import se.su.dsv.scipro.forum.dataobjects.ProjectThread;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ForumBuilder;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.*;

import static org.junit.Assert.*;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class ProjectForumServiceImplTest {

    public static final String SUBJECT = "Subject";
    public static final String CONTENT = "Content";
    @Mock
    AbstractThreadRepository threadRepository;
    @Mock
    ProjectThreadRepository projectThreadRepository;
    @Mock
    ForumPostRepository postRepository;
    @Mock
    ForumPostReadStateRepository readStateRepository;
    @Mock
    FileDescriptionService fileDescriptionService;
    @Mock
    EventBus eventBus;
    @Mock
    private BasicForumService basicForumService;
    @InjectMocks
    ProjectForumServiceImpl service;
    private Project project;

    @Before
    @SuppressWarnings("unchecked")
    public void setUp() throws Exception {
        ProjectType projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "Some project type", "Some description");
        projectType.addModule(ProjectModule.FORUM);
        project = Project.builder().title("Some title").projectType(projectType).build();
        ForumPost post = new ForumPost();
        post.setContent(CONTENT);
        final ForumThread forumThread = new ForumThread();
        forumThread.setSubject(SUBJECT);
        ProjectThread projectThread = new ProjectThread();
        projectThread.setProject(project);
        projectThread.setForumThread(forumThread);
        post.setForumThread(forumThread);
        post.setPostedBy(new UserBuilder().create());

        when(projectThreadRepository.save(isA(ProjectThread.class))).thenAnswer(new Answer<Object>() {
            @Override
            public Object answer(final InvocationOnMock invocation) throws Throwable {
                return invocation.getArguments()[0];
            }
        });
        when(readStateRepository.save(isA(ForumPostReadState.class))).thenAnswer(new Answer<Object>() {
            @Override
            public Object answer(final InvocationOnMock invocation) throws Throwable {
                return invocation.getArguments()[0];
            }
        });
        when(postRepository.save(isA(ForumPost.class))).thenAnswer(new Answer<Object>() {
            @Override
            public Object answer(final InvocationOnMock invocation) throws Throwable {
                return invocation.getArguments()[0];
            }
        });
        when(basicForumService.createReply(any(ForumThread.class), any(User.class), anyString(), anySetOf(FileDescription.class)))
                .thenAnswer(new Answer<ForumPost>() {
                    @Override
                    public ForumPost answer(final InvocationOnMock invocation) throws Throwable {
                        ForumPost forumPost = new ForumPost();
                        forumPost.setForumThread(invocation.getArgumentAt(0, ForumThread.class));
                        forumPost.setPostedBy(invocation.getArgumentAt(1, User.class));
                        forumPost.setContent(invocation.getArgumentAt(2, String.class));
                        forumPost.setAttachments(invocation.getArgumentAt(3, Set.class));
                        forumPost.getForumThread().addPost(forumPost);
                        return forumPost;
                    }
                });
        when(basicForumService.createThread(anyString())).thenAnswer(new Answer<ForumThread>() {
            @Override
            public ForumThread answer(final InvocationOnMock invocation) throws Throwable {
                ForumThread forumThread1 = new ForumThread();
                forumThread1.setSubject(invocation.getArgumentAt(0, String.class));
                return forumThread1;
            }
        });
    }

    @Test
    public void testLoadThread() throws Exception {
        Long threadId = 1L;
        ProjectThread thread = new ForumBuilder().createThread();

        when(projectThreadRepository.findOne(threadId)).thenReturn(thread);

        ProjectThread serviceThread = service.findOne(threadId);

        assertEquals(thread, serviceThread);
    }

    @Test
    public void testCreateThread() throws Exception {
        User user = new UserBuilder().create();

        ProjectThread thread = createAndSaveThread(user, null, null);

        assertNewForumThread(thread, project, user, SUBJECT, CONTENT);
    }

    @Test
    public void testCreateThreadWithPostAttachment() throws Exception {
        User user = new UserBuilder().create();
        ProjectFile fileDescription = new ProjectFile();
        fileDescription.setPath("/foo");

        when(fileDescriptionService.storeFile(
                isA(ProjectFileUpload.class), eq(FileSource.FORUM)))
                .thenReturn(fileDescription);

        ProjectThread thread = createAndSaveThread(user, new byte[]{}, "");

        assertNewForumThread(thread, project, user, SUBJECT, CONTENT, fileDescription);
    }

    @Test
    public void testCreatingThreadSavesIt() throws Exception {
        User user = new UserBuilder().create();
        createAndSaveThread(user, null, null);

        ArgumentCaptor<ProjectThread> captor = ArgumentCaptor.forClass(ProjectThread.class);
        verify(projectThreadRepository).save(captor.capture());

        assertNewForumThread(captor.getValue(), project, user, SUBJECT, CONTENT);
    }

    @Test
    public void testGetPostPageByForumThread() throws Exception {
        List<ForumPost> posts = Collections.singletonList(new ForumBuilder().createPost());
        when(basicForumService.getPosts(any(ForumThread.class))).thenReturn(posts);

        List<ForumPost> servicePage = service.getPosts(mock(ProjectThread.class));

        assertEquals(posts, servicePage);
    }

    @Test
    public void testGetThreadsByProject() throws Exception {
        ProjectThread thread1 = new ForumBuilder().createThread();
        ProjectThread thread2 = new ForumBuilder().createThread();

        List<ProjectThread> threads = Arrays.asList(thread1, thread2);

        when(projectThreadRepository.findByProject(eq(project))).thenReturn(threads);

        List<ProjectThread> serviceThreads = service.getThreads(project);

        assertEquals(threads, serviceThreads);
    }

    private ProjectThread createAndSaveThread(User user, byte[] fileContent, String clientFileName) {
        return service.createThread(project, user, SUBJECT, CONTENT, getFileUploads(fileContent, clientFileName));
    }

    private Set<FileUpload> getFileUploads(byte[] fileContent, String clientFileName) {
        if (fileContent == null || clientFileName == null) {
            return new HashSet<>();
        }
        else {
            FileUpload fileUpload = mock(FileUpload.class);
            when(fileUpload.getFileName()).thenReturn(clientFileName);
            return new HashSet<>(Collections.singletonList(fileUpload));
        }
    }

    private void assertNewForumThread(ProjectThread thread, Project project, User user, String subject, String content) {
        assertNewForumThread(thread, project, user, subject, content, null);
    }

    private void assertNewForumThread(
            ProjectThread thread, Project project, User user, String subject, String content, FileDescription fileDescription) {
        assertEquals("Thread created for the wrong project", project, thread.getProject());
        assertEquals("Thread created by wrong user", user, thread.getForumThread().getCreatedBy());
        assertEquals("Thread created with wrong subject", subject, thread.getForumThread().getSubject());
        assertEquals("Thread does not contain one post", 1, thread.getForumThread().getPostCount());
        ForumPost post = thread.getForumThread().getPosts().get(0);
        assertEquals("Post does not contain content", content, post.getContent());
        assertEquals("Post is not created by thread starter", user, post.getPostedBy());

        if (fileDescription != null) {
            FileDescription attachment = post.getAttachments().iterator().next();
            assertEquals("Post have no attachment", fileDescription, attachment);
        }
    }
}
