package se.su.dsv.scipro.forum.notifications;

import com.google.common.eventbus.EventBus;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import scala.Option;
import se.su.dsv.scipro.forum.ForumPostReadEvent;
import se.su.dsv.scipro.forum.NewGroupForumReplyEvent;
import se.su.dsv.scipro.forum.NewProjectForumReplyEvent;
import se.su.dsv.scipro.forum.dataobjects.ForumPost;
import se.su.dsv.scipro.forum.dataobjects.ForumThread;
import se.su.dsv.scipro.forum.dataobjects.GroupThread;
import se.su.dsv.scipro.forum.dataobjects.ProjectThread;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.NotificationService;
import se.su.dsv.scipro.notifications.dataobject.GroupEvent;
import se.su.dsv.scipro.notifications.dataobject.Notification;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.notifications.dataobject.ProjectForumEvent;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.User;

import java.util.Collections;

import static org.junit.Assert.assertEquals;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static se.su.dsv.scipro.notifications.dataobject.ProjectForumEvent.Event.NEW_FORUM_POST;

@RunWith(MockitoJUnitRunner.class)
public class ForumNotificationsTest {
    @InjectMocks
    private ForumNotifications forumNotifications;

    @Mock
    private EventBus eventBus;
    @Mock
    private NotificationController notificationController;
    @Mock
    private ForumNotificationRepository forumNotificationRepository;
    @Mock
    private NotificationService notificationService;

    private ForumThread forumThread;
    private ForumPost firstReply;
    private ForumPost secondReply;

    @Before
    public void setUp() throws Exception {
        User poster = new User();
        forumThread = new ForumThread();
        firstReply = new ForumPost();
        firstReply.setForumThread(forumThread);
        firstReply.setPostedBy(poster);
        secondReply = new ForumPost();
        secondReply.setForumThread(forumThread);
        secondReply.setPostedBy(poster);
        forumThread.addPost(firstReply);
    }

    @Test
    public void sends_notification_on_new_project_thread() throws Exception {
        ProjectThread projectThread = new ProjectThread();
        projectThread.setForumThread(forumThread);

        NewProjectForumReplyEvent forumReplyEvent = new NewProjectForumReplyEvent(projectThread, firstReply);
        forumNotifications.newProjectForumPost(forumReplyEvent);

        verify(notificationController).notifyProjectForum(eq(NEW_FORUM_POST), any(NotificationSource.class), eq(projectThread.getProject()), any(User.class));
    }

    @Test
    public void sends_notification_on_new_reply() throws Exception {
        GroupThread groupThread = new GroupThread();
        groupThread.setForumThread(forumThread);
        forumThread.addPost(secondReply);

        NewGroupForumReplyEvent forumReplyEvent = new NewGroupForumReplyEvent(groupThread, secondReply);
        forumNotifications.newGroupForumPost(forumReplyEvent);

        verify(notificationController).notifyGroup(eq(groupThread.getGroup()), eq(GroupEvent.Event.MESSAGE_THREAD_REPLY), any(NotificationSource.class), any(User.class));
    }

    @Test
    public void connects_post_and_notification_event() throws Exception {
        ProjectForumEvent projectForumEvent = new ProjectForumEvent();
        Notification notification = new Notification();
        notification.setNotificationEvent(projectForumEvent);

        when(notificationController.notifyProjectForum(any(ProjectForumEvent.Event.class), any(NotificationSource.class), any(Project.class), any(User.class)))
                .thenReturn(Collections.singleton(notification));

        ProjectThread projectThread = new ProjectThread();
        projectThread.setForumThread(forumThread);

        NewProjectForumReplyEvent forumReplyEvent = new NewProjectForumReplyEvent(projectThread, firstReply);
        forumNotifications.newProjectForumPost(forumReplyEvent);

        ArgumentCaptor<ForumNotification> captor = ArgumentCaptor.forClass(ForumNotification.class);
        verify(forumNotificationRepository).save(captor.capture());

        assertEquals(captor.getValue().getForumPost(), firstReply);
        assertEquals(captor.getValue().getNotificationEvent(), projectForumEvent);
    }

    @Test
    public void mark_notification_read() throws Exception {
        User user = new User();
        ProjectForumEvent projectForumEvent = new ProjectForumEvent();
        Notification notification = new Notification();
        notification.setNotificationEvent(projectForumEvent);

        when(forumNotificationRepository.findByForumPost(firstReply))
                .thenReturn(Option.apply(new ForumNotification(firstReply, projectForumEvent)));

        forumNotifications.forumPostRead(new ForumPostReadEvent(firstReply, user));

        verify(notificationService).setRead(user, projectForumEvent, true);
    }
}