package se.su.dsv.scipro.forummail;

import com.google.common.eventbus.EventBus;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import se.su.dsv.scipro.forum.NewProjectForumReplyEvent;
import se.su.dsv.scipro.forum.dataobjects.ForumPost;
import se.su.dsv.scipro.forum.dataobjects.ForumThread;
import se.su.dsv.scipro.forum.dataobjects.ProjectThread;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.springdata.services.UserProfileService;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;

import static org.hamcrest.CoreMatchers.is;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertThat;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class SendForumMailTest {
    @Mock
    private ForumMailRepository forumMailRepository;
    @Mock
    private MailEventService mailEventService;
    @Mock
    private ForumMailSettingsService mailSettingsService;
    @Mock
    private ForumMailFormatter composer;
    @Mock
    private GeneralSystemSettingsService generalSystemSettings;
    @Mock
    private EventBus eventBus;
    @Mock
    private UserProfileService userProfileService;

    private ProjectThread projectThread;
    private ForumPost forumPost;

    private SendForumMail sendForumMail;
    private User coSupervisor;

    @Before
    public void setUp() throws Exception {
        when(generalSystemSettings.getGeneralSystemSettingsInstance()).thenReturn(new GeneralSystemSettings());
        when(mailEventService.save(isA(MailEvent.class))).thenReturn(new MailEvent());
        ForumMailSettings settings = new ForumMailSettings();
        settings.setEnabled(true);
        when(mailSettingsService.getSettings()).thenReturn(settings);

        User poster = newUser(1L);
        final User threadMember = newUser(2L);
        coSupervisor = newUser(3L);
        ProjectType bachelor = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");
        Project project = Project.builder().title("Project").projectType(bachelor).headSupervisor(poster).build();
        project.addProjectParticipant(threadMember);
        project.addCoSupervisor(coSupervisor);
        ForumThread forumThread = new ForumThread();
        projectThread = new ProjectThread();
        projectThread.setProject(project);
        projectThread.setForumThread(forumThread);
        forumPost = new ForumPost();
        forumPost.setForumThread(forumThread);
        forumPost.setPostedBy(poster);
        when(userProfileService.isReceiveForumMail(any(User.class))).thenReturn(true);

        sendForumMail = new SendForumMail(forumMailRepository, mailSettingsService, mailEventService, composer, generalSystemSettings, eventBus, userProfileService);
    }

    @Test
    public void generates_forum_mail_for_every_member() throws Exception {
        sendForumMail.newProjectForumReply(new NewProjectForumReplyEvent(projectThread, forumPost));

        ArgumentCaptor<ForumMailProject> captor = ArgumentCaptor.forClass(ForumMailProject.class);
        verify(forumMailRepository).save(captor.capture());

        ForumMailProject forumMail = captor.getValue();
        assertThat(forumMail.getUser(), is(coSupervisor));
        assertThat(forumMail.getProjectThread(), is(projectThread));
        assertNotNull(forumMail.getMailEvent());
    }

    @Test
    public void uses_correct_body() throws Exception {
        String content = "Reply here!";
        when(composer.formatBody(projectThread.getForumThread())).thenReturn(content);

        sendForumMail.newProjectForumReply(new NewProjectForumReplyEvent(projectThread, forumPost));

        ArgumentCaptor<MailEvent> captor = ArgumentCaptor.forClass(MailEvent.class);
        verify(mailEventService).save(captor.capture());

        assertThat(captor.getValue().getMessageBody(), is(content + "* Project: Project\n* View forum in SciPro here: http://localhost:8080/forum/mail/project/?id=null"));
    }

    private User newUser(final long id) {
        User user = User.builder().firstName("Kalle").lastName("Tester").emailAddress("kalle@example.com").build();
        DomainObjects.injectId(user, id);
        return user;
    }
}