package se.su.dsv.scipro.group;

import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.core.IsCollectionContaining.hasItem;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class GroupServiceImplTest extends IntegrationTest {

    @Inject
    private GroupServiceImpl groupService;
    private User author;
    private User creator;
    private User reviewer;
    private Group group;
    private ProjectType projectType;

    @Before
    public void setUp() throws Exception {
        author = saveUser("author", "name", "author@example.com");
        creator = saveUser("supervisor", "sup", "supervisor@example.com");
        reviewer = saveUser("reviewer", "test", "reviewer@example.com");
        projectType = save(new ProjectType(DegreeLevel.FIRST_CYCLE, "b", "b"));
        User headSupervisor = save(User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build());
        Project project = save(Project.builder().title("title").projectType(projectType).headSupervisor(headSupervisor).build());
        project.addProjectParticipant(author);
        project.addReviewer(reviewer);
        createGroup(project);
    }

    @Test
    public void find_by_author() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        filter.setAuthor(author);
        assertGroup(filter);
    }

    @Test
    public void find_all() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        assertGroup(filter);
    }

    @Test
    public void find_by_creator_fails() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        filter.setCreator(author);
        assertTrue(groupService.findAll(filter).isEmpty());
    }

    @Test
    public void find_by_creator() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        filter.setCreator(creator);
        assertGroup(filter);
    }

    @Test
    public void find_by_non_author_involvee() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        filter.setNonAuthorInvolvee(reviewer);
        assertGroup(filter);
    }

    @Test
    public void find_active_groups() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        filter.setActive(true);
        assertGroup(filter);

        filter.setActive(false);
        assertTrue(groupService.findAll(filter).isEmpty());
    }

    @Test
    public void count() throws Exception {
        GroupService.Filter filter = new GroupService.Filter();
        filter.setAuthor(author);
        assertEquals(1, groupService.count(filter));
    }

    @Test
    public void find_groups_by_project() throws Exception {
        Group another = new Group();
        another.setTitle("hi");
        another.setUser(creator);
        User headSupervisor = save(User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build());
        Project anotherProject = save(Project.builder().title("another").projectType(projectType).headSupervisor(headSupervisor).build());
        another.setProjects(new HashSet<>(Collections.singletonList(anotherProject)));
        save(another);
        GroupService.Filter filter = new GroupService.Filter();
        filter.setProject(anotherProject);
        assertEquals(another, groupService.findAll(filter).get(0));
    }

    private void assertGroup(GroupService.Filter filter) {
        assertThat(groupService.findAll(filter), hasItem(group));
    }

    private void createGroup(Project project) {
        group = new Group();
        group.setTitle("group title");
        group.setProjects(new HashSet<>(Collections.singletonList(project)));
        group.setUser(creator);
        save(group);
    }

    private User saveUser(String firstName, String lastName, String emailAddress) {
        return save(User.builder().firstName(firstName).lastName(lastName).emailAddress(emailAddress).build());
    }
}
