package se.su.dsv.scipro.mail;

import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.TypeSafeMatcher;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.runners.MockitoJUnitRunner;
import org.mockito.stubbing.Answer;
import org.slf4j.Logger;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;

import javax.mail.MessagingException;
import java.util.Arrays;
import java.util.Set;
import java.util.TreeSet;

import static org.hamcrest.CoreMatchers.everyItem;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class MailEventWorkerTest {

    @Before
    public void beforeEachTest() {
        worker = new TestMailEventWorker();
        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(systemSettings);
        when(systemSettings.isMailNotifications()).thenReturn(true);
    }

    @Test
    public void callSendMailUsersOnly() throws Exception {
        setUpMailEventWithUsersOnly();
        worker.doWork();
        verify(mailer).mail(eq(FROM_NAME), eq(FROM_EMAIL), eq(RECIPIENTS_MAILS), eq(SUBJECT), eq(MESSAGE_BODY), any(FileDescription.class));
    }

    @Test
    public void callSendMailUsersOnlyNoFromName() throws Exception {
        setUpMailEventWithUsersOnlyNoFromName();
        worker.doWork();
        verify(mailer).mail(eq(DEFAULT_FROM_NAME), eq(DEFAULT_FROM_EMAIL), eq(RECIPIENTS_MAILS), eq(SUBJECT), eq(MESSAGE_BODY), any(FileDescription.class));
    }

    @Test
    public void callSendMailUsersAndNonUser() throws Exception {
        setUpMailEventWithUsersAndNonUser();
        worker.doWork();
        String[] mails = {RECIPIENTS_MAILS[0], NON_USER_RECIPIENTS.iterator().next()};
        verify(mailer).mail(anyString(), anyString(), eq(mails), anyString(), anyString(), any(FileDescription.class));
    }

    @Test
    public void UseOneTransactionPerMailEvent() throws Exception {
        setUpTwoMailEvents();

        worker.doWork();

        assertEquals(2 + 1, worker.commits);
        assertEquals(2 + 1, worker.begins);
    }

    @Test
    public void marks_all_mail_as_sent() throws Exception {
        setUpTwoMailEvents();
        MailResult result = createSuccessfulResult("1234.567.890@mail.com");
        mockMailResult(result);

        worker.doWork();

        ArgumentCaptor<MailEvent> captor = ArgumentCaptor.forClass(MailEvent.class);
        verify(mailEventService, times(2)).save(captor.capture());

        assertThat(captor.getAllValues(), everyItem(isSent(result.getMessageID())));
    }

    @Test
    public void deletes_mails_without_recipients() throws Exception {
        MailEvent mailEvent = new MailEvent();
        DomainObjects.injectId(mailEvent, 1L);
        mockMailEventService(mailEvent);

        worker.doWork();

        verify(mailEventService).delete(mailEvent);
    }

    @Test
    public void logs_result_of_each_send() throws Exception {
        setUpTwoMailEvents();
        MailResult result = createSuccessfulResult("1234.567.890@mail.com");
        mockMailResult(result);

        worker.doWork();

        verify(result, times(2)).log(isA(Logger.class));
    }

    private static Matcher<MailEvent> isSent(final String messageID) {
        return new TypeSafeMatcher<MailEvent>() {
            @Override
            protected boolean matchesSafely(MailEvent item) {
                return messageID.equals(item.getMessageID());
            }

            @Override
            public void describeTo(Description description) {
                description.appendText("sent with Message-ID ").appendValue(messageID);
            }
        };
    }

    private MailResult createSuccessfulResult(final String messageID) {
        MailResult result = mock(MailResult.class);
        when(result.getMessageID()).thenReturn(messageID);
        return result;
    }

    private void mockMailResult(final MailResult result) throws MessagingException {
        when(mailer.mail(anyString(), anyString(), any(String[].class), anyString(), anyString(), any(FileDescription.class))).thenReturn(result);
    }

    private void setUpMailEventWithUsersAndNonUser() throws Exception {
        MailEvent mailEvent = createMailEventUsersOnlyNoFromName();
        mailEvent.setNonUserRecipients(NON_USER_RECIPIENTS);
        mockMailEventService(mailEvent);
    }

    private void setUpMailEventWithUsersOnlyNoFromName() throws Exception {
        MailEvent mailEvent = createMailEventUsersOnlyNoFromName();
        mockMailEventService(mailEvent);
        when(systemSettings.getMailFromName()).thenReturn(DEFAULT_FROM_NAME);
        when(systemSettings.getSystemFromMail()).thenReturn(DEFAULT_FROM_EMAIL);
    }

    private void setUpMailEventWithUsersOnly() throws Exception {
        MailEvent mailEvent = createMailEventUsersOnlyNoFromName();
        mailEvent.setFromName(FROM_NAME);
        mailEvent.setFromEmail(FROM_EMAIL);
        mockMailEventService(mailEvent);
    }

    private void setUpTwoMailEvents() throws Exception {
        mockMailEventService(createMailEventUsersOnlyNoFromName(), createMailEventUsersOnlyNoFromName());
    }

    private void mockMailEventService(final MailEvent... mailEvents) throws Exception {
        when(mailEventService.findUnsent()).thenReturn(Arrays.asList(mailEvents));
        when(mailEventService.findOne(anyLong())).thenAnswer(new Answer<MailEvent>() {

            @Override
            public MailEvent answer(InvocationOnMock invocation) throws Throwable {
                Long id = (Long) invocation.getArguments()[0];
                for(MailEvent event : mailEvents) {
                    if(event.getId().equals(id)) {
                        return event;
                    }
                }
                return null;            }
        });
        MailResult successfulResult = createSuccessfulResult("123.abcdefg.987@mail.com");
        when(mailer.mail(anyString(), anyString(), any(String[].class), anyString(), anyString(), any(FileDescription.class))).thenReturn(successfulResult);
    }

    private MailEvent createMailEventUsersOnlyNoFromName() {
        MailEvent mailEvent = new MailEvent() {
            {
                setMessageBody(MESSAGE_BODY);
                setRecipients(RECIPIENTS);
                setSubject(SUBJECT);
            }
        };
        DomainObjects.injectId(mailEvent, mailEventsId[eventIndex++]);
        return mailEvent;
    }

    class TestMailEventWorker extends MailEventWorker {
        private int commits = 0;
        public int begins = 0;

        TestMailEventWorker() {
            super(mailEventService, generalSystemSettingsService, mailer);
        }

        @Override
        protected void beginTransaction() {
            begins++;
        }

        @Override
        protected void commitTransaction() {
            commits++;
        }
    }

    private static final Long MAIL_EVENT_ID1 = 999L;
    private static final Long MAIL_EVENT_ID2 = 1337L;
    private static final Long[] mailEventsId = {
            MAIL_EVENT_ID1,
            MAIL_EVENT_ID2
    };
    private int eventIndex = 0;
    private static final String FROM_EMAIL = "fromEmail@example.com";
    private static final String FROM_NAME = "From Name";
    private static final String MESSAGE_BODY = "This is the message body";
    private static final User RECIPIENT_1 = new User() {
        {
            setFirstName("Recipient Firstname");
            setLastName("Recipient Lastname");
            setEmailAddress("recepient@example.com");
        }
    };

    static {
        DomainObjects.injectId(RECIPIENT_1, 42L);
    }

    private static final Set<User> RECIPIENTS = new TreeSet<User>(new User.ByNameComparator()) {
        {
            add(RECIPIENT_1);
        }
    };
    private static final String[] RECIPIENTS_MAILS = {RECIPIENT_1.getEmailAddress()};
    private static final User REPLY_USER_1 = new User() {
        {
            setFirstName("Reply Firstname");
            setLastName("Reply Lastname");
            setEmailAddress("replyto@example.com");
        }
    };

    static {
        DomainObjects.injectId(REPLY_USER_1, 4711L);
    }

    private static final String SUBJECT = "subject of the mail";
    private static final String NON_USER1 = "nonUser1@otherplace.it";
    private static final Set<String> NON_USER_RECIPIENTS = new TreeSet<String>() {
        {
            add(NON_USER1);
        }
    };
    private static final String DEFAULT_FROM_NAME = "Default From Name";
    private static final String DEFAULT_FROM_EMAIL = "default@example.com";

    private TestMailEventWorker worker;
    @Mock
    private GeneralSystemSettingsService generalSystemSettingsService;
    @Mock
    private Mailer mailer;
    @Mock
    private MailEventService mailEventService;
    @Mock
    private GeneralSystemSettings systemSettings;

}
