package se.su.dsv.scipro.match;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.*;
import org.mockito.runners.MockitoJUnitRunner;
import se.su.dsv.scipro.date.DateService;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.dataobject.IdeaEvent;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.system.Program;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.InstanceProvider;
import se.su.dsv.scipro.test.UserBuilder;
import se.su.dsv.scipro.util.Pair;

import javax.persistence.EntityManager;
import java.util.*;

import static org.junit.Assert.assertEquals;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.when;
import static se.su.dsv.scipro.match.IdeaServiceImpl.*;

@RunWith(MockitoJUnitRunner.class)
public class IdeaServiceImplTest {

    @Mock
    private NotificationController notificationController;
    @Mock
    private ApplicationPeriodService applicationPeriodService;
    @Mock
    private MatchService matchService;
    @Mock
    private GeneralSystemSettingsService generalSystemSettingsService;
    @Mock
    private ProjectService projectService;
    @Mock
    private DateService dateService;
    @Mock
    private FirstMeetingRepository firstMeetingRepository;
    @Mock(answer = Answers.RETURNS_MOCKS)
    private EntityManager em;

    IdeaServiceImpl ideaService;
    private User student;
    private User coAuthor;
    private User creator;
    private User supervisor;
    private ApplicationPeriod applicationPeriod;
    private ProjectType bachelor;
    private ProjectType master;

    @Before
    public void setUp() {
        ideaService = new IdeaServiceImpl(new InstanceProvider<>(em), applicationPeriodService, firstMeetingRepository, notificationController, dateService);

        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(new GeneralSystemSettings());

        student = new UserBuilder().create();
        coAuthor = new UserBuilder().create();
        creator = new UserBuilder().create();
        supervisor = new UserBuilder().create();
        applicationPeriod = new ApplicationPeriod();

        bachelor = new ProjectType(ProjectType.BACHELOR, "Bachelor", "Bachelor");
        bachelor.getProjectTypeSettings().setMinAuthors(2);

        master = new ProjectType(ProjectType.MASTER, "Master", "Master");
        master.getProjectTypeSettings().setMaxAuthors(1);
    }

    @Test
    public void testSaveStudentIdeaNotifiesIfCoAuthorOnNewIdea() {
        // given
        Idea idea = new Idea();
        idea.setProjectType(master);
        GeneralSystemSettings settings = DomainObjects.injectId(new GeneralSystemSettings(), 1L);

        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(settings);

        // when
        ideaService.saveStudentIdea(idea, creator, null, coAuthor, new ArrayList<Keyword>(), true);

        // then
        Mockito.verify(notificationController).notifyIdea(eq(idea), eq(IdeaEvent.Event.ADDED_AS_PARTNER), Matchers.isA(NotificationSource.class));
    }

    @Test
    public void testSaveStudentIdeaDoesNotNotifyIfNoNewCoAuthorOnExistingIdea() {
        // given
        Idea idea = new Idea();
        idea.setProjectType(bachelor);

        IdeaParticipation participation = new IdeaParticipation(creator, idea);

        GeneralSystemSettings settings = DomainObjects.injectId(new GeneralSystemSettings(), 1L);

        idea.addIdeaParticipation(participation);

        IdeaParticipation coParticipation = new IdeaParticipation(coAuthor, idea);

        idea.addIdeaParticipation(coParticipation);

        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(settings);

        // when
        ideaService.saveStudentIdea(idea, creator, null, coAuthor, new ArrayList<Keyword>(), false);

        // then
        Mockito.verify(notificationController, Mockito.times(0)).notifyIdea(Matchers.isA(Idea.class), Matchers.isA(IdeaEvent.Event.class), Matchers.isA(NotificationSource.class));
    }

    @Test
    public void testSaveStudentIdeaNotifiesIfNewCoAuthorOnExistingIdea() {
        // given
        Idea idea = new Idea();
        idea.setProjectType(bachelor);

        IdeaParticipation participation = new IdeaParticipation(creator, idea);

        GeneralSystemSettings settings = DomainObjects.injectId(new GeneralSystemSettings(), 1L);

        idea.addIdeaParticipation(participation);

        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(settings);

        // when
        ideaService.saveStudentIdea(idea, creator, null, coAuthor, new ArrayList<Keyword>(), false);

        // then
        Mockito.verify(notificationController).notifyIdea(eq(idea), eq(IdeaEvent.Event.ADDED_AS_PARTNER), Matchers.isA(NotificationSource.class));
    }

    @Test
    public void validateSupervisorAcceptanceAlreadyMatched() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.MATCHED);
        assertValidateSupervisorAcceptance(idea, ALREADY_MATCHED_ERROR, false, true);
    }

    @Test
    public void validateSupervisorAcceptanceNoTargetsLeft() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.UNMATCHED);
        assertValidateSupervisorAcceptance(idea, TARGET_NUMBERS_REACHED_ERROR, false, false);
    }

    @Test
    public void validateSupervisorAcceptance() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.UNMATCHED);
        assertValidateSupervisorAcceptance(idea, SUPERVISING_CONFIRM + idea.getTitle(), true, true);
    }

    @Test
    public void validateStudentAcceptanceNoLongerAvailable() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.MATCHED);

        assertPair(false, NO_LONGER_AVAILABLE_ERROR, ideaService.validateStudentAcceptance(idea, student, coAuthor, applicationPeriod));
    }

    @Test
    public void validateStudentAcceptanceNoPartner() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.UNMATCHED);

        when(applicationPeriodService.hasExemption(student, applicationPeriod)).thenReturn(false);

        assertPair(false, BACHELOR_NEED_PARTNER_ERROR, ideaService.validateStudentAcceptance(idea, student, null, applicationPeriod));

    }

    @Test
    public void validateStudentAcceptanceSelfPartner() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.UNMATCHED);

        assertPair(false, ADD_SELF_AS_PARTNER_ERROR, ideaService.validateStudentAcceptance(idea, student, student, applicationPeriod));
    }

    @Test
    public void validateStudentAcceptance() throws Exception {
        Idea idea = createBachelorIdea(Idea.Status.UNMATCHED);

        assertPair(true, SELECTED_IDEA + idea.getTitle(), ideaService.validateStudentAcceptance(idea, student, coAuthor, applicationPeriod));
    }

    @Test
    public void validateAdminAddAuthors() throws Exception {
        Idea idea = new Idea();
        idea.setType(Idea.Type.SUPERVISOR);

        assertPair(true, VALIDATION_PASSED, ideaService.validateAdminAddAuthors(idea, new TreeSet<>(new User.ByNameComparator())));
    }

    @Test
    public void validateAdminAddAuthorsNoAuthors() throws Exception {
        Idea idea = new Idea();
        idea.setType(Idea.Type.STUDENT);

        assertPair(false, NO_AUTHORS_ERROR, ideaService.validateAdminAddAuthors(idea, new TreeSet<>(new User.ByNameComparator())));
    }

    @Test
    public void validateAdminAddAuthorsHasAuthors() throws Exception {
        Idea idea = new Idea();
        ProjectType projectType = master;
        idea.setProjectType(projectType);

        when(applicationPeriodService.getCurrentPeriod(eq(projectType))).thenReturn(new ApplicationPeriod());

        TreeSet<User> authors = new TreeSet<>(new User.ByNameComparator());
        authors.addAll(Collections.singletonList(student));
        assertPair(true, VALIDATION_PASSED, ideaService.validateAdminAddAuthors(idea, authors));
    }

    @Test
    public void saveSupervisorIdeaNewIdea() throws Exception {
        Idea idea = new Idea();

        assertEquals(idea, ideaService.saveSupervisorIdea(idea, supervisor, Collections.singletonList(new Keyword()), true));
    }

    @Test
    public void saveSupervisorIdeaOldIdea() throws Exception {
        Idea idea = new Idea();
        Match match = new Match();
        match.setStatus(Idea.Status.UNMATCHED);
        idea.setMatch(match);

        assertEquals(idea, ideaService.saveSupervisorIdea(idea, supervisor, Collections.singletonList(new Keyword()), false));
    }

    @Test
    public void acceptIdea() throws Exception {
        Idea idea = new Idea();
        idea.setMatch(new Match());
        GeneralSystemSettings generalSystemSettings = new GeneralSystemSettings();

        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(generalSystemSettings);

        assertEquals(idea, ideaService.acceptIdea(idea, student, null, coAuthor, applicationPeriod));
        Mockito.verify(notificationController).notifyIdea(eq(idea), eq(IdeaEvent.Event.PARTNER_ACCEPT), any(NotificationSource.class));
    }

    @Test
    public void savePartnerProgram(){
        Program program = new Program();
        DomainObjects.injectId(program, 1L);
        User student = new UserBuilder().create();
        student.getPrograms().add(program);
        Idea idea = new Idea();
        IdeaParticipation participation = new IdeaParticipation(student, idea);
        idea.getIdeaParticipations().add(participation);

        idea = ideaService.savePartnerProgram(idea, student, program);

        for (IdeaParticipation npi : idea.getIdeaParticipations()){
            Assert.assertEquals(program, npi.getProgram());
        }
    }

    @Test
    public void inactivate_idea_and_create_match_object() throws Exception {
        Idea idea = mockInactiveIdea();
        User matchCreator = new UserBuilder().create();

        ideaService.inactivateIdea(idea, true, matchCreator);

        ArgumentCaptor<Idea> captor = ArgumentCaptor.forClass(Idea.class);
        Mockito.verify(em).persist(captor.capture());

        Assert.assertTrue(captor.getValue().isInactive());
        Assert.assertEquals(matchCreator, captor.getValue().getMatch().getChangedBy());
    }

    @Test
    public void bachelor_idea_must_have_two_authors() throws Exception {
        when(applicationPeriodService.getCurrentPeriod(any(ProjectType.class))).thenReturn(applicationPeriod);
        SortedSet<User> authors = new TreeSet<>(new User.ByNameComparator());
        authors.add(student);
        assertPair(false, "Selected number of authors is not allowed", ideaService.validateAdminAddAuthors(createBachelorIdea(Idea.Status.UNMATCHED), authors));
    }

    @Test
    public void master_idea_cant_have_two_authors() throws Exception {
        when(applicationPeriodService.getCurrentPeriod(any(ProjectType.class))).thenReturn(applicationPeriod);
        SortedSet<User> authors = new TreeSet<>(new User.ByNameComparator());
        authors.add(student);
        authors.add(coAuthor);
        assertPair(false, "Selected number of authors is not allowed", ideaService.validateAdminAddAuthors(createMasterIdea(Idea.Status.UNMATCHED), authors));
    }

    @Test
    public void bachelor_idea_must_have_two_authors_unless_the_first_one_has_an_exemption() throws Exception {
        when(applicationPeriodService.getCurrentPeriod(any(ProjectType.class))).thenReturn(applicationPeriod);
        when(applicationPeriodService.hasExemption(student, applicationPeriod)).thenReturn(true);
        SortedSet<User> authors = new TreeSet<>(new User.ByNameComparator());
        authors.add(student);
        assertPair(true, VALIDATION_PASSED, ideaService.validateAdminAddAuthors(createBachelorIdea(Idea.Status.UNMATCHED), authors));
    }

    @Test
    public void count_authors_by_application_period_no_status_selected() throws Exception {
        Long expected = 0L;
        Filter params = new Filter();
        params.setStatuses(new ArrayList<Idea.Status>());
        assertEquals(expected, ideaService.countAuthorsByApplicationPeriod(applicationPeriod, params));
    }

    private Idea mockInactiveIdea() {
        Idea idea = new Idea();
        Match match = new Match();
        match.setSupervisor(new UserBuilder().create());
        idea.setMatch(match);
        idea.setInactive(false);
        return idea;
    }

    private Idea createBachelorIdea(Idea.Status status) {
        return createIdea(status, bachelor);
    }

    private Idea createMasterIdea(Idea.Status status) {
        return createIdea(status, master);
    }

    private Idea createIdea(Idea.Status status, ProjectType projectType) {
        Idea idea = new Idea();
        idea.setTitle("idea");
        Match match = new Match();
        match.setStatus(status);
        idea.setMatch(match);
        if (status == Idea.Status.MATCHED) {
            match.setSupervisor(supervisor);
            idea.addIdeaParticipation(new IdeaParticipation(student, idea));
        }
        ApplicationPeriod applicationPeriod1 = new ApplicationPeriod();
        applicationPeriod1.setCourseStartDate(new Date());
        idea.setApplicationPeriod(applicationPeriod1);
        idea.setProjectType(projectType);
        return idea;
    }

    private void assertValidateSupervisorAcceptance(Idea idea, String error, boolean success, boolean hasTargetsLeft) {
        assertPair(success, error, ideaService.validateSupervisorAcceptance(idea, hasTargetsLeft));
    }

    private <H, T> void assertPair(T head, H tail, Pair<T, H> actual) {
        Assert.assertEquals(tail, actual.getTail());
        Assert.assertEquals(head, actual.getHead());
    }

}
