package se.su.dsv.scipro.match;

import com.google.common.collect.ImmutableSet;
import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.TypeSafeMatcher;
import org.junit.Before;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import java.util.Collections;
import java.util.Date;

import static org.hamcrest.CoreMatchers.*;
import static org.hamcrest.MatcherAssert.assertThat;

public class MatchFollowUpServiceTest extends MatchModuleTest {
    public static final PageRequest ALL = new PageRequest(0, Integer.MAX_VALUE, new Sort(Sort.Direction.DESC, "target"));
    @Inject
    private MatchFollowUpService matchFollowUpService;

    private ApplicationPeriod applicationPeriod;
    private ProjectType bachelor;
    private ProjectType master;
    private User bob;
    private User charles;

    @Before
    public void setUp() throws Exception {
        bachelor = save(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"));
        master = save(new ProjectType(DegreeLevel.SECOND_CYCLE, "Master", "Master"));
        ApplicationPeriod applicationPeriod = new ApplicationPeriod("Test period");
        applicationPeriod.setProjectTypes(ImmutableSet.of(bachelor, master));
        applicationPeriod.setStartDate(new Date());
        applicationPeriod.setEndDate(new Date());
        applicationPeriod.setCourseStartDate(new Date());
        this.applicationPeriod = save(applicationPeriod);
        bob = save(User.builder().firstName("Bob").lastName("Supervisor").emailAddress("bob@example.com").roles(Collections.singleton(Roles.SUPERVISOR)).build());
        charles = save(User.builder().firstName("Charles").lastName("Supervisor").emailAddress("charles@example.com").roles(Collections.singleton(Roles.SUPERVISOR)).build());
    }

    @Test
    public void count_unmet_targets() throws Exception {
        giveTargetsTo(bob, bachelor, 2);
        giveTargetsTo(charles, master, 2);

        assertThat(matchFollowUpService.countUnmetTargets(applicationPeriod), is(2L));
    }

    @Test
    public void count_unmet_targets_with_submitted_ideas() throws Exception {
        giveTargetsTo(bob, bachelor, 2);
        submitIdeas(bob, bachelor, 2);

        assertThat(matchFollowUpService.countUnmetTargets(applicationPeriod), is(0L));
    }

    @Test
    public void find_unmet_targets() throws Exception {
        giveTargetsTo(bob, bachelor, 2);

        assertThat(matchFollowUpService.findUnmetTargets(applicationPeriod, ALL), hasItem(forSupervisor(bob)));
        assertThat(matchFollowUpService.findUnmetTargets(applicationPeriod, ALL), not(hasItem(forSupervisor(charles))));
    }

    private static Matcher<? super TargetResult> forSupervisor(final User supervisor) {
        return new TypeSafeMatcher<TargetResult>() {
            @Override
            protected boolean matchesSafely(final TargetResult item) {
                return supervisor.equals(item.getTarget().getUser());
            }

            @Override
            public void describeTo(final Description description) {
                description.appendText("an unmet target result for ").appendValue(supervisor);
            }
        };
    }

    private void submitIdeas(final User supervisor, final ProjectType projectType, final int numberOfIdeas) {
        for (int i = 0; i < numberOfIdeas; i++) {
            Idea idea = new Idea();
            idea.setType(Idea.Type.SUPERVISOR);
            idea.setPublished(true);
            idea.setProjectType(projectType);
            idea.setCreator(supervisor);
            idea.setTitle("Idea #" + i);

            Match match = new Match();
            match.setIdea(idea);
            match.setSupervisor(supervisor);
            match.setStatus(Idea.Status.UNMATCHED);

            idea.setMatch(match);
            save(idea);
        }
    }

    private void giveTargetsTo(final User supervisor, final ProjectType projectType, final int target) {
        Target targetEntity = new Target(supervisor, applicationPeriod, projectType);
        targetEntity.setTarget(target);
        save(targetEntity);
    }
}