package se.su.dsv.scipro.milestones.service.impl;

import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.invocation.InvocationOnMock;
import org.mockito.runners.MockitoJUnitRunner;
import org.mockito.stubbing.Answer;
import se.su.dsv.scipro.milestones.MilestoneActivityTemplateRepository;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.milestones.dataobjects.MilestonePhaseTemplate;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.test.DomainObjects;

import java.util.*;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class MilestoneActivityTemplateServiceImplTest {

    @Mock
    MilestoneActivityTemplateRepository milestoneActivityTemplateRepository;
    @InjectMocks
    MilestoneActivityTemplateServiceImpl mileStoneActivityService;
    private Date date;
    private ProjectType projectType;

    @Before
    public void setUp() throws Exception {
        when(milestoneActivityTemplateRepository.save(isA(MilestoneActivityTemplate.class))).thenAnswer(new Answer<Object>() {
            @Override
            public Object answer(final InvocationOnMock invocation) throws Throwable {
                return invocation.getArguments()[0];
            }
        });
        date = new DateTime().minusDays(1).toDate();
        projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "any", "any");
    }

    @Test
    public void testGetActivities() throws Exception {
        MilestoneActivityTemplate projectActivity = createActivity(MilestoneActivityTemplate.Type.PROJECT, "Project milestone", "This is it..", 0);
        List<MilestoneActivityTemplate> projectList = Collections.singletonList(projectActivity);
        MilestoneActivityTemplate studentActivity = createActivity(MilestoneActivityTemplate.Type.STUDENT, "Student milestone", "Hi?", 0);
        List<MilestoneActivityTemplate> studentList = Collections.singletonList(studentActivity);

        when(milestoneActivityTemplateRepository.findActivities(eq(MilestoneActivityTemplate.Type.PROJECT), isA(ProjectType.class), eq(date))).thenReturn(projectList);
        when(milestoneActivityTemplateRepository.findActivities(eq(MilestoneActivityTemplate.Type.STUDENT), isA(ProjectType.class), eq(date))).thenReturn(studentList);

        List<MilestoneActivityTemplate> studentActivities = mileStoneActivityService.getActivities(MilestoneActivityTemplate.Type.STUDENT, mock(ProjectType.class), date);
        List<MilestoneActivityTemplate> projectActivities = mileStoneActivityService.getActivities(MilestoneActivityTemplate.Type.PROJECT, mock(ProjectType.class), date);

        assertEquals(projectList, projectActivities);
        assertEquals(studentList, studentActivities);
    }

    @Test
    public void testGetAllActivities() throws Exception {
        MilestoneActivityTemplate projectActivity = createActivity(MilestoneActivityTemplate.Type.PROJECT, "Project milestone", "This is it..", 0);
        MilestoneActivityTemplate projectActivity2 = createActivity(MilestoneActivityTemplate.Type.PROJECT, "Project milestone", "This is it..", 1);
        projectActivity2.setDeleted(true);
        List<MilestoneActivityTemplate> activeList = Collections.singletonList(projectActivity);
        List<MilestoneActivityTemplate> listOfAll = Arrays.asList(projectActivity, projectActivity2);

        when(milestoneActivityTemplateRepository.findActivities(eq(MilestoneActivityTemplate.Type.PROJECT), isA(ProjectType.class), eq(date))).thenReturn(activeList);
        when(milestoneActivityTemplateRepository.findAllActivitiesByProjectType(eq(MilestoneActivityTemplate.Type.PROJECT), eq(Collections.singletonList(projectType)))).thenReturn(listOfAll);

        List<ProjectType> projectTypes = new ArrayList<>();
        projectTypes.add(projectType);
        List<MilestoneActivityTemplate> all = mileStoneActivityService.getAllActivities(MilestoneActivityTemplate.Type.PROJECT, projectTypes);

        assertEquals(all, listOfAll);
    }

    @Test
    public void testSave() throws Exception {
        String title = "title";
        String desc = "desc";
        MilestoneActivityTemplate activity = createActivity(MilestoneActivityTemplate.Type.PROJECT, title, desc, 0);
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");

        mileStoneActivityService.save(activity, phase);
        verify(milestoneActivityTemplateRepository, times(1)).save(eq(activity));

        ArgumentCaptor<MilestoneActivityTemplate> captor = ArgumentCaptor.forClass(MilestoneActivityTemplate.class);
        verify(milestoneActivityTemplateRepository).save(captor.capture());

        assertEquals(captor.getValue().getTitle(), title);
        assertEquals(captor.getValue().getDescription(), desc);
        assertEquals(captor.getValue().getType(), MilestoneActivityTemplate.Type.PROJECT);
        assertEquals(captor.getValue().getSortOrder(), 0);
    }

    @Test
    public void testEditActivityShouldNotCareAboutSortOrder() throws Exception {
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        MilestoneActivityTemplate activity = createActivity(MilestoneActivityTemplate.Type.PROJECT, "title", "desc", 0);
        DomainObjects.injectId(activity, 1L);
        activity.setSortOrder(3);
        activity.setMilestonePhaseTemplate(phase);

        mileStoneActivityService.save(activity, phase);

        ArgumentCaptor<MilestoneActivityTemplate> captor = ArgumentCaptor.forClass(MilestoneActivityTemplate.class);
        verify(milestoneActivityTemplateRepository).save(captor.capture());

        assertEquals(captor.getValue().getSortOrder(), 3);
    }

    @Test
    public void testActivityWithZeroIdGetsSortOrderByCount() throws Exception {
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        MilestoneActivityTemplate activity = createActivity(MilestoneActivityTemplate.Type.PROJECT, "title", "desc", 0);
        DomainObjects.injectId(activity, 0L);
        activity.setMilestonePhaseTemplate(phase);

        when(milestoneActivityTemplateRepository.countByMilestonePhase(eq(phase))).thenReturn(4L);

        mileStoneActivityService.save(activity, phase);

        ArgumentCaptor<MilestoneActivityTemplate> captor = ArgumentCaptor.forClass(MilestoneActivityTemplate.class);
        verify(milestoneActivityTemplateRepository).save(captor.capture());

        assertEquals(captor.getValue().getSortOrder(), 4);
    }

    @Test
    public void testActivityWithChangedPhaseGetsSortOrderByCount() throws Exception {
        MilestonePhaseTemplate currentPhase = new MilestonePhaseTemplate("title", "desc");
        DomainObjects.injectId(currentPhase, 2L);
        MilestonePhaseTemplate newPhase = new MilestonePhaseTemplate("new title", "new desc");
        DomainObjects.injectId(newPhase, 3L);
        MilestoneActivityTemplate activity = createActivity(MilestoneActivityTemplate.Type.PROJECT, "title", "desc", 0);
        DomainObjects.injectId(activity, 1L);
        activity.setSortOrder(3);
        activity.setMilestonePhaseTemplate(newPhase);

        when(milestoneActivityTemplateRepository.countByMilestonePhase(eq(newPhase))).thenReturn(1L);

        mileStoneActivityService.save(activity, currentPhase);

        ArgumentCaptor<MilestoneActivityTemplate> captor = ArgumentCaptor.forClass(MilestoneActivityTemplate.class);
        verify(milestoneActivityTemplateRepository).save(captor.capture());

        assertEquals(captor.getValue().getSortOrder(), 1);
    }

    @Test
    public void testFindActivitiesByPhase() throws Exception {
        MilestoneActivityTemplate createSeminar = createActivity(MilestoneActivityTemplate.Type.PROJECT, "Create seminar", "description", 0);
        List<MilestoneActivityTemplate> list = Collections.singletonList(createSeminar);
        when(milestoneActivityTemplateRepository.findByMilestonePhase(isA(MilestonePhaseTemplate.class))).thenReturn(list);

        List<MilestoneActivityTemplate> serviceList = mileStoneActivityService.findByMileStonePhase(mock(MilestonePhaseTemplate.class));

        assertEquals(list, serviceList);
    }

    @Test
    public void testFindActiveActivitiesByPhase() throws Exception {
        List<MilestoneActivityTemplate> list = Collections.singletonList(createActivity(MilestoneActivityTemplate.Type.PROJECT, "Create seminar", "description", 0));
        Date date = new DateTime().minusDays(1).toDate();
        ProjectType bachelor = new ProjectType(DegreeLevel.FIRST_CYCLE, "bachelor", "bachelor");
        Project project = Project.builder().title("title").projectType(bachelor).build();
        project.setDateCreated(date);
        when(milestoneActivityTemplateRepository.findActiveByPhase(isA(MilestonePhaseTemplate.class), eq(bachelor), eq(date))).thenReturn(list);

        List<MilestoneActivityTemplate> serviceList = mileStoneActivityService.findActiveByPhase(mock(MilestonePhaseTemplate.class), project);

        assertEquals(list, serviceList);
    }

    @Test
    public void testCountActivitiesByTypeAndProjectTypeAndDate() throws Exception {
        Long activityCount = 3L;
        Date date = new DateTime().minusDays(1).toDate();
        when(milestoneActivityTemplateRepository.countActivities(eq(MilestoneActivityTemplate.Type.PROJECT), isA(ProjectType.class), eq(date))).thenReturn(activityCount);

        Long serviceCount = mileStoneActivityService.countActivities(MilestoneActivityTemplate.Type.PROJECT, mock(ProjectType.class), date);
        assertEquals(serviceCount, activityCount);
    }

    @Test
    public void countActivitiesByProjectTypeAndDate() throws Exception {
        ProjectType projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");
        Date projectStartDate = new DateTime().minusDays(1).toDate();
        long expected = 3L;

        when(milestoneActivityTemplateRepository.countActivities(projectType, projectStartDate)).thenReturn(expected);

        assertEquals(expected, mileStoneActivityService.countActivities(projectType, projectStartDate));
    }

    @Test
    public void testMoveUpPhase() throws Exception {
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        MilestoneActivityTemplate current = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "title", "desc");
        current.setMilestonePhaseTemplate(phase);
        MilestoneActivityTemplate downActivity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "title", "desc");
        downActivity.setMilestonePhaseTemplate(phase);
        current.setSortOrder(2);
        downActivity.setSortOrder(1);

        when(milestoneActivityTemplateRepository.findBySortOrderAndMilestonePhaseTemplate(eq(1), eq(phase))).thenReturn(downActivity);

        mileStoneActivityService.moveUp(current);

        Assert.assertEquals(current.getSortOrder(), 1);
        Assert.assertEquals(downActivity.getSortOrder(), 2);
    }

    @Test
    public void testMoveUpPhaseOnTopDoesNothing() throws Exception {
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        MilestoneActivityTemplate current = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "title", "desc");
        current.setMilestonePhaseTemplate(phase);
        current.setSortOrder(1);
        when(milestoneActivityTemplateRepository.findBySortOrderAndMilestonePhaseTemplate(eq(0), eq(phase))).thenReturn(null);
        mileStoneActivityService.moveUp(current);
        Assert.assertEquals(current.getSortOrder(), 1);
    }

    @Test
    public void testMoveDownPhase() throws Exception {
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        MilestoneActivityTemplate current = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "title", "desc");
        current.setMilestonePhaseTemplate(phase);
        MilestoneActivityTemplate upActivity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "title", "desc");
        upActivity.setMilestonePhaseTemplate(phase);
        current.setSortOrder(1);
        upActivity.setSortOrder(2);

        when(milestoneActivityTemplateRepository.findBySortOrderAndMilestonePhaseTemplate(eq(2), eq(phase))).thenReturn(upActivity);

        mileStoneActivityService.moveDown(current);

        Assert.assertEquals(current.getSortOrder(), 2);
        Assert.assertEquals(upActivity.getSortOrder(), 1);
    }

    @Test
    public void testMoveDownPhaseOnBottomDoesNothing() throws Exception {
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        MilestoneActivityTemplate current = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "title", "desc");
        current.setMilestonePhaseTemplate(phase);
        current.setSortOrder(1);
        when(milestoneActivityTemplateRepository.findBySortOrderAndMilestonePhaseTemplate(eq(2), eq(phase))).thenReturn(null);
        mileStoneActivityService.moveDown(current);
        Assert.assertEquals(current.getSortOrder(), 1);
    }

    @Test
    public void testGetActivitiesInOrder() throws Exception {
        List<MilestoneActivityTemplate> list = Collections.singletonList(new MilestoneActivityTemplate());
        Date dateCreated = new DateTime().minusDays(1).toDate();
        ProjectType projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "bachelor", "bachelor");
        Project project = Project.builder().title("title").projectType(projectType).build();
        project.setDateCreated(dateCreated);

        when(milestoneActivityTemplateRepository.findAllActivitiesInOrder(MilestoneActivityTemplate.Type.PROJECT, projectType, dateCreated)).thenReturn(list);
        List<MilestoneActivityTemplate> serviceList = mileStoneActivityService.getActivitiesInOrder(MilestoneActivityTemplate.Type.PROJECT, project);
        assertEquals(list, serviceList);
    }


    MilestoneActivityTemplate createActivity(MilestoneActivityTemplate.Type type, String title, String description, int order) {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate();
        activity.setType(type);
        activity.setTitle(title);
        activity.setDescription(description);
        activity.setMilestonePhaseTemplate(new MilestonePhaseTemplate("phase test", "this is for test", 0));
        activity.setProjectTypes(new HashSet<>(Collections.singletonList(projectType)));
        activity.setSortOrder(order);
        return activity;
    }
}
