package se.su.dsv.scipro.notifications;

import com.google.inject.util.Providers;
import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.TypeSafeMatcher;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import se.su.dsv.scipro.data.dataobjects.Member;
import se.su.dsv.scipro.finalseminar.FinalSeminar;
import se.su.dsv.scipro.group.Group;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.match.Idea;
import se.su.dsv.scipro.notifications.dataobject.*;
import se.su.dsv.scipro.notifications.interfaces.NotificationMailFormatter;
import se.su.dsv.scipro.notifications.settings.entities.DeliveryMethod;
import se.su.dsv.scipro.notifications.settings.service.DeliveryConfigurationService;
import se.su.dsv.scipro.notifications.settings.service.ReceiverConfigurationService;
import se.su.dsv.scipro.peer.PeerRequest;
import se.su.dsv.scipro.peer.PeerReview;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.CurrentUser;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import static org.hamcrest.CoreMatchers.everyItem;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class NotificationControllerImplTest {

    @Mock MailEventService mailEventService;
    @Mock NotificationService notificationService;
    @Mock NotificationMailFormatter mailFormatter;
    @Mock ReceiverConfigurationService receiverConfiguration;
    @Mock DeliveryConfigurationService deliveryConfiguration;
    @Mock CurrentUser currentUser;

    NotificationControllerImpl notificationController;

    @Before
    public void setup() {
        notificationController = new NotificationControllerImpl(
                notificationService,
                mailFormatter,
                mailEventService,
                receiverConfiguration, deliveryConfiguration, Providers.of(currentUser));

        when(mailFormatter.format(isA(Notification.class))).thenReturn(new NotificationMail("Subject", "Body"));
        when(receiverConfiguration.isReceiving(isA(Notification.Type.class), isA(Enum.class), isA(Member.Type.class))).thenReturn(true);
        when(deliveryConfiguration.isDelivery(isA(Notification.Type.class), isA(Enum.class), isA(DeliveryMethod.class))).thenReturn(true);
        when(deliveryConfiguration.isDelivery(isA(Notification.Type.class), isA(Enum.class), isA(DeliveryMethod.class), isA(User.class))).thenReturn(true);
    }

    @Test
    public void testCorrectProjectReceivers() throws Exception {
        // given
        Project project = mock(Project.class);

        // when
        when(project.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.AUTHOR);
        notificationController.notifyProject(project, ProjectEvent.Event.CREATED, new NotificationSource());

        // then
        verify(notificationService, times(1)).save(isA(Notification.class));
    }

    @Test
    public void testCorrectSeminarReceivers() throws Exception {
        // given
        FinalSeminar seminar = mock(FinalSeminar.class);

        // when
        when(seminar.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.ACTIVE_PARTICIPANT, Member.Type.OPPONENT);
        notificationController.notifySeminar(seminar, SeminarEvent.Event.ROOM_CHANGED, new NotificationSource());

        // then
        verify(notificationService, times(2)).save(isA(Notification.class));
    }

    @Test
    public void testCorrectIdeaReceivers() throws Exception {
        // given
        Idea idea = mock(Idea.class);

        // when
        when(idea.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.CREATOR, Member.Type.CO_AUTHOR);
        notificationController.notifyIdea(idea, IdeaEvent.Event.EXPORTED_FAIL, new NotificationSource());

        // then
        verify(notificationService, times(2)).save(isA(Notification.class));
    }

    @Test
    public void testCorrectPeerRequestReceivers() throws Exception {
        // given
        PeerRequest request = mock(PeerRequest.class);

        // when
        when(request.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.PEER_REQUESTER, Member.Type.SUPERVISOR);
        notificationController.notifyPeer(request, PeerEvent.Event.REQUEST_EXPIRED, new NotificationSource());

        // then
        verify(notificationService, times(2)).save(isA(Notification.class));
    }

    @Test
    public void testCorrectPeerReviewReceivers() throws Exception {
        // given
        PeerReview review = mock(PeerReview.class);

        // when
        when(review.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.PEER_REVIEWER);
        notificationController.notifyPeer(review, PeerEvent.Event.REVIEW_COMPLETED, new NotificationSource());

        // then
        verify(notificationService, times(1)).save(isA(Notification.class));
    }

    @Test
    public void testCorrectSenderSetForGroup() throws Exception {
        Group group = mock(Group.class);
        User user = new User();
        user.setFirstName("Sender");

        when(group.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.AUTHOR, Member.Type.SUPERVISOR);

        Set<Notification> notifications =
            notificationController.notifyGroup(group,GroupEvent.Event.MESSAGE_THREAD_REPLY,new NotificationSource(),user);

        assertThat(notifications, everyItem(hasSender(user)));
    }

    @Test
    public void testCorrectSenderSetForProject() throws Exception {
        Project project = mock(Project.class);
        User user = new User();
        user.setFirstName("Sender");

        when(project.getMembers()).thenReturn(generateMemberList());
        setReceivers(Member.Type.AUTHOR, Member.Type.SUPERVISOR);

        Set<Notification> notifications =
                notificationController.notifyProjectForum( ProjectForumEvent.Event.NEW_FORUM_POST_COMMENT, new NotificationSource(), project, user);

        assertThat(notifications, everyItem(hasSender(user)));
    }

    @Test
    public void testOnlyChecksUserDeliveryIfSupervisor() throws Exception {
        // given
        Project project = mock(Project.class);

        // when
        when(project.getMembers()).thenReturn(generateMemberList());
        notificationController.notifyProject(project, ProjectEvent.Event.CREATED, new NotificationSource());

        // then
        // there is only one supervisor, setting should be checked for every delivery method
        verify(deliveryConfiguration, times(DeliveryMethod.values().length)).isDelivery(isA(Notification.Type.class), isA(Enum.class), isA(DeliveryMethod.class), isA(User.class));
    }

    private void setReceivers(Member.Type... members) {
        when(receiverConfiguration.isReceiving(isA(Notification.Type.class), isA(Enum.class), isA(Member.Type.class))).thenReturn(false);
        for (Member.Type member : members) {
            when(receiverConfiguration.isReceiving(isA(Notification.Type.class), isA(Enum.class), eq(member))).thenReturn(true);
        }
    }

    /**
     * @return A member list with one user for each member type
     */
    private List<Member> generateMemberList() {
        List<Member> members = new ArrayList<>();
        UserBuilder builder = new UserBuilder();

        for (Member.Type type : Member.Type.values()) {
            User user = builder.setFirstName(type.name()).setLastName(type.name()).create();
            members.add(new Member(user, type));
        }

        return members;
    }

    private Matcher<Notification> hasSender(final User user) {
        return new TypeSafeMatcher<Notification>() {
            @Override
            protected boolean matchesSafely(Notification item) {
                return item.getCausedBy().equals(user);
            }

            @Override
            public void describeTo(Description description) {
                description.appendValue(user);
            }
        };
    }
}
