package se.su.dsv.scipro.notifications;

import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.TypeSafeMatcher;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.notifications.dataobject.Notification;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;

import static org.hamcrest.CoreMatchers.everyItem;
import static org.hamcrest.CoreMatchers.hasItems;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.core.IsCollectionContaining.hasItem;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

public class NotificationServiceImplIntegrationTest extends IntegrationTest {

    @Inject
    private NotificationService notificationService;

    @Test
    public void find_unread_notifications_by_users_with_mail_compilation() throws Exception {
        User user = createUser();
        Notification notification = createNotification(user);
        Collection<Notification> unreadByUsers = notificationService.findUnreadByUsers(Collections.singletonList(user));
        assertThat(unreadByUsers, hasItem(notification));
    }

    @Test
    public void find_unread_returns_empty_result_if_no_users_are_passed_in() throws Exception {
        createNotification(createUser());
        Collection<Notification> unreadByUsers = notificationService.findUnreadByUsers(new ArrayList<User>());
        assertTrue(unreadByUsers.isEmpty());
    }

    @Test
    public void find_all_with_all_filters() throws Exception {
        NotificationService.Filter filter = new NotificationService.Filter();
        User currentUser = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        User anotherUser = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());

        createNotification(currentUser);
        createNotification(anotherUser);

        filter.setCurrentUser(currentUser);
        filter.setFromUser(anotherUser);

        Iterable<Notification> notifications = notificationService.findAll(filter, new PageRequest(0, 3));

        assertThat(notifications, everyItem(hasUser(filter.getCurrentUser())));
        assertThat(notifications, everyItem(comesFrom(filter.getFromUser())));
    }

    @Test
    public void find_all_with_no_filters() throws Exception {
        User currentUser = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        User anotherUser = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());

        Notification first = createNotification(currentUser);
        Notification second = createNotification(anotherUser);

        Iterable<Notification> all = notificationService.findAll(new NotificationService.Filter(), new PageRequest(0, 3));

        assertThat(all, hasItems(first, second));
    }

    @Test
    public void count_with_filter() throws Exception {
        User currentUser = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        User anotherUser = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());

        createNotification(currentUser);
        createNotification(anotherUser);

        assertEquals(notificationService.count(new NotificationService.Filter()), 2L);
    }

    private Matcher<Notification> comesFrom(final User fromUser) {
        return new TypeSafeMatcher<Notification>() {
            @Override
            protected boolean matchesSafely(Notification item) {
                return item.getCausedBy().equals(fromUser);
            }

            @Override
            public void describeTo(Description description) {
                description.appendValue(fromUser);
            }
        };
    }

    private Matcher<Notification> hasUser(final User currentUser) {
        return new TypeSafeMatcher<Notification>() {
            @Override
            protected boolean matchesSafely(Notification item) {
                return item.getUser().equals(currentUser);
            }

            @Override
            public void describeTo(Description description) {
                description.appendValue(currentUser);
            }
        };
    }

    private Notification createNotification(User user) {
        Notification notification = new Notification();
        notification.setUnread(true);
        notification.setUser(user);
        return save(notification);
    }

    private User createUser() {
        User user = User.builder().firstName("first").lastName("last").emailAddress("first@last.com").build();
        return save(user);
    }
}
