package se.su.dsv.scipro.peer;

import org.junit.Before;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.checklist.ChecklistTemplate;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;
import java.util.EnumSet;

import static org.hamcrest.CoreMatchers.*;
import static org.junit.Assert.*;

public class PeerRequestServiceImplTest extends IntegrationTest {

    private static final PageRequest ALL = new PageRequest(0, Long.MAX_VALUE);

    @Inject
    private PeerRequestServiceImpl peerRequestService;

    private User requester;
    private PeerRequest waiting;
    private Project project;
    private PeerRequest cancelled;
    private ProjectType projectType;
    private PeerRequest finished;

    @Before
    public void setUp() throws Exception {
        requester = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        User headSupervisor = save(User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build());
        projectType = save(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"));
        project = Project.builder().title("title").projectType(projectType).headSupervisor(headSupervisor).build();
        project.addProjectParticipant(requester);
        save(project);
        this.cancelled = newRequest(project, requester, RequestStatus.CANCELED);
        this.finished = newRequest(project, requester, RequestStatus.FINISHED);
        this.waiting = newRequest(project, requester);
        addExpiredReview(waiting);
    }

    @Test
    public void findAll() throws Exception {
        assertThat(peerRequestService.findAll(ALL), hasItem(waiting));
    }

    @Test
    public void findAllWithParams() throws Exception {
        PeerRequestService.Filter params = new PeerRequestService.Filter();
        params.setAuthor(save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build()));
        assertThat(peerRequestService.findAll(params, ALL), not(hasItem(waiting)));
    }

    @Test
    public void countWithParams() throws Exception {
        PeerRequestService.Filter filter = new PeerRequestService.Filter();
        filter.setStatuses(EnumSet.of(RequestStatus.CANCELED));
        assertEquals(1, peerRequestService.count(filter));
    }

    @Test
    public void findByStatus() throws Exception {
        RequestStatus status = RequestStatus.CANCELED;

        assertThat(peerRequestService.findByStatus(status), hasItems(cancelled));
    }

    @Test
    public void findByStudentAndProject() throws Exception {
        assertThat(peerRequestService.findByRequesterAndProject(requester, project), hasItems(waiting, cancelled));
    }

    @Test
    public void findAvailableRequests() throws Exception {
        RequestStatus status = RequestStatus.WAITING;
        User reviewer = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        assertThat(peerRequestService.findAvailableRequests(reviewer, status, projectType, ALL), hasItems(waiting));
    }

    @Test
    public void findAvailableRequestsWithoutStatus() throws Exception {
        assertThat(peerRequestService.findAvailableRequests(null, null, projectType, ALL), hasItems(waiting, cancelled));
    }

    @Test
    public void findAvailableRequestsWithoutProjectType() throws Exception {
        RequestStatus status = RequestStatus.WAITING;

        assertThat(peerRequestService.findAvailableRequests(null, status, null, ALL), hasItems(waiting, cancelled));
    }

    @Test
    public void countByStatus() throws Exception {
        assertEquals(1, peerRequestService.countByStatus(RequestStatus.WAITING));
    }

    @Test
    public void countAvailableRequests() throws Exception {
        User reviewer = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        assertEquals(1, peerRequestService.countAvailableRequests(reviewer, RequestStatus.WAITING, projectType));
    }

    @Test
    public void countAvailableRequestsWithoutStatus() throws Exception {
        assertEquals(3, peerRequestService.countAvailableRequests(null, null, projectType));
    }

    @Test
    public void countAvailableRequestsWithoutProjectType() throws Exception {
        assertEquals(3, peerRequestService.countAvailableRequests(null, RequestStatus.WAITING, null));
    }

    @Test
    public void request_without_reviews_gets_deleted() throws Exception {
        peerRequestService.deleteRequest(cancelled);
        assertNotNull(peerRequestService.findOne(cancelled.getId()));
    }

    @Test
    public void waiting_request_with_reviews_gets_cancelled() throws Exception {
        peerRequestService.deleteRequest(waiting);

        assertEquals(RequestStatus.CANCELED, waiting.getStatus());
    }

    @Test
    public void deleteRequestThatIsNotWaitingDoesNothing() throws Exception {
        peerRequestService.deleteRequest(finished);
        assertEquals(finished, peerRequestService.findOne(finished.getId()));
    }

    @Test
    public void findByChecklistTemplate() throws Exception {
        ChecklistTemplate template = new ChecklistTemplate();
        template.setCreator(requester);
        template.setName("My template");
        save(template);

        waiting.setChecklistTemplate(template);

        assertThat(peerRequestService.findByChecklistTemplate(template), hasItem(waiting));
    }

    private PeerRequest newRequest(Project project, User requester, RequestStatus status) {
        PeerRequest peerRequest = new PeerRequest();
        peerRequest.setProject(project);
        peerRequest.setRequester(requester);
        peerRequest.setStatus(status);
        peerRequest.setFile(save(new FileDescription()));
        return save(peerRequest);
    }

    private PeerRequest newRequest(Project project, User requester) {
        return newRequest(project, requester, RequestStatus.WAITING);
    }

    private void addExpiredReview(PeerRequest peerRequest) {
        PeerReview peerReview = new PeerReview();
        peerReview.setPeerRequest(peerRequest);
        peerReview.setProject(project);
        peerReview.setReviewer(requester);
        peerReview.setStatus(PeerReview.ReviewStatus.EXPIRED);
        save(peerReview);
        waiting.getPeerReviews().add(peerReview);
    }
}
