package se.su.dsv.scipro.peer;

import org.apache.commons.lang3.StringUtils;
import org.junit.Before;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;
import java.util.List;

import static org.hamcrest.CoreMatchers.is;
import static org.hamcrest.core.IsCollectionContaining.hasItem;
import static org.junit.Assert.assertThat;

public class PeerReviewServiceImplIntegrationTest extends IntegrationTest {

    public static final int DAYS_TO_SUBMIT_PEER_REVIEW = 7;

    @Inject
    private PeerReviewService peerReviewService;

    private Project project;
    private ProjectType projectType;

    private PeerReview completedReview;
    private PeerReview inProgressReview;

    @Before
    public void setUp() throws Exception {
        createDefaultProjectType();
        createDefaultProject();

        createCompletedReview();
        createInProgressReview();
        createExpiredReview();
    }

    @Test
    public void find_completed_peer_reviews() throws Exception {
        List<PeerReview> completed = peerReviewService.findAllCompleted(new PageRequest(0, Integer.MAX_VALUE));
        assertThat(completed.size(), is(1));
        assertThat(completed, hasItem(completedReview));
    }

    @Test
    public void find_in_progress_peer_reviews() throws Exception {
        List<PeerReview> inProgress = peerReviewService.findAllReviewsInProgress();
        assertThat(inProgress.size(), is(1));
        assertThat(inProgress, hasItem(inProgressReview));
    }

    @Test
    public void count_all_expired_peer_reviews() throws Exception {
        assertThat(peerReviewService.countByStatus(PeerReview.ReviewStatus.EXPIRED), is(1L));
    }

    private void createExpiredReview() {
        PeerReview peerReview = createPeerReview();
        peerReview.expire();
        save(peerReview);
    }

    private void createInProgressReview() {
        PeerReview peerReview = createPeerReview();
        inProgressReview = save(peerReview);
    }

    private void createCompletedReview() {
        PeerReview peerReview = createPeerReview();
        peerReview.setComment(StringUtils.repeat("a;", PeerReview.MINIMUM_COMMENT_LENGTH_IF_NO_ANSWERS));
        peerReview.submit();
        assert peerReview.getStatus() == PeerReview.ReviewStatus.COMPLETED : "Test set up failed";
        completedReview = save(peerReview);
    }

    private PeerReview createPeerReview() {
        PeerReview peerReview = new PeerReview();
        peerReview.setProject(project);
        peerReview.setPeerRequest(createPeerRequest());
        peerReview.setReviewer(createUser());
        return peerReview;
    }

    private PeerRequest createPeerRequest() {
        PeerRequest peerRequest = new PeerRequest();
        peerRequest.setProject(createProject());
        peerRequest.setRequester(createUser());
        peerRequest.setFile(save(new FileDescription()));
        return save(peerRequest);
    }

    private User createUser() {
        User user = User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build();
        return save(user);
    }

    private void createDefaultProject() {
        this.project = createProject();
    }

    private Project createProject() {
        User headSupervisor = save(User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build());
        return save(Project.builder().title("Peer project").projectType(projectType).headSupervisor(headSupervisor).build());
    }

    private void createDefaultProjectType() {
        ProjectType projectType = save(new ProjectType(DegreeLevel.NONE, "name", "description"));
        projectType.getProjectTypeSettings().setNumDaysToSubmitPeerReview(DAYS_TO_SUBMIT_PEER_REVIEW);
        this.projectType = save(projectType);
    }
}
