package se.su.dsv.scipro.peer;


import org.apache.commons.lang3.StringUtils;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.IntegrationTest;

import javax.inject.Inject;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Locale;

import static java.util.Arrays.asList;

public class TestPeerReview extends IntegrationTest {

    @Inject
    private PeerReviewService peerReviewService;
    @Inject
    private PeerRequestService peerRequestService;

    private Project reviewerProject, requesterProject, bachelorProject;
    private ProjectType master;
    private PeerReview review1, review2, review3;
    private PeerRequest request3;
    private PeerRequest request4;
    private User reviewer;
    private User requester;

    @Before
    public void startTransaction() throws Exception {

        ProjectType bachelor = new ProjectType(ProjectType.BACHELOR, "Bachelor", "Bachelor degree thesis project");
        bachelor = save(bachelor);
        master = new ProjectType(ProjectType.MASTER, "Master", "Master degree thesis project");
        master = save(master);

        User employeeUser = User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build();
        employeeUser = save(employeeUser);
        reviewer = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());
        requester = save(User.builder().firstName("Bob").lastName("Sponge").emailAddress("bob@example.com").build());

        reviewerProject = newProject(bachelor, reviewer, employeeUser, ProjectStatus.ACTIVE);
        requesterProject = newProject(master, requester, employeeUser, ProjectStatus.ACTIVE);
        bachelorProject = newProject(bachelor, requester, employeeUser, ProjectStatus.ACTIVE);

        PeerRequest request1 = newRequest(requester, requesterProject, "Request 1", null);
        PeerRequest request2 = newRequest(requester, requesterProject, "Request 2", null);
        request3 = newRequest(requester, requesterProject, "Request 3", null);
        request4 = newRequest(requester, requesterProject, "Request 4", RequestStatus.WAITING);

        review1 = newReview(reviewer, reviewerProject, request1, RequestStatus.FINISHED, false, true);
        review2 = newReview(reviewer, reviewerProject, request2, RequestStatus.FINISHED, false, true);
        review3 = newReview(reviewer, reviewerProject, request3, RequestStatus.TAKEN, false, false);
    }

    @Test
    public void testFindInProgressReviewsByAuthorAndProject() {
        List<PeerReview> inProgressReviews = peerReviewService.findInProgressReviews(reviewer, reviewerProject);
        Assert.assertEquals(1, inProgressReviews.size());
        Assert.assertEquals(asList(review3), inProgressReviews);
    }

    @Test
    public void testFindReceivedCompletedReviewsByProject() {
        List<PeerReview> receivedCompleted = peerReviewService.findReceivedCompletedReviews(requesterProject);
        Assert.assertEquals(2, receivedCompleted.size());
        Assert.assertEquals(asList(review1, review2), receivedCompleted);
    }

    @Test
    public void testCountAllAbortedReviews() {
        Long allAborted = peerReviewService.countByStatus(PeerReview.ReviewStatus.EXPIRED);
        Long expected = 0L;
        Assert.assertEquals(expected, allAborted);

        review2.expire();
        review3.expire();

        allAborted = peerReviewService.countByStatus(PeerReview.ReviewStatus.EXPIRED);
        expected = 2L;
        Assert.assertEquals(expected, allAborted);
    }

    @Test
    public void testFindAllWithParams() {
        PeerReviewService.Filter params = new PeerReviewService.Filter();
        params.setAuthor(reviewer);
        Iterable<PeerReview> reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(asList(review1, review2, review3), reviewList);

        review3.setReviewer(requester);
        reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(asList(review1, review2), reviewList);

        params.setAuthor(null);
        reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(asList(review1, review2, review3), reviewList);

        params.setStatuses(asList(PeerReview.ReviewStatus.IN_PROGRESS));

        reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(asList(review3), reviewList);

        params.setStatuses(asList(PeerReview.ReviewStatus.EXPIRED));
        review3.expire();

        reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(asList(review3), reviewList);
    }

    @Test
    public void testReviewParamsDateFiltering() throws ParseException {
        PeerReviewService.Filter params = new PeerReviewService.Filter();
        SimpleDateFormat date = new SimpleDateFormat("yyyy-MM-dd", Locale.ENGLISH);

        Date january12 = date.parse("2012-01-12");
        Date february15 = date.parse("2012-02-15");
        Date march21 = date.parse("2012-03-21");
        Date april05 = date.parse("2012-04-05");
        Date may02 = date.parse("2012-05-02");
        Date march27 = date.parse("2012-03-27");

        review1.setDateCreated(february15);
        review2.setDateCreated(march21);
        review3.setDateCreated(may02);

        params.setCreatedAfter(january12);
        params.setCreatedBefore(april05);

        Iterable<PeerReview> reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(2, ((Collection<PeerReview>) reviewList).size());
        Assert.assertEquals(asList(review1, review2), reviewList);

        review3.setDateCreated(march27);
        reviewList = peerReviewService.findAll(params, new PageRequest(0, 10));
        Assert.assertEquals(asList(review1, review2, review3), reviewList);
    }

    @Test
    public void findAvailableRequests() {
        List<PeerRequest> requestPage = peerRequestService.findAvailableRequests(reviewer, RequestStatus.WAITING, master, new PageRequest(0, 8));
        List<PeerRequest> requestList = requestPage;
        Assert.assertEquals(1, requestList.size());
        Assert.assertEquals(asList(request4), requestList);

        request3.setStatus(RequestStatus.WAITING);
        request4.setProject(bachelorProject);

        requestPage = peerRequestService.findAvailableRequests(reviewer, RequestStatus.WAITING, master, new PageRequest(0, 8));
        requestList = requestPage;

        Assert.assertEquals(asList(request3), requestList);

    }

    private PeerReview newReview(final User reviewer, final Project project, PeerRequest request, final RequestStatus status, boolean aborted, boolean submitted) {
        PeerReview review = new PeerReview();
        review.setReviewer(reviewer);
        review.setProject(project);
        review.setComment("Hi there!");
        review.setPeerRequest(request);
        request.setStatus(status);
        save(request);
        if (submitted) {
            review.setComment(StringUtils.repeat("a;", PeerReview.MINIMUM_COMMENT_LENGTH_IF_NO_ANSWERS));
            review.submit();
        }
        else if (aborted) {
            review.expire();
        }
        return save(review);
    }

    private PeerRequest newRequest(final User requester, final Project project, final String comment, final RequestStatus status) {
        PeerRequest request = new PeerRequest();
        request.setComment(comment);
        request.setRequester(requester);
        request.setProject(project);
        request.setFile(save(new FileDescription()));
        if (status != null) {
            request.setStatus(status);
        }
        return save(request);
    }

    private Project newProject(final ProjectType projectType, final User student, final User headSupervisor, final ProjectStatus status) {
        Project newProject = Project.builder().
                title("Project title").
                projectType(projectType).
                headSupervisor(headSupervisor).
                projectStatus(status).
                build();
        newProject.addProjectParticipant(student);

        return save(newProject);
    }
}
