package se.su.dsv.scipro.report;

import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ObjectMother;

import java.util.ArrayList;
import java.util.List;

import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.mock;

public class SupervisorMaster15GradeCalculatorTest extends GradeCalculatorTest {

    private SupervisorGradingReport gradingReport;
    private SupervisorMaster15GradeCalculator gradeCalculator;

    @Before
    public void setUp() throws Exception {
        GradingReportTemplate gradingReportTemplate = prepareTemplate();

        gradingReport = gradingReportTemplate.createSupervisorReport(ObjectMother.SOME_MASTER_PROJECT, mock(User.class));
        gradeCalculator = new SupervisorMaster15GradeCalculator();
    }

    @Test
    public void report_with_27_points_gets_A() throws Exception {
        assessAllCriteriaWithMinimumAndTotal(27);
        assertEquals(GradingReport.Grade.A, gradingReport.getGrade(gradeCalculator));
    }

    @Test
    public void report_with_0_on_Ö4_gets_B() throws Exception {
        assessAllCriteriaWithMinimumAndTotalSkipDeadlineCriterion(26);
        assertEquals(GradingReport.Grade.B, gradingReport.getGrade(gradeCalculator));
    }

    @Test
    public void report_with_24_points_gets_B() throws Exception {
        assessAllCriteriaWithMinimumAndTotal(24);
        assertEquals(GradingReport.Grade.B, gradingReport.getGrade(gradeCalculator));
    }

    @Test
    public void report_with_22_points_gets_C() throws Exception {
        assessAllCriteriaWithMinimumAndTotal(22);
        assertEquals(GradingReport.Grade.C, gradingReport.getGrade(gradeCalculator));
    }

    @Test
    public void report_with_20_points_gets_D() throws Exception {
        assessAllCriteriaWithMinimumAndTotal(20);
        assertEquals(GradingReport.Grade.D, gradingReport.getGrade(gradeCalculator));
    }

    @Test
    public void report_with_18_points_gets_E() throws Exception {
        assessAllCriteriaWithMinimumAndTotal(18);
        assertEquals(GradingReport.Grade.E, gradingReport.getGrade(gradeCalculator));
    }

    @Test
    public void a_report_with_less_than_18_points_but_10_completed_project_criteria_with_fx_flag_and_2_completed_individual_criteria_with_fx_flag_gets_Fx() throws Exception {
        assessAllCriteria(0);
        assignPointsToProjectCriteriaWithFxFlag(10);
        assignPointsToTwoIndividualCriteriaWithFxFlag(2);

        assertEquals(GradingReport.Grade.FX, gradingReport.getGrade(gradeCalculator));
    }

    protected GradingReportTemplate prepareTemplate() {
        GradingReportTemplate gradingReportTemplate = new GradingReportTemplate(ObjectMother.MASTER, 15);

        gradingReportTemplate.addProjectCriterion("U1 Sammanfattning", 1, getPointTemplates(1));
        gradingReportTemplate.addProjectCriterion("U2 Introduktion", 1, getPointTemplates(1));
        gradingReportTemplate.addProjectCriterion("U3 Problem", 1, getPointTemplates(1));
        gradingReportTemplate.addProjectCriterion("U4 Frågeställning", 1, getPointTemplates(2));
        gradingReportTemplate.addProjectCriterion("U5 Vetenskaplig förankring", 1, getPointTemplates(2));
        gradingReportTemplate.addProjectCriterion("U6 Metodval", 1, getPointTemplates(2));
        gradingReportTemplate.addProjectCriterion("U7 Metodtillämpning", 1, getPointTemplates(2));
        gradingReportTemplate.addProjectCriterion("U8 Resultat", 2, getPointTemplates(3));
        gradingReportTemplate.addProjectCriterion("U9 Slutsatser och diskussion", 2, getPointTemplates(2));
        gradingReportTemplate.addProjectCriterion("U10 Form, Struktur och språk", 1, getPointTemplates(1));
        gradingReportTemplate.addProjectCriterion("U11 Argumentation", 1, getPointTemplates(1));
        gradingReportTemplate.addProjectCriterion("U12 Källhänvisningar och dokumentation", 1, getPointTemplates(1));
        gradingReportTemplate.addProjectCriterion("U13 Originalitet och signifikans", 1, getPointTemplates(3));

        gradingReportTemplate.addIndividualCriterion("Ö1 Oppositionsrapport", 1, getPointTemplates(2));
        gradingReportTemplate.addIndividualCriterion("Ö2 Presentationer", 1, getPointTemplates(1));
        gradingReportTemplate.addIndividualCriterion("Ö3 Aktivitet vid seminarier och möten", 1, getPointTemplates(1));
        gradingReportTemplate.addIndividualCriterion("Ö4 Deadlines", 0, getPointTemplates(1)).setFx(false);
        gradingReportTemplate.addIndividualCriterion("Ö5 Revision efter slutseminarium", 0, getPointTemplates(1)).setFx(false);

        return gradingReportTemplate;
    }

    private List<GradingCriterionPointTemplate> getPointTemplates(int maxPoint) {
        List<GradingCriterionPointTemplate> gradingCriterionPointTemplates = new ArrayList<>();
        for (int i = 0; i <= maxPoint; i++) {
            gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(i).description("Description").build());
        }
        return gradingCriterionPointTemplates;
    }

    protected void assessAllCriteriaWithMinimumAndTotalSkipDeadlineCriterion(int remaining) {
        for (GradingCriterion gradingCriterion : getGradingReport().getGradingCriteria()) {
            int toAssign = Math.min(remaining, gradingCriterion.getPointsRequiredToPass());
            if (!gradingCriterion.getTitle().equals("Ö4 Deadlines")) {
                int leftAfterAssignment = assignAtMost(toAssign, gradingCriterion);
                remaining -= (toAssign - leftAfterAssignment);
            }
        }

        for (GradingCriterion gradingCriterion : getGradingReport().getGradingCriteria()) {
            remaining = assignAtMost(remaining, gradingCriterion);
        }
    }

    @Override
    protected GradingReport getGradingReport() {
        return gradingReport;
    }

    @Override
    protected GradeCalculator getGradeCalculator() {
        return gradeCalculator;
    }
}
