package se.su.dsv.scipro.reviewing;

import com.google.common.base.Function;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import org.springframework.data.domain.PageRequest;
import scala.Option;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;

import static org.hamcrest.CoreMatchers.any;
import static org.hamcrest.CoreMatchers.hasItem;
import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class ReviewingServiceImplTest extends ReviewingModuleTest {
    @Inject
    private ReviewingServiceImpl reviewingService;
    @Inject
    private FinalSeminarApprovalServiceImpl finalSeminarApprovalService;
    @Inject
    private RoughDraftApprovalServiceImpl roughDraftApprovalService;

    private Project project;
    private User reviewer;

    @Before
    public void setUp() throws Exception {
        reviewer = save(User.builder().firstName("Ronny").lastName("Reviewer").emailAddress("reviewer@dsv.su.se").build());
        User headSupervisor = save(User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build());
        ProjectType bachelor = save(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"));
        project = save(Project.builder().title("My project").projectType(bachelor).headSupervisor(headSupervisor).build());
        project.addReviewer(reviewer);
    }

    @Test
    public void count() throws Exception {
        finalSeminarApprovalService.requestApproval(project, createFileUpload(), "test");
        assertEquals(1, reviewingService.count(getFilter(ReviewerApproval.Step.FINAL_SEMINAR_APPROVAL)));
    }

    private MyReviewService.Filter getFilter(ReviewerApproval.Step step) {
        MyReviewService.Filter filter = new MyReviewService.Filter();
        filter.setUser(reviewer);
        filter.setStep(step);
        return filter;
    }

    @Test
    public void find_undecided() throws Exception {
        finalSeminarApprovalService.requestApproval(project, createFileUpload(), "test");
        assertThat(reviewingService.findAll(getFilter(ReviewerApproval.Step.FINAL_SEMINAR_APPROVAL), new PageRequest(0, Long.MAX_VALUE)), hasItem(any(FinalSeminarApproval.class)));
    }

    @Test
    public void find_undecided_rough_draft_approvals() throws Exception {
        roughDraftApprovalService.requestApproval(project, createFileUpload(), "test");
        assertThat(reviewingService.findAll(getFilter(ReviewerApproval.Step.ROUGH_DRAFT_APPROVAL), new PageRequest(0, Long.MAX_VALUE)), hasItem(any(RoughDraftApproval.class)));
    }

    @Test
    public void count_undecided_rough_drafts() throws Exception {
        roughDraftApprovalService.requestApproval(project, createFileUpload(), "test");
        assertEquals(reviewingService.countUndecidedRoughDraft(reviewer), 1);
    }

    @Test
    public void count_undecided_final_seminar() throws Exception {
        finalSeminarApprovalService.requestApproval(project, createFileUpload(), "test");
        assertEquals(reviewingService.countUndecidedFinalSeminar(reviewer), 1);
    }

    @Test
    public void find_approval() throws Exception {
        FinalSeminarApproval approval = finalSeminarApprovalService.requestApproval(project, createFileUpload(), "test");
        assertEquals(reviewingService.findApproval(1L), approval);
    }

    @Test
    public void back_and_forth() throws Exception {
        assertTrue(finalSeminarApprovalService.requiresUpload(project));
        FinalSeminarApproval firstFinalSeminarApproval = finalSeminarApprovalService.requestApproval(project, createFileUpload(), "test");
        assertFalse(finalSeminarApprovalService.requiresUpload(project));

        reviewingService.reject(firstFinalSeminarApproval, "Very bad", Option.apply(createFileUpload()), Option.<FileDescription>empty());

        assertTrue(finalSeminarApprovalService.requiresUpload(project));
        FinalSeminarApproval secondFinalSeminarApproval = finalSeminarApprovalService.requestApproval(project, createFileUpload(), "test");
        assertFalse(finalSeminarApprovalService.requiresUpload(project));

        reviewingService.approve(secondFinalSeminarApproval, "Very good", Option.<FileUpload>empty(), Option.<FileDescription>empty());
        assertFalse(finalSeminarApprovalService.requiresUpload(project));
    }

    private FileUpload createFileUpload() {
        FileUpload fileUpload = mock(FileUpload.class);
        when(fileUpload.handleData(Matchers.any(Function.class))).thenReturn(null);
        return fileUpload;
    }

}
