package se.su.dsv.scipro.milestones.service

import com.google.common.eventbus.EventBus
import org.junit.runner.RunWith
import org.junit.{Before, Test}
import org.mockito.Matchers.{eq => is}
import org.mockito.Mockito._
import org.mockito.runners.MockitoJUnitRunner
import org.mockito.{Matchers, Mock, Mockito}
import se.su.dsv.scipro.finalseminar._
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate
import se.su.dsv.scipro.peer.{FirstPeerReviewCompletedEvent, PeerRequest, PeerReview, SecondPeerReviewCompletedEvent}
import se.su.dsv.scipro.project.Project
import se.su.dsv.scipro.report.{GradingReport, ReviewerGradingReportSubmittedEvent, SupervisorGradingReport, SupervisorGradingReportSubmittedEvent}
import se.su.dsv.scipro.system.User
import se.su.dsv.scipro.test.ObjectMother.{SOME_PROJECT, SOME_USER}

@RunWith(classOf[MockitoJUnitRunner]) class MilestoneActivatorTest {

  @Mock private var mileStoneService: MileStoneService = _
  @Mock private var milestoneActivityTemplateService: MilestoneActivityTemplateService = _
  @Mock private var eventBus: EventBus = _

  private var milestoneActivator: MilestoneActivator = _

  private var opposition: FinalSeminarOpposition = _
  private var respondent: FinalSeminarRespondent = _
  private var participation: FinalSeminarActiveParticipation = _
  private var peerReview: PeerReview = _
  private var gradingReport: GradingReport = _

  @Before def setUp(): Unit = {
    opposition = createOpposition
    respondent = createRespondent
    participation = createParticipation
    peerReview = createPeerReview
    gradingReport = createGradingReport
    milestoneActivator = new MilestoneActivator(milestoneActivityTemplateService, mileStoneService, eventBus)
  }

  @Test def registers_itself_with_the_event_bus(): Unit = {
    verify(eventBus).register(milestoneActivator)
  }

  @Test def activates_milestone_for_opposition_approved_if_there_is_one():Unit = {
    val event: OppositionApprovedEvent = new OppositionApprovedEvent(opposition)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.oppositionApproved(event)
    verifyIndividualMilestoneActivated(event, milestoneActivityTemplate)
  }

  @Test def does_nothing_if_no_milestone_for_the_opposition_approved_event():Unit = {
    milestoneActivator.oppositionApproved(new OppositionApprovedEvent(opposition))
    verify(mileStoneService, never).setConfirmed(Matchers.any(classOf[Project]), Matchers.any(classOf[User]), Matchers.any(classOf[MilestoneActivityTemplate]), Matchers.anyBoolean)
  }

  @Test def activates_milestone_for_participation_approved_events():Unit = {
    val event: ParticipationApprovedEvent = new ParticipationApprovedEvent(participation)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.participationApproved(event)
    verifyIndividualMilestoneActivated(event, milestoneActivityTemplate)
  }

  @Test def activates_milestone_for_respondent_approved_events():Unit = {
    val event: RespondentApprovedEvent = new RespondentApprovedEvent(respondent)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.respondentApproved(event)
    verifyIndividualMilestoneActivated(event, milestoneActivityTemplate)
  }

  @Test def deactivates_milestone_when_opposition_is_failed():Unit = {
    val event: OppositionFailedEvent = new OppositionFailedEvent(opposition)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.oppositionFailed(event)
    verifyMilestoneDeactivated(event, milestoneActivityTemplate)
  }

  @Test def deactivates_milestone_when_respondent_is_failed():Unit = {
    val event: RespondentFailedEvent = new RespondentFailedEvent(respondent)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.respondentFailed(event)
    verifyMilestoneDeactivated(event, milestoneActivityTemplate)
  }

  @Test def deactivates_milestone_when_participation_is_failed():Unit = {
    val event: ParticipationFailedEvent = new ParticipationFailedEvent(participation)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.participationFailed(event)
    verifyMilestoneDeactivated(event, milestoneActivityTemplate)
  }

  @Test def activates_milestone_when_peer_review_event_is_submitted():Unit = {
    val event: FirstPeerReviewCompletedEvent = new FirstPeerReviewCompletedEvent(peerReview)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.peerReviewCompleted(event)
    verifyIndividualMilestoneActivated(event, milestoneActivityTemplate)
  }

  @Test def activates_milestone_when_second_peer_review_event_is_submitted():Unit = {
    val event: SecondPeerReviewCompletedEvent = new SecondPeerReviewCompletedEvent(peerReview)
    val milestoneActivityTemplate: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.secondPeerReviewCompleted(event)
    verifyIndividualMilestoneActivated(event, milestoneActivityTemplate)
  }

  @Test def activates_milestone_when_thesis_uploaded():Unit = {
    val seminar: FinalSeminar = new FinalSeminar
    seminar.setProject(SOME_PROJECT)
    val event: FinalSeminarThesisUploadedEvent = new FinalSeminarThesisUploadedEvent(seminar)
    val activity: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.seminarThesisUploaded(event)
    verifyProjectMilestoneActivated(activity)
  }

  @Test def activates_milestone_when_final_seminar_created():Unit = {
    val seminar: FinalSeminar = new FinalSeminar
    seminar.setProject(SOME_PROJECT)
    val event: FinalSeminarCreatedEvent = new FinalSeminarCreatedEvent(seminar)
    val activity: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.finalSeminarCreated(event)
    verifyProjectMilestoneActivated(activity)
  }

  @Test def activates_milestone_when_supervisor_submits_grading_report():Unit = {
    val supervisorGradingReport: SupervisorGradingReport = mock(classOf[SupervisorGradingReport])
    when(supervisorGradingReport.getProject).thenReturn(SOME_PROJECT)
    when(supervisorGradingReport.getUser).thenReturn(SOME_USER)
    val event: SupervisorGradingReportSubmittedEvent = new SupervisorGradingReportSubmittedEvent(supervisorGradingReport)
    val activity: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.supervisorGradingReportSubmitted(event)
    verify(mileStoneService).setConfirmed(is(SOME_PROJECT), is(SOME_USER), is(activity), is(true))
  }

  @Test def deactivates_milestone_when_final_seminar_deleted():Unit = {
    val seminar: FinalSeminar = new FinalSeminar
    seminar.setProject(SOME_PROJECT)
    val event: FinalSeminarDeletedEvent = new FinalSeminarDeletedEvent(seminar)
    val activity: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.finalSeminarDeleted(event)
    verifyProjectMilestoneDeactivated(activity)
  }

  @Test def activates_milestone_when_reviewer_submits_grading_report():Unit = {
    val event: ReviewerGradingReportSubmittedEvent = new ReviewerGradingReportSubmittedEvent(gradingReport)
    val activity: MilestoneActivityTemplate = mockMilestoneForEvent(event)
    milestoneActivator.reviewerGradingReportSubmitted(event)
    verifyProjectMilestoneActivated(activity)
  }

  private def verifyIndividualMilestoneActivated[A : Authorship](authorship: A, milestoneActivityTemplate: MilestoneActivityTemplate): Unit = {
    verifyIndividualMilestoneConfirmed(authorship, milestoneActivityTemplate, activated = true)
  }

  private def verifyIndividualMilestoneConfirmed[A](authorship: A, milestoneActivityTemplate: MilestoneActivityTemplate, activated: Boolean)(implicit $ev: Authorship[A]): Unit = {
    verify(mileStoneService).setConfirmed(is($ev.project(authorship)), is($ev.author(authorship)), is(milestoneActivityTemplate), is(activated))
  }

  private def verifyProjectMilestoneActivated(activity: MilestoneActivityTemplate): Unit = {
    verify(mileStoneService).setConfirmed(is(SOME_PROJECT), is(activity), is(true))
  }

  private def verifyProjectMilestoneDeactivated(activity: MilestoneActivityTemplate): Unit = {
    verify(mileStoneService).setConfirmed(is(SOME_PROJECT), is(activity), is(false))
  }

  private def verifyMilestoneDeactivated[A : Authorship](authorship: A, milestoneActivityTemplate: MilestoneActivityTemplate): Unit = {
    verifyMilestoneConfirmed(authorship, milestoneActivityTemplate, activated = false)
  }

  private def verifyMilestoneConfirmed[A](authorship: A, milestoneActivityTemplate: MilestoneActivityTemplate, activated: Boolean)(implicit $ev: Authorship[A]): Unit = {
    verify(mileStoneService).setConfirmed(is($ev.project(authorship)), is($ev.author(authorship)), is(milestoneActivityTemplate), is(activated))
  }

  private def createRespondent: FinalSeminarRespondent = {
    val finalSeminar: FinalSeminar = new FinalSeminar
    finalSeminar.setProject(SOME_PROJECT)
    new FinalSeminarRespondent(SOME_USER, finalSeminar)
  }

  private def createParticipation: FinalSeminarActiveParticipation = {
    val finalSeminarActiveParticipation: FinalSeminarActiveParticipation = new FinalSeminarActiveParticipation
    finalSeminarActiveParticipation.setUser(SOME_USER)
    finalSeminarActiveParticipation.setProject(SOME_PROJECT)
    finalSeminarActiveParticipation
  }

  private def createOpposition: FinalSeminarOpposition = {
    val finalSeminarOpposition: FinalSeminarOpposition = new FinalSeminarOpposition
    finalSeminarOpposition.setProject(SOME_PROJECT)
    finalSeminarOpposition.setUser(SOME_USER)
    finalSeminarOpposition
  }

  private def mockMilestoneForEvent[A](event: A)(implicit $ev: NamedEvent[A]): MilestoneActivityTemplate = {
    val milestoneActivityTemplate: MilestoneActivityTemplate = new MilestoneActivityTemplate
    $ev.name(event) foreach { name =>
      when(milestoneActivityTemplateService.findByEvent(name)).thenReturn(milestoneActivityTemplate)
    }
    milestoneActivityTemplate
  }

  private def createPeerReview: PeerReview = {
    val peerReview: PeerReview = new PeerReview
    peerReview.setProject(SOME_PROJECT)
    peerReview.setReviewer(SOME_USER)
    peerReview.setPeerRequest(new PeerRequest)
    peerReview
  }

  private def createGradingReport: GradingReport = {
    val mock: GradingReport = Mockito.mock(classOf[GradingReport])
    when(mock.getProject).thenReturn(SOME_PROJECT)
    mock
  }
}