package se.su.dsv.scipro.integration.daisy.workers;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import scala.Option;
import se.su.dsv.scipro.io.ExternalImporter;
import se.su.dsv.scipro.io.dto.Person;
import se.su.dsv.scipro.io.dto.Thesis;
import se.su.dsv.scipro.io.http.DaisyAPI;
import se.su.dsv.scipro.io.impl.ImporterTransactions;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.system.UserService;
import se.su.dsv.scipro.system.Username;
import se.su.dsv.scipro.workerthreads.AbstractWorker;

import javax.inject.Inject;
import java.util.List;
import java.util.Set;

public class UserImportWorker extends AbstractWorker {
    private static final Logger LOGGER = LoggerFactory.getLogger(UserImportWorker.class);
    private static final long DSV = 4L;

    private final DaisyAPI daisyAPI;
    private final ImporterTransactions importerTransactions;
    private final ExternalImporter externalImporter;
    private final UserService userService;

    @Inject
    public UserImportWorker(DaisyAPI daisyAPI,
                            ImporterTransactions importerTransactions,
                            ExternalImporter externalImporter,
                            UserService userService) {
        this.daisyAPI = daisyAPI;
        this.importerTransactions = importerTransactions;
        this.externalImporter = externalImporter;
        this.userService = userService;
    }

    @Override
    protected void doWork() {

        long startDate = System.currentTimeMillis();
        LOGGER.info("Starting execution for scheduled remote user import of all remote supervisors");

        importUserDetails();
        importSupervisors();
        importSupervisorUnits();

        long endDate = System.currentTimeMillis();
        long executionTimeSecs = (endDate - startDate) / 1000;
        LOGGER.info("Finished running in {} hours, {} minutes, {} seconds",
                executionTimeSecs / 3600,
                (executionTimeSecs % 3600) / 60,
                executionTimeSecs % 60);
    }

    private void importUserDetails() {
        for (final User user : userService.findAll()) {
            if (user.getIdentifier() != null) {
                try {
                    Option<Person> idPerson = daisyAPI.findPersonById(user.getIdentifier());
                    if (idPerson.isDefined()) {
                        Person person = idPerson.get();
                        if (person.getEmail() != null) {
                            importerTransactions.importPerson(person);
                        }
                        else {
                            LOGGER.info("Skipping user because of no e-mail in Daisy: {}", person);
                        }
                    }
                    else {
                        for (Username username : user.getUsernames()) {
                            Option<Person> usernamePerson = daisyAPI.findByUsername(username.getUsername());
                            if (usernamePerson.isDefined()) {
                                user.setIdentifier(usernamePerson.get().getId());
                                User duplicatedUser = userService.findByExternalIdentifier(usernamePerson.get().getId());
                                if (duplicatedUser != null) {
                                    LOGGER.info("User {} ({}) already exist in SciPro", user, user.getIdentifier());
                                }
                                userService.save(user);
                                importerTransactions.importPerson(usernamePerson.get());
                                break;
                            }
                        }
                    }
                } catch (RuntimeException e) {
                    LOGGER.info("Failed to re-import user {} ({})", user, user.getIdentifier(), e);
                }
            }
        }
    }

    private void importSupervisorUnits() {
        externalImporter.importSupervisorUnits();
    }

    private void importSupervisors() {
        List<Person> supervisors = daisyAPI.getSupervisors(DSV);
        for (Person supervisor : supervisors) {
            try {
                importerTransactions.importSupervisor(supervisor);
                importAllProjectsForSupervisor(supervisor);
            } catch (RuntimeException e) {
                LOGGER.error("Failed to fetch projects for supervisor: {} {} ({})",
                        supervisor.getFirstName(),
                        supervisor.getLastName(),
                        supervisor.getId(),
                        e);
            }
        }
    }

    private void importAllProjectsForSupervisor(Person supervisor) {
        Set<Thesis> projects = daisyAPI.getProjects(supervisor.getId());
        for (Thesis project : projects) {
            try {
                importProject(project);
            } catch (RuntimeException e) {
                LOGGER.error("Failed to import project: {} ({})",
                        project.getTitle(),
                        project.getId(),
                        e);
            }
        }
    }

    private void importProject(Thesis project) {
        importerTransactions.importProject(project);
    }
}
