package se.su.dsv.scipro.io.impl;

import com.google.inject.persist.Transactional;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.io.ExternalImporter;
import se.su.dsv.scipro.io.dto.Person;
import se.su.dsv.scipro.io.dto.Thesis;
import se.su.dsv.scipro.io.dto.Unit;
import se.su.dsv.scipro.io.exceptions.ExternalImportException;
import se.su.dsv.scipro.io.http.DaisyAPI;
import se.su.dsv.scipro.springdata.services.UnitService;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.system.UserService;

import javax.inject.Inject;
import javax.ws.rs.ClientErrorException;
import java.util.HashSet;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;

/**
 * Importer implementation, uses standard HTTP requests and Json-wrappers.
 */
public class ExternalImporterDaisyImpl implements ExternalImporter {
    public static final Long DSV = 4L;

    private static final Logger LOGGER = LoggerFactory.getLogger(ExternalImporterDaisyImpl.class);

    private final UserService userService;
    private final UnitService unitService;
    private final DaisyAPI client;
    private final ImporterTransactions importerTransactions;

    @Inject
    public ExternalImporterDaisyImpl(
            UserService userService,
            UnitService unitService,
            DaisyAPI client,
            ImporterTransactions importerTransactions) {
        this.userService = userService;
        this.unitService = unitService;
        this.client = client;
        this.importerTransactions = importerTransactions;
    }

    @Override
    public User importUser(final String userName) {
        try {
            Person person = client.findByUsername(userName).get();
            return importPerson(person);
        } catch (ClientErrorException | NoSuchElementException e) {
            throw new ExternalImportException(e);
        }
    }

    private User importPerson(Person person) {
        User user;

        if (client.getSupervisors(DSV).contains(person)){
            user = importerTransactions.importSupervisor(person);
        } else {
            user = importerTransactions.importStudent(person);
        }

        importProjects(person);

        return user;
    }

    @Override
    @Transactional
    public void importSupervisorUnits() {
        Set<Unit> unitSet = client.getSubUnits(DSV);

        addUnexistingUnitsAsUnits(unitSet);

        Set<Person> allSupervisorsOnUnits = new HashSet<>();
        for (final Unit unitDTO : unitSet) {
            List<Person> supervisorSet = client.getSupervisors(unitDTO.getId());
            allSupervisorsOnUnits.addAll(supervisorSet);
            if (!supervisorSet.isEmpty()) {
                for (Person supervisorDTO : supervisorSet) {
                    addUnitToSupervisor(supervisorDTO, unitDTO);
                }
            } else {
                LOGGER.info("Unit: " + unitDTO.getName() + " has no supervisors attached.");
            }
        }

        // Compare the list of supervisors with remote units with the list of
        // all supervisors in scipro, to make sure to remove units from the
        // users that doesn't have one in daisy.
        LOGGER.debug("Number of supervisors on remote units: " + allSupervisorsOnUnits.size());
        removeUnitFromSupervisorsNotInSet(allSupervisorsOnUnits);
    }

    private void addUnitToSupervisor(Person supervisorDTO, se.su.dsv.scipro.io.dto.Unit unitDTO) {
        User supervisor = userService.findByExternalIdentifier(supervisorDTO.getId());
        if (supervisor != null) {
            if (supervisor.getUnit() != null) {
                LOGGER.info(supervisor.getFullName() + " is already attached to a unit and this is being overwritten.");
            }
            se.su.dsv.scipro.system.Unit unitToAdd = unitService.findByIdentifier(unitDTO.getId());
            LOGGER.debug("Adding unit: " + unitToAdd.getTitle() + " to supervisor: " + supervisor.getFullName());
            supervisor.setUnit(unitToAdd);
        } else {
            LOGGER.debug("Can't find supervisor: " + supervisorDTO.getId());
        }
    }

    private void removeUnitFromSupervisorsNotInSet(Set<Person> supervisorSet) {
        final Set<User> supervisorsWithRemoteUnit = new HashSet<>();
        final Set<User> currentSupervisors = new HashSet<>();

        for (User supervisor : userService.findAll()) {
            currentSupervisors.add(supervisor);
        }

        //Create SciPro Employees from DTOs.
        for (Person supervisorDTO : supervisorSet) {
            User supervisorUser = userService.findByExternalIdentifier(supervisorDTO.getId());
            supervisorsWithRemoteUnit.add(supervisorUser);
        }

        //Remove units from supervisors that is not in the remote set.
        for (User supervisor : currentSupervisors) {
            if (!supervisorsWithRemoteUnit.contains(supervisor)) {
                LOGGER.debug(supervisor.getFullName() + " does not have a remote unit. Searching for local units to remove.");
                supervisor.setUnit(null);
                LOGGER.debug("Removed units from: " + supervisor.getFullName());
            }
        }
    }

    private void addUnexistingUnitsAsUnits(final Set<se.su.dsv.scipro.io.dto.Unit> unitsFromDTO) {

        Set<se.su.dsv.scipro.io.dto.Unit> remoteUnits = new HashSet<>(unitsFromDTO);

        //Add units to database
        for (final se.su.dsv.scipro.io.dto.Unit unitDTO : remoteUnits) {
            LOGGER.info("************UNIT CLOSED?********** " + unitDTO.isClosed());
            if (!unitDTO.isClosed()) {
                se.su.dsv.scipro.system.Unit unit = unitService.findByIdentifier(unitDTO.getId());
                if (unit == null) {
                    LOGGER.info("External unit: '" + unitDTO + "' has no local representation, creating");
                    unit = new se.su.dsv.scipro.system.Unit();
                    unit.setIdentifier(unitDTO.getId());
                    unit.setTitle(unitDTO.getName());
                    unitService.save(unit);
                } else if (!unit.getTitle().equals(unitDTO.getName())) {
                    LOGGER.info("Unit " + unit.getTitle() + " has changed name, renaming to: " + unitDTO.getName());
                    unit.setTitle(unitDTO.getName());
                } else {
                    LOGGER.info("Unit " + unit.getTitle() + " already exists, skipping");
                }
            } else {
                se.su.dsv.scipro.system.Unit unit = unitService.findByIdentifier(unitDTO.getId());
                if (unit != null) {
                    LOGGER.info("Unit " + unit.getTitle() + " is closed, deleting");
                    unitService.delete(unit.getId());
                }
            }
        }
    }

    private void importProjects(final Person person) {
        Set<Thesis> fetchedProjects = client.getProjects(person.getId());
        for (Thesis project : fetchedProjects) {
            importProject(project);
        }
    }

    private void importProject(Thesis project) {
        importerTransactions.importProject(project);
    }
}
