package se.su.dsv.scipro.integration.daisy.workers;

import com.google.inject.persist.UnitOfWork;
import com.google.inject.util.Providers;
import com.mysema.query.types.Predicate;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import scala.Option;
import se.su.dsv.scipro.finalseminar.FinalSeminarService;
import se.su.dsv.scipro.io.ExternalExporter;
import se.su.dsv.scipro.io.dto.*;
import se.su.dsv.scipro.io.facade.ExporterFacade;
import se.su.dsv.scipro.io.http.DaisyAPI;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectRepo;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.Unit;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.workerthreads.WorkerData;
import se.su.dsv.scipro.workerthreads.WorkerDataService;

import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import static org.junit.Assert.assertEquals;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyLong;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@RunWith(MockitoJUnitRunner.class)
public class ProjectExporterTest {

    public static final long SUPERVISOR_IDENTIFIER = 99L;
    public static final long AUTHOR_IDENTIFIER = 2342L;

    private @Mock ProjectRepo projectRepo;
    private @Mock ExporterFacade exporterFacade;
    private @Mock DaisyAPI daisyAPI;
    private @Mock ExternalExporter externalExporter;
    private @Mock EntityManager entityManager;
    private @Mock EntityTransaction entityTransaction;
    private @Mock WorkerDataService workerDataService;
    private @Mock UnitOfWork unitOfWork;
    private @Mock FinalSeminarService finalSeminarService;

    private ProjectExporter projectExporter;

    private Project newProject;
    private Project exportedProject;
    private User author;

    @Before
    public void setUp() throws Exception {
        projectExporter = new ProjectExporter(projectRepo, exporterFacade, daisyAPI, externalExporter, finalSeminarService);
        projectExporter.setTxManager(Providers.of(entityManager));
        projectExporter.setWorkerDataService(workerDataService);
        projectExporter.setUnitOfWork(unitOfWork);
        when(workerDataService.save(any(WorkerData.class))).thenReturn(new WorkerData());
        when(entityManager.getTransaction()).thenReturn(entityTransaction);

        Unit unit = new Unit();
        unit.setIdentifier(239478L);

        User supervisor = User.builder().firstName("Foo").lastName("Bar").emailAddress("foo@bar.com").build();
        supervisor.setIdentifier(SUPERVISOR_IDENTIFIER);
        supervisor.setUnit(unit);

        author = User.builder().firstName("First name").lastName("Last name").emailAddress("foo@bar.com").build();
        author.setIdentifier(AUTHOR_IDENTIFIER);
        author.setId(234L);

        ProjectType bachelor = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");
        newProject = Project.builder()
                .title("New project")
                .projectType(bachelor)
                .headSupervisor(supervisor)
                .build();

        ProjectType master = new ProjectType(DegreeLevel.SECOND_CYCLE, "Master", "Master");
        exportedProject = Project.builder()
                .title("Exported project")
                .projectType(master)
                .headSupervisor(supervisor)
                .identifier(8L)
                .build();

        when(projectRepo.findAll(any(Predicate.class)))
                .thenReturn(Arrays.asList(newProject, exportedProject));
        when(daisyAPI.getThesis(anyLong())).thenReturn(Option.apply(new Thesis()));
    }

    @Test
    public void creates_new_project_if_no_identifier() throws Exception {
        projectExporter.run();

        verify(exporterFacade).exportProject(newProject, newProject.getHeadSupervisor().getUnit());
    }

    @Test
    public void updates_existing_projects() throws Exception {
        projectExporter.run();

        ArgumentCaptor<Thesis> captor = ArgumentCaptor.forClass(Thesis.class);
        verify(daisyAPI).updateThesis(captor.capture());
        Thesis thesis = captor.getValue();
        assertSameData(exportedProject, thesis);
    }

    @Test
    public void removes_removed_authors() throws Exception {
        Person person = new Person();
        person.setId(5454L);
        StudentProjectParticipant author = new StudentProjectParticipant();
        author.setPerson(person);
        when(daisyAPI.getAuthors(exportedProject.getIdentifier())).thenReturn(Collections.singleton(author));

        projectExporter.run();

        verify(daisyAPI).deleteAuthor(exportedProject.getIdentifier(), person.getId());
    }

    @Test
    public void adds_new_author() throws Exception {
        exportedProject.addProjectParticipant(author);
        when(daisyAPI.getAuthors(exportedProject.getIdentifier())).thenReturn(Collections.<StudentProjectParticipant>emptySet());

        projectExporter.run();

        verify(externalExporter).addStudentToProject(exportedProject, author);
    }

    @Test
    public void deletes_wrong_head_supervisor() throws Exception {
        Person person = new Person();
        person.setId(234234L);

        ProjectParticipant headSupervisor = new ProjectParticipant();
        headSupervisor.setRole(Role.SUPERVISOR);
        headSupervisor.setPerson(person);

        Set<ProjectParticipant> contributors = new HashSet<>();
        contributors.add(headSupervisor);

        when(daisyAPI.getContributors(exportedProject.getIdentifier())).thenReturn(contributors);

        projectExporter.run();

        verify(daisyAPI).deleteContributor(exportedProject.getIdentifier(), person.getId());
    }

    @Test
    public void adds_missing_co_supervisor_to_project() throws Exception {
        User user = User.builder().firstName("co").lastName("supervisor").emailAddress("co@supervisor.com").build();
        user.setIdentifier(789L);
        user.setId(7890L);
        exportedProject.addCoSupervisor(user);

        projectExporter.run();

        verify(externalExporter).addContributorToProject(exportedProject, user, Role.ASSISTANT_SUPERVISOR);
    }

    private static void assertSameData(Project project, Thesis thesis) {
        assertEquals(project.getTitle(), thesis.getTitle());
        assertEquals(project.getProjectStatus() == ProjectStatus.COMPLETED, thesis.isFinished());
        assertEquals(project.getProjectStatus() == ProjectStatus.INACTIVE, thesis.isAborted());
        assertEquals(Level.valueOf(project.getProjectType().getDegreeLevel().name()), thesis.getLevel());
        assertEquals(project.getHeadSupervisor().getUnit().getIdentifier(), thesis.getUnit().getId());
    }
}
