package se.su.dsv.scipro.checklists;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.AjaxLink;
import org.apache.wicket.ajax.markup.html.form.AjaxButton;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.*;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.apache.wicket.validation.validator.StringValidator;
import se.su.dsv.scipro.admin.pages.AbstractAdminProjectPage;
import se.su.dsv.scipro.checklist.ChecklistCategory;
import se.su.dsv.scipro.checklist.ChecklistCategoryRepo;
import se.su.dsv.scipro.checklist.ChecklistTemplate;
import se.su.dsv.scipro.checklist.ChecklistTemplateService;
import se.su.dsv.scipro.components.menuhighlighting.MenuHighlightAdminChecklist;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.reusable.SciProUtilities;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.ProjectTypeService;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.List;

public class AdminEditChecklistTemplatePage extends AbstractAdminProjectPage implements MenuHighlightAdminChecklist {

    public static final String FEEDBACK = "feedback";
    public static final String FORM = "form";
    public static final String TITLE = "title";
    public static final String CATEGORIES = "categories";
    public static final String PROJECT_TYPES = "projectTypes";
    public static final String DESCRIPTION = "description";

    @Inject
    private ChecklistCategoryRepo checklistCategoryRepo;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private ChecklistTemplateService checklistTemplateService;

    public AdminEditChecklistTemplatePage(PageParameters pp) {
        add(new FencedFeedbackPanel(FEEDBACK, this));

        final IModel<ChecklistTemplate> model = new DetachableServiceModel<>(checklistTemplateService);
        if (pp.get(PageParameterKeys.MAP.get(ChecklistTemplate.class)).isNull()){
            model.setObject(new ChecklistTemplate());
        } else {
            model.setObject(checklistTemplateService.findOne(pp.get(PageParameterKeys.MAP.get(ChecklistTemplate.class)).toLong()));
        }

        add(new EditForm(FORM, model));
    }

    protected class EditForm extends Form<ChecklistTemplate> {

        List<String> questionList = new ArrayList<>();
        private final WebMarkupContainer wmc;

        public EditForm(String id, final IModel<ChecklistTemplate> model) {
            super(id, model);

            questionList.addAll(getModelObject().getQuestions());

            RequiredTextField<String> title = new RequiredTextField<>(TITLE, new PropertyModel<String>(getModel(), "name"));
            title.add(StringValidator.minimumLength(ChecklistTemplate.MIN_TITLE_LENGTH));
            add(title);

            add(new TextArea<>(DESCRIPTION, new PropertyModel<String>(model, "description")));

            add(new CheckBoxMultipleChoice<>(CATEGORIES, new PropertyModel<List<ChecklistCategory>>(model, "categories"), checklistCategoryRepo.findAll(), new ChoiceRenderer<>("categoryName", "id"))
                    .setLabelPosition(AbstractChoice.LabelPosition.WRAP_AFTER));

            add(new CheckBoxMultipleChoice<>(PROJECT_TYPES, new PropertyModel<List<ProjectType>>(model, "projectTypes"), getRelevantProjectTypes(), new ChoiceRenderer<>("name", "id"))
                    .setLabelPosition(AbstractChoice.LabelPosition.WRAP_BEFORE));

            wmc = new WebMarkupContainer("wmc");
            wmc.setOutputMarkupId(true);
            add(wmc);

            wmc.add(new NewQuestionForm("newQuestionForm"));

            wmc.add(new ListView<String>("questions", new PropertyModel<List<String>>(this, "questionList")) {
                @Override
                protected void populateItem(final ListItem<String> item) {

                    item.add(new AjaxLink<Void>("up") {
                        @Override
                        public void onClick(AjaxRequestTarget target) {
                            swapQuestions(item.getIndex(), item.getIndex()-1);
                            target.add(wmc);
                        }
                        @Override
                        protected void onConfigure() {
                            setVisibilityAllowed(item.getIndex()!=0);
                        }
                    });

                    item.add(new AjaxLink<Void>("down") {
                        @Override
                        public void onClick(AjaxRequestTarget target) {
                            swapQuestions(item.getIndex(), item.getIndex()+1);
                            target.add(wmc);
                        }
                        @Override
                        protected void onConfigure() {
                            setVisibilityAllowed(item.getIndex()!=(getList().size()-1));
                        }
                    });

                    item.add(new EditQuestionForm("edit", item.getModel()));
                    item.add(new AjaxLink<Void>("delete") {
                        @Override
                        public void onClick(AjaxRequestTarget target) {
                            questionList.remove(item.getModelObject());
                            target.add(wmc);
                        }
                    });
                }
            });
        }

        private void swapQuestions(int index1, int index2){
            String s1 = questionList.get(index1);
            String s2 = questionList.get(index2);

            questionList.set(index2, s1);
            questionList.set(index1, s2);
        }

        private IModel<List<ProjectType>> getRelevantProjectTypes() {
            return new LoadableDetachableModel<List<ProjectType>>() {
                @Override
                protected List<ProjectType> load() {
                    return projectTypeService.findWithModule(ProjectModule.CHECKLIST);
                }
            };
        }

        @Override
        protected void onSubmit() {
            ChecklistTemplate template = EditForm.this.getModelObject();

            template.clearQuestions();
            template.setQuestions(questionList);

            if (template.getCreator()==null){
                template.setCreator(SciProSession.get().getUser());
            }
            if (template.getTemplateNumber() == ChecklistTemplate.DEFAULT_TEMPLATE_NUMBER){
                template.setTemplateNumber(SciProUtilities.safeLongToInt(checklistTemplateService.count()));
            }
            setModelObject(checklistTemplateService.save(EditForm.this.getModelObject()));
            success(getString("success"));
        }

        @Override
        protected void onError() {
            error(getString("submissionFailed"));
        }

        private class NewQuestionForm extends Form<String> {
            public NewQuestionForm(String id) {
                super(id);
                final TextField<String> newQuestion = new TextField<>("newQuestion", Model.of(""));
                add(newQuestion);
                add(new AjaxButton("addNewQuestion") {
                    @Override
                    protected void onSubmit(AjaxRequestTarget target, Form<?> form) {
                        questionList.add(newQuestion.getModelObject());
                        newQuestion.setModelObject("");
                        target.add(wmc);
                    }
                });
            }
        }

    }

    private static class EditQuestionForm extends Form<String> {

        private boolean editMode = false;

        public EditQuestionForm(String id, IModel<String> model) {
            super(id, model);

            final WebMarkupContainer wmc = new WebMarkupContainer("wmc");
            wmc.setOutputMarkupId(true);
            add(wmc);

            wmc.add(new Label("questionLabel", getModel()){
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(!editMode);
                }
            });

            wmc.add(new AjaxLink<Void>("edit") {
                @Override
                public void onClick(AjaxRequestTarget target) {
                    editMode = !editMode;
                    target.add(wmc);
                }
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(!editMode);
                }
            });

            wmc.add(new TextField<String>("questionField", getModel()) {
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(editMode);
                }
            });

            wmc.add(new AjaxButton("doEdit") {
                @Override
                protected void onSubmit(AjaxRequestTarget target, Form<?> form) {
                    editMode = !editMode;
                    target.add(wmc);
                }
                @Override
                protected void onConfigure() {
                    setVisibilityAllowed(editMode);
                }
            });
        }
    }
}