package se.su.dsv.scipro.economy;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.behavior.AttributeAppender;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.form.RequiredTextField;
import org.apache.wicket.markup.html.form.TextField;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.list.PropertyListView;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.CompoundPropertyModel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.validation.IValidatable;
import org.apache.wicket.validation.IValidator;
import org.apache.wicket.validation.ValidationError;
import se.su.dsv.scipro.components.DateLabel;
import se.su.dsv.scipro.components.OppositeVisibility;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.security.auth.ProjectModuleComponent;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.util.AjaxConfirmationLink;

import javax.inject.Inject;
import java.util.List;

@ProjectModuleComponent(ProjectModule.BUDGET)
public class EconomyPanel extends GenericPanel<Project> {

    public static final String FEEDBACK = "feedback";
    public static final String BALANCE = "balance";
    public static final String TRANSACTIONS = "transactions";
    public static final String AMOUNT = "amount";
    public static final String COMMENT = "comment";
    public static final String NO_TRANSACTIONS = "noTransactions";
    public static final String FORM = "form";
    public static final String USER = "user";
    public static final String DELETE = "delete";
    public static final String CONTAINER = "container";

    @Inject
    private TransactionService transactionService;

    private final ListView<Transaction> transactions;

    public EconomyPanel(String id, IModel<Project> model) {
        super(id, model);
        final WebMarkupContainer container = new WebMarkupContainer(CONTAINER);
        container.setOutputMarkupId(true);
        add(container);

        container.add(new FencedFeedbackPanel(FEEDBACK, this));
        container.add(new Label(BALANCE, getBalance()));
        transactions = new PropertyListView<Transaction>(TRANSACTIONS, getTransactions()) {
            @Override
            protected void populateItem(final ListItem<Transaction> item) {
                item.add(new Label(AMOUNT));
                item.add(new Label(COMMENT));
                item.add(new DateLabel("dateCreated", DateStyle.DATETIME));
                item.add(new UserLinkPanel(USER));
                AjaxConfirmationLink<Transaction> delete = new AjaxConfirmationLink<Transaction>(DELETE, item.getModel(), getString("deleteSure")) {
                    @Override
                    public void onClick(AjaxRequestTarget target) {
                        target.add(container);
                        transactionService.delete(item.getModelObject());
                        info(getString("deleted"));
                        transactions.setModel(getTransactions());
                    }
                };
                item.add(delete);
            }

            @Override
            protected void onConfigure() {
                setVisibilityAllowed(!getTransactions().getObject().isEmpty());
            }
        };
        container.add(transactions);
        container.add(new WebMarkupContainer(NO_TRANSACTIONS).add(new OppositeVisibility(transactions)));
        TransactionForm form = new TransactionForm(FORM, Model.of(new Transaction()));
        container.add(form);
        String toggleScript = String.format("$('#%s').slideToggle();", form.getMarkupId());
        container.add(new WebMarkupContainer("add").add(AttributeAppender.replace("onclick", toggleScript)));
    }

    private LoadableDetachableModel<List<Transaction>> getTransactions() {
        return new LoadableDetachableModel<List<Transaction>>() {
            @Override
            protected List<Transaction> load() {
                return transactionService.findByProject(getModelObject());
            }
        };
    }

    private LoadableDetachableModel<String> getBalance(){
        return new LoadableDetachableModel<String>() {
            @Override
            protected String load() {
                int balance = 0;
                for (Transaction t : getTransactions().getObject()){
                    balance = balance + t.getAmount();
                }
                return String.valueOf(balance);
            }
        };
    }

    private class TransactionForm extends Form<Transaction> {
        public TransactionForm(String id, IModel<Transaction> model) {
            super(id, new CompoundPropertyModel<>(model));
            TextField<Integer> amount = new RequiredTextField<>(AMOUNT, Integer.class);
            amount.add(new NegativeBalanceValidator());
            add(amount);
            add(new TextField<>(COMMENT));
        }

        @Override
        protected void onSubmit() {
            super.onSubmit();
            getModelObject().setProject(EconomyPanel.this.getModelObject());
            getModelObject().setUser(SciProSession.get().getUser());
            transactionService.save(getModelObject());
            this.setModelObject(new Transaction());
            info(getString("saved"));
        }
    }

    private class NegativeBalanceValidator implements IValidator<Integer> {
        @Override
        public void validate(IValidatable<Integer> validatable) {
            if ((Integer.parseInt(getBalance().getObject()) + validatable.getValue()) < 0){
                validatable.error(new ValidationError(this));
            }
        }
    }
}
