package se.su.dsv.scipro.finalseminar;

import com.google.common.eventbus.EventBus;
import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.form.AjaxSubmitLink;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.form.FormComponent;
import org.apache.wicket.markup.html.form.TextArea;
import org.apache.wicket.markup.html.form.TextField;
import org.apache.wicket.markup.html.link.Link;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.EmptyPanel;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.ResourceModel;
import org.apache.wicket.validation.validator.RangeValidator;
import org.apache.wicket.validation.validator.StringValidator;
import se.su.dsv.scipro.components.ListAdapterModel;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.report.GradingReportService;
import se.su.dsv.scipro.report.OppositionReportService;
import se.su.dsv.scipro.report.SupervisorGradingReport;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.ProjectTypeService;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.JavascriptEventConfirmation;

import javax.inject.Inject;
import java.util.Date;
import java.util.List;
import java.util.Set;

public class SeminarOppositionPanel extends Panel {

    public static final String USER = "user";
    public static final String OPPONENTS = "opponents";
    public static final String REMOVE = "remove";
    public static final String FORM = "form";
    public static final String POINTS = "points";
    public static final int MIN_POINTS = 0;
    public static final int MAX_POINTS = 2;
    public static final String GRADING_FEEDBACK = "gradingFeedback";
    public static final int FEEDBACK_MAX_LENGTH = 2000;
    public static final String SUBMIT = "submit";
    public static final String FEEDBACK = "feedback";
    public static final String CONTAINER = "container";
    public static final String GRADE_CONTAINER = "gradeContainer";

    @Inject
    private FinalSeminarOppositionService finalSeminarOppositionService;
    @Inject
    private EventBus eventBus;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private FinalSeminarService finalSeminarService;
    @Inject
    private GradingReportService gradingReportService;
    @Inject
    private OppositionReportService oppositionReportService;

    private IModel<FinalSeminar> seminar;
    private FeedbackPanel feedbackPanel;
    private final WebMarkupContainer oppositionContainer;
    private final ListView<FinalSeminarOpposition> opponents;

    public SeminarOppositionPanel(String id, final IModel<FinalSeminar> seminar) {
        super(id, seminar);
        this.seminar = seminar;

        oppositionContainer = new WebMarkupContainer(CONTAINER);
        oppositionContainer.setOutputMarkupId(true);

        feedbackPanel = new FencedFeedbackPanel(FEEDBACK, this);
        feedbackPanel.setOutputMarkupId(true);
        add(feedbackPanel);

        final IModel<List<FinalSeminarOpposition>> oppositions = new ListAdapterModel<>(new PropertyModel<Set<FinalSeminarOpposition>>(seminar, "oppositions"));

        add(new Label("noOpponents", new ResourceModel("no.opponents")) {
            @Override
            protected void onConfigure() {
                setVisible(oppositions.getObject().isEmpty());
            }
        });

        opponents = getOpponentsList(oppositions);
        oppositionContainer.add(opponents);
        add(oppositionContainer);
    }

    private ListView<FinalSeminarOpposition> getOpponentsList(final IModel<List<FinalSeminarOpposition>> oppositions) {
        return new ListView<FinalSeminarOpposition>(OPPONENTS, oppositions) {
            @Override
            protected void populateItem(final ListItem<FinalSeminarOpposition> item) {
                final FinalSeminarOpposition opposition = item.getModelObject();
                item.add(getUserLinkPanel(item));

                WebMarkupContainer gradeContainer = getGradeContainer(opposition);
                gradeContainer.add(new Label("gradedPoints", new PropertyModel<>(opposition, POINTS)));
                gradeContainer.add(new Label("gradedFeedback", new PropertyModel<>(opposition, FEEDBACK)));
                item.add(gradeContainer);

                item.add(getRemoveLink(item));

                item.add(getFinalSeminarOppositionForm(item));

                if (gradingModuleIsOnForProjectType()){
                    item.add(new SeminarOppositionReportPanel("report", item.getModel()));
                } else {
                    item.add(new EmptyPanel("report"));
                }
            }
        };
    }

    private UserLinkPanel getUserLinkPanel(ListItem<FinalSeminarOpposition> item) {
        return new UserLinkPanel(USER, new PropertyModel<User>(item.getModel(), "user"));
    }

    private WebMarkupContainer getGradeContainer(final FinalSeminarOpposition opposition) {
        return new WebMarkupContainer(GRADE_CONTAINER) {
            @Override
            protected void onConfigure() {
                super.onConfigure();
                FinalSeminarGrade grade = opposition.getGrade();
                setVisibilityAllowed(grade != null && opposition.getPoints() != null && opposition.getFeedback() != null && isEmployee());
            }
        };
    }

    private Link<FinalSeminarOpposition> getRemoveLink(final ListItem<FinalSeminarOpposition> item) {
        return new Link<FinalSeminarOpposition>(REMOVE, item.getModel()) {
            {
                add(new JavascriptEventConfirmation("onclick", SeminarOppositionPanel.this.getString("confirm")));
            }

            @Override
            public void onClick() {
                SeminarOppositionPanel.this.success(getString("removed", item.getModel()));
                FinalSeminar finalSeminar = getModelObject().getFinalSeminar();
                finalSeminar.removeOpposition(getModelObject());
                finalSeminarService.save(finalSeminar);
                opponents.setModel(new ListAdapterModel<>(new PropertyModel<Set<FinalSeminarOpposition>>(seminar, "oppositions")));
            }

            @Override
            protected void onConfigure() {
                super.onConfigure();
                if (gradingModuleIsOnForProjectType()){
                setVisibilityAllowed((isHeadSupervisor() || isAdmin()) && !hasSubmittedOppositionReport(item.getModelObject()));
                } else {
                    setVisibilityAllowed((isHeadSupervisor() || isAdmin())); //cant do the above check from this panel
                }
            }
        };
    }

    private FinalSeminarOppositionForm getFinalSeminarOppositionForm(ListItem<FinalSeminarOpposition> item) {
        return new FinalSeminarOppositionForm(SeminarOppositionPanel.FORM, item.getModel());
    }

    private boolean startDateHasPassed() {
        return seminar.getObject().getStartDate().before(new Date());
    }

    private boolean isAdmin() {
        return SciProSession.get().authorizedForRole(Roles.SYSADMIN) && SciProSession.get().hasActualRole(Roles.SYSADMIN);
    }

    private boolean isHeadSupervisor() {
        return SciProSession.get().getUser().equals(seminar.getObject().getProject().getHeadSupervisor());
    }

    private boolean isEmployee() {
        return SciProSession.get().authorizedForRole(Roles.SUPERVISOR);
    }

    private class FinalSeminarOppositionForm extends Form<FinalSeminarOpposition> {
        public FinalSeminarOppositionForm(String id, final IModel<FinalSeminarOpposition> finalSeminarOpposition) {
            super(id, finalSeminarOpposition);

            FormComponent<Integer> pointsField = new TextField<>(POINTS, new PropertyModel<Integer>(finalSeminarOpposition, POINTS))
                    .add(RangeValidator.range(MIN_POINTS, MAX_POINTS))
                    .setType(Integer.class).setRequired(true);
            add(pointsField);

            TextArea<Object> feedback = new TextArea<>(GRADING_FEEDBACK, new PropertyModel<>(finalSeminarOpposition, FEEDBACK));
            feedback.add(StringValidator.maximumLength(FEEDBACK_MAX_LENGTH));
            feedback.setRequired(true);
            add(feedback);

            add(new AjaxSubmitLink(SUBMIT) {
                @Override
                protected void onSubmit(AjaxRequestTarget target, Form<?> form) {
                    super.onSubmit(target, form);

                    if (getModelObject().getPoints().equals(0)){
                        finalSeminarOpposition.getObject().setGrade(FinalSeminarGrade.NOT_APPROVED);
                        eventBus.post(new OppositionFailedEvent(finalSeminarOpposition.getObject()));
                    } else {
                        finalSeminarOpposition.getObject().setGrade(FinalSeminarGrade.APPROVED);
                        eventBus.post(new OppositionApprovedEvent(finalSeminarOpposition.getObject()));
                    }
                    finalSeminarOppositionService.save(finalSeminarOpposition.getObject());
                    boolean updated = true;
                    if (gradingModuleIsOnForProjectType()){
                        SupervisorGradingReport report = gradingReportService.getSupervisorGradingReport(finalSeminarOpposition.getObject().getProject(), finalSeminarOpposition.getObject().getUser());
                        updated = gradingReportService.updateOppositionCriteria(report, finalSeminarOpposition.getObject());
                    }
                    success(getString(updated ? "feedback.opponent.updated" : "feedback.opponent.not.updated", finalSeminarOpposition));
                    target.add(feedbackPanel);
                    target.add(oppositionContainer);
                }

                @Override
                protected void onError(AjaxRequestTarget target, Form<?> form) {
                    target.add(feedbackPanel);
                }
            });
        }

        @Override
        protected void onConfigure() {
            super.onConfigure();
            setVisibilityAllowed(startDateHasPassed() && getModelObject().getPoints() == null && getModelObject().getFeedback() == null && isHeadSupervisor());
        }
    }

    private boolean gradingModuleIsOnForProjectType(){
        return projectTypeService.findWithModule(ProjectModule.GRADING).contains(seminar.getObject().getProjectType());
    }

    private boolean hasSubmittedOppositionReport(FinalSeminarOpposition opposition) {
        return oppositionReportService.findOrCreateReport(opposition).isSubmitted();
    }
}
