package se.su.dsv.scipro.firstmeeting;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.form.AjaxSubmitLink;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.basic.MultiLineLabel;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.form.FormComponent;
import org.apache.wicket.markup.html.form.RequiredTextField;
import org.apache.wicket.markup.html.form.TextArea;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.CompoundPropertyModel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.ResourceModel;
import org.apache.wicket.util.io.IClusterable;
import org.apache.wicket.validation.validator.DateValidator;
import scala.Option;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.date.DateService;
import se.su.dsv.scipro.match.*;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.dataobject.IdeaEvent;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.session.SciProSession;

import javax.inject.Inject;
import java.util.Date;

import static java.lang.String.format;

public class FirstMeetingPanel extends GenericPanel<Idea> {

    public static final String MEETING_FORM = "meetingForm";
    public static final String SUBMIT_BUTTON = "saveButton";
    public static final String ROOM_FIELD = "room";
    public static final String DESCRIPTION = "description";
    public static final String FIRST_MEETING_DATE = "firstMeetingDate";
    public static final String ROOM_HELP = "roomHelp";
    public static final String SUPERVISOR_WMC = "supervisorWMC";
    public static final String STUDENT_WMC = "studentWMC";
    public static final String DATE_LABEL = "dateLabel";
    public static final String ROOM_LABEL = "roomLabel";
    public static final String DESCRIPTION_LABEL = "descriptionLabel";
    public static final String FEEDBACK = "feedback";

    @Inject
    private FirstMeetingRepository firstMeetingRepository;
    @Inject
    private NotificationController notificationController;
    @Inject
    private DateService dateService;

    public void onSaved(AjaxRequestTarget target){}

    public FirstMeetingPanel(String id, IModel<Idea> model) {
        super(id, model);
        final IModel<FirstMeeting> firstMeetingModel = getMeeting();

        WebMarkupContainer supervisorWMC = new WebMarkupContainer(SUPERVISOR_WMC) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(isSupervisor(firstMeetingModel.getObject()));
            }
        };
        supervisorWMC.add(new MeetingForm(MEETING_FORM, firstMeetingModel));
        add(supervisorWMC);

        WebMarkupContainer studentWMC = new WebMarkupContainer(STUDENT_WMC) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(!isSupervisor(firstMeetingModel.getObject()));
            }
        };
        studentWMC.add(new DateLabel(DATE_LABEL, new PropertyModel<Date>(model, "firstMeeting.firstMeetingDate"), DateStyle.DATETIME));
        studentWMC.add(new Label(ROOM_LABEL, new PropertyModel<String>(model, "firstMeeting.room")));
        studentWMC.add(new MultiLineLabel(DESCRIPTION_LABEL, new PropertyModel<String>(model, "firstMeeting.description")));
        add(studentWMC);
    }

    private IModel<FirstMeeting> getMeeting() {
        Idea idea = getModelObject();
        final FirstMeeting firstMeeting = idea.getFirstMeeting() != null
                ? idea.getFirstMeeting()
                : new FirstMeeting(getCourseStartDate(), idea);
        return new DetachableModel<>(firstMeeting, new FirstMeetingPersister());
    }

    private Date getCourseStartDate() {
        ApplicationPeriod period = getModelObject().getApplicationPeriod();
        return period != null ? period.getCourseStartDate() : new Date();
    }

    @Override
    protected void onConfigure() {
        super.onConfigure();
        Idea.Status status = getModelObject().getMatchStatus();
        setVisible(status == Idea.Status.COMPLETED || status == Idea.Status.MATCHED);
    }

    private class MeetingForm extends Form<FirstMeeting> {
        public MeetingForm(String id, final IModel<FirstMeeting> model) {
            super(id, new CompoundPropertyModel<>(model));
            final FeedbackPanel feedback = new FeedbackPanel(FEEDBACK);
            feedback.setOutputMarkupId(true);
            add(feedback);

            FormComponent<Date> date = new BootstrapDateTimeComponent(FIRST_MEETING_DATE);
            date.setRequired(true);
            date.add(DateValidator.minimum(new Date()));
            add(date);
            add(new TextArea<String>(DESCRIPTION));
            add(new RequiredTextField<String>(ROOM_FIELD));
            add(new InfoPanel(ROOM_HELP, new ResourceModel("roomInfo")));
            add(new AjaxSubmitLink(SUBMIT_BUTTON) {
                @Override
                protected void onSubmit(final AjaxRequestTarget target, final Form<?> form) {
                    target.add(feedback);
                    onSaved(target);
                }

                @Override
                protected void onError(AjaxRequestTarget target, Form<?> form) {
                    target.add(feedback);
                }
            });
        }

        @Override
        protected void onSubmit() {
            saveAndNotify();
            info(format(getString("meeting.scheduled", FirstMeetingPanel.this.getModel()),
                    dateService.format(getModelObject().getFirstMeetingDate(), DateStyle.DATETIME)));
        }

        private void saveAndNotify() {
            firstMeetingRepository.save(getModelObject());
            NotificationSource source = new NotificationSource();
            String date = dateService.format(getModelObject().getFirstMeetingDate(), DateStyle.DATETIME);
            String room = getModelObject().getRoom();
            source.setMessage(date + "\n" + getString("room") + ": " + room);
            source.setAdditionalMessage(getModelObject().getDescription());
            notificationController.notifyIdea(FirstMeetingPanel.this.getModelObject(),
                    IdeaEvent.Event.FIRST_MEETING, source);
        }
    }

    private boolean isSupervisor(FirstMeeting firstMeeting) {
        return firstMeeting.getIdea().getMatch().getSupervisor().equals(SciProSession.get().getUser());
    }

    private class FirstMeetingPersister implements Persister<FirstMeeting>, IClusterable {
        @Override
        public Option<FirstMeeting> load(final long id) {
            return Option.apply(firstMeetingRepository.findOne(id));
        }

        @Override
        public Option<Long> id(final FirstMeeting firstMeeting) {
            return Option.apply(firstMeeting.getId());
        }
    }
}
