package se.su.dsv.scipro.notifications.panels;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.form.AjaxFormComponentUpdatingBehavior;
import org.apache.wicket.extensions.markup.html.repeater.data.grid.ICellPopulator;
import org.apache.wicket.extensions.markup.html.repeater.data.sort.SortOrder;
import org.apache.wicket.extensions.markup.html.repeater.data.table.AbstractColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.IColumn;
import org.apache.wicket.extensions.markup.html.repeater.data.table.export.AbstractExportableColumn;
import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.html.panel.Fragment;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.model.*;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import se.su.dsv.scipro.basepages.MenuPage;
import se.su.dsv.scipro.components.DateColumn;
import se.su.dsv.scipro.components.ExportableDataPanel;
import se.su.dsv.scipro.components.UserAutoComplete;
import se.su.dsv.scipro.data.dataobjects.Member;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.dataproviders.FilteredDataProvider;
import se.su.dsv.scipro.forum.panels.AbstractReadStatePanel;
import se.su.dsv.scipro.notifications.NotificationService;
import se.su.dsv.scipro.notifications.dataobject.Notification;
import se.su.dsv.scipro.notifications.pages.NotificationLandingPage;
import se.su.dsv.scipro.notifications.pages.SupervisorNotificationSettingsPage;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.pages.ProjectDetailsPage;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.supervisor.pages.SupervisorProjectDetailsPage;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.AjaxConfirmationLink;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.text.MessageFormat;
import java.util.Arrays;
import java.util.List;

public class NotificationDataPanel extends Panel {

    @Inject
    NotificationService notificationService;
    @Inject
    ProjectService projectService;

    /**
     * No-op method. Override if you want to.
     */
    public void unreadStateChanged(AjaxRequestTarget target) {
    }

    public NotificationDataPanel(String id) {
        super(id);
        addNotificationsPanel();
        addMarkReadLink();
        addSettingsPageLink();
        setOutputMarkupId(true);
    }

    private void addMarkReadLink() {
        add(new AjaxConfirmationLink<Void>("markRead", "Are you sure you want to mark ALL unread notifications as read? This action can NOT be undone.") {
            @Override
            public void onClick(AjaxRequestTarget target) {
                notificationService.markRead(SciProSession.get().getUser());
                target.add(NotificationDataPanel.this);
                unreadStateChanged(target);
            }
        });
    }

    private void addSettingsPageLink() {
        BookmarkablePageLink<Void> settingsLink = new BookmarkablePageLink<>("settingsLink", SupervisorNotificationSettingsPage.class);
        add(settingsLink.setVisible(SciProSession.get().getUser().hasRole(Roles.SUPERVISOR)));
    }

    private IModel<NotificationService.Filter> getFilter() {
        NotificationService.Filter filter = new NotificationService.Filter();
        filter.setCurrentUser(SciProSession.get().getUser());
        return Model.of(filter);
    }

    private List<? extends IColumn<Notification, String>> columns() {
        return Arrays.asList(
                createColumnUnread(),
                new DateColumn<Notification>(new StringResourceModel("column.date", this, null), "dateCreated", "dateCreated", DateStyle.DATETIME),
                createColumnSubject(),
                createColumnFrom(),
                createColumnRole(),
                new ProjectColumn(new StringResourceModel("column.project", NotificationDataPanel.this, null))
        );
    }


    private void addNotificationsPanel() {
        IModel<NotificationService.Filter> filter = getFilter();
        add(new FilterForm("filterForm", filter));
        SortableDataProvider<Notification, String> provider = new FilteredDataProvider<>(notificationService, filter);
        provider.setSort("dateCreated", SortOrder.DESCENDING);
        add(new ExportableDataPanel<>("dataPanel", columns(), provider));
    }

    private IColumn<Notification, String> createColumnUnread() {
        return new AbstractColumn<Notification, String>(new ResourceModel("column.unread"), "unread") {
            @Override
            public void populateItem(Item<ICellPopulator<Notification>> cellItem, String componentId, final IModel<Notification> rowModel) {
                cellItem.add(new AbstractReadStatePanel(componentId) {
                    @Override
                    protected boolean isRead() {
                        return !rowModel.getObject().isUnread();
                    }

                    @Override
                    protected void toggleReadState(final AjaxRequestTarget target) {
                        Notification notification = rowModel.getObject();
                        notification.setUnread(!notification.isUnread());
                        notificationService.save(notification);
                        unreadStateChanged(target);
                    }
                });
            }
        };
    }

    private IColumn<Notification, String> createColumnRole() {
        return new AbstractColumn<Notification, String>(new StringResourceModel("column.role", NotificationDataPanel.this, null)) {
            @Override
            public void populateItem(Item<ICellPopulator<Notification>> cellItem, String componentId, IModel<Notification> rowModel) {
                User user = rowModel.getObject().getCausedBy();
                if (user == null) {
                    cellItem.add(new Label(componentId, new StringResourceModel("role.SYSTEM", NotificationDataPanel.this, null)));
                } else {
                    Project project = rowModel.getObject().getProject();
                    Member.Type type = project != null ? projectService.getMemberType(project, user) : null;
                    if (type == null) {
                        cellItem.add(new Label(componentId, new StringResourceModel("unknown", NotificationDataPanel.this, null)));
                    } else {
                        cellItem.add(new Label(componentId, new StringResourceModel("role.${name}", NotificationDataPanel.this, Model.of(type))));
                    }
                }
            }
        };
    }

    private IColumn<Notification, String> createColumnFrom() {
        return new AbstractColumn<Notification, String>(new StringResourceModel("column.from", NotificationDataPanel.this, null)) {
            @Override
            public void populateItem(Item<ICellPopulator<Notification>> cellItem, String componentId, IModel<Notification> rowModel) {
                User user = rowModel.getObject().getCausedBy();

                if (user != null) {
                    cellItem.add(new UserLinkPanel(componentId, new PropertyModel<User>(rowModel, "causedBy")));
                } else {
                    cellItem.add(new Label(componentId, "SciPro"));
                }
            }
        };
    }

    private IColumn<Notification, String> createColumnSubject() {
        return new AbstractExportableColumn<Notification, String, String>(new StringResourceModel("column.subject", this, null)) {
            @Override
            public IModel<String> getDataModel(IModel<Notification> model) {
                return new StringResourceModel("${type}.${event}.title",
                        NotificationDataPanel.this, model, MessageFormat.format("Missing property: {0}.{1}.title",
                        model.getObject().getType(), model.getObject().getEvent()),
                        model.getObject().getTitle(), model.getObject().getSource(), model.getObject().getAdditionalSource());
            }

            @Override
            public void populateItem(final Item<ICellPopulator<Notification>> cellItem, final String componentId, final IModel<Notification> rowModel) {
                final PageParameters parameters = new PageParameters();
                parameters.set("id", rowModel.getObject().getNotificationEvent().getId());

                final BookmarkablePageLink<Void> link = new BookmarkablePageLink<>("link", NotificationLandingPage.class, parameters);
                link.setBody(getDataModel(rowModel));

                final Fragment fragment = new Fragment(componentId, "subject", NotificationDataPanel.this);
                fragment.add(link);

                cellItem.add(fragment);
            }
        };
    }

    private class ProjectColumn extends AbstractExportableColumn<Notification, String, String> {

        public ProjectColumn(IModel<String> displayModel) {
            super(displayModel);
        }

        @Override
        public IModel<String> getDataModel(final IModel<Notification> rowModel) {
            return new AbstractReadOnlyModel<String>() {
                @Override
                public String getObject() {
                    Project project = rowModel.getObject().getProject();
                    return project != null ? project.getTitle() : getString("unknown");
                }
            };
        }

        @Override
        public void populateItem(final Item<ICellPopulator<Notification>> cellItem, final String componentId, final IModel<Notification> rowModel) {
            PageParameters pp = new PageParameters();
            final Project project = rowModel.getObject().getNotificationEvent().getProject();
            if (project != null) {
                pp.add(PageParameterKeys.MAP.get(Project.class), project.getId());
            }
            Fragment fragment = new Fragment(componentId, "projectTitle", NotificationDataPanel.this);
            BookmarkablePageLink<Object> link = new BookmarkablePageLink<Object>("projectLink", getProjectDetailsPageClass(project), pp) {
                @Override
                protected void onConfigure() {
                    setEnabled(shallBeEnable(project));
                }
            };
            link.setBody(getDataModel(rowModel));
            if (project == null) {
                link.setEnabled(false);
            }
            fragment.add(link);
            cellItem.add(fragment);
        }

        private boolean shallBeEnable(Project project) {
            if (project != null) {
                User user = SciProSession.get().getUser();
                return project.isParticipant(user) || project.isSupervisor(user) || project.isCoSupervisor(user);
            }
            return false;
        }

        private Class<? extends MenuPage> getProjectDetailsPageClass(Project project) {
            User user = SciProSession.get().getUser();
            if (project != null && (project.isSupervisor(user) || project.isCoSupervisor(user))) {
                return SupervisorProjectDetailsPage.class;
            }
            return ProjectDetailsPage.class;
        }
    }

    private class FilterForm extends Form<NotificationService.Filter> {
        public FilterForm(String id, IModel<NotificationService.Filter> filter) {
            super(id, new CompoundPropertyModel<>(filter));
            UserAutoComplete fromUser = new UserAutoComplete("fromUser");
            fromUser.add(new AjaxFormComponentUpdatingBehavior("onchange") {
                @Override
                protected void onUpdate(AjaxRequestTarget target) {
                    target.add(NotificationDataPanel.this);
                }
            });
            add(fromUser);
        }
    }
}
