package se.su.dsv.scipro.notifications.settings.panels;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.attributes.AjaxCallListener;
import org.apache.wicket.ajax.attributes.AjaxRequestAttributes;
import org.apache.wicket.ajax.markup.html.form.AjaxCheckBox;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.StringResourceModel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.notifications.dataobject.Notification;
import se.su.dsv.scipro.notifications.settings.entities.DeliveryMethod;
import se.su.dsv.scipro.notifications.settings.service.DeliveryConfigurationService;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.User;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

public class TypeDeliveryConfigurationPanel extends Panel {

    private static final Logger LOGGER = LoggerFactory.getLogger(TypeDeliveryConfigurationPanel.class);

    static final String METHODS = "methods";
    static final String METHOD = "method";
    static final String EVENTS = "events";
    static final String EVENT = "event";
    static final String DESCRIPTION = "description";

    @Inject
    DeliveryConfigurationService configurationService;

    protected final Notification.Type type;
    private Enum[] eventValues;
    private List<? extends Enum> events;

    public TypeDeliveryConfigurationPanel(final String id, final Notification.Type type, final Enum[] events) {
        super(id);

        this.type = type;
        this.eventValues = events.clone();

    }

    @Override
    protected void onInitialize() {
        super.onInitialize();

        events = filterEvents(eventValues);

        add(new ListView<DeliveryMethod>(METHODS, deliveryMethods()) {
            @Override
            protected void populateItem(ListItem<DeliveryMethod> item) {
                item.add(new Label(METHOD, new StringResourceModel("${name}", TypeDeliveryConfigurationPanel.this, item.getModel())));
            }
        });

        add(new ListView<Enum>(EVENTS, events) {
            @Override
            protected void populateItem(ListItem<Enum> item) {
                item.add(new Label(EVENT, Model.of(item.getModelObject().name())));
                item.add(new Label(DESCRIPTION, getString(item.getModelObject().getClass().getEnclosingClass().getSimpleName() + "." + item.getModelObject().name())));
                item.add(new EventRow(METHODS, item.getModelObject()));
            }
        });
    }

    @Override
    protected void onConfigure() {
        super.onConfigure();
        setVisibilityAllowed(!events.isEmpty());
    }

    protected void setDeliverySetting(Enum event, DeliveryMethod method, boolean setting) {
        LOGGER.info("{} is changing {} for {}.{} to {}", SciProSession.get().getUser(), method, type, event, setting);
        configurationService.setDelivery(type, event, method, setting, getUser());
    }

    protected boolean getDeliverySetting(Enum event, DeliveryMethod method) {
        return configurationService.isDelivery(type, event, method, getUser());
    }

    /**
     * @return the user whose delivery configuration is being edited
     *         default value is {@code null} for system wide configuration
     */
    protected User getUser() {
        return null;
    }

    protected boolean isMethodEnabled(DeliveryMethod method) {
        return true;
    }

    protected boolean includeEvent(Enum event) {
        return true;
    }

    private List<? extends Enum> filterEvents(final Enum[] events) {
        List<Enum> filteredEvents = new ArrayList<>(events.length);
        for (Enum event : events) {
            if (includeEvent(event)) {
                filteredEvents.add(event);
            }
        }
        return filteredEvents;
    }

    private List<DeliveryMethod> deliveryMethods() {
        return Arrays.asList(DeliveryMethod.values());
    }

    private class EventRow extends ListView<DeliveryMethod> {

        private final Enum event;

        EventRow(String id, final Enum event) {
            super(id, deliveryMethods());

            this.event = event;
        }
        @Override
        protected void populateItem(final ListItem<DeliveryMethod> item) {
            boolean enabled = getDeliverySetting(event, item.getModelObject());
            item.add(new MethodCheckbox(METHOD, Model.of(enabled), event, item.getModelObject()));
        }

    }

    private final class MethodCheckbox extends AjaxCheckBox {
        private final Enum event;

        private final DeliveryMethod method;

        private MethodCheckbox(final String id, final IModel<Boolean> model, Enum event, DeliveryMethod method) {
            super(id, model);
            this.event = event;
            this.method = method;
        }

        @Override
        protected void onUpdate(AjaxRequestTarget target) {
            setDeliverySetting(event, method, getModelObject());
        }
        @Override
        protected void onConfigure() {
            setEnabled(isMethodEnabled(method));
        }

        @Override
        protected void updateAjaxAttributes(final AjaxRequestAttributes attributes) {
            super.updateAjaxAttributes(attributes);
            AjaxCallListener ajaxCallListener = new AjaxCallListener();
            ajaxCallListener.onPrecondition("if (!confirm('Are you sure?')) {window.location.reload(); return false;}");
            attributes.getAjaxCallListeners().add(ajaxCallListener);
        }
    }
}
