package se.su.dsv.scipro.peer;

import org.apache.wicket.extensions.markup.html.repeater.util.SortableDataProvider;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.link.Link;
import org.apache.wicket.markup.html.navigation.paging.PagingNavigator;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.markup.repeater.Item;
import org.apache.wicket.markup.repeater.data.DataView;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.StringResourceModel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.data.domain.PageRequest;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.notifications.NotificationController;
import se.su.dsv.scipro.notifications.dataobject.NotificationSource;
import se.su.dsv.scipro.notifications.dataobject.PeerEvent;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.JavascriptEventConfirmation;

import javax.inject.Inject;
import java.util.Arrays;
import java.util.Iterator;

public class PeerRequestSelectionPanel extends GenericPanel<Project> {

    public static final int REQUESTS_PER_PAGE = 8;
    public static final String CURRENTLY_AWAITING = "currentlyAwaiting";
    public static final String DAYS_TO_SUBMIT = "daysToSubmit";
    public static final String RECENTLY_REVIEWED = "recentlyReviewed";
    public static final String AVAILABLE_REQUESTS = "availableRequests";
    public static final String TITLE = "title";
    public static final String REQUEST_DETAILS = "requestDetails";
    public static final String ACCEPT = "accept";
    private static final Logger LOGGER = LoggerFactory.getLogger(PeerRequestSelectionPanel.class);
    @Inject
    private PeerReviewService peerReviewService;
    @Inject
    private PeerRequestService peerRequestService;
    @Inject
    private PeerPortal peerPortal;
    @Inject
    private NotificationController notificationController;

    public PeerRequestSelectionPanel(String id, IModel<Project> model) {
        super(id, model);
        add(new FeedbackPanel("feedback"));
        addInfoBox();
        addRequestList();
    }

    private void addRequestList() {
        DataView<PeerRequest> requestList = new DataView<PeerRequest>(AVAILABLE_REQUESTS, new RequestProvider(), REQUESTS_PER_PAGE) {
            @Override
            protected void populateItem(Item<PeerRequest> item) {
                item.add(new Label(TITLE, new PropertyModel<>(item.getModel(), "project.title")));
                item.add(new RequestDetailsPanel(REQUEST_DETAILS, item.getModel()));
                item.add(new AcceptLink(ACCEPT, item.getModel()));
            }
        };
        add(requestList);
        add(new PagingNavigator("pagingNavigator", requestList));
    }

    private void addInfoBox() {
        add(new Label(CURRENTLY_AWAITING, new StringResourceModel("currentlyAwaiting", null,
                peerRequestService.countAvailableRequests(getUser(), RequestStatus.WAITING, getModelObject().getProjectType()))));
        add(new Label(DAYS_TO_SUBMIT, new StringResourceModel("daysToSubmit", new PropertyModel<>(getModel(), "projectType.projectTypeSettings"))));
        add(new Label(RECENTLY_REVIEWED, new StringResourceModel("recentlyReviewed", new PropertyModel<>(getModel(), "projectType.projectTypeSettings"))));
    }

    @Override
    protected void onConfigure() {
        setVisibilityAllowed(noReviewInProgress() && hasMoreRequestsThanReviews());
    }


    private boolean hasMoreRequestsThanReviews() {
        return countRequests() > countReviews();
    }

    private long countReviews() {
        PeerReviewService.Filter reviewParams = new PeerReviewService.Filter();
        reviewParams.setAuthor(getUser());
        reviewParams.setAuthorProject(getModelObject());
        reviewParams.setStatuses(Arrays.asList(PeerReview.ReviewStatus.COMPLETED, PeerReview.ReviewStatus.IN_PROGRESS));
        return peerReviewService.count(reviewParams);
    }

    private long countRequests() {
        PeerRequestService.Filter requestParams = new PeerRequestService.Filter();
        requestParams.setAuthor(getUser());
        requestParams.setStatuses(Arrays.asList(RequestStatus.WAITING, RequestStatus.FINISHED, RequestStatus.TAKEN));
        requestParams.setProject(getModelObject());
        return peerRequestService.count(requestParams);
    }

    private boolean noReviewInProgress() {
        return peerReviewService.findInProgressReviews(getUser(), getModelObject()).isEmpty();
    }

    private User getUser() {
        return SciProSession.get().getUser();
    }

    private final class AcceptLink extends Link<PeerRequest> {

        private AcceptLink(String id, IModel<PeerRequest> model) {
            super(id, model);
            add(new JavascriptEventConfirmation("onclick", new StringResourceModel("confirmAccept", model)));
        }

        @Override
        public void onClick() {
            try {
                PeerReview review = peerPortal.acceptReview(getModelObject(), getUser(), PeerRequestSelectionPanel.this.getModelObject());
                notificationController.notifyPeer(review, PeerEvent.Event.REVIEW_ACCEPTED, new NotificationSource());
            } catch (PartOfProjectException ignored) {
                String partOfProjectException = getString("partOfProjectException");
                getSession().error(partOfProjectException);
                LOGGER.info(partOfProjectException, ignored);
            } catch (RecentlyReviewedException ignored) {
                String recentlyReviewedException = getString("recentlyReviewedException");
                getSession().error(recentlyReviewedException);
                LOGGER.info(recentlyReviewedException, ignored);
            } catch (RequestTakenException ignored) {
                String requestTakenException = getString("requestTakenException");
                getSession().error(requestTakenException);
                LOGGER.info(requestTakenException, ignored);
            } catch (Exception ignored) {
                String unknownError = getString("unknownError");
                getSession().error(unknownError);
                LOGGER.warn(unknownError, ignored);
            }
            setResponsePage(PeerPortalPage.class);
        }
    }

    private class RequestProvider extends SortableDataProvider<PeerRequest, String> {

        @Override
        public Iterator<? extends PeerRequest> iterator(long first, long count) {
            return peerRequestService.findAvailableRequests(getUser(), RequestStatus.WAITING, getModelObject().getProjectType(), new PageRequest(first, count)).iterator();
        }

        @Override
        public long size() {
            return peerRequestService.countAvailableRequests(getUser(), RequestStatus.WAITING, getModelObject().getProjectType());
        }

        @Override
        public IModel<PeerRequest> model(PeerRequest object) {
            return new DetachableServiceModel<>(peerRequestService, object);
        }

    }
}
