/**
 *
 */
package se.su.dsv.scipro.peer;

import org.apache.wicket.Page;
import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.form.AjaxFormComponentUpdatingBehavior;
import org.apache.wicket.ajax.markup.html.form.AjaxCheckBox;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.DropDownChoice;
import org.apache.wicket.markup.html.link.Link;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.*;
import org.joda.time.DateTime;
import se.su.dsv.scipro.components.*;
import se.su.dsv.scipro.date.DeactivatableDatePickerPanel;
import se.su.dsv.scipro.peer.PeerReview.ReviewStatus;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.*;

import javax.inject.Inject;
import java.io.Serializable;
import java.util.*;

public class PeerReviewStatisticsPanel extends Panel {

    @Inject
    private PeerReviewService peerReviewService;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private LanguageService languageService;


    private PeerReviewService.Filter filter = new PeerReviewService.Filter();
    private WebMarkupContainer wmc;
    private final PageSupplier pageSupplier;

    public PeerReviewStatisticsPanel(String id, PageSupplier pageSupplier) {
        super(id);
        this.pageSupplier = pageSupplier;
        this.setOutputMarkupId(true);
        setDefaultModel(new CompoundPropertyModel<>(filter));

        addFilterControls();
        setupStatisticDetailsPart();
    }

    private void addFilterControls() {
        prepareDefaultFilter();
        addProjectTypesFilter();
        addOtherFilter();
        addAuthorFilter();
        addLanguageFilter();
        addDateFilter();
    }

    private void prepareDefaultFilter() {
        filter.setProjectTypes(projectTypeService.findByDegreeLevels(EnumSet.of(DegreeLevel.FIRST_CYCLE, DegreeLevel.SECOND_CYCLE)));
        filter.setCreatedBefore(new Date());
        filter.setCreatedAfter(DateTime.now().minusMonths(6).toDate());
    }

    private void addAuthorFilter() {
        final AuthorAutoComplete authorField = new AuthorAutoComplete("filterAuthor") {

            @Override
            protected void action(AjaxRequestTarget pTarget, User newSelection) {
                filter.setAuthor(newSelection);
                pTarget.add(wmc);
            }
        };
        add(authorField);
    }

    private void addOtherFilter() {
        final AjaxCheckBox supervisorProjects = new AjaxCheckBox("supervisorCheckBox", Model.of(false)) {

            @Override
            protected void onUpdate(AjaxRequestTarget target) {
                if (this.getModelObject()) {
                    User supervisor = SciProSession.get().getUser();
                    filter.setSupervisor(supervisor);
                } else {
                    filter.setSupervisor(null);
                }
                target.add(wmc);
            }
        };
        add(supervisorProjects);
    }

    private void addDateFilter() {
        add(new DeactivatableDatePickerPanel("datePanel",
                new PropertyModel<Date>(filter, "createdAfter"),
                new PropertyModel<Date>(filter, "createdBefore"),
                Model.of(true)) {
            @Override
            protected void datesChanged(AjaxRequestTarget target) {
                target.add(wmc);
            }
        });
    }

    private void addLanguageFilter() {
        List<Language> languages = languageService.findAll();
        final PropertyModel<Language> languagePropertyModel = new PropertyModel<>(filter, "language");
        final DropDownChoice<Language> languageFilter = new DropDownChoice<>("filterLanguage",
                languagePropertyModel, languages, new ChoiceRenderer<>("name", "id"));
        languageFilter.add(new AjaxFormComponentUpdatingBehavior("onchange") {

            @Override
            protected void onUpdate(AjaxRequestTarget target) {
                target.add(wmc);
            }
        });

        filter.setLanguage(languagePropertyModel.getObject());
        add(languageFilter);
    }

    private void addProjectTypesFilter() {
        add(new AjaxCheckBoxMultipleChoice<ProjectType>("projectTypes",
                projectTypeService.findWithModule(ProjectModule.PEER), new ChoiceRenderer<ProjectType>("name", "id")) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                target.add(wmc);
            }
        });
    }

    private void setupStatisticDetailsPart() {
        wmc = new WebMarkupContainer("container");
        wmc.setOutputMarkupId(true);
        add(wmc);
        wmc.add(new Label("totalPeerReviews", new LoadableDetachableModel<Long>() {
            @Override
            protected Long load() {
                return peerReviewService.count(filter);
            }
        }));
        wmc.add(new Link<Void>("chartLink") {
            @Override
            public void onClick() {
                setResponsePage(pageSupplier.chartPage(filter));
            }
        });
        ListView<ReviewStatus> statusListView = new ListView<ReviewStatus>("statusList", Arrays.asList(ReviewStatus.values())) {
            @Override
            protected void populateItem(final ListItem<ReviewStatus> item) {
                item.add(new Label("numberOfPeerReviews", peerReviewService.countByStatus(item.getModelObject(), filter)));
                Link detailsLink = new Link("detailsLink") {
                    @Override
                    public void onClick() {
                        setResponsePage(pageSupplier.statusPage(item.getModel(), filter));
                    }
                };
                detailsLink.add(new Label("status", new ResourceModel(item.getModelObject().name())));
                item.add(detailsLink);
                item.add(new Label("withOutChecklistTmp", peerReviewService.countByStatusAndNoCheckListTemplate(filter, item.getModelObject())));
            }
        };
        wmc.add(statusListView);
    }

    public interface PageSupplier extends Serializable {
        Page statusPage(IModel<PeerReview.ReviewStatus> model, PeerReviewService.Filter filter);
        Page chartPage(PeerReviewService.Filter filter);
    }
}
