package se.su.dsv.scipro.peer;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.Button;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.form.TextArea;
import org.apache.wicket.markup.html.form.upload.FileUpload;
import org.apache.wicket.markup.html.form.upload.FileUploadField;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.PropertyListView;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.CompoundPropertyModel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.util.ListModel;
import scala.Option;
import se.su.dsv.scipro.checklists.TrafficLightPanel;
import se.su.dsv.scipro.components.AutoSavingBehaviour;
import se.su.dsv.scipro.components.DateLabel;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.files.WicketFileUpload;
import se.su.dsv.scipro.grading.ScrollingSaveButtonPanel;
import se.su.dsv.scipro.util.JavascriptEventConfirmation;

import javax.inject.Inject;

public class PerformReviewPanel extends GenericPanel<PeerReview> {

    public static final String REVIEW_FORM = "reviewForm";
    public static final String SUBMIT_BUTTON = "submitButton";
    public static final String COMMENT = "comment";
    public static final String ANSWERS = "answers";
    public static final String QUESTION = "question";
    public static final String MOTIVATION = "motivation";
    public static final String ANSWER = "answer";
    public static final String FILE = "reviewFileUpload";
    public static final String SAVE_BUTTON = "saveButton";
    public static final String DEADLINE = "deadline";
    public static final String REQUEST_DETAILS = "requestDetails";

    @Inject
    private PerformReviewService performReviewService;

    public PerformReviewPanel(String id, IModel<PeerReview> peerReview) {
        super(id, peerReview);

        PropertyModel<PeerRequest> peerRequest = new PropertyModel<>(getModel(), "peerRequest");
        add(new ReviewPageRequestDetailsPanel(REQUEST_DETAILS, peerRequest));
        add(new PerformReviewForm(REVIEW_FORM, getModel()));
    }

    @Override
    protected void onConfigure() {
        setVisible(getModelObject() != null);
    }

    private class PerformReviewForm extends Form<PeerReview> {

        private final FileUploadField reviewFileUpload;

        public PerformReviewForm(String id, final IModel<PeerReview> reviewModel) {
            super(id, new CompoundPropertyModel<>(reviewModel));
            setMultiPart(true);

            add(new DateLabel(DEADLINE, DateStyle.DATETIME));

            add(new PropertyListView<Answer>(ANSWERS) {
                @Override
                protected void populateItem(ListItem<Answer> item) {
                    item.add(new Label(QUESTION));
                    item.add(new TrafficLightPanel(ANSWER));
                    item.add(new TextArea<>(MOTIVATION).add(new AutoSavingBehaviour() {
                        @Override
                        protected void onUpdate(final AjaxRequestTarget target) {
                            performReviewService.save(PerformReviewForm.this.getModelObject());
                        }
                    }));
                }

                @Override
                protected void onConfigure() {
                    setVisible(!getModelObject().isEmpty());
                }
            });

            add(new TextArea<String>(COMMENT) {
                {
                    add(new AutoSavingBehaviour() {
                        @Override
                        protected void onUpdate(final AjaxRequestTarget target) {
                            performReviewService.save(PerformReviewForm.this.getModelObject());
                        }
                    });
                }
                @Override
                protected void onConfigure() {
                    setVisible(PerformReviewForm.this.getModelObject().getAnswers().isEmpty());
                }
            });

            reviewFileUpload = new FileUploadField(FILE, new ListModel<FileUpload>());
            add(reviewFileUpload);

            addButtonWithJavaScriptConformation();
            add(new ScrollingSaveButtonPanel(SAVE_BUTTON, this));
        }

        private void addButtonWithJavaScriptConformation() {
            Button submitButton = new Button(SUBMIT_BUTTON) {
                @Override
                public void onSubmit() {
                    try {
                        Option<se.su.dsv.scipro.file.FileUpload> fileUploadOption =
                                WicketFileUpload.ofOptional(reviewFileUpload.getFileUpload());
                        performReviewService.submit(PerformReviewForm.this.getModelObject(), fileUploadOption);
                        success(getString("submitted"));
                        PerformReviewForm.this.detach();
                    } catch (MissingAnswerException ignored) {
                        error(getString("missing.answer"));
                    } catch (TooShortCommentException ignored) {
                        error(getString("too.short.comment"));
                    }
                }
            };
            submitButton.add(new JavascriptEventConfirmation("onclick",
                    "Please confirm that you have completed the review and taken care to highlight both strengths and weaknesses."));
            add(submitButton);
        }

        @Override
        public void onSubmit() {
            if (!hasError() && getModelObject() != null) {
                setModelObject(performReviewService.save(getModelObject()));
                info(getString("review.saved.successfully"));
            }
        }
    }
}
