package se.su.dsv.scipro.phd;

import org.apache.wicket.behavior.AttributeAppender;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.form.upload.FileUploadField;
import org.apache.wicket.markup.html.link.Link;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.ResourceModel;
import se.su.dsv.scipro.components.ListAdapterModel;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.files.WicketFileUpload;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.repository.panels.ViewAttachmentPanel;
import se.su.dsv.scipro.security.auth.ProjectModuleComponent;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.util.JavascriptEventConfirmation;

import javax.inject.Inject;
import java.util.List;
import java.util.Set;

@ProjectModuleComponent(ProjectModule.STUDY_PLAN)
public class StudyPlanPanel extends GenericPanel<Project> {
    public static final String BASIC_CONTRACT_FORM = "basicContractForm";
    public static final String BASIC_CONTRACT_UPLOAD = "basicContractUpload";
    public static final String ANNUAL_REVIEWS = "annualReviews";
    public static final String NO_REVIEWS = "noReviews";
    public static final String ANNUAL_REVIEW_FORM = "annualReviewForm";
    public static final String ANNUAL_REVIEW_UPLOAD = "annualReviewUpload";
    public static final String ANNUAL_REVIEW = "annualReview";
    public static final String DELETE = "delete";
    public static final String ADD = "add";
    public static final String CONTAINER = "container";

    @Inject
    private StudyPlanService studyPlanService;

    private ListView<AnnualReview> annualReviews;
    private WebMarkupContainer container;

    public StudyPlanPanel(String id, final IModel<Project> model) {
        super(id, model);
        add(new FencedFeedbackPanel("feedback", this));
        addContainer();
        addCurrentStudyPlan();
        addAdd();
        addBasicContractForm();
        addAnnualReviewForm();
        addAnnualReviewList();
        addNoAnnualReviewLabel();
    }

    private void addCurrentStudyPlan() {
        add(new ViewAttachmentPanel("current", new PropertyModel<FileDescription>(getStudyPlan(), "basicContract"), DateStyle.DATE));
        add(new DeleteLink("deleteCurrent", getStudyPlan()));
        add(new Label("noCurrent", new ResourceModel("no.current")) {
            @Override
            protected void onConfigure() {
                setVisible(getStudyPlan().getObject().getId() == null);
            }
        });
    }

    private void addAdd() {
        String toggleScript = String.format("$('#%s').slideToggle();", container.getMarkupId());
        add(new WebMarkupContainer(ADD).add(AttributeAppender.replace("onclick", toggleScript)));
    }

    private void addContainer() {
        container = new WebMarkupContainer(CONTAINER);
        container.setOutputMarkupPlaceholderTag(true);
        add(container);
    }

    private void addBasicContractForm() {
        final FileUploadField basicContractUpload = new FileUploadField(BASIC_CONTRACT_UPLOAD);
        basicContractUpload.setRequired(true);
        Form basicContractForm = new Form<Void>(BASIC_CONTRACT_FORM) {
            @Override
            protected void onSubmit() {
                studyPlanService.addBasicContract(WicketFileUpload.from(basicContractUpload.getFileUpload()), getStudyPlan().getObject());
                this.success(getString("storedBasicContract"));
            }

            @Override
            protected void onConfigure() {
                setVisibilityAllowed(getStudyPlan().getObject().getBasicContract() == null);
            }
        };
        add(basicContractForm);
        basicContractForm.add(basicContractUpload);
    }

    private void addAnnualReviewForm() {
        final FileUploadField annualReviewUpload = new FileUploadField(ANNUAL_REVIEW_UPLOAD);
        annualReviewUpload.setRequired(true);

        Form annualReviewForm = new Form<Void>(ANNUAL_REVIEW_FORM) {
            @Override
            protected void onSubmit() {
                studyPlanService.addAnnualReview(WicketFileUpload.from(annualReviewUpload.getFileUpload()), getStudyPlan().getObject());
                this.success(getString("storedAnnualReview"));
            }

            @Override
            protected void onConfigure() {
                setVisibilityAllowed(getStudyPlan().getObject().getBasicContract() != null);
            }
        };
        container.add(annualReviewForm);

        annualReviewForm.add(annualReviewUpload);
    }

    private void addAnnualReviewList() {
        annualReviews = new ListView<AnnualReview>(ANNUAL_REVIEWS, getAnnualReviews()) {
            @Override
            protected void populateItem(final ListItem<AnnualReview> item) {
                item.add(new ViewAttachmentPanel(ANNUAL_REVIEW, new PropertyModel<FileDescription>(item.getModel(), "annualReview")));
                item.add(new Link<Void>(DELETE) {
                    @Override
                    public void onClick() {
                        studyPlanService.removeAnnualReview(getStudyPlan().getObject(), item.getModelObject());
                        annualReviews.detach();
                        StudyPlanPanel.this.success(getString("deletedAnnualReview"));
                    }
                }.add(new JavascriptEventConfirmation("onclick", new ResourceModel("annualReviewDelete"))));
            }

            @Override
            protected void onConfigure() {
                setVisibilityAllowed(getStudyPlan().getObject().getBasicContract() != null);
            }
        };
        add(annualReviews);
    }

    private void addNoAnnualReviewLabel() {
        add(new Label(NO_REVIEWS, new ResourceModel("noAnnualReviews")) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(getStudyPlan().getObject().getAnnualReviews().isEmpty());
            }
        });
    }

    private IModel<StudyPlan> getStudyPlan() {
        return new LoadableDetachableModel<StudyPlan>() {
            @Override
            protected StudyPlan load() {
                StudyPlan studyPlan = studyPlanService.findByProject(getModelObject());
                return studyPlan!= null ? studyPlan : new StudyPlan(getModelObject());
            }
        };
    }

    private IModel<List<AnnualReview>> getAnnualReviews() {
        return new ListAdapterModel<>(new PropertyModel<Set<AnnualReview>>(getStudyPlan(), "annualReviews"));
    }

    private class DeleteLink extends Link<StudyPlan> {

        public DeleteLink(String id, IModel<StudyPlan> model) {
            super(id, model);
            add(new JavascriptEventConfirmation("onclick", new ResourceModel("basicContractDelete")));
        }

        @Override
        public void onClick() {
            studyPlanService.delete(getModelObject());
            StudyPlanPanel.this.detach();
            StudyPlanPanel.this.success(getString("deletedBasicContract"));
        }

        @Override
        protected void onConfigure() {
            setVisible(getModelObject().getId() != null);
        }
    }
}