package se.su.dsv.scipro.project.panels;

import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.*;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import se.su.dsv.scipro.components.AutoHidingListView;
import se.su.dsv.scipro.components.ListAdapterModel;
import se.su.dsv.scipro.date.DateService;
import se.su.dsv.scipro.profile.UserLinkPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.pages.SupervisorEditProjectPage;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.List;
import java.util.Set;

public class ProjectHeaderPanel extends GenericPanel<Project> {
	static final String TITLE = "title";
	static final String STATUS = "projectStatus";
	static final String PROJECT_TYPE = "projectType";
	static final String START_DATE = "startDate";
	static final String SUPERVISOR = "headSupervisor";
	static final String AUTHOR_LIST = "authorList";
	static final String AUTHOR = "author";
	static final String CO_SUPERVISOR_LIST = "coSupervisorList";
	static final String CO_SUPERVISOR = "coSupervisor";
	static final String REVIEWER_LIST = "reviewerList";
	static final String REVIEWER = "reviewer";

	static final String EDIT_LINK = "editLink";
	static final String CREDITS = "credits";

	@Inject
	private DateService dateService;

	public ProjectHeaderPanel(String id, IModel<Project> model) {
		super(id, new CompoundPropertyModel<>(model));

		add(new Label(TITLE));
		add(new Label(PROJECT_TYPE, new PropertyModel<String>(model, "projectType.name")));
		add(new Label(CREDITS) {
			@Override
			protected void onConfigure() {
				setVisible(getModelObject().getCredits() > 0);
			}
		});

		addDetails();
		addEditLink(model);
		addMisc(model);

		addAuthors(model);
		addCoSupervisors(model);
		addReviewer(model);
	}

	private void addDetails() {
        add(new Label(STATUS));
    }

	private void addEditLink(final IModel<Project> model) {
		PageParameters pp = new PageParameters();
		pp.add(PageParameterKeys.MAP.get(Project.class), model.getObject().getId());

		BookmarkablePageLink<Void> editLink = new BookmarkablePageLink<Void>(EDIT_LINK, SupervisorEditProjectPage.class, pp){
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(isHeadSupervisor(model.getObject()));
            }
        };
		add(editLink);
	}

	private void addMisc(IModel<Project> model) {
		add(new Label(START_DATE, getStartDate(model)));
		add(new UserLinkPanel(SUPERVISOR, new PropertyModel<User>(model, "headSupervisor")));
	}

	private IModel<String> getStartDate(IModel<Project> model) {
		if (model.getObject().getDateCreated() == null) {
			return new StringResourceModel("noStartDate", ProjectHeaderPanel.this, null);
		}
		return new Model<>(dateService.format(model.getObject().getDateCreated()));
	}

	private void addAuthors(IModel<Project> model) {
		add(new ListView<User>(AUTHOR_LIST, new ListAdapterModel<>(new PropertyModel<Set<User>>(model, "projectParticipants"))) {
			@Override
			protected void populateItem(ListItem<User> item) {
				item.add(new UserLinkPanel(AUTHOR, item.getModel()));
			}
		});
	}

	private void addCoSupervisors(IModel<Project> model) {
		IModel<List<User>> coSupervisorList =
				new ListAdapterModel<>(new PropertyModel<Set<User>>(model, "coSupervisors"));
		add(new AutoHidingListView<User>(CO_SUPERVISOR_LIST, coSupervisorList) {
			@Override
			protected void populateItem(ListItem<User> item) {
				item.add(new UserLinkPanel(CO_SUPERVISOR, item.getModel()));
			}
		});
	}

	private void addReviewer(IModel<Project> model) {
		add(new AutoHidingListView<User>(REVIEWER_LIST, new ListAdapterModel<>(new PropertyModel<Set<User>>(model, "reviewers"))) {
			@Override
			protected void populateItem(ListItem<User> item) {
				item.add(new UserLinkPanel(REVIEWER, item.getModel()));
			}
		});
	}

	private boolean isHeadSupervisor(Project project) {
		User headSupervisor = project.getHeadSupervisor();
		User user = SciProSession.get().getUser();
		return headSupervisor.equals(user);
	}
}
