package se.su.dsv.scipro.project.panels;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.behavior.AttributeAppender;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.EnumLabel;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.EnumChoiceRenderer;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.PropertyListView;
import org.apache.wicket.markup.html.panel.GenericPanel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import se.su.dsv.scipro.components.AjaxCheckBoxMultipleChoice;
import se.su.dsv.scipro.icons.ImageObject;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.project.StateOfMind;
import se.su.dsv.scipro.project.pages.ProjectDetailsPage;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

public class ProjectsOverviewPanel extends GenericPanel<List<Project>> {

    public static final String TABLE_CONTAINER = "tableContainer";
    public static final String NO_ACTIVE_PROJECTS_LABEL = "noActiveProjectsLabel";

    @Inject
    private ProjectService projectService;
    private ProjectService.Filter filter = new ProjectService.Filter();
    private WebMarkupContainer container;

    public ProjectsOverviewPanel(String id) {
        super(id);
        setModel(getProjects());
        prepareFilter();

        container = new WebMarkupContainer(TABLE_CONTAINER);
        container.setOutputMarkupId(true);
        add(container);

        setUpNoProjectLabel();
        setUpNoActiveProjectsLabel();
        setUpStatusFilter();
        setUpProjectList();
    }

    private void setUpNoActiveProjectsLabel() {
        container.add(new Label(NO_ACTIVE_PROJECTS_LABEL, "You don't have any active projects") {
            @Override
            protected void onConfigure() {
                super.onConfigure();
                setVisibilityAllowed(!hasActiveProject() && !hasCompletedProject());
            }
        });
    }

    private boolean hasActiveProject() {
        return !projectService.findByUserAndStatus(SciProSession.get().getUser(), ProjectStatus.ACTIVE).isEmpty();
    }

    private boolean hasCompletedProject() {
        return !projectService.findByUserAndStatus(SciProSession.get().getUser(), ProjectStatus.COMPLETED).isEmpty();
    }

    private void prepareFilter() {
        filter.setAuthorUser(SciProSession.get().getUser());
        filter.setStatuses(defaultStatusFilter());
    }

    private static Collection<ProjectStatus> defaultStatusFilter() {
        List<ProjectStatus> statuses = new ArrayList<>();
        statuses.add(ProjectStatus.ACTIVE);
        return statuses;
    }

    private IModel<List<Project>> getProjects() {
        return new LoadableDetachableModel<List<Project>>() {
            @Override
            protected List<Project> load() {
                return projectService.findAll(filter);
            }
        };
    }

    private void setUpStatusFilter() {
        add(new AjaxCheckBoxMultipleChoice<ProjectStatus>("statuses", new PropertyModel<Collection<ProjectStatus>>(filter, "statuses"), Arrays.asList(ProjectStatus.ACTIVE, ProjectStatus.COMPLETED), new EnumChoiceRenderer<ProjectStatus>(this)) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                target.add(container);
            }
        });
    }

    private void setUpNoProjectLabel() {
        container.add(new Label("emptyLabel", "You don't have any projects with the selected project status.") {
            @Override
            protected void onConfigure() {
                super.onConfigure();
                setVisible(getModelObject().isEmpty() && (hasActiveProject() || hasCompletedProject()));
            }
        });
    }

    private void setUpProjectList() {
        container.add(new PropertyListView<Project>("listView", getModel()) {
            @Override
            protected void populateItem(ListItem<Project> item) {
                Project p = item.getModelObject();

                if (p.getStateOfMind() == StateOfMind.FINE) {
                    item.add(new ImageObject("state", ImageObject.TWENTYFOUR, ImageObject.LIGHT_GREEN));
                } else if (p.getStateOfMind() == StateOfMind.NEUTRAL) {
                    item.add(new ImageObject("state", ImageObject.TWENTYFOUR, ImageObject.LIGHT_YELLOW));
                } else {
                    item.add(new ImageObject("state", ImageObject.TWENTYFOUR, ImageObject.LIGHT_RED));
                }

                item.add(new Label("projectType.name"));

                PageParameters pp = new PageParameters();
                pp.add(PageParameterKeys.MAP.get(Project.class), p.getId());
                BookmarkablePageLink<Void> projectLink = new BookmarkablePageLink<>("projectLink", ProjectDetailsPage.class, pp);
                Label title = new Label("title", p.getTitle().length() < 31 ? p.getTitle() : p.getTitle().substring(0, 31) + "...");
                title.add(new AttributeAppender("title", Model.of(p.getTitle())));
                projectLink.add(title);
                item.add(projectLink);

                boolean first = true;
                StringBuilder authors = new StringBuilder("");
                for (User user : p.getProjectParticipants()) {
                    if (first) {
                        first = false;
                        authors.append(user.getFirstName()).append(" ").append(user.getLastName());
                    } else {
                        authors.append(", ").append(user.getFirstName()).append(" ").append(user.getLastName());
                    }
                }
                item.add(new Label("authors", authors));
                item.add(new Label("supervisor", p.getHeadSupervisor().getFullName()));
                item.add(new EnumLabel<ProjectStatus>("projectStatus"));
                add(item);
            }
        });
    }
}
