package se.su.dsv.scipro.projectstate;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.markup.html.form.AjaxButton;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.basic.MultiLineLabel;
import org.apache.wicket.markup.html.form.Form;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.util.ListModel;
import se.su.dsv.scipro.components.AjaxDropDownChoice;
import se.su.dsv.scipro.components.CounterTextArea;
import se.su.dsv.scipro.components.DateLabel;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.forum.ProjectForumService;
import se.su.dsv.scipro.icons.ImageObject;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.StateOfMind;
import se.su.dsv.scipro.security.auth.ProjectModuleComponent;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.ProjectModule;

import javax.inject.Inject;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashSet;

import static se.su.dsv.scipro.data.enums.DateStyle.DATETIME;

@ProjectModuleComponent(ProjectModule.PROJECT_STATE)
public abstract class ProjectStatePanel extends Panel {

    public static final int MAX_CHARS = 255;

    @Inject
    ProjectService projectService;
    @Inject
    ProjectForumService projectForumService;

    private final IModel<Project> model;
    private ImageObject green;
    private ImageObject yellow;
    private ImageObject red;
    private AjaxDropDownChoice<StateOfMind> dropDown;
    private DateLabel date;
    private Form form;
    private MultiLineLabel reasonLabel;

    /**
     * This is arguably a better method for controlling visibility of components compared to looking at what page this
     * panel has been added to (supervisor page or author page) because implementing checks for the page increases
     * adds dependencies and increases coupling. Instead, its easy to simply send "false" to this method if this panel
     * is added to a supervisor page, and "true" if it is a author page.
     */
    public abstract boolean isAuthorView();

    public ProjectStatePanel(String id, final IModel<Project> model) {
        super(id, model);
        this.model = model;
        addLights();
        addStateLabel();
        addUpdatedLabel();
        addDropDown();
        addReasonLabel();
        addForm();
    }

    private void addLights() {
        green = new ImageObject(GREEN, ImageObject.TWENTYFOUR + ImageObject.LIGHT_GREEN) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(model.getObject().getStateOfMind().equals(StateOfMind.FINE));
            }
        };
        add(green);
        green.setOutputMarkupPlaceholderTag(true);

        yellow = new ImageObject(YELLOW, ImageObject.TWENTYFOUR + ImageObject.LIGHT_YELLOW) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(model.getObject().getStateOfMind().equals(StateOfMind.NEUTRAL));
            }
        };
        add(yellow);
        yellow.setOutputMarkupPlaceholderTag(true);

        red = new ImageObject(RED, ImageObject.TWENTYFOUR + ImageObject.LIGHT_RED) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(model.getObject().getStateOfMind().equals(StateOfMind.NEEDHELP));
            }
        };
        add(red);
        red.setOutputMarkupPlaceholderTag(true);
    }

    private void addStateLabel() {
        add(new Label(STATE, model.getObject().getStateOfMind().getDescription()) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(!isAuthorView());
            }
        });
    }

    private void addUpdatedLabel() {
        date = new DateLabel(DATE, new Model<>(model.getObject().getStateOfMindDate()), DATETIME);
        add(date);
        date.setOutputMarkupId(true);
    }

    private void addReasonLabel() {
        reasonLabel = new MultiLineLabel(REASON_LABEL, Model.of(model.getObject().getStateOfMindReason())) {
            @Override
            protected void onConfigure() {
                setVisibilityAllowed(model.getObject().getStateOfMind().equals(StateOfMind.NEEDHELP));
            }
        };
        add(reasonLabel);
        reasonLabel.setOutputMarkupPlaceholderTag(true);
    }

    private void addForm() {
        form = new Form(FORM);
        add(form);
        form.setOutputMarkupPlaceholderTag(true);
        form.setVisibilityAllowed(false);

        final CounterTextArea reasonArea = new CounterTextArea(REASON_AREA, Model.of(""), MAX_CHARS);
        reasonArea.setRequired(true);
        form.add(reasonArea);

        form.add(new AjaxButton(SUBMIT) {
            @Override
            protected void onSubmit(AjaxRequestTarget target, Form<?> form) {
                model.getObject().setStateOfMind(dropDown.getModelObject());
                model.getObject().setStateOfMindDate(new Date());
                model.getObject().setStateOfMindReason(reasonArea.getModelObject());
                reasonLabel.setDefaultModel(Model.of(model.getObject().getStateOfMindReason()));
                model.setObject(projectService.save(model.getObject()));
                date = new DateLabel(DATE, new Model<>(model.getObject().getStateOfMindDate()), DATETIME);
                getParent().getParent().replace(date);
                date.setOutputMarkupId(true);
                form.setVisibilityAllowed(false);
                if (model.getObject().hasModule(ProjectModule.FORUM)) {
                    projectForumService.createThread(model.getObject(), SciProSession.get().getUser(), getString("forumSubject"), reasonArea.getModelObject(), new HashSet<FileUpload>());
                }
                addComponents(target);
            }

            @Override
            protected void onError(AjaxRequestTarget target, Form<?> form) {
                addComponents(target);
            }
        });
    }

    private void addDropDown() {
        dropDown = new AjaxDropDownChoice<StateOfMind>(DROPDOWN, Model.of(model.getObject().getStateOfMind()), new ListModel<>(new ArrayList<>(Arrays.asList(StateOfMind.values())))) {
            @Override
            public void onNewSelection(AjaxRequestTarget target, StateOfMind objectSelected) {
                if (!objectSelected.equals(StateOfMind.NEEDHELP)) {
                    form.setVisibilityAllowed(false);
                    model.getObject().setStateOfMind(objectSelected);
                    model.getObject().setStateOfMindDate(new Date());
                    model.getObject().setStateOfMindReason("");
                    reasonLabel.setDefaultModel(Model.of(model.getObject().getStateOfMindReason()));
                    model.setObject(projectService.save(model.getObject()));
                    date = new DateLabel(DATE, new Model<>(model.getObject().getStateOfMindDate()), DATETIME);
                    getParent().replace(date);
                    date.setOutputMarkupId(true);
                } else {
                    form.setVisibilityAllowed(true);
                }
                addComponents(target);
            }

            @Override
            protected void onConfigure() {
                setVisibilityAllowed(isAuthorView());
            }
        };
        add(dropDown);
    }

    private void addComponents(AjaxRequestTarget target) {
        target.add(date);
        target.add(green);
        target.add(yellow);
        target.add(red);
        target.add(form);
        target.add(reasonLabel);
    }

    public static final String STATE = "state";
    public static final String DROPDOWN = "dropDown";
    public static final String DATE = "date";
    public static final String GREEN = "green";
    public static final String YELLOW = "yellow";
    public static final String RED = "red";
    public static final String FORM = "form";
    public static final String REASON_AREA = "reasonArea";
    public static final String SUBMIT = "submit";
    public static final String REASON_LABEL = "reasonLabel";
}
