package se.su.dsv.scipro.reviewer.timeline;

import com.google.common.base.Function;
import com.google.common.collect.FluentIterable;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.LoadableDetachableModel;
import scala.runtime.AbstractFunction0;
import scala.runtime.AbstractFunction1;
import se.su.dsv.scipro.forum.dataobjects.ForumPost;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.reviewing.*;

import javax.inject.Inject;
import java.util.*;

public class TimelinePanel extends Panel {
    @Inject
    private ReviewerInteractionService reviewerInteractionService;
    @Inject
    private FinalSeminarApprovalService finalSeminarApprovalService;
    @Inject
    private RoughDraftApprovalService roughDraftApprovalService;

    public TimelinePanel(final String id, final IModel<Project> model) {
        super(id, model);

        IModel<List<Event>> events = new LoadableDetachableModel<List<Event>>() {
            @Override
            protected List<Event> load() {
                Collection<Event> events = getForumEvents(model.getObject());

                Collection<Decision> finalSeminarDecisions = getFinalSeminarDecisions(model.getObject());

                Collection<Event> finalSeminarEvents = expandMultiple(finalSeminarDecisions);

                Collection<Decision> roughDraftDecisions = getRoughDraftDecisions(model.getObject());

                Collection<Event> roughDraftEvents = expandMultiple(roughDraftDecisions);

                List<Event> ev = new ArrayList<>();
                ev.addAll(events);
                ev.addAll(finalSeminarEvents);
                ev.addAll(roughDraftEvents);

                Collections.sort(ev, Collections.reverseOrder(new Comparator<Event>() {
                    @Override
                    public int compare(final Event o1, final Event o2) {
                        return o1.when().compareTo(o2.when());
                    }
                }));

                return ev;
            }
        };

        add(new ListView<Event>("events", events) {
            @Override
            protected void populateItem(final ListItem<Event> item) {
                item.add(item.getModelObject().component("event", item.getModel()));
            }
        });
    }

    private Collection<Event> getForumEvents(final Project project) {
        return FluentIterable.from(reviewerInteractionService.getPosts(project))
                .transform(new Function<ForumPost, Event>() {
                    @Override
                    public Event apply(final ForumPost input) {
                        return new Event.Message(input);
                    }
                })
                .toList();
    }

    private Collection<Decision> getFinalSeminarDecisions(final Project project) {
        return finalSeminarApprovalService.findBy(project)
                .map(new AbstractFunction1<FinalSeminarApproval, List<Decision>>() {
                    @Override
                    public List<Decision> apply(final FinalSeminarApproval approval) {
                        return approval.getHistory();
                    }
                })
                .getOrElse(new AbstractFunction0<List<Decision>>() {
                    @Override
                    public List<Decision> apply() {
                        return Collections.emptyList();
                    }
                });
    }

    private Collection<Decision> getRoughDraftDecisions(final Project project) {
        return roughDraftApprovalService.findBy(project)
                .map(new AbstractFunction1<RoughDraftApproval, List<Decision>>() {
                    @Override
                    public List<Decision> apply(final RoughDraftApproval approval) {
                        return approval.getHistory();
                    }
                })
                .getOrElse(new AbstractFunction0<List<Decision>>() {
                    @Override
                    public List<Decision> apply() {
                        return Collections.emptyList();
                    }
                });
    }

    private Collection<Event> expandMultiple(final Collection<Decision> finalSeminarDecisions) {
        return FluentIterable.from(finalSeminarDecisions)
                .transformAndConcat(new Function<Decision, List<Event>>() {
                    @Override
                    public List<Event> apply(final Decision input) {
                        return expand(input);
                    }
                })
                .toList();
    }

    private List<Event> expand(final Decision decision) {
        switch (decision.getStatus()) {
            case APPROVED:
                return Arrays.<Event>asList(new Event.Requested(decision), new Event.Approved(decision));
            case REJECTED:
                return Arrays.<Event>asList(new Event.Requested(decision), new Event.Denied(decision));
            case UNDECIDED:
                return Collections.<Event>singletonList(new Event.Requested(decision));
            default:
                throw new Error();
        }
    }
}
