package se.su.dsv.scipro.statistics;

import org.apache.wicket.Page;
import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.ajax.form.OnChangeAjaxBehavior;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.form.ChoiceRenderer;
import org.apache.wicket.markup.html.form.EnumChoiceRenderer;
import org.apache.wicket.markup.html.link.Link;
import org.apache.wicket.markup.html.list.ListItem;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.markup.html.panel.Panel;
import org.apache.wicket.model.*;
import org.joda.time.DateTime;
import se.su.dsv.scipro.components.AjaxCheckBoxMultipleChoice;
import se.su.dsv.scipro.components.BootstrapDateField;
import se.su.dsv.scipro.components.BootstrapDatePicker;
import se.su.dsv.scipro.date.DeactivatableDatePickerPanel;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.milestones.service.MileStoneService;
import se.su.dsv.scipro.milestones.service.MilestoneActivityTemplateService;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.ProjectTypeService;

import javax.inject.Inject;
import java.io.Serializable;
import java.util.Arrays;
import java.util.Date;
import java.util.EnumSet;
import java.util.List;

public class MilestoneStatisticsPanel extends Panel {

    public static final String PROJECT_ACTIVITY_LIST = "projectActivityList";
    public static final String PROJECT_ACTIVITY_NAME = "projectActivityName";
    public static final String INDIVIDUAL_ACTIVITY_LIST = "individualActivityList";
    public static final String INDIVIDUAL_ACTIVITY_NAME = "individualActivityName";
    public static final String COMPLETED_PROJECTS = "completedProjects";
    public static final String COMPLETED_AUTHORS = "completedAuthors";
    public static final String PROJECT_TYPE_FILTER = "projectTypes";
    public static final String CONTAINER = "container";
    public static final String DATE_FILTER = "dateFilter";
    public static final String MILESTONE_DATE_FILTER = "mileStoneCompletedBefore";
    public static final String TOTAL_PROJECTS = "totalProjects";
    public static final String DETAILS_LINK = "detailsLink";
    public static final String PROJECT_STATUSES = "projectStatuses";
    private final PageSupplier pageSupplier;

    @Inject
    private MilestoneActivityTemplateService milestoneActivityTemplateService;
    @Inject
    private MileStoneService mileStoneService;
    @Inject
    private ProjectTypeService projectTypeService;
    @Inject
    private ProjectService projectService;

    private MileStoneService.Filter filter = new MileStoneService.Filter();
    private final WebMarkupContainer container;

    public MilestoneStatisticsPanel(String id, PageSupplier pageSupplier) {
        super(id);
        this.pageSupplier = pageSupplier;
        setDefaultModel(new CompoundPropertyModel<>(filter));

        container = new WebMarkupContainer(CONTAINER);
        container.setOutputMarkupId(true);
        add(container);

        addFilterControls();
        addProjectMileStoneList();
        addIndividualMileStoneList();
        addChartLink();
    }

    private void addChartLink() {
        container.add(new Link<Void>("chartLink") {
            @Override
            public void onClick() {
                setResponsePage(pageSupplier.chartPage(filter));
            }
        });
    }

    private void addFilterControls() {
        prepareDefaultFilter();
        addProjectTypeFilter();
        addProjectDateFilter();
        addProjectStatusFilter();
        addMileStoneDateFilter();
    }

    private void prepareDefaultFilter() {
        filter.setProjectTypes(projectTypeService.findByDegreeLevels(EnumSet.of(DegreeLevel.FIRST_CYCLE, DegreeLevel.SECOND_CYCLE)));
        filter.setProjectStatuses(EnumSet.of(ProjectStatus.ACTIVE));
        filter.setMileStoneCompletedBefore(new Date());
        filter.setProjectStartedBefore(new Date());
        filter.setProjectStartedAfter(DateTime.now().minusMonths(6).toDate());
    }

    private void addProjectTypeFilter() {
        add(new AjaxCheckBoxMultipleChoice<ProjectType>(PROJECT_TYPE_FILTER,
                projectTypeService.findWithModule(ProjectModule.MILESTONES), new ChoiceRenderer<ProjectType>("name", "id")) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                target.add(container);
            }
        });
    }

    private void addProjectDateFilter() {
        add(new DeactivatableDatePickerPanel(DATE_FILTER,
                new PropertyModel<Date>(filter, "projectStartedAfter"),
                new PropertyModel<Date>(filter, "projectStartedBefore"),
                Model.of(true)) {
            @Override
            protected void datesChanged(AjaxRequestTarget target) {
                target.add(container);
            }
        });
    }

    private void addProjectStatusFilter() {
        add(new AjaxCheckBoxMultipleChoice<ProjectStatus>(PROJECT_STATUSES,
                Arrays.asList(ProjectStatus.values()), new EnumChoiceRenderer<ProjectStatus>(this)) {
            @Override
            public void onUpdate(AjaxRequestTarget target) {
                target.add(container);
            }
        });
    }

    private void addMileStoneDateFilter() {
        BootstrapDateField mileStoneCompletedBefore = new BootstrapDateField(MILESTONE_DATE_FILTER, new PropertyModel<Date>(filter, "mileStoneCompletedBefore"));
        mileStoneCompletedBefore.add(new OnChangeAjaxBehavior() {
            @Override
            protected void onUpdate(final AjaxRequestTarget target) {
                target.add(container);
            }
        });
        mileStoneCompletedBefore.add(new BootstrapDatePicker());
        add(mileStoneCompletedBefore);
    }

    private void addProjectMileStoneList() {
        container.add(new Label(TOTAL_PROJECTS, countTotalProjects()));
        container.add(new ListView<MilestoneActivityTemplate>(PROJECT_ACTIVITY_LIST, getActivities(MilestoneActivityTemplate.Type.PROJECT)) {
            @Override
            protected void populateItem(ListItem<MilestoneActivityTemplate> item) {
                Link<MilestoneActivityTemplate> link = new Link<MilestoneActivityTemplate>(DETAILS_LINK, item.getModel()) {
                    @Override
                    public void onClick() {
                        setResponsePage(pageSupplier.projectPage(getModel(), filter));
                    }
                };
                link.add(new Label(PROJECT_ACTIVITY_NAME, new PropertyModel<String>(item.getModel(), MilestoneActivityTemplate.TITLE)));
                item.add(link);
                item.add(new Label(COMPLETED_PROJECTS, countCompleted(item.getModel())));
            }
        });
    }

    private void addIndividualMileStoneList() {
        container.add(new ListView<MilestoneActivityTemplate>(INDIVIDUAL_ACTIVITY_LIST, getActivities(MilestoneActivityTemplate.Type.STUDENT)) {
            @Override
            protected void populateItem(ListItem<MilestoneActivityTemplate> item) {
                Link<MilestoneActivityTemplate> link = new Link<MilestoneActivityTemplate>(INDIVIDUAL_ACTIVITY_NAME, item.getModel()) {
                    @Override
                    public void onClick() {
                        setResponsePage(pageSupplier.individualPage(getModel(), filter));
                    }
                };
                item.add(link);
                link.setBody(new PropertyModel<String>(item.getModel(), MilestoneActivityTemplate.TITLE));
                item.add(new Label(COMPLETED_AUTHORS, countCompleted(item.getModel())));
            }
        });
    }

    private IModel<Long> countCompleted(final IModel<MilestoneActivityTemplate> activityModel) {
        return new LoadableDetachableModel<Long>() {
            @Override
            protected Long load() {
                return mileStoneService.countCompletedMileStones(activityModel.getObject(), filter);
            }
        };
    }

    private IModel<?> countTotalProjects() {
        return new LoadableDetachableModel<Object>() {
            @Override
            protected Object load() {
                return projectService.count(convertFilter(filter));
            }
        };
    }

    private static ProjectService.Filter convertFilter(MileStoneService.Filter filter) {
        ProjectService.Filter filterParams = new ProjectService.Filter();
        filterParams.setProjectTypes(filter.getProjectTypes());
        filterParams.setStatuses(filter.getProjectStatuses());
        filterParams.setCreatedAfter(filter.getProjectStartedAfter());
        filterParams.setCreatedBefore(filter.getProjectStartedBefore());
        return filterParams;
    }

    private IModel<List<MilestoneActivityTemplate>> getActivities(final MilestoneActivityTemplate.Type activityType) {
        return new LoadableDetachableModel<List<MilestoneActivityTemplate>>() {
            @Override
            protected List<MilestoneActivityTemplate> load() {
                return milestoneActivityTemplateService.getAllActivities(activityType, filter.getProjectTypes());
            }
        };
    }

    public interface PageSupplier extends Serializable {
        Page individualPage(IModel<MilestoneActivityTemplate> model, MileStoneService.Filter filter);
        Page projectPage(IModel<MilestoneActivityTemplate> model, MileStoneService.Filter filter);
        Page chartPage(MileStoneService.Filter filter);
    }
}
