package se.su.dsv.scipro.user;

import org.apache.wicket.ajax.AjaxRequestTarget;
import org.apache.wicket.extensions.markup.html.form.palette.Palette;
import org.apache.wicket.feedback.FeedbackMessage;
import org.apache.wicket.feedback.IFeedbackMessageFilter;
import org.apache.wicket.markup.html.WebMarkupContainer;
import org.apache.wicket.markup.html.form.*;
import org.apache.wicket.markup.html.panel.ComponentFeedbackPanel;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.apache.wicket.model.PropertyModel;
import org.apache.wicket.model.util.ListModel;
import org.apache.wicket.request.Url;
import org.apache.wicket.request.cycle.RequestCycle;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.apache.wicket.util.string.StringValue;
import org.apache.wicket.validation.IValidatable;
import org.apache.wicket.validation.IValidator;
import org.apache.wicket.validation.ValidationError;
import org.apache.wicket.validation.validator.EmailAddressValidator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import se.su.dsv.scipro.components.AjaxCheckBoxMultipleChoice;
import se.su.dsv.scipro.components.menuhighlighting.MenuHighlightAdminInternalUsers;
import se.su.dsv.scipro.data.DetachableServiceModel;
import se.su.dsv.scipro.loginlogout.pages.SetPasswordPage;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.match.ProgramService;
import se.su.dsv.scipro.reusable.SciProUtilities;
import se.su.dsv.scipro.security.auth.Authorization;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.*;
import se.su.dsv.scipro.util.PageParameterKeys;

import javax.inject.Inject;
import java.util.*;

@Authorization(authorizedRoles = {Roles.ADMIN})
public class AdminEditUserPage extends AbstractAdminUsersPage implements MenuHighlightAdminInternalUsers {

    private static final Logger LOGGER = LoggerFactory.getLogger(AdminEditUserPage.class);

    @Inject
    private UserService userService;
    @Inject
    private ResearchAreaService researchAreaService;
    @Inject
    private ProgramService programService;
    @Inject
    private HashService hashService;
    @Inject
    private MailEventService mailEventService;
    @Inject
    private UserFacade userFacade;
    @Inject
    private LanguageService languageService;
    @Inject
    private UserNameService userNameService;

    private Palette<ResearchArea> researchAreaPalette;
    private Palette<Program> programPalette;

    public AdminEditUserPage(final PageParameters pp) {
        StringValue userId = pp.get(PageParameterKeys.MAP.get(User.class));
        User user;
        if (userId.isNull() || userId.isEmpty()) {
            user = new User();
        } else {
            user = userService.findOne(userId.toLong());
        }
        IModel<User> userModel = new DetachableServiceModel<>(userService, user);
        add(new UserForm("form", userModel));
    }

    private class UserForm extends Form<User> {

        private IModel<Username> usernameModel;

        public UserForm(String id, IModel<User> model) {
            super(id, model);

            RequiredTextField<String> firstName = new RequiredTextField<>("firstName", new PropertyModel<String>(model, "firstName"));
            add(firstName);
            add(new ComponentFeedbackPanel("firstNameFeedback", firstName));

            RequiredTextField<String> lastName = new RequiredTextField<>("lastName", new PropertyModel<String>(model, "lastName"));
            add(lastName);
            add(new ComponentFeedbackPanel("lastNameFeedback", lastName));

            RequiredTextField<String> email = new RequiredTextField<String>("email", new PropertyModel<String>(model, "emailAddress")) {
                @Override
                protected void onConfigure() {
                    super.onConfigure();
                    if (UserForm.this.getModelObject().getId() == null){
                        setEnabled(true);
                    } else {
                        setEnabled(showChangeUsernameForm(UserForm.this.getModelObject()));
                    }
                }
            };
            email.add(new UniqueValidator());
            email.add(EmailAddressValidator.getInstance());
            add(email);
            add(new ComponentFeedbackPanel("emailFeedback", email));

            final WebMarkupContainer researchAreaWMC = new WebMarkupContainer("researchAreaWMC") {
                @Override
                protected void onConfigure() {
                    super.onConfigure();
                    setVisible(getModelObject().getRoles().contains(Roles.SUPERVISOR));
                }
            };
            researchAreaWMC.setOutputMarkupPlaceholderTag(true);
            researchAreaPalette = new Palette<>("researchAreas",
                    new ListModel<>(new ArrayList<>(getModelObject().getActiveResearchAreas())),
                    new ListModel<>(researchAreaService.findNotDeleted()),
                    new ChoiceRenderer<ResearchArea>("title", "id"),
                    10,
                    false
            );
            researchAreaWMC.add(researchAreaPalette);
            add(researchAreaWMC);

            final WebMarkupContainer programWMC = new WebMarkupContainer("programWMC") {
                @Override
                protected void onConfigure() {
                    super.onConfigure();
                    setVisible(getModelObject().getRoles().contains(Roles.AUTHOR));
                }
            };
            programWMC.setOutputMarkupPlaceholderTag(true);
            programPalette = new Palette<>("programs",
                    new ListModel<>(new ArrayList<>(getModelObject().getPrograms())),
                    new ListModel<>(programService.findAll()),
                    new ChoiceRenderer<Program>("name", "id"),
                    10,
                    false
            );
            programWMC.add(programPalette);
            add(programWMC);

            List<Roles> roles = Arrays.asList(Roles.values());
            add(new AjaxCheckBoxMultipleChoice<Roles>("roles", new PropertyModel<Set<Roles>>(model, "roles"), roles, new EnumChoiceRenderer<Roles>(this)) {
                @Override
                public void onUpdate(AjaxRequestTarget target) {
                    target.add(researchAreaWMC);
                    target.add(programWMC);
                }
            });

            add(new FeedbackPanel("feedback", new IFeedbackMessageFilter() {
                @Override
                public boolean accept(FeedbackMessage message) {
                    return !message.isRendered();
                }
            }));

            add(new Button("saveAddButton") {
                @Override
                public void onAfterSubmit() {
                    setResponsePage(AdminEditUserPage.class);
                }
            });
        }

        @Override
        protected void onValidate() {
            super.onValidate();
            if (showChangeUsernameForm(getModelObject())){
                usernameModel = Model.of(userNameService.findByUsername(getModelObject().getEmailAddress()));
            }
        }

        @Override
        protected void onSubmit() {
            super.onSubmit();

            if (getModelObject().getRoles().contains(Roles.AUTHOR)) {
                getModelObject().setPrograms(new HashSet<>(SciProUtilities.copyIterator(programPalette.getSelectedChoices())));
            }
            if (getModelObject().getRoles().contains(Roles.SUPERVISOR)) {
                getModelObject().setResearchAreas(new HashSet<>(SciProUtilities.copyIterator(researchAreaPalette.getSelectedChoices())));

            }

            if (getModelObject().getLanguages().isEmpty()) {
                getModelObject().setLanguages(new HashSet<>(languageService.findAll()));
            }

            LOGGER.info("{} is updating user: {}", SciProSession.get().getUser(), getModelObject());

            if (getModelObject().getId() == null) { //new user
                setModelObject(userFacade.createUser(getModelObject()));
                // TODO: The following should be done when creating the user and not as an optional separate process...
                Hash hash = hashService.createForUser(getModelObject());
                sendActivateAccountMail(getModelObject(), hash);
            } else { //existing user

                if (usernameModel.getObject()!=null){
                    Username username = usernameModel.getObject();
                    username.setUsername(getModelObject().getEmailAddress());
                    usernameModel.setObject(userNameService.save(username));
                }

                setModelObject(userService.save(getModelObject()));
            }

            getSession().info("Saved user " + getModelObject().getFullName());
            setResponsePage(AdminInternalUsersPage.class);
        }

        private class UniqueValidator implements IValidator<String> {
            @Override
            public void validate(IValidatable<String> validation) {
                String value = validation.getValue();
                User username = userService.findByUsername(value);
                if (username != null && !username.equals(getModelObject())) {
                    validation.error(new ValidationError(this));
                }
            }
        }
    }

    private void sendActivateAccountMail(final User user, final Hash hash) {
        final PageParameters pp = new PageParameters();
        pp.add("hash", hash.getHash());
        final String setPasswordURL = RequestCycle.get().getUrlRenderer().renderFullUrl(Url.parse(urlFor(SetPasswordPage.class, pp)));
        final String subject = getString("activate.account.subject");
        final String body = getString("activate.account.content") + " " + setPasswordURL;

        final MailEvent mailEvent = new MailEvent();
        mailEvent.getRecipients().add(user);
        mailEvent.setSubject(subject);
        mailEvent.setMessageBody(body);

        mailEventService.save(mailEvent);
    }

    private boolean showChangeUsernameForm(User user){
        return user.getId()!=null && user.getIdentifier()==null && user.getUsernames().size()==1 && usernameSameAsEmail(user);
    }

    private boolean usernameSameAsEmail(User user){
        boolean result = false;
        for (Username username : user.getUsernames()){
            result = username.getUsername().equals(user.getEmailAddress());
        }
        return result;
    }
}