package se.su.dsv.scipro.datatables.project;

import org.apache.wicket.Component;
import org.apache.wicket.ajax.form.AjaxFormComponentUpdatingBehavior;
import org.apache.wicket.behavior.AbstractAjaxBehavior;
import org.apache.wicket.behavior.Behavior;
import org.apache.wicket.markup.html.form.FormComponent;
import org.apache.wicket.util.tester.WicketTesterHelper;
import org.apache.wicket.util.visit.IVisit;
import org.apache.wicket.util.visit.IVisitor;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.springframework.data.domain.Pageable;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.datatables.AjaxCheckboxWrapper;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ObjectMother;

import java.util.Arrays;
import java.util.Collections;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.mockito.Mockito.*;

public class ProjectDataPanelTest extends SciProTest {

    private ProjectDataPanel panel;

    @Test
    public void renders() throws Exception {
        startPanel();
    }

    @Test
    public void clicking_checkbox_calls_inactivate_service_method() throws Exception {
        mockProject();
        startPanel();
        inactivateProject();
    }

    @Test
    public void change_reviewer() throws Exception {
        User reviewer = User.builder().firstName("reviewer").lastName("review").emailAddress("rev@reviewer.com").build();
        reviewer.setId(74L);
        reviewer.addRole(Roles.REVIEWER);
        when(userRepo.findOne(74L)).thenReturn(reviewer);
        mockProject();
        startPanel();
        selectReviewer();

        ArgumentCaptor<Project> captor = ArgumentCaptor.forClass(Project.class);
        verify(projectService).save(captor.capture());

        assertEquals(captor.getValue().getReviewer(), reviewer);
    }

    @Test
    public void change_reviewer_removes_old_reviewers() throws Exception {
        User reviewer = User.builder().firstName("reviewer").lastName("review").emailAddress("rev@reviewer.com").build();
        reviewer.addRole(Roles.REVIEWER);
        reviewer.setId(74L);
        when(userRepo.findOne(74L)).thenReturn(reviewer);
        User oldReviewer = mockProjectWithReviewer();
        startPanel();
        selectReviewer();

        ArgumentCaptor<Project> captor = ArgumentCaptor.forClass(Project.class);
        verify(projectService).save(captor.capture());

        assertFalse(captor.getValue().getReviewers().contains(oldReviewer));
    }

    @Test
    public void change_reviewer_fails_if_selecting_the_current_supervisor() throws Exception {
        when(userRepo.findOne(74L)).thenReturn(ObjectMother.SOME_USER);
        mockProject();
        startPanel();
        selectReviewer();
        verify(projectService, never()).save(any(Project.class));
    }

    private void selectReviewer() {
        FormComponent reviewerSelection = (FormComponent) panel.get("dataTable:body:rows:1:cells:7:cell:eac");
        tester.getRequest().getPostParameters().setParameterValue(reviewerSelection.getInputName(), "74");
        Behavior behavior = WicketTesterHelper.findBehavior(reviewerSelection, AjaxFormComponentUpdatingBehavior.class);
        tester.executeBehavior((AbstractAjaxBehavior) behavior);
    }

    private void inactivateProject() {
        panel.visitChildren(AjaxCheckboxWrapper.class, new IVisitor<Component, Object>() {
            @Override
            public void component(Component object, IVisit<Object> visit) {
                tester.executeAjaxEvent(path(object.getPageRelativePath(), AjaxCheckboxWrapper.CHECKBOX), "onclick");
                verify(projectService).toggleActiveStatus(any(Project.class), eq(false));
                visit.stop();
            }
        });
    }

    private void mockProject() {
        Project project = Project.builder().title("title").projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "bach", "b")).projectStatus(ProjectStatus.ACTIVE).build();
        project.setHeadSupervisor(ObjectMother.SOME_USER);
        when(projectService.findAll(any(ProjectService.Filter.class), any(Pageable.class))).thenReturn(Collections.singletonList(project));
        when(projectService.count(any(ProjectService.Filter.class))).thenReturn(1L);
    }

    private User mockProjectWithReviewer() {
        Project project = Project.builder().title("title").projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "bach", "b")).projectStatus(ProjectStatus.ACTIVE).build();
        project.setHeadSupervisor(ObjectMother.SOME_USER);
        User oldReviewer = User.builder().firstName("old").lastName("reviewer").emailAddress("old@reviewer.com").build();
        oldReviewer.setId(7676L);
        project.addReviewer(oldReviewer);
        when(projectService.findAll(any(ProjectService.Filter.class), any(Pageable.class))).thenReturn(Collections.singletonList(project));
        when(projectService.count(any(ProjectService.Filter.class))).thenReturn(1L);
        return oldReviewer;
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new ProjectDataPanel("id"));
    }
}
