package se.su.dsv.scipro.finalseminar;

import com.google.common.collect.Lists;
import org.apache.wicket.markup.html.link.BookmarkablePageLink;
import org.apache.wicket.model.Model;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import org.mockito.Mockito;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.report.OppositionReport;
import se.su.dsv.scipro.security.auth.roles.Roles;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.Collections;
import java.util.List;
import java.util.TreeSet;

public class SeminarOppositionReportPanelTest extends SciProTest {

    private SeminarOppositionReportPanel panel;
    private List<FinalSeminarOpposition> opponents;
    private FinalSeminar finalSeminar;
    private FinalSeminarOpposition opposition;
    private User coSupervisorUser;
    private User respondentUser;
    private User opponentUser;
    private User reviewerUser;
    private Project project;

    @Before
    public void setUp() throws Exception {
        opposition = new FinalSeminarOpposition();
        opponentUser = new UserBuilder().create();
        opposition.setUser(opponentUser);
        DomainObjects.injectId(opposition, 1L);
        opponents = Lists.newArrayList(opposition);

        respondentUser = new UserBuilder().create();

        finalSeminar = new FinalSeminar();
        opposition.setFinalSeminar(finalSeminar);
        finalSeminar.setStartDate(new DateTime().minusDays(1).toDate());
        finalSeminar.setOppositions(opponents);

        User supervisorUser = new UserBuilder().create();
        TreeSet<User> projectParticipants = new TreeSet<>(new User.ByNameComparator());
        projectParticipants.addAll(Collections.singletonList(respondentUser));
        project = Project.builder().title("title")
                .projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, ProjectType.BACHELOR_STRING, ProjectType.BACHELOR_STRING))
                .headSupervisor(supervisorUser)
                .projectParticipants(projectParticipants)
                .build();

        coSupervisorUser = new UserBuilder().create();
        project.addCoSupervisor(coSupervisorUser);

        finalSeminar.setProject(project);


        setLoggedInAs(supervisorUser);
        Mockito.when(finalSeminarOppositionService.findAll(Matchers.isA(FinalSeminarOppositionService.FilterParams.class))).thenReturn(opponents);

        OppositionReport oppositionReport = Mockito.mock(OppositionReport.class);
        Mockito.when(oppositionReportService.findOrCreateReport(opposition)).thenReturn(oppositionReport);
    }


    @Test
    public void supervisor_allowed_to_see_opposition_report() throws Exception {
        setRoles(Roles.SUPERVISOR);
        startPanel();
        tester.assertVisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void co_supervisor_allowed_to_see_opposition_report() throws Exception {
        setLoggedInAs(coSupervisorUser);
        setRoles(Roles.SUPERVISOR);
        startPanel();
        tester.assertVisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void normal_student_not_allowed_to_see_opposition_report() throws Exception {
        addReviewerToProject(project);
        setLoggedInAs(new UserBuilder().create());
        setRoles(Roles.AUTHOR);
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void respondent_allowed_to_see_opposition_report() throws Exception {
        setLoggedInAs(respondentUser);
        setRoles(Roles.AUTHOR);
        startPanel();
        tester.assertVisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void opposition_author_allowed_to_see_opposition_report() throws Exception {
        setLoggedInAs(opponentUser);
        setRoles(Roles.AUTHOR);
        startPanel();
        tester.assertVisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void reviewer_allowed_to_see_opposition_report() throws Exception {
        addReviewerToProject(project);
        setLoggedInAs(reviewerUser);
        setRoles(Roles.SUPERVISOR);
        startPanel();
        tester.assertVisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void admin_allowed_to_see_opposition_report() throws Exception {
        addReviewerToProject(project);
        setLoggedInAs(new UserBuilder().create());
        setRoles(Roles.SYSADMIN);
        startPanel();
        tester.assertVisible(path(panel, "wmc", "oppositionReportLabel"));
    }

    @Test
    public void has_download_pdf_panel() throws Exception {
        startPanel();
        tester.assertComponent(path(panel, "wmc", "downloadPdfPanel"), DownloadPdfReportPanel.class);
    }

    @Test
    public void download_pdf_panel_invisible_for_normal_student() throws Exception {
        addReviewerToProject(project);
        setLoggedInAs(new UserBuilder().create());
        setRoles(Roles.AUTHOR);
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "downloadPdfPanel"));
    }

    @Test
    public void has_opposition_report_link() throws Exception {
        setLoggedInAs(opponentUser);
        finalSeminar.setDocument(new FileDescription());
        startPanel();
        tester.assertComponent(path(panel, "wmc", "oppositionReportLink"), BookmarkablePageLink.class);
    }

    @Test
    public void opposition_report_link_invisible_if_thesis_not_submitted() throws Exception {
        setLoggedInAs(opponentUser);
        finalSeminar.setDocument(null);
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "oppositionReportLink"));
    }

    @Test
    public void opposition_report_link_invisible_if_not_opponent() throws Exception {
        finalSeminar.setDocument(new FileDescription());
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "oppositionReportLink"));
    }

    @Test
    public void opposition_report_link_invisible_if_report_submitted() throws Exception {
        finalSeminar.setDocument(new FileDescription());
        setLoggedInAs(opponentUser);
        OppositionReport oppositionReport = Mockito.mock(OppositionReport.class);
        Mockito.when(oppositionReport.isSubmitted()).thenReturn(true);
        Mockito.when(oppositionReportService.findOrCreateReport(opposition)).thenReturn(oppositionReport);
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "oppositionReportLink"));
    }

    @Test
    public void has_view_attachment_panel() throws Exception {
        FileDescription opponentReport = new FileDescription();
        DomainObjects.injectId(opponentReport, 1L);
        opposition.setOpponentReport(opponentReport);
        startPanel();
        tester.assertModelValue(path(panel, "wmc", "reportDownload"), opposition.getOpponentReport());
    }

    @Test
    public void old_opposition_report_attachment_not_visible_if_not_existing() throws Exception {
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "reportDownload"));
    }

    @Test
    public void old_opposition_report_attachment_not_visible_if_not_allowed_to_see_it() throws Exception {
        setLoggedInAs(new UserBuilder().create());
        setRoles(Roles.AUTHOR);
        FileDescription opponentReport = new FileDescription();
        DomainObjects.injectId(opponentReport, 1L);
        opposition.setOpponentReport(opponentReport);
        startPanel();
        tester.assertInvisible(path(panel, "wmc", "reportDownload"));
    }

    @Test
    public void report_attachment_visible_if_allowed_to_see_and_existing() throws Exception {
        mockPanelWithOppositionReport(opponentUser, true, true);
        tester.assertVisible(path(panel, "wmc", "downloadAttachment"));
    }

    @Test
    public void report_attachment_not_visible_if_allowed_to_see_and_not_existing() throws Exception {
        mockPanelWithOppositionReport(opponentUser, false, true);
        tester.assertInvisible(path(panel, "wmc", "downloadAttachment"));
    }

    @Test
    public void report_attachment_not_visible_if_not_allowed_to_see() throws Exception {
        setRoles(Roles.AUTHOR);
        mockPanelWithOppositionReport(new UserBuilder().create(), true, true);
        tester.assertInvisible(path(panel, "wmc", "downloadAttachment"));
    }

    @Test
    public void report_attachment_not_visible_if_not_submitted() throws Exception {
       mockPanelWithOppositionReport(opponentUser, true, false);
       tester.assertInvisible(path(panel, "wmc", "downloadAttachment"));
    }

    private void mockPanelWithOppositionReport(User user, boolean hasAttachment, boolean submitted) {
        finalSeminar.setDocument(new FileDescription());
        setLoggedInAs(user);
        OppositionReport oppositionReport = Mockito.mock(OppositionReport.class);
        if (hasAttachment) {
            createFile(oppositionReport);
        }
        Mockito.when(oppositionReport.isSubmitted()).thenReturn(submitted);
        Mockito.when(oppositionReportService.findOrCreateReport(opposition)).thenReturn(oppositionReport);
        startPanel();
    }

    private void createFile(OppositionReport oppositionReport) {
        FileDescription file = new FileDescription();
        file.setId(1L);
        Mockito.when(oppositionReport.getAttachment()).thenReturn(file);
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new SeminarOppositionReportPanel("id", Model.of(opposition)));
    }

    private void addReviewerToProject(Project project) {
        reviewerUser = new UserBuilder().create();
        project.addReviewer(reviewerUser);
    }
}