package se.su.dsv.scipro.finalseminar;

import org.apache.wicket.Component;
import org.apache.wicket.feedback.FencedFeedbackPanel;
import org.apache.wicket.model.Model;
import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import org.mockito.Mockito;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.repository.panels.ViewAttachmentPanel;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.ObjectMother;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.Date;

import static org.mockito.Matchers.anyLong;
import static org.mockito.Matchers.isA;
import static org.mockito.Mockito.when;

public class SeminarThesisPanelTest extends SciProTest {

    SeminarThesisPanel panel;
    private FileDescription fileDescription;

    @Before
    public void setUp() throws Exception {
        when(dateService.getFormat(isA(DateStyle.class))).thenReturn("yyyy-MM-dd");
        when(daysService.workDaysAhead(isA(Date.class), isA(Integer.class))).thenReturn(new Date());
        when(daysService.workDaysAhead((Date) Matchers.eq(null), isA(Integer.class))).thenReturn(new Date());

        fileDescription = new FileDescription();
        fileDescription.setIdentifier("1");
        fileDescription.setId(1L);
    }

    @Test
    public void Renders() throws Exception {
        startPanel();

        assert panel != null : "Panel does not render";
    }

    @Test
    public void RendersWithThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(fileDescription);

        // when
        startPanel(seminar);

        // then
        assert panel != null : "Panel does not render with thesis";
    }

    @Test
    public void DeadlineInformationContainsDeadlineDate() throws Exception {
        // given
        DateTime seminarDate = new DateTime(2012, 4, 21, 10, 10, 0, 0);
        int daysAhead = 10;

        FinalSeminarSettings settings = new FinalSeminarSettings();
        settings.setDaysAheadToUploadThesis(daysAhead);

        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(null);
        seminar.setStartDate(seminarDate.toDate());

        // when
        when(finalSeminarSettingsService.getInstance()).thenReturn(settings);
        when(daysService.workDaysAhead(isA(Date.class), isA(Integer.class))).thenReturn(seminarDate.minusDays(daysAhead).toDate());
        when(finalSeminarService.thesisUploadDeadline(isA(FinalSeminar.class))).thenReturn(seminarDate.minusDays(daysAhead).toDate());

        startPanel(seminar);

        // then
        Component dateComponent = panel.get(path(SeminarThesisPanel.DEADLINE_INFORMATION, SeminarThesisPanel.DATE));
        Assert.assertNotNull("No date component", dateComponent);

        Object modelObject = dateComponent.getDefaultModelObject();
        Assert.assertTrue("Date components model is not a date", modelObject instanceof Date);

        Date date = (Date) modelObject;
        Assert.assertEquals("Does not display correct date", seminarDate.minusDays(daysAhead).toDate(), date);
    }

    @Test
    public void ShowsUploadIfAuthorAndNoThesis() throws Exception {
        User loggedInUser = new UserBuilder().create();
        loggedInUser.setId(7L);
        setLoggedInAs(loggedInUser);
        // given
        Project project = Project.builder().title("title").projectType(ObjectMother.BACHELOR).build();
        project.setId(5L);
        project.addProjectParticipant(loggedInUser);
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(null);
        seminar.setProject(project);


        when(projectService.findOne(project.getId())).thenReturn(project);
        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.UPLOAD).isVisibleInHierarchy();
        assert isVisible : "Upload form not shown";
    }

    @Test
    public void HidesUploadIfNotAuthorAndNoThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(null);

        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.UPLOAD).isVisibleInHierarchy();
        assert !isVisible : "Upload form shown for non-authors";
    }

    @Test
    public void HidesUploadIfThesis() throws Exception {
        User loggedIn = new UserBuilder().create();
        setLoggedInAs(loggedIn);
        Project project = Project.builder().title("hej").projectType(ObjectMother.BACHELOR).build();
        project.setId(7L);
        project.addProjectParticipant(loggedIn);
        when(projectService.findOne(project.getId())).thenReturn(project);
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(fileDescription);
        seminar.setProject(project);

        // when
        when(finalSeminarService.hasThesis(isA(FinalSeminar.class))).thenReturn(true);
        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.UPLOAD).isVisibleInHierarchy();
        assert !isVisible : "Upload form shown when thesis already uploaded";
    }

    @Test
    public void ShowsLabelIfMissingThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(null);

        // when
        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.NO_THESIS).isVisibleInHierarchy();
        assert isVisible : "No thesis label not shown";
    }

    @Test
    public void ShowsDownloadLinkIfThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        FileDescription document = fileDescription;
        document.setPath("/foo");
        seminar.setDocument(document);

        // when
        startPanel(seminar);

        // then
        tester.assertComponent(path(panel.getId(), SeminarThesisPanel.DOWNLOAD_THESIS), ViewAttachmentPanel.class);
    }

    @Test
    public void HidesDownloadLinkIfNoThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(null);

        // when
        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.DOWNLOAD_THESIS).isVisibleInHierarchy();
        assert !isVisible : "Download link shown despite no thesis";
    }

    @Test
    public void HidesNoThesisLabelIfThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(fileDescription);

        // when
        when(finalSeminarService.hasThesis(isA(FinalSeminar.class))).thenReturn(true);
        startPanel(seminar);


        // then
        boolean isVisible = panel.get(SeminarThesisPanel.NO_THESIS).isVisibleInHierarchy();
        assert !isVisible : "No thesis label shown when exists";
    }

    @Test
    public void ShowsDeleteLinkIfSupervisorAndThesis() throws Exception {
        User headSupervisor = new UserBuilder().create();
        Project p = Project.builder().title("title").projectType(ObjectMother.BACHELOR).build();
        p.setId(6L);
        p.setHeadSupervisor(headSupervisor);
        setLoggedInAs(headSupervisor);
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(fileDescription);
        seminar.setProject(p);

        // when
        when(projectService.findOne(anyLong())).thenReturn(p);
        when(finalSeminarService.hasThesis(isA(FinalSeminar.class))).thenReturn(true);
        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.DELETE_THESIS).isVisibleInHierarchy();
        assert isVisible : "Delete link not shown";
    }

    @Test
    public void HidesDeleteLinkIfNoThesis() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(null);

        // when
        startPanel(seminar);

        // then
        boolean isVisible = panel.get(SeminarThesisPanel.DELETE_THESIS).isVisibleInHierarchy();
        assert !isVisible : "Delete link shown when no thesis";
    }

    @Test
    public void ClickingDeleteDeletesThesisReport() throws Exception {
        // given
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(fileDescription);
        seminar.setDocumentUploadDate(new Date());
        seminar.setDocument(fileDescription);
        seminar.setProject(mockProject());

        when(finalSeminarService.hasThesis(isA(FinalSeminar.class))).thenReturn(true);

        // when
        startPanel(seminar);
        when(finalSeminarService.deleteThesis(Matchers.eq(seminar))).thenReturn(seminar);
        tester.clickLink(panel.get(SeminarThesisPanel.DELETE_THESIS));

        // then
        Mockito.verify(finalSeminarService).deleteThesis(Matchers.eq(seminar));
    }

    private Project mockProject() {
        User headSupervisor = new UserBuilder().create();
        Project proj = Project.builder().title("hej").projectType(ObjectMother.BACHELOR).headSupervisor(headSupervisor).build();
        proj.setId(77L);
        setLoggedInAs(headSupervisor);
        when(projectService.findOne(anyLong())).thenReturn(proj);
        return proj;
    }

    @Test
    public void HidesDeleteLinkIfNotSupervisor() throws Exception {
        FinalSeminar seminar = new FinalSeminar();
        seminar.setDocument(fileDescription);

        startPanel(seminar);

        boolean isVisible = panel.get(SeminarThesisPanel.DELETE_THESIS).isVisibleInHierarchy();
        assert !isVisible : "Delete link shown for non-supervisors";
    }

    @Test
    public void has_fenced_feedback_panel() throws Exception {
        startPanel();
        tester.assertComponent(path(panel, SeminarThesisPanel.FEEDBACK), FencedFeedbackPanel.class);
    }

    @Test
    public void deadline_information_not_visible_if_seminar_is_not_created_yet() throws Exception {
        startPanel();
        tester.assertInvisible(path(panel, SeminarThesisPanel.DEADLINE_INFORMATION));
    }

    @Test
    public void deadline_information_visible_if_seminar_exists() throws Exception {
        FinalSeminar seminar = new FinalSeminar();
        DomainObjects.injectId(seminar, 45L);
        startPanel(seminar);
        tester.assertVisible(path(panel, SeminarThesisPanel.DEADLINE_INFORMATION));
    }

    private void startPanel() {
        startPanel(new FinalSeminar());
    }

    private void startPanel(final FinalSeminar seminar) {
        if (seminar.getProject() == null) {
            ProjectType projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "Some project type", "Some description");
            Project project = Project.builder().title("Some title").projectType(projectType).build();
            seminar.setProject(project);
            DomainObjects.injectId(project, 74L);
            when(projectService.findOne(project.getId())).thenReturn(project);
        }

        panel = tester.startComponentInPage(new SeminarThesisPanel("panel", Model.of(seminar)));
    }
}
