package se.su.dsv.scipro.finalthesis;

import org.apache.wicket.model.Model;
import org.apache.wicket.util.tester.FormTester;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import org.mockito.Mockito;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.forum.dataobjects.ForumPost;
import se.su.dsv.scipro.forum.dataobjects.ForumThread;
import se.su.dsv.scipro.forum.dataobjects.ProjectThread;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.repository.panels.ViewAttachmentPanel;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.Collections;
import java.util.Date;

import static org.mockito.Matchers.anySetOf;
import static org.mockito.Matchers.eq;
import static se.su.dsv.scipro.finalthesis.ApproveFinalThesisPanel.*;

public class ApproveFinalThesisPanelTest extends SciProTest {

    private static final Project SOME_PROJECT = Project.builder()
            .title("Title")
            .projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"))
            .build();
    public static final String SOME_CONTENT = "content";
    private ApproveFinalThesisPanel panel;

    private User reviewerUser;

    @Before
    public void before() throws Exception {
        reviewerUser = new UserBuilder().create();

        SOME_PROJECT.setId(1L);
        SOME_PROJECT.addReviewer(reviewerUser);
        Mockito.when(finalSeminarService.hasHadFinalSeminar(SOME_PROJECT)).thenReturn(true);
    }

    @Test
    public void approve() throws Exception {
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        tester.clickLink(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER, APPROVE));

        Mockito.verify(finalThesisService).approve(eq(SOME_PROJECT));
    }

    @Test
    public void rejection_click() throws Exception {
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        tester.clickLink(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER, REJECT));

        tester.assertVisible(path(panel, THESIS_CONTAINER, REJECT_CONTAINER, REJECT_PANEL));
        tester.assertInvisible(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER));
    }

    @Test
    public void cancel_rejection_panel() throws Exception {
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        tester.clickLink(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER, REJECT));
        tester.clickLink(path(panel, THESIS_CONTAINER, REJECT_CONTAINER, REJECT_PANEL, RejectFinalThesisPanel.FORM, RejectFinalThesisPanel.CANCEL_REJECTION));

        tester.assertInvisible(path(panel, THESIS_CONTAINER, REJECT_CONTAINER, REJECT_PANEL));
        tester.assertVisible(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER));
    }

    @Test
    public void reject_notify_and_start_forum_thread_on_reject() throws Exception {
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        ForumThread forumThread = new ForumThread();
        forumThread.setId(1L);
        ProjectThread projectThread = new ProjectThread();
        projectThread.setForumThread(forumThread);

        ForumPost post = new ForumPost();
        post.setId(1L);
        post.setForumThread(forumThread);

        forumThread.addPost(post);

        User user = new UserBuilder().create();
        setLoggedInAs(user);

        Mockito.when(projectForumService.createThread(eq(SOME_PROJECT),
                eq(user),
                Matchers.anyString(),
                eq(SOME_CONTENT),
                anySetOf(FileUpload.class)))
                .thenReturn(projectThread);

        tester.clickLink(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER, REJECT));

        FormTester formTester = tester.newFormTester(path(panel, THESIS_CONTAINER, REJECT_CONTAINER, REJECT_PANEL, RejectFinalThesisPanel.FORM));
        formTester.setValue(path(RejectFinalThesisPanel.CONTENT), SOME_CONTENT);
        formTester.submit();

        Mockito.verify(finalThesisService).reject(eq(SOME_PROJECT));
        Mockito.verify(projectForumService).createThread(SOME_PROJECT,
                SciProSession.get().getUser(),
                RejectFinalThesisPanel.REJECTED_SUBJECT,
                SOME_CONTENT,
                Collections.<FileUpload>emptySet());
    }

    @Test
    public void shows_panel_if_final_seminar_has_taken_place() throws Exception {
        Mockito.when(finalThesisService.getFinalThesisFileDescription(SOME_PROJECT)).thenReturn(null);
        startPanel();

        tester.assertVisible(path(panel));
    }

    @Test
    public void hides_panel_if_final_seminar_has_not_taken_place() throws Exception {
        Mockito.when(finalSeminarService.hasHadFinalSeminar(SOME_PROJECT)).thenReturn(false);

        startPanel();

        tester.assertInvisible(path(panel));
    }

    @Test
    public void hides_panel_if_user_is_reviewer_and_final_thesis_is_not_approved() throws Exception {
        setLoggedInAs(reviewerUser);

        mockHasFinalThesisAndStartPanel(FinalThesis.Status.REJECTED);

        startPanel();

        tester.assertInvisible(path(panel));
    }

    @Test
    public void shows_panel_if_user_is_reviewer_and_final_thesis_is_approved() throws Exception {
        setLoggedInAs(reviewerUser);

        mockHasFinalThesisAndStartPanel(FinalThesis.Status.APPROVED);

        startPanel();

        tester.assertVisible(path(panel));
    }

    @Test
    public void final_thesis_is_uploaded() throws Exception {
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        tester.assertVisible(path(panel, THESIS_CONTAINER));
        tester.assertInvisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER));
    }

    @Test
    public void final_thesis_is_not_uploaded_shows_upload_form() throws Exception {
        Mockito.when(finalThesisService.isUploadAllowed(SOME_PROJECT)).thenReturn(true);
        Mockito.when(finalThesisService.hasFinalThesis(SOME_PROJECT)).thenReturn(false);

        startPanel();

        tester.assertInvisible(path(panel, THESIS_CONTAINER));
        tester.assertVisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER));
        tester.assertComponent(path(panel, WAITING_FOR_AUTHOR_CONTAINER, UPLOAD_FORM_PANEL), FinalThesisUploadFormPanel.class);
        tester.assertVisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.STATUS_NO_THESIS_LABEL));
        tester.assertInvisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.REJECTED_DATE_LABEL));
        tester.assertInvisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.STATUS_REJECTED_LABEL));
        tester.assertInvisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.REJECTED_INFO_LABEL));
    }

    @Test
    public void rejected_final_thesis_shows_upload_form_and_info() throws Exception {
        Mockito.when(finalThesisService.isUploadAllowed(SOME_PROJECT)).thenReturn(true);
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.REJECTED);

        tester.assertInvisible(path(panel, THESIS_CONTAINER));
        tester.assertVisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER));
        tester.assertComponent(path(panel, WAITING_FOR_AUTHOR_CONTAINER, UPLOAD_FORM_PANEL), FinalThesisUploadFormPanel.class);
        tester.assertInvisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.STATUS_NO_THESIS_LABEL));
        tester.assertVisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.REJECTED_DATE_LABEL));
        tester.assertVisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.STATUS_REJECTED_LABEL));
        tester.assertVisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER, ApproveFinalThesisPanel.REJECTED_INFO_LABEL));
    }

    @Test
    public void shows_download_link_information() throws Exception {
        FinalThesis finalThesis = mockFinalThesisWithFileDescription();
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        String downloadPath = path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER, FILE_PANEL);
        tester.assertComponent(downloadPath, ViewAttachmentPanel.class);
        tester.assertModelValue(downloadPath, finalThesis.getFileDescription());
    }

    @Test
    public void shows_date_of_uploaded_file() throws Exception {
        FinalThesis finalThesis = mockHasFinalThesisAndStartPanel(FinalThesis.Status.NO_DECISION);

        tester.assertModelValue(path(panel, THESIS_CONTAINER, THESIS_DECISION_CONTAINER, DATE_LABEL), finalThesis.getDateCreated());
    }

    @Test
    public void shows_grading_form() throws Exception {
        mockHasFinalThesisAndStartPanel(FinalThesis.Status.APPROVED);

        tester.assertInvisible(path(panel, THESIS_CONTAINER));
        tester.assertInvisible(path(panel, WAITING_FOR_AUTHOR_CONTAINER));
        tester.assertVisible(path(panel, ApproveFinalThesisPanel.APPROVED_THESIS_CONTAINER));
    }

    @Test
    public void has_no_rejection_date() throws Exception {
        Mockito.when(finalThesisService.isUploadAllowed(SOME_PROJECT)).thenReturn(true);
        Mockito.when(finalThesisService.hasFinalThesis(SOME_PROJECT)).thenReturn(false);
        Mockito.when(finalThesisService.getRejectedDate(SOME_PROJECT)).thenReturn(null);
        Mockito.when(dateService.format(null)).thenThrow(new NullPointerException());

        startPanel();
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new ApproveFinalThesisPanel("id", Model.of(SOME_PROJECT)));
    }

    private FinalThesis mockHasFinalThesisAndStartPanel(FinalThesis.Status status) {
        Mockito.when(finalThesisService.hasFinalThesis(SOME_PROJECT)).thenReturn(true);
        FinalThesis finalThesis = mockFinalThesisWithFileDescription();
        finalThesis.setStatus(status);

        startPanel();
        return finalThesis;
    }

    private FinalThesis mockFinalThesisWithFileDescription() {
        FileDescription fileDescription = new FileDescription();
        fileDescription.setName("this is the file");
        fileDescription.setIdentifier("file");
        fileDescription.setId(1L);

        FinalThesis finalThesis = new FinalThesis();
        finalThesis.setProject(SOME_PROJECT);
        finalThesis.setFileDescription(fileDescription);
        finalThesis.setDateCreated(new Date());

        SOME_PROJECT.addReviewer(reviewerUser);
        Mockito.when(finalThesisService.getFinalThesisFileDescription(SOME_PROJECT)).thenReturn(fileDescription);
        Mockito.when(finalThesisService.findByProject(SOME_PROJECT)).thenReturn(finalThesis);
        return finalThesis;
    }
}
