package se.su.dsv.scipro.finalthesis;

import org.apache.wicket.model.Model;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import org.mockito.Mockito;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.files.FileDescriptionPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.test.ObjectMother;

import java.util.Date;

import static se.su.dsv.scipro.finalthesis.FinalThesis.Status;
import static se.su.dsv.scipro.finalthesis.FinalThesisPanel.*;

public class FinalThesisPanelTest extends SciProTest {

    public static final ProjectType BACHELOR = new ProjectType(DegreeLevel.FIRST_CYCLE, ProjectType.BACHELOR_STRING, ProjectType.BACHELOR_STRING);
    private Project project = new Project();

    private FinalThesisPanel panel;
    public static final FinalThesis SOME_FINAL_THESIS = new FinalThesis();

    @Before
    public void setUp() {
        Mockito.when(finalThesisService.findByProject(project)).thenReturn(SOME_FINAL_THESIS);
        Mockito.when(finalSeminarService.hasHadFinalSeminar(Matchers.any(Project.class))).thenReturn(true);
        Mockito.when(finalThesisService.isUploadAllowed(Matchers.any(Project.class))).thenReturn(true);
    }

    @Test
    public void download_link_is_hidden_if_no_file() throws Exception {
        Mockito.when(finalThesisService.hasFinalThesis(project)).thenReturn(false);

        startPanel(BACHELOR);

        tester.assertInvisible(path(DOWNLOAD));
    }

    @Test
    public void download_link_is_visible_if_a_final_thesis_is_uploaded() throws Exception {
        mockExistingFinalThesis();

        startPanel(BACHELOR);

        tester.assertVisible(path(DOWNLOAD));
    }

    @Test
    public void download_link_is_a_file_description_panel() throws Exception {
        FileDescription fileDescription = mockExistingFinalThesis();

        startPanel(BACHELOR);

        tester.assertComponent(path(DOWNLOAD), FileDescriptionPanel.class);
        tester.assertModelValue(path(DOWNLOAD), fileDescription);
    }

    @Test
    public void revision_panel_is_shown_when_final_thesis_was_rejected() throws Exception {
        SOME_FINAL_THESIS.setStatus(Status.REJECTED);

        Mockito.when(finalThesisService.findByProject(project)).thenReturn(SOME_FINAL_THESIS);

        startPanel(BACHELOR);

        tester.assertVisible(path(FinalThesisPanel.REVISION_PANEL));
        tester.assertInvisible(path(DOWNLOAD));
    }

    @Test
    public void revision_panel_is_hidden() throws Exception {
        mockFinalThesisStatus(Status.NO_DECISION);

        startPanel(BACHELOR);

        tester.assertInvisible(path(FinalThesisPanel.REVISION_PANEL));
    }

    @Test
    public void revision_panel_shows_correct_date() throws Exception {
        SOME_FINAL_THESIS.setDateRejected(new Date());

        mockFinalThesisStatus(Status.REJECTED);

        startPanel(BACHELOR);

        tester.assertModelValue(path(REVISION_PANEL, FinalThesisPanel.REJECTED_DATE_LABEL), SOME_FINAL_THESIS.getDateRejected());
    }

    @Test
    public void approved_panel_is_hidden() throws Exception {
        mockFinalThesisStatus(Status.NO_DECISION);

        startPanel(BACHELOR);

        tester.assertInvisible(path(APPROVED_PANEL));
    }

    @Test
    public void approved_panel_is_shown() throws Exception {
        mockFinalThesisStatus(Status.APPROVED);
        Mockito.when(finalThesisService.isUploadAllowed(project)).thenReturn(false);

        startPanel(BACHELOR);

        tester.assertVisible(path(APPROVED_PANEL));
        tester.assertInvisible(path(UPLOAD_PANEL));
        tester.assertInvisible(path(DOWNLOAD));
    }

    @Test
    public void approved_date() throws Exception {
        SOME_FINAL_THESIS.setDateApproved(new Date());
        mockFinalThesisStatus(Status.APPROVED);
        Mockito.when(finalThesisService.isUploadAllowed(project)).thenReturn(false);

        startPanel(BACHELOR);

        tester.assertModelValue(path(APPROVED_PANEL, FinalThesisPanel.APPROVED_DATE), SOME_FINAL_THESIS.getDateApproved());
    }

    @Test
    public void no_decision_panel_is_shown() throws Exception {
        mockFinalThesisStatus(Status.NO_DECISION);
        Mockito.when(finalThesisService.isUploadAllowed(project)).thenReturn(false);

        startPanel(BACHELOR);

        tester.assertVisible(path(FinalThesisPanel.NO_DECISION_PANEL));
    }

    @Test
    public void no_decision_panel_is_hidden() throws Exception {
        mockFinalThesisStatus(Status.APPROVED);
        Mockito.when(finalThesisService.isUploadAllowed(project)).thenReturn(false);

        startPanel(BACHELOR);

        tester.assertInvisible(path(FinalThesisPanel.NO_DECISION_PANEL));
    }

    @Test
    public void visible_for_first_cycle() throws Exception {
        startPanel(BACHELOR);
        tester.assertVisible(path());
    }

    @Test
    public void visible_for_second_cycle() throws Exception {
        startPanel(new ProjectType(DegreeLevel.SECOND_CYCLE, ProjectType.MASTER_STRING, ProjectType.MASTER_STRING));
        tester.assertVisible(path());
    }

    private void mockFinalThesisStatus(Status status) {
        SOME_FINAL_THESIS.setStatus(status);

        Mockito.when(finalThesisService.findByProject(project)).thenReturn(SOME_FINAL_THESIS);
    }

    private FileDescription mockExistingFinalThesis() {
        FileDescription fileDescription = new FileDescription();
        fileDescription.setIdentifier("identifier");
        fileDescription.setUploader(ObjectMother.SOME_USER);
        fileDescription.setId(1L);
        Mockito.when(finalThesisService.hasFinalThesis(project)).thenReturn(true);
        Mockito.when(finalThesisService.getFinalThesisFileDescription(project)).thenReturn(fileDescription);
        return fileDescription;
    }

    @Override
    protected String path(Object... parts) {
        return super.path(panel, super.path(parts));
    }

    private void startPanel(ProjectType projectType) {
        project = Project.builder().title("title").projectType(projectType).build();
        panel = tester.startComponentInPage(new FinalThesisPanel("id", Model.of(project)));
    }
}
