package se.su.dsv.scipro.firstmeeting;

import org.apache.wicket.model.Model;
import org.apache.wicket.util.tester.FormTester;
import org.joda.time.DateTime;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import org.mockito.Mock;
import org.mockito.Mockito;
import se.su.dsv.scipro.SciProSessionTest;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.data.enums.DateStyle;
import se.su.dsv.scipro.match.*;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.UserBuilder;

import java.text.SimpleDateFormat;
import java.util.Date;

import static java.lang.String.format;
import static se.su.dsv.scipro.firstmeeting.FirstMeetingPanel.*;

public class FirstMeetingPanelTest extends SciProTest {

    @Mock
    private FirstMeetingRepository firstMeetingRepository;

    private FirstMeetingPanel panel;
    private FirstMeeting firstMeeting;
    private Idea idea;

    @Before
    public void setUp() throws Exception {
        Date courseStartDate = new Date();

        idea = createCompletedIdea(courseStartDate);
        firstMeeting = new FirstMeeting(courseStartDate, idea);
    }

    private Idea createCompletedIdea(Date courseStartDate) {
        Idea idea = createUnmatchedIdea(courseStartDate);
        idea.setProject(new Project());
        return idea;
    }

    private Idea createUnmatchedIdea(Date courseStartDate) {
        Idea idea = new Idea();
        ApplicationPeriod applicationPeriod = new ApplicationPeriod("name");
        applicationPeriod.setCourseStartDate(courseStartDate);
        idea.setApplicationPeriod(applicationPeriod);
        idea.setTitle("blabla");

        User user = new UserBuilder().create();

        setLoggedInAs(user);

        Match match = new Match();
        match.setIdea(idea);
        idea.setMatch(match);
        match.setStatus(Idea.Status.MATCHED);
        match.setSupervisor(user);

        return idea;
    }

    protected void setLoggedInAs(User user) {
        setLoggedIn(user != null);
        SciProSessionTest session = (SciProSessionTest) tester.getSession();
        session.setUser(user);
    }

    @Test
    public void renders() throws Exception {
        startPanel();
        Assert.assertNotNull(panel);
    }

    @Test
    public void meeting_form_is_not_visible_when_idea_is_unmatched() throws Exception {
        Date courseStartDate = new Date();
        Idea unmatched = createUnmatchedIdea(courseStartDate);

        startPanel(unmatched);

        tester.assertInvisible(path(SUPERVISOR_WMC, MEETING_FORM));
    }

    @Test
    public void submit_successful_form() throws Exception {
        startPanel();
        scheduleMeetingInThreeDays();

        tester.assertNoErrorMessage();
        Mockito.verify(firstMeetingRepository).save(Matchers.any(FirstMeeting.class));
    }

    @Test
    public void shows_feedback_that_meeting_is_scheduled_and_authors_notified() throws Exception {
        startPanel();
        Date date = createDateThreeDaysFromNow();
        String formattedDate = new SimpleDateFormat(dateService.getFormat(DateStyle.DATETIME)).format(date);
        Mockito.when(dateService.format(Matchers.any(Date.class), Matchers.eq(DateStyle.DATETIME))).thenReturn(formattedDate);

        scheduleMeeting(date);

        tester.assertInfoMessages(format(panel.getString("meeting.scheduled", panel.getModel()), formattedDate));
    }

    @Test
    public void not_able_to_submit_a_date_before_today() throws Exception {
        startPanel();
        Date notAllowedDate = new DateTime().minusDays(3).toDate();
        submitForm(notAllowedDate);

        tester.assertErrorMessages(panel.getString("DateValidator.minimum"));
        Mockito.verify(firstMeetingRepository, Mockito.never()).save(firstMeeting);
    }

    private Date scheduleMeetingInThreeDays() {
        Date inThreeDays = createDateThreeDaysFromNow();

        return scheduleMeeting(inThreeDays);
    }

    private Date scheduleMeeting(Date allowedDate) {
        submitForm(allowedDate);

        return allowedDate;
    }

    private Date createDateThreeDaysFromNow() {
        return new DateTime().plusDays(3).toDate();
    }

    private void submitForm(Date meetingDate) {
        FormTester formTester = tester.newFormTester(path(SUPERVISOR_WMC, MEETING_FORM));
        String dateString = new SimpleDateFormat(dateService.getFormat(DateStyle.DATE)).format(meetingDate);
        String timeString = new SimpleDateFormat(dateService.getFormat(DateStyle.TIME)).format(meetingDate);
        formTester.setValue(ROOM_FIELD, "a room");
        formTester.setValue(super.path(FIRST_MEETING_DATE, "date"), dateString);
        formTester.setValue(super.path(FIRST_MEETING_DATE, "time"), timeString);
        tester.executeAjaxEvent(path(SUPERVISOR_WMC, MEETING_FORM, SUBMIT_BUTTON), "onclick");
    }

    private void startPanel() {
        startPanel(idea);
    }

    private void startPanel(Idea passedIdea) {
        panel = tester.startComponentInPage(new FirstMeetingPanel("id", Model.of(passedIdea)));
    }

    @Override
    protected String path(Object... parts) {
        return super.path(panel, super.path(parts));
    }
}
