package se.su.dsv.scipro.grading;

import org.apache.wicket.AttributeModifier;
import org.apache.wicket.Component;
import org.apache.wicket.feedback.FeedbackMessage;
import org.apache.wicket.model.Model;
import org.apache.wicket.util.tester.FormTester;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.finalthesis.FinalThesis;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.report.*;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.UserBuilder;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.core.IsCollectionContaining.hasItem;
import static org.hamcrest.core.StringContains.containsString;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;
import static se.su.dsv.scipro.grading.FillOutGradingReportPanel.*;

public class FillOutGradingReportPanelTest extends SciProTest {

    public static final ProjectType BACHELOR = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");
    public static final Project SOME_PROJECT = Project.builder().title("Grading report project").projectType(BACHELOR).build();
    public static final User SOME_USER = new UserBuilder().setFirstName("Author").setLastName("One").create();
    public static final User SOME_USER_2 = new UserBuilder().setFirstName("Author").setLastName("two").create();

    private SupervisorGradingReport gradingReport1;
    private SupervisorGradingReport gradingReport2;

    static {
        setUpProject();
    }

    private FillOutGradingReportPanel panel;

    @Before
    public void setUp() throws Exception {
        prepareReports();
        when(gradingReportService.getSupervisorGradingReports(SOME_PROJECT)).thenReturn(Arrays.asList(gradingReport1, gradingReport2));
    }

    @Test
    public void renders() throws Exception {
        startPanel();
    }

    @Test
    public void saves_correct_points() throws Exception {
        startPanel();

        int points = 1;
        assignPointsToCriteria(points);

        ArgumentCaptor<GradingReport> captor = ArgumentCaptor.forClass(GradingReport.class);
        verify(gradingReportService, atLeastOnce()).save(captor.capture());

        assertEquals(points, gradingReport1.getGradingCriteria().get(0).getPoints().intValue());
    }

    @Test
    public void saves_correct_comment() throws Exception {
        String comment = "Very good!";

        startPanel();
        assignAssessmentToCriteria(comment);

        ArgumentCaptor<GradingReport> captor = ArgumentCaptor.forClass(GradingReport.class);
        verify(gradingReportService, atLeastOnce()).save(captor.capture());

        assertEquals(comment, gradingReport1.getGradingCriteria().get(0).getFeedback());
    }

    @Test
    public void saves_correct_plagiarism_feedback() throws Exception {
        String feedback = "Plagiarism!!";

        startPanel();
        assignPlagiarismFeedback(feedback);

        ArgumentCaptor<GradingReport> captor = ArgumentCaptor.forClass(GradingReport.class);
        verify(gradingReportService, atLeastOnce()).save(captor.capture());

        assertEquals(feedback, gradingReport1.getPlagiarismFeedback());
    }

    @Test
    public void shows_feedback_when_saving_report() throws Exception {
        startPanel();
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        formTester.submit("save");

        List<Serializable> messages = tester.getMessages(FeedbackMessage.SUCCESS);
        assertThat(messages, hasItem(containsString(panel.getString("saved"))));
    }

    @Test
    public void shows_existing_plagiarism_feedback() throws Exception {
        String feedback = "plagiarism!";
        gradingReport1.setPlagiarismFeedback(feedback);
        startPanel();

        tester.assertModelValue(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, PLAGIARISM_FEEDBACK), feedback);
    }

    @Test
    public void click_download_writes_to_pdf() throws Exception {
        gradingReport1.submit();

        startPanel();

        tester.clickLink(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, PDF_DOWNLOAD));

        assertEquals("application/pdf", tester.getContentTypeFromResponseHeader());
    }


    @Test
    public void plagiarism_field_is_disabled_after_submit() throws Exception {
        gradingReport1.submit();

        startPanel();

        tester.assertDisabled(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, PLAGIARISM_FEEDBACK));
    }

    @Test
    public void first_tab_is_active() throws Exception {
        startPanel();

        Component authorTab = tester.getComponentFromLastRenderedPage(path(NAVIGATION_TABS, 0));
        List<AttributeModifier> behaviors = authorTab.getBehaviors(AttributeModifier.class);
        assertEquals(1, behaviors.size());
    }

    @Test
    public void second_tab_is_inactive() throws Exception {
        startPanel();

        Component authorTab = tester.getComponentFromLastRenderedPage(path(NAVIGATION_TABS, 1));
        List<AttributeModifier> behaviors = authorTab.getBehaviors(AttributeModifier.class);
        assertEquals(0, behaviors.size());
    }

    @Test
    public void tab_link_has_href_modifier() throws Exception {
        startPanel();

        tester.assertModelValue(path(NAVIGATION_TABS, 0, FillOutGradingReportPanel.TAB_LINK), "#author0");
    }

    @Test
    public void tab_pane_has_id_modifier() throws Exception {
        startPanel();

        Component link = tester.getComponentFromLastRenderedPage(path(GRADING_REPORTS, 0));
        assertEquals("author0", link.getMarkupId());
    }

    @Test
    public void first_tab_pane_has_class_modifier() throws Exception {
        startPanel();

        Component link = tester.getComponentFromLastRenderedPage(path(GRADING_REPORTS, 0));
        List<AttributeModifier> behaviors = link.getBehaviors(AttributeModifier.class);
        assertEquals("class", behaviors.get(0).getAttribute());
    }

    @Test
    public void second_tab_does_not_have_active_class_behaviour() throws Exception {
        startPanel();

        Component link = tester.getComponentFromLastRenderedPage(path(GRADING_REPORTS, 1));
        List<AttributeModifier> behaviors = link.getBehaviors(AttributeModifier.class);
        assertEquals(0, behaviors.size());
    }

    @Test
    public void tab_have_copy_link_for_all_authors() throws Exception {
        startPanel();

        tester.assertListView(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, OTHER_REPORTS), new ArrayList<>(SOME_PROJECT.getProjectParticipants()));
    }

    @Test
    public void copy_results_to_co_author() throws Exception {
        when(gradingReportService.getSupervisorGradingReport(SOME_PROJECT, SOME_USER_2)).thenReturn(gradingReport2);
        startPanel();

        tester.clickLink(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, OTHER_REPORTS, 1, COPY));

        verify(gradingReportService).copyReport(gradingReport1, gradingReport2);
    }

    @Test
    public void copy_link_is_not_visible_for_single_authors() throws Exception {
        startPanelWithOneAuthor();

        tester.assertInvisible(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, OTHER_REPORTS, 0, COPY));
    }

    @Test
    public void copy_link_is_not_visible_if_report_has_been_submitted() throws Exception {
        gradingReport1.submit();
        startPanel();
        tester.assertInvisible(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, OTHER_REPORTS, 0, COPY));
    }

    @Test
    public void copying_results_shows_feedback() throws Exception {
        startPanel();

        tester.clickLink(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, OTHER_REPORTS, 1, COPY));

        List<Serializable> messages = tester.getMessages(FeedbackMessage.SUCCESS);
        assertThat(messages, hasItem(containsString(panel.getString("copy"))));
    }

    @Test
    public void move_to_reviewing_step_after_initial_assessment() throws Exception {
        startPanel();
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        formTester.submit(super.path(REPORT_TAB, "nextStep", "dialog", "initialAssessmentDone"));
        verify(gradingReportService).initialAssessmentDone(gradingReport1);
    }

    @Test
    public void input_disabled_after_submitting() throws Exception {
        gradingReport1.submit();

        startPanel();

        tester.assertDisabled(path(GRADING_REPORTS, 0, FORM, REPORT_TAB, CRITERIA_PANEL));
    }

    @Test
    public void finalize_grading_report() throws Exception {
        gradingReport1.setState(GradingReport.State.REVIEWING);
        startPanel();
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        formTester.submit(super.path(REPORT_TAB, "nextStep", "dialog", "finalize"));
        verify(gradingReportService).submitReport(gradingReport1);
    }

    @Test
    public void trying_to_move_to_reviewing_step_without_finishing_initial_assessment_should_show_error() throws Exception {
        doThrow(IllegalStateException.class).when(gradingReportService).initialAssessmentDone(any(GradingReport.class));
        startPanel();
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        formTester.submit(super.path(REPORT_TAB, "nextStep", "dialog", "initialAssessmentDone"));
        tester.assertErrorMessages(panel.getString("need.to.assess.all.criteria"));
    }

     private void assignAssessmentToCriteria(String assessment) {
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        int i = 0;
        for (GradingCriterion ignored : gradingReport1.getGradingCriteria()) {
            formTester.setValue(super.path(REPORT_TAB, CRITERIA_PANEL, CriteriaPanel.CRITERIA, i, CriteriaPanel.FEEDBACK), assessment);
            i++;
        }
        formTester.submit();
    }

    private void assignPlagiarismFeedback(String feedback) {
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        formTester.setValue(super.path(REPORT_TAB, PLAGIARISM_FEEDBACK), feedback);
        formTester.submit();
    }

    private void assignPointsToCriteria(int points) {
        FormTester formTester = tester.newFormTester(path(GRADING_REPORTS, 0, FORM));
        int i = 0;
        for (GradingCriterion ignored : gradingReport1.getGradingCriteria()) {
            formTester.setValue(super.path(REPORT_TAB, CRITERIA_PANEL, CriteriaPanel.CRITERIA, i, CriteriaPanel.POINTS), Integer.toString(points));
            i++;
        }
        formTester.submit();
    }

    @Override
    protected String path(Object... parts) {
        return super.path(panel, super.path(parts));
    }

    private static void setUpProject() {
        SOME_PROJECT.setHeadSupervisor(new UserBuilder().setFirstName("Head").setLastName("Supervisor").create());
        SOME_PROJECT.addProjectParticipant(SOME_USER);
        SOME_PROJECT.addProjectParticipant(SOME_USER_2);
        SOME_PROJECT.addReviewer(new UserBuilder().create());
    }

    private void prepareReports() {
        GradingReportTemplate gradingReportTemplate = new GradingReportTemplate(BACHELOR, 30);
        gradingReportTemplate.addProjectCriterion("title", 0, getPointTemplates(2));
        gradingReport1 = gradingReportTemplate.createSupervisorReport(SOME_PROJECT, SOME_USER);
        gradingReport2 = gradingReportTemplate.createSupervisorReport(SOME_PROJECT, SOME_USER_2);
        for (GradingCriterion gradingCriterion : gradingReport1.getGradingCriteria()) {
            gradingCriterion.setPoints(1);
            gradingCriterion.setFeedback("Feedback");
        }
        gradingReport1.setPlagiarismFeedback("feedback");
    }

    private List<GradingCriterionPointTemplate> getPointTemplates(int maxPoint) {
        List<GradingCriterionPointTemplate> gradingCriterionPointTemplates = new ArrayList<>();
        for (int i = 0; i <= maxPoint; i++) {
            gradingCriterionPointTemplates.add(new GradingCriterionPointTemplate.Builder().point(i).description("Description").build());
        }
        return gradingCriterionPointTemplates;
    }

    private void startPanel() {
        createFinalThesis(SOME_PROJECT);
        panel = tester.startComponentInPage(new FillOutGradingReportPanel("id", Model.of(SOME_PROJECT)));
    }

    private void startPanelWithOneAuthor() {
        Project project = Project.builder().title("title").projectType(BACHELOR).build();
        project.addProjectParticipant(SOME_USER);
        GradingReportTemplate gradingReportTemplate = new GradingReportTemplate(BACHELOR, 30);
        SupervisorGradingReport gradingReport = gradingReportTemplate.createSupervisorReport(project, SOME_USER);
        when(gradingReportService.getSupervisorGradingReports(project)).thenReturn(Collections.singletonList(gradingReport));

        createFinalThesis(project);

        panel = tester.startComponentInPage(new FillOutGradingReportPanel("id", Model.of(project)));
    }

    private void createFinalThesis(Project project) {
        FinalThesis finalThesis = new FinalThesis();
        finalThesis.setId(1L);
        FileDescription fileDescription = new FileDescription();
        fileDescription.setId(1L);
        finalThesis.setFileDescription(fileDescription);
        when(finalThesisService.findByProject(project)).thenReturn(finalThesis);
    }
}
