package se.su.dsv.scipro.match;

import org.apache.wicket.ajax.form.AjaxFormChoiceComponentUpdatingBehavior;
import org.apache.wicket.behavior.AbstractAjaxBehavior;
import org.apache.wicket.model.Model;
import org.apache.wicket.util.tester.FormTester;
import org.apache.wicket.util.tester.WicketTesterHelper;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.system.*;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.*;

import static org.mockito.Mockito.when;
import static se.su.dsv.scipro.match.ProjectIdeaSubmissionPanel.*;

public class ProjectIdeaSubmissionPanelTest extends SciProTest {

    private User user;
    private ProjectIdeaSubmissionPanel panel;
    private Idea idea;

    @Before
    public void setUp() throws Exception {
        user = new UserBuilder().create();
        user.addProgram(new Program());
        when(userService.findOne(Matchers.isA(Long.class))).thenReturn(user);
        setLoggedInAs(user);
        Language english = new Language();
        DomainObjects.injectId(english, 1L);
        when(languageService.findAll()).thenReturn(Collections.singletonList(english));

        idea = createIdea(2, 2);
    }

    @Test
    public void renders() throws Exception {
        startPanel(idea);
    }

    @Test
    public void submit_form_without_selecting_partner_on_bachelor_level_fails() throws Exception {
        startPanel(idea);
        fillInFormAndSubmit();
        tester.assertFeedback(feedbackPath("coAuthorFeedback"), panel.getString(CO_AUTHOR_CHOICE + ".Required"));
    }

    @Test
    public void submit_form_without_selecting_partner_on_bachelor_level_passes_if_exemption_exists() throws Exception {
        when(applicationPeriodService.hasExemption(Matchers.eq(user), Matchers.any(ApplicationPeriod.class))).thenReturn(true);
        startPanel(idea);
        fillInFormAndSubmit();
        tester.assertFeedback(feedbackPath("coAuthorFeedback"));
    }

    @Test
    public void submit_form_with_yourself_as_co_author_fails() throws Exception {
        when(userRepo.findOne(user.getId())).thenReturn(user);
        startPanel(idea);
        fillInFormAndSubmit(user.getId(), true);
        tester.assertFeedback(feedbackPath("coAuthorFeedback"), panel.getString("CoAuthorValidator.addedYourself"));
    }

    @Test
    public void submit_form_when_co_author_already_participating_on_another_idea() throws Exception {
        User coAuthor = mockCoAuthor();
        when(ideaService.authorParticipatingOnActiveIdea(coAuthor, idea.getApplicationPeriod())).thenReturn(true);
        startPanel(idea);
        fillInFormAndSubmit(coAuthor.getId(), true);
        tester.assertFeedback(feedbackPath("coAuthorFeedback"), panel.getString("CoAuthorValidator.coAuthorParticipating"));
    }

    @Test
    public void submit_form_when_project_type_max_authors_is_exceeded_fails() throws Exception {
        User coAuthor = mockCoAuthor();
        startPanel(createIdea(1, 1));
        fillInFormAndSubmit(coAuthor.getId(), true);
        tester.assertFeedback(feedbackPath("coAuthorFeedback"), panel.getString("CoAuthorValidator.toManyAuthors"));
    }

    @Test
    public void submit_form_without_keywords_fails() throws Exception {
        idea.setKeywords(new HashSet<Keyword>());
        startPanel(idea);
        fillInFormAndSubmit(mockCoAuthor().getId(), true);
        tester.assertFeedback(feedbackPath("areaContainer", "keywordsFeedback"), panel.getString("keywordError"));
    }

    @Test
    public void submit_form_with_to_many_keywords_fails() throws Exception {
        idea.setKeywords(new HashSet<>(getKeywords(6)));
        startPanel(idea);
        fillInFormAndSubmit(mockCoAuthor().getId(), true);
        tester.assertFeedback(feedbackPath("areaContainer", "keywordsFeedback"), panel.getString("keywordError"));
    }

    @Test
    public void submit_form_without_checking_contract_checkbox() throws Exception {
        startPanel(idea);
        fillInFormAndSubmit(mockCoAuthor().getId(), false);
        tester.assertFeedback(feedbackPath("contractFeedback"), panel.getString("CheckBoxRequiredValidator"));
    }

    private List<Keyword> getKeywords(int size) {
        List<Keyword> keywords = new ArrayList<>();
        for (int i = 0; i < size; i++) {
            Keyword keyword = new Keyword();
            DomainObjects.injectId(keyword, (long) i);
            keywords.add(keyword);
        }
        return keywords;
    }

    private User mockCoAuthor() {
        User coAuthor = new UserBuilder().create();
        when(userRepo.findOne(coAuthor.getId())).thenReturn(coAuthor);
        return coAuthor;
    }

    private String feedbackPath(final Object... feedbackId) {
        return path(panel, "ideaForm", path(feedbackId));
    }

    private void fillInFormAndSubmit() {
        FormTester formTester = fillInForm(null, true);
        formTester.submit();
    }

    private void fillInFormAndSubmit(Long id, boolean contract) {
        FormTester formTester = fillInForm(id, contract);
        formTester.submit();
    }

    private FormTester fillInForm(Long authorId, boolean contract) {
        FormTester formTester = tester.newFormTester(path(panel, IDEA_FORM));
        formTester.select(PROJECT_TYPE, 0);
        executeAjaxBehavior(formTester);
        formTester.setValue("title", "title");
        formTester.setValue(path("what", "input"), "what");
        formTester.setValue(path("why", "input"), "why");
        formTester.setValue(path("theory", "input"), "how");
        formTester.setValue(path("practical", "input"), "how");
        formTester.select("languages", 0);
        formTester.select(path("areaContainer", "area"), 0);
        formTester.setValue("contract", contract);
        if (authorId != null) {
            formTester.setValue(CO_AUTHOR_CHOICE, String.valueOf(authorId));
        }
        return formTester;
    }

    private void executeAjaxBehavior(FormTester formTester) {
        AbstractAjaxBehavior behavior = (AbstractAjaxBehavior) WicketTesterHelper.findBehavior(
                tester.getComponentFromLastRenderedPage(path(panel, IDEA_FORM, PROJECT_TYPE)), AjaxFormChoiceComponentUpdatingBehavior.class);
        tester.executeBehavior(behavior);
        formTester.select(PROJECT_TYPE, 0);
    }

    private void startPanel(Idea theIdea) {
        ResearchArea researchArea = new ResearchArea();
        researchArea.setId(5L);
        when(researchAreaService.findNotDeleted()).thenReturn(Collections.singletonList(researchArea));
        panel = tester.startComponentInPage(new ProjectIdeaSubmissionPanel("id", Model.of(theIdea), true));
    }

    private Idea createIdea(int minAuthors, int maxAuthors) {
        Idea idea = new Idea();
        idea.setKeywords(new HashSet<>(Collections.singletonList(new Keyword("keyword!"))));
        idea.setApplicationPeriod(createApplicationPeriod(minAuthors, maxAuthors));
        return idea;
    }

    private ApplicationPeriod createApplicationPeriod(int minAuthors, int maxAuthors) {
        ApplicationPeriod applicationPeriod = new ApplicationPeriod();
        applicationPeriod.setId(1L);
        applicationPeriod.setProjectTypes(new HashSet<>(Collections.singletonList(createProjectType(minAuthors, maxAuthors))));
        return applicationPeriod;
    }

    private ProjectType createProjectType(int minAuthors, int maxAuthors) {
        ProjectType projectType = new ProjectType(DegreeLevel.FIRST_CYCLE, "bachelor", "bachelor");
        projectType.setId(3L);
        ProjectTypeSettings projectTypeSettings = new ProjectTypeSettings(projectType);
        projectTypeSettings.setMinAuthors(minAuthors);
        projectTypeSettings.setMaxAuthors(maxAuthors);
        projectType.setProjectTypeSettings(projectTypeSettings);
        return projectType;
    }
}

