package se.su.dsv.scipro.milestones;

import org.apache.wicket.ajax.markup.html.form.AjaxSubmitLink;
import org.apache.wicket.feedback.FeedbackMessage;
import org.apache.wicket.markup.html.form.*;
import org.apache.wicket.markup.html.panel.FeedbackPanel;
import org.apache.wicket.model.Model;
import org.apache.wicket.util.tester.FormTester;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Matchers;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.internal.verification.VerificationModeFactory;
import org.springframework.data.domain.Sort;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.milestones.dataobjects.MilestonePhaseTemplate;
import se.su.dsv.scipro.system.Event;
import se.su.dsv.scipro.system.EventService;
import se.su.dsv.scipro.system.ProjectType;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

public class EditMilestoneActivityPanelTest extends SciProTest {

    @Mock
    private EventService eventService;

    EditMilestoneActivityPanel panel;
    MilestoneActivityTemplate activity;
    List<ProjectType> projectTypes;

    @Before
    public void setUp() throws Exception {
        projectTypes = Collections.singletonList(new ProjectType(ProjectType.BACHELOR, "bachelor", "master"));
        Mockito.when(projectTypeService.findAllActive()).thenReturn(projectTypes);
        MilestonePhaseTemplate phase = new MilestonePhaseTemplate("title", "desc");
        Mockito.when(mileStonePhaseTemplateService.findAll(Matchers.isA(Sort.class))).thenReturn(Collections.singletonList(phase));
        activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT, "Activity title", "Description");
        activity.setMilestonePhaseTemplate(phase);

        startPanel(activity);
    }

    @Test
    public void testRenders() throws Exception {
        Assert.assertNotNull(panel);
    }

    @Test
    public void testPanelContainsFeedback() throws Exception {
        tester.assertComponent(path(panel.getId(), EditMilestoneActivityPanel.FEEDBACK), FeedbackPanel.class);
    }

    @Test
    public void testPanelContainsForm() throws Exception {
        tester.assertComponent(path(panel.getId(), EditMilestoneActivityPanel.EDIT_FORM), Form.class);
    }

    @Test
    public void testFormContainsTitleField() throws Exception {
        tester.assertComponent(formPath(EditMilestoneActivityPanel.TITLE), RequiredTextField.class);
        tester.assertModelValue(formPath(EditMilestoneActivityPanel.TITLE), activity.getTitle());
    }

    @Test
    public void testFormContainsDescriptionArea() throws Exception {
        tester.assertComponent(formPath(EditMilestoneActivityPanel.DESCRIPTION), TextArea.class);
        tester.assertModelValue(formPath(EditMilestoneActivityPanel.DESCRIPTION), activity.getDescription());
    }

    @Test
    public void testFormContainsCheckboxMultipleChoice() throws Exception {
        tester.assertComponent(formPath(EditMilestoneActivityPanel.PROJECT_TYPES), CheckBoxMultipleChoice.class);
    }

    @Test
    public void testFormContainsEditableByStudentChoice() throws Exception {
        tester.assertComponent(formPath(EditMilestoneActivityPanel.EDITABLE_BY_AUTHORS), CheckBox.class);
    }

    @Test
    public void testFormContainsSaveButton() throws Exception {
        tester.assertComponent(formPath(EditMilestoneActivityPanel.SAVE), AjaxSubmitLink.class);
    }

    @Test
    public void testSubmitInvalidForm() throws Exception {
        fillOutForm("", "description");

        tester.clickLink(formPath(EditMilestoneActivityPanel.SAVE));

        Assert.assertFalse(tester.getMessages(FeedbackMessage.ERROR).isEmpty());

        tester.assertComponentOnAjaxResponse(path(panel.getId(), EditMilestoneActivityPanel.FEEDBACK));
    }

    @Test
    public void testSubmitCorrectForm() throws Exception {
        fillOutForm("Title", "Description");

        tester.clickLink(formPath(EditMilestoneActivityPanel.SAVE));

        tester.assertNoErrorMessage();
        Mockito.verify(milestoneActivityTemplateService, VerificationModeFactory.times(1)).save(Matchers.isA(MilestoneActivityTemplate.class), Matchers.isA(MilestonePhaseTemplate.class));
    }

    @Test
    public void testSubmitFormSavesCorrectInformation() throws Exception {
        String titleString = "this is a title";
        fillOutForm(titleString, "and this is a description");

        tester.clickLink(formPath(EditMilestoneActivityPanel.SAVE));

        MilestoneActivityTemplate saved = captureSave();

        Assert.assertEquals(titleString, saved.getTitle());
    }

    @Test
    public void change_activated_by_event() throws Exception {
        Event event = Mockito.mock(Event.class);
        Mockito.when(eventService.findAll()).thenReturn(Collections.singletonList(event));

        startPanel(activity);

        FormTester formTester = tester.newFormTester(formPath());
        formTester.select(EditMilestoneActivityPanel.EVENT, 0);
        formTester.setValue(EditMilestoneActivityPanel.PROJECT_TYPES, "on"); // required for form submit
        formTester.submit();

        MilestoneActivityTemplate milestoneActivityTemplate = captureSave();
        Assert.assertEquals(event, milestoneActivityTemplate.getActivatedBy());
    }

    private MilestoneActivityTemplate captureSave() {
        ArgumentCaptor<MilestoneActivityTemplate> captor = ArgumentCaptor.forClass(MilestoneActivityTemplate.class);
        Mockito.verify(milestoneActivityTemplateService).save(captor.capture(), Matchers.isA(MilestonePhaseTemplate.class));
        return captor.getValue();
    }

    private void fillOutForm(String title, String desc) {
        FormTester formTester = tester.newFormTester(formPath());
        formTester.setValue(EditMilestoneActivityPanel.TITLE, title);
        formTester.setValue(EditMilestoneActivityPanel.DESCRIPTION, desc);
        formTester.select(EditMilestoneActivityPanel.MILE_STONE_PHASE_TEMPLATE, 0);
        formTester.setValue(EditMilestoneActivityPanel.PROJECT_TYPES, "on");
    }

    private void startPanel(final MilestoneActivityTemplate activity) {
        panel = tester.startComponentInPage(new EditMilestoneActivityPanel("panel", Model.of(activity)));
    }

    private String formPath(String... paths) {
        return path(panel, EditMilestoneActivityPanel.EDIT_FORM, path((Object[]) paths));
    }
}
