package se.su.dsv.scipro.milestones.panels;

import org.apache.wicket.model.IModel;
import org.apache.wicket.model.Model;
import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.milestones.dataobjects.Milestone;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.Arrays;
import java.util.Collections;
import java.util.TreeSet;

import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.*;
import static se.su.dsv.scipro.milestones.panels.MileStoneCheckBoxPanel.CHECKBOX;
import static se.su.dsv.scipro.milestones.panels.MileStoneCheckBoxPanel.CREATE_SEMINAR;

public class MilestoneCheckBoxPanelTest extends SciProTest {

    private MileStoneCheckBoxPanel panel;
    private Project project;

    @Before
    public void setUp() throws Exception {
        project = Project.builder()
                .title("some title")
                .projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"))
                .build();
        DomainObjects.injectId(project, 1L);

        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);

        Milestone milestone = new Milestone(project, new MilestoneActivityTemplate());
        DomainObjects.injectId(milestone, 2L);

        when(mileStoneService.getMileStone(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(milestone);
        when(mileStoneService.getMileStone(isA(Project.class), isA(User.class), isA(MilestoneActivityTemplate.class))).thenReturn(milestone);
        startPanel(Model.of(activity), MilestoneActivityTemplate.Type.PROJECT);
    }

    @Test
    public void testRenders() throws Exception {
        assert panel != null;
    }

    @Test
    public void testShowCheckBoxIfActivityIsCorrectTypeAndHasNoException() throws Exception {
        startPanel(Model.of(new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT)), MilestoneActivityTemplate.Type.PROJECT);
        boolean isVisible = panel.get(CHECKBOX).isVisibleInHierarchy();
        assertTrue(isVisible);
    }

    @Test
    public void testHideCheckBoxIfActivityIsWrongType() throws Exception {
        startPanel(Model.of(new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT)), MilestoneActivityTemplate.Type.STUDENT);
        boolean isVisible = panel.get(CHECKBOX).isVisibleInHierarchy();
        assertFalse(isVisible);
    }

    @Test
    public void testClickOnProjectCheckBoxCallsCorrectServiceMethod() throws Exception {
        String checkBoxPath = path(panel.getId(), CHECKBOX);
        tester.executeAjaxEvent(checkBoxPath, "onclick");
        verify(mileStoneService, times(1)).setConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class), eq(true));
    }

    @Test
    public void testClickOnUnconfirmedStudentCheckBoxChangesToConfirmed() throws Exception {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.STUDENT);
        when(mileStoneService.isConfirmed(isA(Project.class), isA(User.class), isA(MilestoneActivityTemplate.class))).thenReturn(false);
        startPanel(Model.of(activity), Model.of(new User()));

        tester.executeAjaxEvent(path(panel.getId(), CHECKBOX), "onclick");
        verify(mileStoneService, times(1)).setConfirmed(isA(Project.class), isA(User.class), isA(MilestoneActivityTemplate.class), eq(true));
    }

    @Test
    public void testClickOnConfirmedStudentCheckBoxChangesToUnconfirmed() throws Exception {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.STUDENT);
        when(mileStoneService.isConfirmed(isA(Project.class), isA(User.class), isA(MilestoneActivityTemplate.class))).thenReturn(true);
        startPanel(Model.of(activity), Model.of(new User()));

        tester.executeAjaxEvent(path(panel.getId(), CHECKBOX), "onclick");
        verify(mileStoneService, times(1)).setConfirmed(isA(Project.class), isA(User.class), isA(MilestoneActivityTemplate.class), eq(false));
    }

    @Test
    public void testClickOnProjectCheckBoxSendsCorrectNotificationEvent() throws Exception {
        String checkBoxPath = path(panel.getId(), CHECKBOX);
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);
        when(mileStoneService.isConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(true);
        startPanel(Model.of(activity), MilestoneActivityTemplate.Type.PROJECT);
        tester.executeAjaxEvent(checkBoxPath, "onclick");
        verify(mileStoneService, times(1)).setConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class), eq(false));

        when(mileStoneService.isConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(false);
        startPanel(Model.of(activity), MilestoneActivityTemplate.Type.PROJECT);
        tester.executeAjaxEvent(checkBoxPath, "onclick");
        verify(mileStoneService, times(1)).setConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class), eq(true));
    }

    @Test
    public void testShowSeminarLinkIfActivityHasSeminarExceptionAndMileStoneIsNotConfirmed() throws Exception {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);
        activity.setCode(MilestoneActivityTemplate.CREATE_SEMINAR);
        when(mileStoneService.isConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(false);
        startPanel(Model.of(activity), MilestoneActivityTemplate.Type.PROJECT);

        boolean isVisible = panel.get(CREATE_SEMINAR).isVisibleInHierarchy();
        assertTrue(isVisible);
    }

    @Test
    public void testHideSeminarLinkIfActivityDoesNotHaveSeminarException() throws Exception {
        when(mileStoneService.isConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(false);
        startPanel(Model.of(new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT)), MilestoneActivityTemplate.Type.PROJECT);

        boolean isVisible = panel.get(CREATE_SEMINAR).isVisibleInHierarchy();
        assertFalse(isVisible);
    }

    @Test
    public void testHideSeminarLinkIfActivityUsWrongType() throws Exception {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);
        activity.setCode(MilestoneActivityTemplate.CREATE_SEMINAR);
        when(mileStoneService.isConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(false);
        startPanel(Model.of(activity), MilestoneActivityTemplate.Type.STUDENT);

        boolean isVisible = panel.get(CREATE_SEMINAR).isVisibleInHierarchy();
        assertFalse(isVisible);
    }

    @Test
    public void testHideSeminarLinkIfMileStoneIsConfirmed() throws Exception {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);
        activity.setCode(MilestoneActivityTemplate.CREATE_SEMINAR);
        when(mileStoneService.isConfirmed(isA(Project.class), isA(MilestoneActivityTemplate.class))).thenReturn(true);
        startPanel(Model.of(activity), MilestoneActivityTemplate.Type.PROJECT);

        boolean isVisible = panel.get(CREATE_SEMINAR).isVisibleInHierarchy();
        assertFalse(isVisible);
    }

    private void startPanel(final IModel<MilestoneActivityTemplate> activityModel, final IModel<User> studentModel) {
        panel = tester.startComponentInPage(new MileStoneCheckBoxPanel("panel", Model.of(project), activityModel, studentModel, MilestoneActivityTemplate.Type.STUDENT));
    }

    private void startPanel(final IModel<MilestoneActivityTemplate> activityModel, final MilestoneActivityTemplate.Type type) {
        panel = tester.startComponentInPage(new MileStoneCheckBoxPanel("panel", Model.of(project), activityModel, type));
    }

}
