package se.su.dsv.scipro.milestones.panels;

import org.apache.wicket.markup.html.basic.Label;
import org.apache.wicket.markup.html.list.ListView;
import org.apache.wicket.model.Model;
import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.TreeSet;

import static org.junit.Assert.assertFalse;
import static org.mockito.Matchers.eq;
import static org.mockito.Matchers.isA;
import static org.mockito.Mockito.when;

public class MilestoneSummaryPanelTest extends SciProTest {

    MilestoneSummaryPanel panel;
    Project project;
    private User studentUser;
    private User supervisorUser;

    @Before
    public void setUp() throws Exception {

        supervisorUser = new UserBuilder().setFirstName("super").setLastName("visor").create();

        studentUser = new UserBuilder().setFirstName("author").setLastName("1").create();
        TreeSet<User> projectParticipants = new TreeSet<>(new User.ByNameComparator());
        projectParticipants.addAll(Collections.singletonList(studentUser));
        project = Project.builder()
                .title("some title")
                .projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor"))
                .projectParticipants(projectParticipants)
                .headSupervisor(supervisorUser)
                .build();
        DomainObjects.injectId(project, 1L);
        when(milestoneActivityTemplateService.countActivities(project.getProjectType(), project.getDateCreated())).thenReturn(1L);

        MilestoneActivityTemplate notEditableByStudent = new MilestoneActivityTemplate();
        notEditableByStudent.setEditableByStudents(false);

        MilestoneActivityTemplate editableByStudent = new MilestoneActivityTemplate();
        editableByStudent.setEditableByStudents(true);

        when(milestoneActivityTemplateService.getActivitiesInOrder(MilestoneActivityTemplate.Type.PROJECT, project)).thenReturn(Arrays.asList(notEditableByStudent, editableByStudent));

        startPanel();
    }

    @Test
    public void testRenders() throws Exception {
        assert panel != null;
    }

    @Test
    public void testPanelContainsProjectProgress() throws Exception {
        tester.assertComponent(path(panel, MilestoneSummaryPanel.PROJECT_PROGRESS), MileStoneProgressPanel.class);
    }

    @Test
    public void testPanelContainsListOfStudents() throws Exception {
        String listPath = path(panel, MilestoneSummaryPanel.STUDENT_LIST);
        tester.assertComponent(listPath, ListView.class);
        tester.assertListView(listPath, new ArrayList<>(project.getProjectParticipants()));
    }

    @Test
    public void testStudentContainsNameLabel() throws Exception {
        tester.assertComponent(studentPath(MilestoneSummaryPanel.STUDENT_NAME), Label.class);
        tester.assertLabel(studentPath(MilestoneSummaryPanel.STUDENT_NAME), project.getProjectParticipants().first().getFullName());
    }

    @Test
    public void testEachStudentContainsProgress() throws Exception {
        tester.assertComponent(studentPath(MilestoneSummaryPanel.STUDENT_PROGRESS), MileStoneProgressPanel.class);
    }

    @Test
    public void testPanelContainsListOfUpcomingActivities() throws Exception {
        MilestoneActivityTemplate activity1 = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);
        DomainObjects.injectId(activity1, 1L);
        MilestoneActivityTemplate activity2 = new MilestoneActivityTemplate(MilestoneActivityTemplate.Type.PROJECT);
        DomainObjects.injectId(activity2, 2L);
        when(milestoneActivityTemplateService.getActivitiesInOrder(MilestoneActivityTemplate.Type.PROJECT, project)).thenReturn(Arrays.asList(activity1, activity2));
        when(mileStoneService.isConfirmed(isA(Project.class), eq(activity1))).thenReturn(true);
        when(mileStoneService.isConfirmed(isA(Project.class), eq(activity2))).thenReturn(false);

        startPanel();

        tester.assertListView(path(panel, MilestoneSummaryPanel.UPCOMING_LIST), Collections.singletonList(activity2));
    }

    @Test
    public void invisibleWhenNoActiveActivities() throws Exception {
        when(milestoneActivityTemplateService.countActivities(project.getProjectType(), project.getDateCreated())).thenReturn(0L);
        startPanel();
        assertFalse(panel.isVisibilityAllowed());
    }

    private String studentPath(String... paths) {
        return path(panel, MilestoneSummaryPanel.STUDENT_LIST, 0, path((Object[]) paths));
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new MilestoneSummaryPanel("panel", Model.of(project)));
    }
}
