package se.su.dsv.scipro.peer;

import org.apache.wicket.feedback.FeedbackMessage;
import org.apache.wicket.model.Model;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ObjectMother;
import se.su.dsv.scipro.util.PageParameterKeys;

import java.io.Serializable;
import java.util.*;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.core.IsCollectionContaining.hasItem;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static se.su.dsv.scipro.peer.MyPeerPanel.*;
import static se.su.dsv.scipro.test.ObjectMother.SOME_PROJECT;

public class MyPeerPanelTest extends SciProTest {

    public static final String PANEL = "id";
    public static final String SOME_PROJECT_TITLE = "someProjectTitle";
    public static final long REVIEW_ID = 5L;
    private Date requestDate;
    private Date reviewDate;
    private MyPeerPanel panel;

    @Before
    public void setUp() throws Exception {
        requestDate = new Date();
        reviewDate = new DateTime().plusDays(3).toDate();
        mockExistingPeers(getRequests(RequestStatus.WAITING), getReviews());
        startPanel();
    }

    @Test
    public void lists_my_requests() throws Exception {
        tester.assertListView(path(PANEL, REQUEST_LIST), getRequests(RequestStatus.WAITING));
    }

    @Test
    public void lists_my_reviews() throws Exception {
        tester.assertListView(path(PANEL, REVIEW_LIST), getReviews());
    }

    @Test
    public void requests_lists_date_created() throws Exception {
        tester.assertModelValue(path(PANEL, REQUEST_LIST, 0, REQUEST_DATE), requestDate);
    }

    @Test
    public void reviews_lists_date_created() throws Exception {
        tester.assertModelValue(path(PANEL, REVIEW_LIST, 0, REVIEW_DATE), reviewDate);
    }

    @Test
    public void review_list_shows_title_of_request_project() throws Exception {
        tester.assertLabel(path(PANEL, REVIEW_LIST, 0, REVIEW_LINK, PROJECT_TITLE), SOME_PROJECT_TITLE);
    }

    @Test
    public void click_on_review_title_takes_reviewer_to_the_review() throws Exception {
        PageParameters parameters = new PageParameters();
        parameters.add(PageParameterKeys.MAP.get(PeerReview.class), REVIEW_ID);
        tester.assertBookmarkablePageLink(path(PANEL, REVIEW_LIST, 0, REVIEW_LINK), PeerPortalPage.class, parameters);
        tester.assertBookmarkablePageLink(path(PANEL, REVIEW_LIST, 1, REVIEW_LINK), ViewPeerReviewPage.class, parameters);
    }

    @Test
    public void request_list_shows_request_status() throws Exception {
        tester.assertModelValue(path(PANEL, REQUEST_LIST, 0, REQUEST_LINK, REQUEST_STATUS), getPeerRequest().getStatus());
    }

    @Test
    public void review_list_shows_review_status() throws Exception {
        tester.assertModelValue(path(PANEL, REVIEW_LIST, 0, REVIEW_STATUS), getReviews().get(0).getStatus());
    }

    @Test
    public void cancel_request_calls_service() throws Exception {
        tester.clickLink(path(PANEL, REQUEST_LIST, 0, CANCEL));
        verify(peerRequestService).deleteRequest(eq(getPeerRequest()));
    }

    @Test
    public void cancel_generates_feedback() throws Exception {
        tester.clickLink(path(PANEL, REQUEST_LIST, 0, CANCEL));
        List<Serializable> messages = tester.getMessages(FeedbackMessage.SUCCESS);
        assertThat(messages, hasItem(panel.getString("deleted")));
    }

    @Test
    public void cancel_link_is_only_visible_for_awaiting_requests() throws Exception {
        mockExistingPeers(getRequests(RequestStatus.TAKEN), getReviews());
        startPanel();
        tester.assertInvisible(path(PANEL, REQUEST_LIST, 0, CANCEL));
    }

    @Test
    public void no_requests_label_visible_if_no_requests_exist() throws Exception {
        mockExistingPeers(new ArrayList<PeerRequest>(), getReviews());
        startPanel();
        tester.assertVisible(path(PANEL, NO_REQUESTS));
    }

    @Test
    public void no_reviews_label_visible_if_no_reviews_exist() throws Exception {
        mockExistingPeers(new ArrayList<PeerRequest>(), new ArrayList<PeerReview>());
        startPanel();
        tester.assertVisible(path(PANEL, NO_REVIEWS));
    }

    @Test
    public void request_is_clickable_if_review_is_completed() throws Exception {
        mockExistingPeers(getRequests(RequestStatus.FINISHED), getReviews());
        startPanel();
        tester.assertEnabled(path(PANEL, REQUEST_LIST, 0, REQUEST_LINK));
    }

    @Test
    public void request_is_not_clickable_if_review_is_completed() throws Exception {
        tester.assertDisabled(path(PANEL, REQUEST_LIST, 0, REQUEST_LINK));
    }

    private PeerRequest getPeerRequest() {
        return getRequests(RequestStatus.WAITING).get(0);
    }

    private void mockExistingPeers(List<PeerRequest> requests, List<PeerReview> reviews) {
        User me = SciProSession.get().getUser();
        when(peerRequestService.findByRequesterAndProject(me, SOME_PROJECT)).thenReturn(requests);
        when(peerReviewService.findByReviewerAndProject(me, SOME_PROJECT)).thenReturn(reviews);
    }

    private List<PeerReview> getReviews() {
        PeerReview inProgress = createReview();
        inProgress.setPeerRequest(createRequest(Project.builder().title(SOME_PROJECT_TITLE).projectType(ObjectMother.BACHELOR).build(), RequestStatus.WAITING));
        PeerReview complete = createReview();
        complete.setStatus(PeerReview.ReviewStatus.COMPLETED);
        complete.setPeerRequest(createRequest(Project.builder().title(SOME_PROJECT_TITLE).projectType(ObjectMother.BACHELOR).build(), RequestStatus.WAITING));
        return Arrays.asList(inProgress, complete);
    }

    private List<PeerRequest> getRequests(RequestStatus status) {
        PeerRequest peerRequest = createRequest(SOME_PROJECT, status);
        peerRequest.setPeerReviews(Collections.singletonList(createReview()));
        return Collections.singletonList(peerRequest);
    }

    private PeerReview createReview() {
        PeerReview peerReview = new PeerReview();
        peerReview.setId(REVIEW_ID);
        peerReview.setDateCreated(reviewDate);
        return peerReview;
    }

    private PeerRequest createRequest(Project project, RequestStatus status) {
        PeerRequest peerRequest = new PeerRequest();
        peerRequest.setProject(project);
        peerRequest.setDateCreated(requestDate);
        peerRequest.setStatus(status);
        return peerRequest;
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new MyPeerPanel(PANEL, Model.of(SOME_PROJECT)));
    }
}
