package se.su.dsv.scipro.peer;

import org.apache.wicket.model.Model;
import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.session.SciProSession;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.test.DomainObjects;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import static org.mockito.Matchers.any;
import static org.mockito.Mockito.when;

public class PeerPortalHeaderTest extends SciProTest {
    private Project project;
    private PeerPortalHeader panel;

    @Before
    public void setUp() throws Exception {
        project = Project.builder().title("Title").projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor")).build();
        DomainObjects.injectId(project, 234L);
    }

    @Test
    public void renders() throws Exception {
        startPanel();
    }

    @Test
    public void shows_review_box_if_user_has_more_requests_than_reviews() throws Exception {
        mockMoreRequestsThanReviews();
        startPanel();
        tester.assertVisible(path(panel, "review"));
    }

    @Test
    public void hides_review_box_if_user_has_review_in_progress() throws Exception {
        mockMoreRequestsThanReviews();
        mockReviewInProgress();
        startPanel();
        tester.assertInvisible(path(panel, "review"));
    }

    @Test
    public void has_link_to_perform_review_if_user_has_more_requests_than_reviews() throws Exception {
        mockMoreRequestsThanReviews();
        startPanel();
        tester.assertBookmarkablePageLink(path(panel, "review", "perform"), PeerPortalPage.class, PeerPortalPage.getPageParameters(project));
    }

    @Test
    public void request_box_is_hidden_if_user_has_more_requests_than_reviews() throws Exception {
        mockMoreRequestsThanReviews();
        startPanel();
        tester.assertInvisible(path(panel, "request"));
    }

    @Test
    public void request_box_hidden_if_review_in_progress() throws Exception {
        mockMoreReviewsThanRequests();
        mockReviewInProgress();
        startPanel();
        tester.assertInvisible(path(panel, "request"));
    }

    @Test
    public void shows_request_box_if_user_needs_to_upload_a_request() throws Exception {
        mockMoreReviewsThanRequests();
        startPanel();
        tester.assertVisible(path(panel, "request"));
    }

    @Test
    public void review_box_is_hidden_if_user_needs_to_upload_a_request() throws Exception {
        mockMoreReviewsThanRequests();
        startPanel();
        tester.assertInvisible(path(panel, "review"));
    }

    @Test
    public void has_peer_request_submission_dialog() throws Exception {
        startPanel();
        tester.assertComponent(path(panel, "submit"), PeerRequestSubmissionPanel.class);
    }

    @Test
    public void shows_review_in_progress_box_if_review_in_progress() throws Exception {
        mockReviewInProgress();
        startPanel();
        tester.assertVisible(path(panel, "reviewInProgress"));
    }

    @Test
    public void hides_review_in_progress_box_if_no_review_in_progress() throws Exception {
        mockNoReviewInProgress();
        startPanel();
        tester.assertInvisible(path(panel, "reviewInProgress"));
    }

    private void mockMoreRequestsThanReviews() {
        when(peerRequestService.count(any(PeerRequestService.Filter.class))).thenReturn(100L);
    }

    private void mockMoreReviewsThanRequests() {
        when(peerReviewService.count(any(PeerReviewService.Filter.class))).thenReturn(100L);
    }

    private void mockReviewInProgress() {
        when(peerReviewService.findInProgressReviews(SciProSession.get().getUser(), project)).thenReturn(Collections.singletonList(new PeerReview()));
    }

    private void mockNoReviewInProgress() {
        when(peerReviewService.findInProgressReviews(SciProSession.get().getUser(), project)).thenReturn(new ArrayList<PeerReview>());
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new PeerPortalHeader("id", Model.of(project)));
    }
}
