package se.su.dsv.scipro.peer;

import org.apache.wicket.model.Model;
import org.apache.wicket.util.file.File;
import org.apache.wicket.util.tester.FormTester;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.checklist.ChecklistTemplate;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.Language;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.UserBuilder;

import java.net.URL;
import java.util.Arrays;
import java.util.Collections;

import static org.junit.Assert.assertEquals;
import static org.mockito.Matchers.isA;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class PeerRequestSubmissionPanelTest extends SciProTest {

    private Project project = Project.builder().title("project").projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "bachelor", "bachelor")).projectStatus(ProjectStatus.ACTIVE).build();
    private PeerRequestSubmissionPanel panel;
    private User user;
    private ChecklistTemplate background;
    private Language language;

    @Before
    public void setUp() throws Exception {
        user = new UserBuilder().create();
        setLoggedInAs(user);
        project.addProjectParticipant(user);
        DomainObjects.injectId(project, 2134L);
        background = new ChecklistTemplate("Background", user);
        when(checklistTemplateService.findPeerRequestChecklists(project)).thenReturn(Collections.singletonList(background));
        language = new Language();
        language.setId(657863L);
        when(languageService.findAll()).thenReturn(Collections.singletonList(language));
    }

    @Test
    public void renders_with_correct_user() throws Exception {
        startPanel();
        tester.assertVisible(path(panel));
    }

    @Test
    public void panel_is_not_visible_if_project_is_not_active() throws Exception {
        project.setProjectStatus(ProjectStatus.COMPLETED);
        startPanel();
        tester.assertInvisible(path(panel));
    }

    @Test
    public void submit_form_without_upload() throws Exception {
        startPanel();
        FormTester formTester = tester.newFormTester(path(panel, "requestForm"));
        formTester.select("language", 0 );
        formTester.submit();
        tester.assertErrorMessages(panel.getString(("fileUpload.Required")));
    }

    @Test
    public void submit_new_peer_request() throws Exception {
        String comment = "my comment";

        startPanel();
        FormTester formTester = tester.newFormTester(path(panel, "requestForm"));
        formTester.setValue("comment", comment);
        formTester.select(path("listContainer", "radioGroup"), 1);
        formTester.setFile("fileUpload", getTestFile(), "text/plain");
        formTester.select("language", 0 );
        formTester.submit();

        tester.assertNoErrorMessage();

        ArgumentCaptor<PeerRequest> captor = ArgumentCaptor.forClass(PeerRequest.class);
        verify(peerPortal).storePeerRequest(isA(FileUpload.class), captor.capture());

        PeerRequest peerRequest = captor.getValue();
        assertEquals(user, peerRequest.getRequester());
        assertEquals(comment, peerRequest.getComment());
        assertEquals(background, peerRequest.getChecklistTemplate());
        assertEquals(project, peerRequest.getProject());
    }

    private static File getTestFile() throws Exception {
        URL resource = Thread.currentThread().getContextClassLoader().getResource("peer_request.txt");
        assert resource != null;
        return new File(resource.toURI());
    }

    private void startPanel() {
        panel = tester.startComponentInPage(new PeerRequestSubmissionPanel("id", Model.of(project)));
    }
}
