package se.su.dsv.scipro.peer;

import org.apache.wicket.markup.html.WebPage;
import org.apache.wicket.model.Model;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.joda.time.DateTime;
import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ObjectMother;
import se.su.dsv.scipro.util.PageParameterKeys;

import java.util.*;

import static org.mockito.Mockito.when;
import static se.su.dsv.scipro.peer.PeerSummaryPanel.*;

public class PeerSummaryPanelTest extends SciProTest {

    public static final String PANEL = "id";

    private PeerReview reviewReceived;
    private PeerReview reviewPerformed;
    private Project project;

    @Before
    public void setUp() throws Exception {
        User author = User.builder().firstName("author").lastName("author").emailAddress("author@author.com").build();
        author.setId(4L);
        reviewReceived = createReview(1L, new DateTime().plusDays(1).toDate(), User.builder().firstName("reviewer").lastName("1").emailAddress("reviewer@example.com").build());
        reviewPerformed = createReview(2L, new DateTime().plusDays(3).toDate(), User.builder().firstName("reviewer").lastName("2").emailAddress("reviewer@example.com").build());
        project = Project.builder().title("title").projectType(ObjectMother.BACHELOR).build();
        project.addProjectParticipant(author);
        setLoggedInAs(author);
        startPanel(performedReviews(), receivedReviews());
    }

    @Test
    public void lists_reviews_for_completed_requests() throws Exception {
        tester.assertListView(path(PANEL, RECEIVED_REVIEWS), receivedReviews());
    }

    @Test
    public void lists_completed_reviews_by_project_authors() throws Exception {
        tester.assertListView(path(PANEL, PERFORMED_REVIEWS), performedReviews());
    }

    @Test
    public void lists_have_links_to_reviews() throws Exception {
        assertReviewPageLink(ViewPeerReviewPage.class);
    }

    @Test
    public void links_to_supervisor_page_if_not_project_participant() throws Exception {
        setLoggedInAs(ObjectMother.SOME_USER);
        startPanel(performedReviews(), receivedReviews());
        assertReviewPageLink(SupervisorViewPeerReviewPage.class);
    }

    @Test
    public void lists_show_date_created() throws Exception {
        tester.assertModelValue(path(PANEL, RECEIVED_REVIEWS, 0, REVIEW_LINK, DATE_CREATED), reviewReceived.getDateCreated());
        tester.assertModelValue(path(PANEL, PERFORMED_REVIEWS, 0, REVIEW_LINK, DATE_CREATED), reviewPerformed.getDateCreated());
    }

    @Test
    public void lists_show_reviewer_name() throws Exception {
        tester.assertLabel(path(PANEL, RECEIVED_REVIEWS, 0, REVIEWER), reviewReceived.getReviewer().getFullName());
        tester.assertLabel(path(PANEL, PERFORMED_REVIEWS, 0, REVIEWER), reviewPerformed.getReviewer().getFullName());
    }

    @Test
    public void no_reviews_label_not_visible_if_reviews_exist() throws Exception {
        tester.assertInvisible(path(PANEL, NO_PERFORMED));
        tester.assertInvisible(path(PANEL, NO_RECEIVED));
    }

    @Test
    public void no_reviews_label_visible_if_empty() throws Exception {
        startPanel(new ArrayList<PeerReview>(), new ArrayList<PeerReview>());
        tester.assertVisible(path(PANEL, NO_PERFORMED));
        tester.assertVisible(path(PANEL, NO_RECEIVED));
    }

    private List<PeerReview> performedReviews() {
        return Collections.singletonList(reviewPerformed);
    }

    private void assertReviewPageLink(Class<? extends WebPage> pageClass) {
        PageParameters parameters = new PageParameters();
        parameters.add(PageParameterKeys.MAP.get(PeerReview.class), reviewReceived.getId());
        tester.assertBookmarkablePageLink(path(PANEL, RECEIVED_REVIEWS, 0, REVIEW_LINK), pageClass, parameters);
    }

    private List<PeerReview> receivedReviews() {
        return Collections.singletonList(reviewReceived);
    }

    private PeerReview createReview(long id, Date dateCreated, User reviewer) {
        PeerReview peerReview = new PeerReview();
        peerReview.setDateCreated(dateCreated);
        peerReview.setId(id);
        peerReview.setReviewer(reviewer);
        return peerReview;
    }

    private void startPanel(List<PeerReview> performed, List<PeerReview> received) {
        when(peerReviewService.findCompletedReviews(project)).thenReturn(performed);
        when(peerReviewService.findReceivedCompletedReviews(project)).thenReturn(received);
        tester.startComponentInPage(new PeerSummaryPanel(PANEL, Model.of(project)));
    }

}
