package se.su.dsv.scipro.project.panels;

import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.ProjectService;
import se.su.dsv.scipro.project.ProjectStatus;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.ObjectMother;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static java.util.Arrays.asList;
import static org.junit.Assert.assertEquals;
import static org.mockito.Mockito.*;

public class ProjectsOverviewPanelTest extends SciProTest {

    public static final ProjectType BACHELOR = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");

    private User author;
    private List<Project> projects;

    @Before
    public void setUp() throws Exception {
        projects = Collections.singletonList(buildProject());
        when(projectService.findAll(anyFilter())).thenReturn(projects);

        author = User.builder().firstName("Kalle").lastName("Testare").emailAddress("kalle@dsv.su.se").build();
        setLoggedInAs(author);
    }

    @Test
    public void renders() throws Exception {
        tester.startComponentInPage(ProjectsOverviewPanel.class);
    }

    @Test
    public void only_shows_project_where_the_current_user_is_author() throws Exception {
        tester.startComponentInPage(ProjectsOverviewPanel.class);
        assertEquals(author, filter().getAuthorUser());
    }

    @Test
    public void lists_correct_projects() throws Exception {
        tester.startComponentInPage(ProjectsOverviewPanel.class);
        tester.assertListView(path("tableContainer", "listView"), projects);
    }

    @Test
    public void shows_empty_label_if_there_are_no_projects_matching_filter() throws Exception {
        when(projectService.findAll(anyFilter())).thenReturn(Collections.<Project>emptyList());
        when(projectService.findByUserAndStatus(author, ProjectStatus.COMPLETED)).thenReturn(Collections.singletonList(ObjectMother.SOME_PROJECT));
        tester.startComponentInPage(ProjectsOverviewPanel.class);
        tester.assertVisible(path("tableContainer", "emptyLabel"));
    }

    @Test
    public void hides_empty_label_if_there_are_projects() throws Exception {
        tester.startComponentInPage(ProjectsOverviewPanel.class);
        tester.assertInvisible(path("tableContainer", "emptyLabel"));
    }

    private static ProjectService.Filter anyFilter() {
        return any(ProjectService.Filter.class);
    }

    private Project buildProject() {
        Project project = Project.builder().title("My project").projectType(BACHELOR).build();
        project.setHeadSupervisor(User.builder().firstName("Susan").lastName("Supervisor").emailAddress("susan@dsv.su.se").build());
        DomainObjects.injectId(project, 1L);
        return project;
    }

    private ProjectService.Filter filter() {
        ArgumentCaptor<ProjectService.Filter> captor = ArgumentCaptor.forClass(ProjectService.Filter.class);
        verify(projectService, atLeastOnce()).findAll(captor.capture());
        return captor.getValue();
    }
}
