package se.su.dsv.scipro.reviewer;

import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.apache.wicket.util.file.File;
import org.apache.wicket.util.tester.FormTester;
import org.junit.Before;
import org.junit.Test;
import org.mockito.ArgumentCaptor;
import scala.Option;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.file.FileUpload;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.panels.ProjectHeaderPanel;
import se.su.dsv.scipro.reviewing.FinalSeminarApproval;
import se.su.dsv.scipro.reviewing.ReviewerApproval;
import se.su.dsv.scipro.reviewing.RoughDraftApproval;
import se.su.dsv.scipro.supervisor.panels.FinalSeminarApprovalProcessPanel;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;

import java.net.URISyntaxException;
import java.util.Date;

import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.*;

public class ReviewerApprovalDecisionPageTest extends SciProTest {
    private Project project;
    private ReviewerApproval finalSeminarApproval;

    @Before
    public void setUp() throws Exception {
        project = createProject();
        setLoggedInAs(project.getReviewer());
        finalSeminarApproval = new FinalSeminarApproval(project, mock(FileDescription.class),"test", new Date());
        DomainObjects.injectId(finalSeminarApproval, 123L);
        when(myReviewService.findApproval(finalSeminarApproval.getId())).thenReturn(finalSeminarApproval);
        when(finalSeminarApprovalService.findBy(project)).thenReturn(Option.<FinalSeminarApproval>empty());
        when(roughDraftApprovalService.findBy(project)).thenReturn(Option.<RoughDraftApproval>empty());
        startPage();
    }

    @Test
    public void can_be_initiated_with_page_paremeters() throws Exception {
        tester.assertRenderedPage(ReviewerApprovalDecisionPage.class);
        tester.assertModelValue("decision", finalSeminarApproval);
    }

    @Test
    public void shows_project_information() throws Exception {
        tester.assertComponent("projectHeader", ProjectHeaderPanel.class);
        tester.assertModelValue("projectHeader", project);
    }

    @Test
    public void reject_final_seminar() throws Exception {
        String feedback = "The results are incomprehensible";

        FormTester formTester = tester.newFormTester("decision");
        formTester.setValue("feedback", feedback);
        formTester.submit("reject");

        verify(reviewerDecisionService).reject(finalSeminarApproval, feedback, Option.<FileUpload>empty(), Option.<FileDescription>empty());
    }

    @SuppressWarnings("unchecked")
    @Test
    public void approve_final_seminar() throws Exception {
        String feedback = "This is good enough";
        File attachment = testFile();

        FormTester formTester = tester.newFormTester("decision");
        formTester.setValue("feedback", feedback);
        formTester.setFile("attachment", attachment, "application/pdf");
        formTester.submit("approve");

        ArgumentCaptor<Option> captor = ArgumentCaptor.forClass(Option.class);
        verify(reviewerDecisionService).approve(eq(finalSeminarApproval), eq(feedback), captor.capture(), eq(Option.<FileDescription>empty()));

        assertTrue(captor.getValue().isDefined());
    }

    @Test
    public void shows_details_about_the_final_seminar_approval() throws Exception {
        tester.assertComponent("details", FinalSeminarApprovalProcessPanel.class);
        tester.assertModelValue("details", finalSeminarApproval);
    }

    @Test
    public void form_is_hidden_if_a_decision_has_been_made() throws Exception {
        finalSeminarApproval.approve("Very good", Option.<FileDescription>empty(), Option.<FileDescription>empty());
        startPage();

        tester.assertInvisible("decision");
    }

    private Project createProject() {
        ProjectType bachelor = new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor");
        User supervisor = User.builder().firstName("Kalle").lastName("Tester").emailAddress("kalle@dsv.su.se").build();
        User reviewer = User.builder().firstName("John").lastName("Doe").emailAddress("john@example.com").build();
        reviewer.setId(34L);
        Project build = Project.builder().title("My project").projectType(bachelor).headSupervisor(supervisor).build();
        build.addReviewer(reviewer);
        build.setId(87L);
        return build;
    }

    private void startPage() {
        PageParameters pageParameters = ReviewerApprovalDecisionPage.pageParametersFor(finalSeminarApproval);
        tester.startPage(ReviewerApprovalDecisionPage.class, pageParameters);
    }

    private static File testFile() throws URISyntaxException {
        return new File(ReviewerApprovalDecisionPageTest.class.getResource("thesis.pdf").toURI());
    }
}