package se.su.dsv.scipro.reviewer;

import org.apache.wicket.Page;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Matchers;
import scala.Option;
import se.su.dsv.scipro.PageTest;
import se.su.dsv.scipro.file.FileDescription;
import se.su.dsv.scipro.finalthesis.FinalThesis;
import se.su.dsv.scipro.grading.ReviewerFillOutGradingReportPanel;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.report.GradingReport;
import se.su.dsv.scipro.report.GradingReportTemplate;
import se.su.dsv.scipro.report.ReviewerGradingReport;
import se.su.dsv.scipro.report.SupervisorGradingReport;
import se.su.dsv.scipro.reviewing.FinalSeminarApproval;
import se.su.dsv.scipro.reviewing.RoughDraftApproval;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.DomainObjects;
import se.su.dsv.scipro.test.UserBuilder;
import se.su.dsv.scipro.util.PageParameterKeys;

import java.util.Collections;

import static org.mockito.Mockito.when;

public class ReviewerGradingPageTest extends PageTest {

    public static final ProjectType BACHELOR = new ProjectType(DegreeLevel.FIRST_CYCLE, "bachelor", "bachelor");
    private ReviewerGradingReport reviewerReport;
    private SupervisorGradingReport supervisorReport;

    @Override
    protected Class<? extends Page> getPage() {
        return ReviewerGradingPage.class;
    }

    @Before
    public void setUp() throws Exception {
        final User emp = new UserBuilder().setId(3434L).create();
        setLoggedInAs(emp);
        final User author1 = User.builder().firstName("First").lastName("Last").emailAddress("author@example.com").build();
        author1.setId(234L);

        Project project = Project.builder().title("title").projectType(BACHELOR).build();
        project.addReviewer(emp);
        project.setHeadSupervisor(new UserBuilder().create());
        project.addProjectParticipant(author1);
        DomainObjects.injectId(project, 66L);

        GradingReportTemplate gradingReportTemplate = new GradingReportTemplate(BACHELOR, 30);
        supervisorReport = gradingReportTemplate.createSupervisorReport(project, author1);
        reviewerReport = gradingReportTemplate.createReviewerReport(project);

        when(projectService.findOne(Matchers.anyLong())).thenReturn(project);
        when(gradingReportService.getReviewerGradingReport(project)).thenReturn(reviewerReport);
        when(gradingReportService.getSupervisorGradingReports(project)).thenReturn(Collections.singletonList(supervisorReport));
        when(finalSeminarApprovalService.findBy(project)).thenReturn(Option.<FinalSeminarApproval>empty());
        when(roughDraftApprovalService.findBy(project)).thenReturn(Option.<RoughDraftApproval>empty());

        createFinalThesis();
    }

    private void createFinalThesis() {
        FinalThesis finalThesis = new FinalThesis();
        finalThesis.setId(1L);
        FileDescription fileDescription = new FileDescription();
        fileDescription.setId(1L);
        finalThesis.setFileDescription(fileDescription);
        when(finalThesisService.findByProject(Matchers.<Project>any())).thenReturn(finalThesis);
    }

    @Test
    public void has_fill_out_reviewer_grading_report_panel_if_reviewer() throws Exception {
        reviewerIsOnInitialStep();
        startPage();
        tester.assertComponent("fillOutPanel", ReviewerFillOutGradingReportPanel.class);
    }

    @Test
    public void does_not_show_supervisor_reports_if_initial_grading_is_not_done() throws Exception {
        reviewerIsOnInitialStep();
        startPage();
        tester.assertInvisible(path("supervisorReports", 0, "supervisorReport"));
    }

    @Test
    public void does_not_show_supervisor_reports_if_supervisor_initial_grading_is_not_done() throws Exception {
        reviewerIsOnReviewingStep();
        supervisorIsOnInitialStep();
        startPage();
        tester.assertInvisible(path("supervisorReports", 0, "supervisorReport"));
    }

    @Test
    public void shows_supervisor_reports_if_both_reviewer_and_supervisor_has_done_initial_grading() throws Exception {
        reviewerIsOnReviewingStep();
        supervisorIsOnReviewingStep();
        startPage();
        tester.assertVisible(path("supervisorReports", 0, "supervisorReport"));
    }

    private void supervisorIsOnReviewingStep() {
        supervisorReport.setState(GradingReport.State.REVIEWING);
    }

    private void reviewerIsOnReviewingStep() {
        reviewerReport.setState(GradingReport.State.REVIEWING);
    }

    private void supervisorIsOnInitialStep() {
        supervisorReport.setState(GradingReport.State.INITIAL);
    }

    private void reviewerIsOnInitialStep() {
        reviewerReport.setState(GradingReport.State.INITIAL);
    }

    @Override
    protected PageParameters getPageParameters() {
        PageParameters pageParameters = new PageParameters();
        pageParameters.add(PageParameterKeys.MAP.get(Project.class), 1L);
        return pageParameters;
    }
}