package se.su.dsv.scipro.statistics;

import org.apache.wicket.Page;
import org.apache.wicket.markup.html.form.FormComponent;
import org.apache.wicket.model.IModel;
import org.junit.Before;
import org.junit.Test;
import org.mockito.Mockito;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.date.DatePickerPanel;
import se.su.dsv.scipro.date.DeactivatableDatePickerPanel;
import se.su.dsv.scipro.milestones.dataobjects.MilestoneActivityTemplate;
import se.su.dsv.scipro.milestones.dataobjects.MilestonePhaseTemplate;
import se.su.dsv.scipro.milestones.service.MileStoneService;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectType;

import java.util.*;

import static java.util.Arrays.asList;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.eq;
import static se.su.dsv.scipro.statistics.MilestoneStatisticsPanel.*;

public class MilestoneStatisticsPanelTest extends SciProTest {

    private List<MilestoneActivityTemplate> projectActivities;
    private List<MilestoneActivityTemplate> individualActivities;
    private MilestoneActivityTemplate projectActivity;
    private MilestoneActivityTemplate individualActivity;
    private List<ProjectType> projectTypes;

    @Before
    public void setUp() throws Exception {
        projectTypes = new ArrayList<>(Collections.singletonList(new ProjectType(DegreeLevel.FIRST_CYCLE, "any", "any")));
        initActivities();
        Mockito.when(projectTypeService.findByDegreeLevels(EnumSet.of(DegreeLevel.FIRST_CYCLE, DegreeLevel.SECOND_CYCLE))).thenReturn(projectTypes);
        tester.startComponentInPage(new MilestoneStatisticsPanel("panel", new MilestoneStatisticsPanel.PageSupplier() {
            @Override
            public Page individualPage(IModel<MilestoneActivityTemplate> model, MileStoneService.Filter filter) {
                return new AdminIndividualMilestoneDetailsPage(model, filter);
            }

            @Override
            public Page projectPage(IModel<MilestoneActivityTemplate> model, MileStoneService.Filter filter) {
                return new AdminProjectMilestoneDetailsPage(model, filter);
            }

            @Override
            public Page chartPage(MileStoneService.Filter filter) {
                return new AdminMilestoneChartPage(filter);
            }
        }));
    }

    @Test
    public void contains_list_of_project_milestone_activities() throws Exception {
        tester.assertListView(path("panel", CONTAINER, PROJECT_ACTIVITY_LIST), projectActivities);
    }

    @Test
    public void project_milestones_have_number_of_completed_projects() throws Exception {
        Long completed = 2L;
        Mockito.when(mileStoneService.countCompletedMileStones(eq(projectActivity), any(MileStoneService.Filter.class))).thenReturn(completed);

        tester.assertModelValue(path("panel", CONTAINER, PROJECT_ACTIVITY_LIST, 0, COMPLETED_PROJECTS), completed);
    }

    @Test
    public void contains_list_of_individual_milestone_activities() throws Exception {
        tester.assertListView(path("panel", CONTAINER, INDIVIDUAL_ACTIVITY_LIST), individualActivities);
    }

    @Test
    public void individual_milestones_have_number_of_completed_authors() throws Exception {
        Long completed = 2L;
        Mockito.when(mileStoneService.countCompletedMileStones(eq(individualActivity), any(MileStoneService.Filter.class))).thenReturn(completed);

        tester.assertModelValue(path("panel", CONTAINER, INDIVIDUAL_ACTIVITY_LIST, 0, COMPLETED_AUTHORS), completed);
    }

    @Test
    public void updating_project_type_filter_renders_container() throws Exception {
        tester.executeAjaxEvent(getProjectTypeCheckGroup(), "click");
        tester.assertComponentOnAjaxResponse(path("panel", CONTAINER));
    }

    @Test
    public void updating_project_status_filter_renders_container() throws Exception {
        tester.executeAjaxEvent(getProjectStatusCheckGroup(), "click");
        tester.assertComponentOnAjaxResponse(path("panel", CONTAINER));
    }

    @Test
    public void updating_milestone_date_filter_renders_container() throws Exception {
        tester.executeAjaxEvent(getMileStoneDateField(), "inputchange");
        tester.assertComponentOnAjaxResponse(path("panel", CONTAINER));
    }

    @Test
    public void updating_project_date_filter_start_date() throws Exception {
        enableDateFilter();
        tester.executeAjaxEvent(getProjectDateField(DatePickerPanel.START_DATE), "inputchange");
        tester.assertComponentOnAjaxResponse(path("panel", CONTAINER));
    }

    @Test
    public void updating_project_date_filter_end_date() throws Exception {
        enableDateFilter();
        tester.executeAjaxEvent(getProjectDateField(DatePickerPanel.END_DATE), "inputchange");
        tester.assertComponentOnAjaxResponse(path("panel", CONTAINER));
    }

    @Test
    public void disabling_date_filter_disables_date_picker() throws Exception {
        disableDateFilter();
        tester.assertDisabled(path("panel", DATE_FILTER, DeactivatableDatePickerPanel.DATE_PICKER));
    }

    @Test
    public void enabling_date_filter_enables_date_picker() throws Exception {
        disableDateFilter();
        tester.assertDisabled(path("panel", DATE_FILTER, DeactivatableDatePickerPanel.DATE_PICKER));
    }

    @Test
    public void activities_have_link_to_project_statistics_page() throws Exception {
        tester.clickLink(path("panel", CONTAINER, PROJECT_ACTIVITY_LIST, 0, DETAILS_LINK));
        tester.assertRenderedPage(AdminProjectMilestoneDetailsPage.class);
    }

    @Test
    public void go_to_individual_milestone_activity_details() throws Exception {
        tester.clickLink(path("panel", CONTAINER, INDIVIDUAL_ACTIVITY_LIST, 0, INDIVIDUAL_ACTIVITY_NAME));

        tester.assertRenderedPage(AdminIndividualMilestoneDetailsPage.class);
    }

    @Test
    public void link_to_chart_page() throws Exception {
        tester.clickLink(path("panel", CONTAINER, "chartLink"));
        tester.assertRenderedPage(AdminMilestoneChartPage.class);
    }

    private void initActivities() {
        projectActivity = createActivity(MilestoneActivityTemplate.Type.PROJECT, "Project milestone", "This is it..", 0);
        projectActivities = asList(projectActivity);
        individualActivity = createActivity(MilestoneActivityTemplate.Type.STUDENT, "Student milestone", "Hi?", 0);
        individualActivities = asList(individualActivity);

        Mockito.when(milestoneActivityTemplateService.getAllActivities(eq(MilestoneActivityTemplate.Type.PROJECT), eq(projectTypes))).thenReturn(projectActivities);
        Mockito.when(milestoneActivityTemplateService.getAllActivities(eq(MilestoneActivityTemplate.Type.STUDENT), eq(projectTypes))).thenReturn(individualActivities);
    }

    private String getMileStoneDateField() {
        return path("panel", MILESTONE_DATE_FILTER);
    }

    private String getProjectTypeCheckGroup() {
        return path("panel", PROJECT_TYPE_FILTER);
    }

    private String getProjectStatusCheckGroup() {
        return path("panel", PROJECT_STATUSES);
    }

    private String getProjectDateField(String dateField) {
        return path("panel", DATE_FILTER, DeactivatableDatePickerPanel.DATE_PICKER, DatePickerPanel.DATE_FORM, dateField);
    }

    private void enableDateFilter() {
        setDateFilterEnabled(true);
    }

    private void disableDateFilter() {
        setDateFilterEnabled(false);
    }

    private void setDateFilterEnabled(boolean enabled) {
        FormComponent<?> dateFilterToggle = (FormComponent<?>) tester.getComponentFromLastRenderedPage(path("panel", DATE_FILTER, DeactivatableDatePickerPanel.TOGGLE));
        tester.getRequest().setParameter(dateFilterToggle.getInputName(), Boolean.toString(enabled));
        tester.executeAjaxEvent(dateFilterToggle, "click");
    }

    MilestoneActivityTemplate createActivity(MilestoneActivityTemplate.Type type, String title, String description, int order) {
        MilestoneActivityTemplate activity = new MilestoneActivityTemplate();
        activity.setType(type);
        activity.setTitle(title);
        activity.setDescription(description);
        activity.setMilestonePhaseTemplate(new MilestonePhaseTemplate("phase test", "this is for test", 0));
        activity.setProjectTypes(new HashSet<>(projectTypes));
        activity.setSortOrder(order);
        return activity;
    }
}
