package se.su.dsv.scipro.supervisor.panels;

import org.apache.wicket.model.Model;
import org.apache.wicket.request.mapper.parameter.PageParameters;
import org.junit.Before;
import org.junit.Test;
import se.su.dsv.scipro.SciProTest;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.project.pages.AbstractProjectDetailsPage;
import se.su.dsv.scipro.system.DegreeLevel;
import se.su.dsv.scipro.system.ProjectModule;
import se.su.dsv.scipro.system.ProjectType;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ObjectMother;
import se.su.dsv.scipro.test.UserBuilder;

import java.util.Date;

import static org.hamcrest.core.IsNot.not;
import static org.hamcrest.core.StringContains.containsString;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertThat;
import static org.mockito.Matchers.any;
import static org.mockito.Mockito.when;
import static se.su.dsv.scipro.supervisor.panels.SupervisorSubTabMenuPanel.*;

public class SupervisorSubTabMenuPanelTest extends SciProTest {

    public static final Project SOME_BACHELOR_PROJECT = Project.builder().title("Title").projectType(new ProjectType(DegreeLevel.FIRST_CYCLE, "Bachelor", "Bachelor")).build();
    public static final Project SOME_MASTER_PROJECT = Project.builder().title("Title").projectType(new ProjectType(DegreeLevel.SECOND_CYCLE, "Master", "Master")).build();
    public static final Project ICT4D_PROJECT = Project.builder().title("ICT4D").projectType(new ProjectType(DegreeLevel.NONE, "ICT4D", "ICT4D")).build();
    private User reviewer;

    @Before
    public void setUp() throws Exception {
        reviewer = new UserBuilder().create();
    }

    @Test
    public void renders() throws Exception {
        assertNotNull(startPanel(SOME_BACHELOR_PROJECT));
    }

    @Test
    public void milestone_tab_is_hidden_if_no_activities_are_available() throws Exception {
        when(milestoneActivityTemplateService.countActivities(any(ProjectType.class), any(Date.class))).thenReturn(0L);
        assertNotVisible(MILESTONES, SOME_BACHELOR_PROJECT);
    }

    @Test
    public void milestone_is_visible_if_activities_are_available() throws Exception {
        when(milestoneActivityTemplateService.countActivities(any(ProjectType.class), any(Date.class))).thenReturn(1L);
        assertVisible(MILESTONES, SOME_BACHELOR_PROJECT);
    }

    @Test
    public void final_seminar_tab_is_hidden_if_no_seminar_module() throws Exception {
        setModules(ProjectModule.CHECKLIST);
        assertNotVisible(FINAL_SEMINAR, ICT4D_PROJECT);
    }

    @Test
    public void final_seminar_tab_is_visible_if_first_cycle() throws Exception {
        assertVisible(FINAL_SEMINAR, SOME_BACHELOR_PROJECT);
    }

    @Test
    public void final_seminar_tab_is_visible_if_second_cycle() throws Exception {
        assertVisible(FINAL_SEMINAR, SOME_MASTER_PROJECT);
    }

    @Test
    public void reviewer_interaction_tab_is_invisible_if_not_head_supervisor() throws Exception {
        assertNotVisible(INTERACT_WITH_REVIEWER, SOME_BACHELOR_PROJECT);
    }

    @Test
    public void reviewer_interaction_tab_is_visible_if_head_supervisor() throws Exception {
        Project project = mockProjectWithSupervisorLoggedIn();
        project.getProjectType().setReviewed(true);
        assertVisible(INTERACT_WITH_REVIEWER, project);
    }

    @Test
    public void reviewer_interaction_is_not_visible_if_no_reviewer_is_attached() throws Exception {
        Project project = mockProjectWithSupervisorLoggedIn();
        project.getProjectType().setReviewed(false);
        assertNotVisible(INTERACT_WITH_REVIEWER, project);
    }

    private Project mockProjectWithSupervisorLoggedIn() {
        User supervisor = new UserBuilder().create();
        setLoggedInAs(supervisor);
        Project project = Project.builder().title("test").projectType(ObjectMother.BACHELOR).headSupervisor(supervisor).build();
        project.addReviewer(reviewer);
        return project;
    }

    private String renderedPanel(Project project) {
        startPanel(project);
        return tester.getLastResponseAsString();
    }

    private void assertVisible(String menuChoice, Project project) {
        assertThat(renderedPanel(project), containsString(menuChoice));
    }

    private void assertNotVisible(String menuChoice, Project project) {
        assertThat(renderedPanel(project), not(containsString(menuChoice)));
    }

    private SupervisorSubTabMenuPanel startPanel(Project project) {
        return tester.startComponentInPage(new SupervisorSubTabMenuPanel("id", AbstractProjectDetailsPage.class, new PageParameters(), Model.of(project)));
    }
}
