package se.su.dsv.scipro.workerthreads;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.notifications.NotificationService;
import se.su.dsv.scipro.notifications.dataobject.Notification;
import se.su.dsv.scipro.notifications.dataobject.ProjectEvent;
import se.su.dsv.scipro.notifications.dataobject.SeminarEvent;
import se.su.dsv.scipro.notifications.interfaces.NotificationMailFormatter;
import se.su.dsv.scipro.notifications.settings.entities.DeliveryMethod;
import se.su.dsv.scipro.notifications.settings.service.DeliveryConfigurationService;
import se.su.dsv.scipro.project.Project;
import se.su.dsv.scipro.springdata.services.UserProfileService;
import se.su.dsv.scipro.system.User;
import se.su.dsv.scipro.test.ObjectMother;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class NotificationCompilationWorkerTest {

    NotificationCompilationWorker worker;

    @Mock NotificationService notificationService;
    @Mock NotificationMailFormatter formatter;
    @Mock MailEventService mailEventService;
    @Mock GeneralSystemSettingsService systemSettingsService;
    @Mock DeliveryConfigurationService deliveryConfiguration;
    @Mock UserProfileService userProfileService;

    @Before
    public void setUp() throws Exception {
        worker = spy(new NotificationCompilationWorker());
        formatter = mock(NotificationMailFormatter.class, RETURNS_MOCKS);

        worker.notificationService = notificationService;
        worker.mailEventService = mailEventService;
        worker.systemSettingsService = systemSettingsService;
        worker.notificationMailFormatter = formatter;
        worker.deliveryConfiguration = deliveryConfiguration;
        worker.userProfileService = userProfileService;

        when(systemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(new GeneralSystemSettings());

        doNothing().when(worker).beginTransaction();
        doNothing().when(worker).commitTransaction();
        doNothing().when(worker).rollbackTransaction();
    }

    @Test
    public void test_adherence_to_type_setting() {
        User user1 = ObjectMother._User.create(13L);
        Project project = ObjectMother._Project.create(14L);

        ProjectEvent projectData = new ProjectEvent();
        projectData.setProject(project);
        projectData.setEvent(ProjectEvent.Event.STATE_CHANGED);

        Notification notification1 = new Notification();
        notification1.setUser(user1);
        notification1.setNotificationEvent(projectData);

        Notification notification2 = new Notification();
        notification2.setUser(user1);
        notification2.setNotificationEvent(new SeminarEvent());

        List<Notification> notificationList = new ArrayList<>();
        notificationList.add(notification1);
        notificationList.add(notification2);

        List<User> userList = Collections.singletonList(user1);
        when(userProfileService.findUsersWithMailCompilation()).thenReturn(userList);
        when(notificationService.findUnreadByUsers(userList)).thenReturn(notificationList);

        when(deliveryConfiguration.isDelivery(eq(Notification.Type.PROJECT), isA(Enum.class), eq(DeliveryMethod.MAIL), isA(User.class))).thenReturn(true);
        when(deliveryConfiguration.isDelivery(eq(Notification.Type.FINAL_SEMINAR), isA(Enum.class), eq(DeliveryMethod.MAIL), isA(User.class))).thenReturn(false);

        worker.doWork();

        verify(worker.mailEventService, times(1)).save(isA(MailEvent.class));
    }

    @Test
    public void include_notification_only_once() {
        User user = ObjectMother._User.create(14L);
        Project project = ObjectMother._Project.create(15L);

        ProjectEvent data = new ProjectEvent();
        data.setEvent(ProjectEvent.Event.CREATED);
        data.setProject(project);

        Notification notification = new Notification();
        notification.setNotificationEvent(data);
        notification.setUser(user);

        List<Notification> notificationList = new ArrayList<>();
        notificationList.add(notification);

        List<User> users = Collections.singletonList(user);
        when(userProfileService.findUsersWithMailCompilation()).thenReturn(users);
        when(notificationService.findUnreadByUsers(users)).thenReturn(notificationList);
        when(deliveryConfiguration.isDelivery(eq(Notification.Type.PROJECT), isA(Enum.class), eq(DeliveryMethod.MAIL), isA(User.class))).thenReturn(true);

        worker.doWork();
        worker.doWork();

        verify(worker.mailEventService, times(1)).save(isA(MailEvent.class));
    }
}
