package se.su.dsv.scipro.workerthreads;

import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.runners.MockitoJUnitRunner;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings;
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettingsService;
import se.su.dsv.scipro.mail.MailEvent;
import se.su.dsv.scipro.mail.MailEventService;
import se.su.dsv.scipro.plagiarism.TurnitinSettings;
import se.su.dsv.scipro.plagiarism.TurnitinSettingsService;

import java.util.ArrayList;
import java.util.Date;

import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.*;

@RunWith(MockitoJUnitRunner.class)
public class TurnItInExpirationWorkerTest {

    private static final String MOCK_MAIL = "example@example.com";

    @Mock
    MailEventService mailEventService;
    @Mock
    GeneralSystemSettingsService generalSystemSettingsService;
    @Mock
    TurnitinSettingsService turnitinSettingsService;

    TurnItInExpirationWorker worker;

    @Before
    public void setUp() throws Exception {
        worker = new TurnItInExpirationWorker(generalSystemSettingsService, turnitinSettingsService, mailEventService) {
            @Override
            protected void beginTransaction() {
            }

            @Override
            protected void commitTransaction() {
            }

            @Override
            protected void rollbackTransaction() {
            }
        };

        when(generalSystemSettingsService.getGeneralSystemSettingsInstance()).thenReturn(new GeneralSystemSettings());
    }

    @Test
    public void testDoesNotMailBeforeFourWeeks() throws Exception {
        // given
        when(turnitinSettingsService.getSettings()).thenReturn(withExpirationInDays(4 * 7 + 1));

        // when
        worker.doWork();

        // then
        verify(mailEventService, never()).save(isA(MailEvent.class));
    }

    @Test
    public void testSendsMailAtFourWeeks() throws Exception {
        // given
        setExpiresIn(4 * 7);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    @Test
    public void testSendsMailAtThreeWeeks() throws Exception {
        // given
        setExpiresIn(3 * 7);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    @Test
    public void testSubjectContainsDaysToExpiration() throws Exception {
        // given
        int days = 21;
        setExpiresIn(days);

        // when
        worker.doWork();

        // then
        ArgumentCaptor<MailEvent> captor = ArgumentCaptor.forClass(MailEvent.class);
        verify(mailEventService).save(captor.capture());

        assertTrue("Does not contain days to expiration", captor.getValue().getSubject().contains(String.valueOf(days)));
    }

    @Test
    public void testSendsMailAtTwoWeeks() throws Exception {
        // given
        setExpiresIn(2 * 7);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    @Test
    public void testSendsMailAtOneWeek() throws Exception {
        // given
        setExpiresIn(7);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    @Test
    public void testSendsMailDayBefore() throws Exception {
        // given
        setExpiresIn(1);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    @Test
    public void testSendsMailOnExpirationDay() throws Exception {
        // given
        setExpiresIn(0);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    @Test
    public void testSendsMailWeeklyAfterExpiration() throws Exception {
        // given
        setExpiresIn(-7);

        // when
        worker.doWork();

        // then
        verify(mailEventService).save(isA(MailEvent.class));
    }

    private void setExpiresIn(int days) {



        when(turnitinSettingsService.getSettings()).thenReturn(withExpirationInDays(days));
    }

    private TurnitinSettings withExpirationInDays(int days) {
        Date expirationDate = new Date();
        expirationDate.setTime(days * 24L * 60 * 60 * 1000 + expirationDate.getTime());
        TurnitinSettings turnitinSettings = new TurnitinSettings();
        turnitinSettings.setExpirationDate(expirationDate);
        turnitinSettings.setExpirationContent("");

        ArrayList<String> mailAddresses = new ArrayList<>();
        mailAddresses.add(MOCK_MAIL);

        turnitinSettings.setExpirationMails(mailAddresses);
        return turnitinSettings;
    }
}
