package se.su.dsv.scipro.finalseminar

import java.text.SimpleDateFormat
import java.util
import java.util.{Calendar, Date}

import org.apache.wicket.feedback.FeedbackMessage
import org.apache.wicket.markup.html.form.RequiredTextField
import org.apache.wicket.model.Model
import org.apache.wicket.util.tester.FormTester
import org.hamcrest.CoreMatchers.hasItem
import org.joda.time.{MutableDateTime, ReadableDateTime}
import org.junit.Assert.{assertEquals, assertFalse, assertThat}
import org.junit.{Before, Test}
import org.mockito.Mockito.{verify, verifyNoMoreInteractions, when}
import org.mockito.{ArgumentCaptor, Matchers, Mockito}
import se.su.dsv.scipro.SciProTest
import se.su.dsv.scipro.data.enums.DateStyle
import se.su.dsv.scipro.finalseminar.SeminarCRUDPanel._
import se.su.dsv.scipro.generalsystemsettings.GeneralSystemSettings
import se.su.dsv.scipro.notifications.dataobject.{NotificationSource, SeminarEvent}
import se.su.dsv.scipro.project.Project
import se.su.dsv.scipro.reusable.SciProUtilities
import se.su.dsv.scipro.security.auth.roles.Roles
import se.su.dsv.scipro.system.{Language, ProjectType, User}
import se.su.dsv.scipro.test.DomainObjects

class SeminarCRUDPanelTest extends SciProTest {
  private var panel: SeminarCRUDPanel = _
  private var seminar: FinalSeminar = _
  private var supervisor: User = _
  private var english: Language = _
  private var languages: util.List[Language] = _

  @Before def setup_context(): Unit = {
    english = new Language
    DomainObjects.injectId(english, 1L)
    languages = java.util.Arrays.asList(english)
    prepareSeminar()
    when(generalSystemSettingsService.getGeneralSystemSettingsInstance).thenReturn(new GeneralSystemSettings)
    when(finalSeminarSettingsService.getInstance).thenReturn(new FinalSeminarSettings)
    when(finalSeminarService.save(seminar)).thenReturn(seminar)
    when(languageService.findAll).thenReturn(languages)
    when(finalSeminarCreationService.isCreationAllowed(seminar.getProject)).thenReturn(NoReviewer)
    setLoggedInAs(supervisor)
    startPanel()
  }

  @Test def changing_room_should_send_notification() = {
    DomainObjects.injectId(seminar, 3254L)
    val formTester: FormTester = newFormTester
    formTester.setValue(ROOM, "some new room")
    formTester.submit
    verifyNotificationSent(SeminarEvent.Event.ROOM_CHANGED)
  }

  @Test def not_changing_anything_does_not_send_any_notifications() = {
    DomainObjects.injectId(seminar, 3254L)
    val formTester: FormTester = newFormTester
    formTester.submit
    tester.assertNoErrorMessage()
    verify(notificationController, Mockito.never).notifySeminar(Matchers.any(classOf[FinalSeminar]), Matchers.any(classOf[SeminarEvent.Event]), Matchers.any(classOf[NotificationSource]))
  }

  @Test def shows_feedback_when_saving_seminar() = {
    val formTester: FormTester = newFormTester
    formTester.submit
    assertThat(tester.getMessages(FeedbackMessage.SUCCESS), hasItem(panel.getString("final.seminar.updated")))
  }

  @Test def panel_should_contain_max_participant() = {
    tester.assertComponent(path(FORM, MAX_PARTICIPANTS), classOf[RequiredTextField[_]])
  }

  @Test def panel_should_contain_max_opponents() = {
    tester.assertComponent(path(FORM, MAX_OPPONENTS), classOf[RequiredTextField[_]])
  }

  @Test def panel_not_visible_when_student() = {
    setRoles(Roles.AUTHOR)
    setLoggedInAs(User.builder.firstName("Sture").lastName("Student").emailAddress("sture@example.com").build)
    startPanel()
    tester.assertInvisible(path())
  }

  @Test def delete_seminar() = {
    DomainObjects.injectId(seminar, 5L)
    startPanel()
    tester.clickLink(path(FORM, DELETE))
    verify(finalSeminarService).delete(seminar)
  }

  @Test def shows_cancelled_information_if_seminar_is_lazy_deleted() = {
    seminar.setCancelled(true)
    startPanel()
    tester.assertVisible(path(FORM, CANCELLED))
  }

  @Test def uncancels_seminar_on_successful_submit() = {
    seminar.setCancelled(true)
    startPanel()
    val formTester: FormTester = newFormTester
    formTester.submit
    val captor: ArgumentCaptor[FinalSeminar] = ArgumentCaptor.forClass(classOf[FinalSeminar])
    verify(finalSeminarService).save(captor.capture)
    assertFalse(captor.getValue.isCancelled)
  }

  @Test def delete_button_is_only_visible_if_seminar_has_an_id() = {
    tester.assertInvisible(path(FORM, DELETE))
  }

  @Test def shows_create_if_seminar_has_no_id() = {
    tester.assertLabel(path(FORM, SUBMIT), panel.getString("create"))
  }

  @Test def shows_update_if_seminar_has_an_id() = {
    DomainObjects.injectId(seminar, 23234L)
    startPanel()
    tester.assertLabel(path(FORM, SUBMIT), panel.getString("update"))
  }

  @Test def creating_final_seminar_sends_created_event_only() = {
    val formTester: FormTester = newFormTester
    formTester.submit
    verifyNotificationSent(SeminarEvent.Event.CREATED)
    verifyNoMoreInteractions(notificationController)
  }

  @Test def shows_feedback_on_delete() = {
    DomainObjects.injectId(seminar, 32525L)
    startPanel()
    tester.clickLink(path(FORM, DELETE))
    assertThat(tester.getMessages(FeedbackMessage.SUCCESS), hasItem(panel.getString("deleted")))
  }

  @Test def testInvalidMaxOpponentsSubmission() = {
    val formTester: FormTester = newFormTester
    formTester.setValue(MAX_OPPONENTS, Integer.toString(seminar.getMinOpponents - 1))
    formTester.submit
    tester.assertErrorMessages("The selected number of max oppositions may not be lower than " + seminar.getMinOpponents)
  }

  @Test def should_update_domain_object() = {
    setRoles(Roles.SYSADMIN)
    startPanel()
    val formTester: FormTester = newFormTester
    val opponents: Int = 52
    formTester.setValue(MAX_OPPONENTS, String.valueOf(opponents))
    val participants: Int = 5
    formTester.setValue(MAX_PARTICIPANTS, String.valueOf(participants))
    val room: String = "secret room"
    formTester.setValue(ROOM, room)
    val calendar: Calendar = Calendar.getInstance
    calendar.set(calendar.get(Calendar.YEAR) + 1, Calendar.APRIL, 23, 7, 32, 0)
    calendar.set(Calendar.MILLISECOND, 0)
    val dateString: String = getDateString(calendar)
    formTester.setValue(super.path(DATE, "date"), dateString)
    formTester.setValue(super.path(DATE, "time"), "07:32")
    formTester.submit
    tester.assertNoErrorMessage()
    val persistedSeminar: FinalSeminar = seminar
    assertEquals(calendar.getTime, persistedSeminar.getStartDate)
    assertEquals(opponents.toLong, persistedSeminar.getMaxOpponents.toLong)
    assertEquals(participants.toLong, persistedSeminar.getMaxParticipants.toLong)
    assertEquals(room, persistedSeminar.getRoom)
  }

  @Test def presentation_language_selected_if_only_one_language_exists_otherwise_null(): Unit = {
    tester.assertModelValue(path(FORM, PRESENTATION_LANGUAGE), english)
  }

  @Test def report_language_selected_if_only_one_language_exists(): Unit = {
    tester.assertModelValue(path(FORM, REPORT_LANGUAGE), english)
  }

  @Test def move_seminar_link_is_only_visible_if_seminar_has_an_id(): Unit = {
    tester.assertInvisible(path(FORM, MOVE_SEMINAR_LINK))
  }

  protected override def path(parts: AnyRef*): String = {
    super.path(panel, super.path(parts: _*))
  }

  private def verifyNotificationSent(event: SeminarEvent.Event): Unit = {
    verify(notificationController).notifySeminar(Matchers.eq(seminar), Matchers.eq(event), Matchers.any(classOf[NotificationSource]))
  }

  private def setDateValue(formTester: FormTester, path: String, date: ReadableDateTime): Unit = {
    formTester.setValue(path, date.toString(DateStyle.DATETIME.getFormat))
    ()
  }

  private def getDateString(calendar: Calendar): String = {
    new SimpleDateFormat(dateService.getFormat(DateStyle.DATE)).format(calendar.getTime)
  }

  private def newFormTester: FormTester = {
    tester.newFormTester(path(FORM))
  }

  private def startPanel() = {
    panel = tester.startComponentInPage(new SeminarCRUDPanel("panel", Model.of(seminar)))
  }

  private def prepareSeminar() = {
    val bachelor: ProjectType = new ProjectType(ProjectType.BACHELOR, "Bachelor", "Bachelor")
    bachelor.getProjectTypeSettings.setMinOpponentsOnFinalSeminar(2)
    supervisor = User.builder.firstName("Hasse").lastName("Handledare").emailAddress("hasse@example.com").build
    DomainObjects.injectId(supervisor, 234L)
    val project: Project = new Project
    project.setId(1L)
    project.setProjectType(bachelor)
    project.setHeadSupervisor(supervisor)
    seminar = new FinalSeminar
    seminar.setProject(project)
    seminar.setStartDate(oneYearFromNow)
    seminar.setRoom("6405")
    seminar.setPresentationLanguage(english)
    seminar.setReportLanguage(english)
  }

  private def thirtyDaysFromNow: ReadableDateTime = {
    val dateTime: MutableDateTime = new MutableDateTime
    dateTime.addDays(30)
    dateTime
  }

  private def oneYearFromNow: Date = {
    val startDate: MutableDateTime = new MutableDateTime
    startDate.addDays(365)
    SciProUtilities.setTimeOfDayTo0000(startDate.toDate)
  }
}